/*
 TSAFE Prototype: A decision support tool for air traffic controllers
 Copyright (C) 2003  Gregory D. Dennis

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package tsafe.engine;

import tsafe.data.*;
import java.util.*;

/**
 * ConformanceMonitor.java
 * Determines to what degree a flight is conforming to its planned route
 */
class ConformanceMonitor {
    
    /**
     * Calculator for distances and lat/long to x,y conversion
     */
    private EngineCalculator calc;

    /**
     * Holds conformance monitoring parameters
     */
    private EngineParameters params;

    /**
     * Sole Constructor
     */
    public ConformanceMonitor(EngineCalculator calc, EngineParameters params) {
        this.calc = calc;
        this.params = params;
    }

    /**
     * Blunder Detection Algorithm
     * Return true if the ftObserved is too far from ftExpected
     */
    public boolean isBlundering(FlightTrack ftObserved, FlightTrack ftExpected) {
        double residual = generateResidual(ftObserved, ftExpected);
        return residual >= params.cmResidualThreshold;
    }

    /**
     * Residual Generator
     */
    private double generateResidual(FlightTrack ftObserved, FlightTrack ftExpected) {
        int numFactors = 0;
        double sum = 0;

        // Add the lateral factor if it is turned on
        if (params.cmLateralWeightOn) {
            double lateralDev = calc.distanceLL(ftObserved.getLatitude(), ftObserved.getLongitude(),
                                                ftExpected.getLatitude(), ftExpected.getLongitude());
            double lateralResidual = lateralDev / params.cmLateralThreshold;
            sum += lateralResidual;
            numFactors++;

            System.out.println();
            System.out.println("Observed position:  " + ftObserved.getLatitude() + "/" + ftObserved.getLongitude());
            System.out.println("Expected position:  " + ftExpected.getLatitude() + "/" + ftExpected.getLongitude());
            System.out.println("Lateral deviation:  " + lateralDev);
            System.out.println("Lateral threashold: " + params.cmLateralThreshold);
        }

        // Add the vertical factor if it is turned on
        if (params.cmVerticalWeightOn) {
            double verticalDev = Math.abs(ftObserved.getAltitude() - ftExpected.getAltitude());
            double verticalResidual = verticalDev / params.cmVerticalThreshold;
            sum += verticalResidual;
            numFactors++;

            System.out.println();
            System.out.println("Observed altitude:   " + ftObserved.getAltitude());
            System.out.println("Expected altitude:   " + ftExpected.getAltitude());
            System.out.println("Vertical deviation:  " + verticalDev);
            System.out.println("Vertical threashold: " + params.cmVerticalThreshold);
        }

        // Add the angular factor if it is on
        if (params.cmAngularWeightOn) {
            double angularDev = Math.abs(ftObserved.getHeading() - ftExpected.getHeading());
            //while (angularDev >  2 * Math.PI) angularDev -= 2 * Math.PI;
            sum += angularDev / params.cmAngularThreshold;
            numFactors++;

            System.out.println();
            System.out.println("Observed heading:   " + ftObserved.getHeading());
            System.out.println("Expected heading:   " + ftExpected.getHeading());
            System.out.println("Angular deviation:  " + angularDev);
            System.out.println("Angular threashold: " + params.cmAngularThreshold);
        }

        // Add the speed factor if it is on
        if (params.cmSpeedWeightOn) {
            double speedDev = Math.abs(ftObserved.getAltitude() - ftExpected.getAltitude());
            sum += speedDev / params.cmSpeedThreshold;
            numFactors++;

            System.out.println();
            System.out.println("Observed speed:   " + ftObserved.getSpeed());
            System.out.println("Expected speed:   " + ftExpected.getSpeed());
            System.out.println("Speed deviation:  " + speedDev);
            System.out.println("Speed threashold: " + params.cmSpeedThreshold);
        }

        // Normalize the summation
        // If zero factors are involved, define the residual to be zero
        System.out.println();
        System.out.println("Total Sum: " + sum + "  Num Factors: " + numFactors);
        System.out.println("Conformance Residual: " + (numFactors == 0 ? 0 : sum / numFactors));
        
        return numFactors == 0 ? 0 : sum / (double)numFactors;
    }
}