/*
 TSAFE Prototype: A decision support tool for air traffic controllers
 Copyright (C) 2003  Gregory D. Dennis

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package tsafe.engine;

import tsafe.data.*;
import java.util.*;

/**
 * Finds a flights position along its route
 */
class RouteTracker {

    /** Calculator for distances and lat/long to x,y conversion */
    private EngineCalculator calc;

    /** Sole Constructor */
    public RouteTracker(EngineCalculator calc) {
        this.calc = calc;
    }

    /**
     * Finds the point along the route that is closest to the flight track
     * We assume that every route has at least 2 fixes.
     */
    public RouteTrack findRouteTrack(FlightTrack ft, FlightPlan fp) {
        Point2D flightPoint = new Point2D(ft.getLatitude(), ft.getLongitude());

        // We will find the route position on each route segment
        // and take the route position that is closest to the flight
        Iterator fixIter = fp.getRoute().fixIterator();

        // If route does not have any fixess
        // return flights current position
        if (!fixIter.hasNext()) {
            //System.out.println("ROUTE DOES NOT HAVE ANY POINTS");
            return new RouteTrack(null, null, ft);
        }
        Fix prevFix = (Fix)fixIter.next();

        // If route does not has exactly one fix,
        // make that fix the prev and next
        if (!fixIter.hasNext()) {
            //System.out.println("ROUTE HAS EXACTLY ONE POINT");
            return new RouteTrack(prevFix, prevFix, ft);
        }

        // Minimums to keep track of
        double minDeviation = Double.MAX_VALUE;
        Fix minPrevFix = null, minNextFix = null;
        Point2D minSnapPoint = null;
        
        while (fixIter.hasNext()) {
            Fix nextFix = (Fix)fixIter.next();
            Point2D snapPoint = snapPointToRouteSegment(flightPoint, prevFix, nextFix, false);
            double snapDistance = calc.distanceLL(snapPoint, flightPoint);

            // found a new minimum
            if (snapDistance < minDeviation) {
                minPrevFix = prevFix;
                minNextFix = nextFix;
                minSnapPoint = snapPoint;
                minDeviation = snapDistance;
            }
            
            prevFix = nextFix;
        }

        PointXY flightXY = calc.toXY(flightPoint);
        PointXY prevXY = calc.toXY(minPrevFix);
        PointXY nextXY = calc.toXY(minNextFix);

        return new RouteTrack(minPrevFix, minNextFix,
                              minSnapPoint.getLatitude(), minSnapPoint.getLongitude(),
                              fp.getAssignedAltitude(), ft.getTime(), fp.getAssignedSpeed(),
                              calc.angleLL(minPrevFix, minNextFix));
    }

    /**
     * Added by Bohuslav Krena.    
     * Creates a list of closest point for each segment of the route.
     */
    public LinkedList getSegmentsTracks(FlightTrack ft, FlightPlan fp) {
        Point2D flightPoint = new Point2D(ft.getLatitude(), ft.getLongitude());
        LinkedList resultSegments = new LinkedList();

        // We will find the route position on each route segment
        // and take the route position that is closest to the flight
        Iterator fixIter = fp.getRoute().fixIterator();

        // If route does not have any fixess do nothing.
        if (!fixIter.hasNext()) {
            return resultSegments;
        }
        Fix prevFix = (Fix)fixIter.next();

        // If route does not has exactly one fix,
        // make that fix the prev and next
        if (!fixIter.hasNext()) {
            resultSegments.add(new RouteTrack(prevFix, prevFix, ft));
            return resultSegments;
        }

        // Minimums to keep track of
        double minDeviation = Double.MAX_VALUE;
        Fix minPrevFix = null, minNextFix = null;
        Point2D minSnapPoint = null;
        
        while (fixIter.hasNext()) {
            Fix nextFix = (Fix)fixIter.next();
            Point2D snapPoint = snapPointToRouteSegment(flightPoint, prevFix, nextFix, false);
            double snapDistance = calc.distanceLL(snapPoint, flightPoint);

            minPrevFix = prevFix;
            minNextFix = nextFix;
            minSnapPoint = snapPoint;
            minDeviation = snapDistance;

            PointXY flightXY = calc.toXY(flightPoint);
            PointXY prevXY = calc.toXY(minPrevFix);
            PointXY nextXY = calc.toXY(minNextFix);

            resultSegments.add(new RouteTrack(minPrevFix, minNextFix,
                              minSnapPoint.getLatitude(), minSnapPoint.getLongitude(),
                              fp.getAssignedAltitude(), ft.getTime(), fp.getAssignedSpeed(),
                              calc.angleLL(minPrevFix, minNextFix)));
            
            prevFix = nextFix;
        }

        return resultSegments;
    }

    /**
     * Finds closest point on straight line between l1 and l2 to p
     */
    private Point2D snapPointToRouteSegment(Point2D latLonPoint, Point2D fix1, Point2D fix2, boolean print) {
        PointXY p = calc.toXY(latLonPoint);
        PointXY l1 = calc.toXY(fix1);
        PointXY l2 = calc.toXY(fix2);
    
        // Find linear equation for line segment (y = mx + b)
        double m = (l2.getY() - l1.getY()) / (l2.getX() - l1.getX());
        double b = l1.getY() - (m * l1.getX());
        
        // Find linear equation for perpendicular line
        double mP = -1 / m;
        double bP = p.getY() - (mP * p.getX());
        
        // Find where line segment and perpendicular intersect
        double x = (b - bP) / (mP - m);
        double y = (m * x) + b;
        PointXY intersectXY = new PointXY(x, y);
        
        // Check if this point does indeed lie on the line segment.
        // If so, it must be the closest point on the line segment to p
        if (((l1.getX() < x && x < l2.getX()) || (l2.getX() < x && x < l1.getX())) &&
            ((l1.getY() < y && y < l2.getY()) || (l2.getY() < y && y < l1.getY()))) {
            return calc.toLL(intersectXY);
        }
    
        // If this intersection point does not lie on the line segment,
        // the closest point on the line segment to p must be an end point
        // Find the minimum distance to an end point
        double dist1 = calc.distanceXY(p, l1);
        double dist2 = calc.distanceXY(p, l2);
        PointXY minXY = dist1 < dist2 ? l1 : l2;
        return calc.toLL(minXY);
    }
}
