/*
 TSAFE Prototype: A decision support tool for air traffic controllers
 Copyright (C) 2003  Gregory D. Dennis

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package tsafe.engine;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Map;

import tsafe.data.Flight;
import tsafe.data.FlightPlan;
import tsafe.data.FlightTrack;
import tsafe.data.Trajectory;
import tsafe.engine.EngineParameters;
import tsafe.engine.RouteTracker;
import tsafe.engine.ConformanceMonitor;

/**
 * Handles the control flow between the algorithmic units of TSAFE
 */
public class TsafeEngine extends Thread {

    /**
     * Engine parameters
     */
    private EngineParameters parameters;

    /**
     * Engine components
     */
    private RouteTracker routeTracker;
    private ConformanceMonitor confMonitor;
    private TrajectorySynthesizer trajSynth;

    /**
     * Collection of flights over which algorithms calculate
     */
    private Collection flights = new LinkedList();
   
    /**
     * Private engine constructor
     */
    public TsafeEngine(EngineCalculator calc) {
        this.parameters    = new EngineParameters();
        this.routeTracker  = new RouteTracker(calc);
        this.confMonitor   = new ConformanceMonitor(calc, parameters);
        this.trajSynth     = new TrajectorySynthesizer(calc, parameters);
    }

    /**
     * Return the set of engine parameters
     */
    public EngineParameters getParameters() {
        return this.parameters;
    }

    /**
     * Set the flights over which the engine is to compute
     */
    public void setFlights(Collection flights) {
        this.flights = flights;
    }

    /****************************************************************************
     * Runs the TSAFE Engine
     * Performs Conformance Monitoring and Trajectory Synthesis
     ****************************************************************************/
    public void start() {
        // Instantiate an empty collection of blunders,
        // an empty flight2traj map, and a copy of the current flights
        //Collection copyFlights = new LinkedList(flights);
        Collection blunders = new LinkedList();
        Map flight2TrajMap = new HashMap();

        // For each flight:
        // 1) If it has no flight plan, assign it a dr traj and continue
        // 2) If it has a flight plan, determine if it is blundering
        // 3) If it is, assign its dr trajectory as its predicted trajectory
        //    If it isn't, assign its route trajectory as its predicted trajectory
        Iterator flightIter = flights.iterator(); //copyFlights.iterator();

        while(flightIter.hasNext()) {
            Flight flight = (Flight)flightIter.next();
            FlightTrack ft = flight.getFlightTrack();
            FlightPlan fp = flight.getFlightPlan();

            // If the flight doesn't have a flight plan, assign it a dr trajectory
            // Don't check its conformance, just continue to the next flight
            if (fp == null) {
                Trajectory drTraj = trajSynth.getDeadReckoningTrajectory(ft);
                flight2TrajMap.put(flight, drTraj);
                continue;
            }

            // Determine if flight is blundering by comparing its actual track to its route track
            RouteTrack rt = routeTracker.findRouteTrack(ft, fp);
            boolean blundering = confMonitor.isBlundering(ft, rt);

//INSTRUMENTATION BEGIN
{
	// Check whether any segment of the plan conforms with current track
	boolean conforming = false;
	LinkedList segments = routeTracker.getSegmentsTracks(ft, fp);
	ListIterator segIter = segments.listIterator();
	int segNum = 1;
	while (segIter.hasNext()) {
	    rt = (RouteTrack)segIter.next();
	    boolean currentTrackConforms = ! confMonitor.isBlundering(ft, rt); 
	    conforming = conforming || currentTrackConforms;
	}
	
	// Assertion that should be verified by symbolic execution.
	boolean correctBehaviour = (blundering != conforming);
	if (! correctBehaviour) {
		assert false;
	}
}
//INSTRUMENTATION END
            

            // If the flight is bludering, add it to the set of blunders
            // and assign a dead reckoning trajectory as its assigned trajectory            
            if (blundering) {
                blunders.add(flight);

                Trajectory drTraj = trajSynth.getDeadReckoningTrajectory(ft);
                flight2TrajMap.put(flight, drTraj);
            }
            // If the flight is conforming, synthesize a route trajectory for it,
            // assuming it's current track is its route track
            else {
                Trajectory rtTraj = trajSynth.getRouteTrajectory(rt, fp.getRoute());
                flight2TrajMap.put(flight, rtTraj);
            }
        }
    }
}
