/*
 TSAFE Prototype: A decision support tool for air traffic controllers
 Copyright (C) 2003  Gregory D. Dennis

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package tsafe.engine;

import java.util.LinkedList;
import java.util.ListIterator;

import tsafe.SymbolicExecutionSuccess;
import tsafe.data.Flight;
import tsafe.data.FlightPlan;
import tsafe.data.FlightTrack;
import tsafe.data.Trajectory;
import tsafe.engine.EngineParameters;
import tsafe.engine.RouteTracker;
import tsafe.engine.ConformanceMonitor;

/**
 * Handles the control flow between the algorithmic units of TSAFE
 */
public class TsafeEngine extends Thread {

	/**
	 * Engine parameters
	 */
	private EngineParameters parameters;

	/**
	 * Engine components
	 */
	private RouteTracker routeTracker;
	private ConformanceMonitor confMonitor;
	private TrajectorySynthesizer trajSynth;

	/**
	 * flight over which algorithms calculate
	 */
	private Flight flight;

	/**
	 * Private engine constructor
	 */
	public TsafeEngine(EngineCalculator calc) {
		this.parameters    = new EngineParameters();
		this.routeTracker  = new RouteTracker(calc);
		this.confMonitor   = new ConformanceMonitor(calc, parameters);
		this.trajSynth     = new TrajectorySynthesizer(calc, parameters);
	}

	/****************************************************************************
	 * Runs the TSAFE Engine
	 * Performs Conformance Monitoring and Trajectory Synthesis
	 ****************************************************************************/
	Trajectory flight2Traj;
	public void start() {
		// flight:
		// 1) If it has no flight plan, assign it a dr traj and continue
		// 2) If it has a flight plan, determine if it is blundering
		// 3) If it is, assign its dr trajectory as its predicted trajectory
		//    If it isn't, assign its route trajectory as its predicted trajectory

		FlightTrack ft = flight.getFlightTrack();
		FlightPlan fp = flight.getFlightPlan();

		// If the flight doesn't have a flight plan, assign it a dr trajectory
		if (fp == null) {
			Trajectory drTraj = trajSynth.getDeadReckoningTrajectory(ft);
			flight2Traj = drTraj;
		} else {
			// Determine if flight is blundering by comparing its actual track to its route track
			RouteTrack rt = routeTracker.findRouteTrack(ft, fp);
			boolean blundering = confMonitor.isBlundering(ft, rt);

			//INSTRUMENTATION BEGIN
			{
				// Check whether any segment of the plan conforms with current track
				boolean conforming = false;
				LinkedList segments = routeTracker.getSegmentsTracks(ft, fp);
				ListIterator segIter = segments.listIterator();
				if (segIter.hasNext()) {  //bounded iteration with bound 1
					rt = (RouteTrack)segIter.next();
					boolean currentTrackConforms = ! confMonitor.isBlundering(ft, rt); 
					conforming = conforming || currentTrackConforms;
					//assume (!segIter.hasNext())
					if (segIter.hasNext())
						throw new SymbolicExecutionSuccess();
				}
				boolean correctBehaviour = (blundering != conforming);
				if (! correctBehaviour) {
					assert false;
				}
			}
			//INSTRUMENTATION END


			// If the flight is blundering, add it to the set of blunders
			// and assign a dead reckoning trajectory as its assigned trajectory            
			if (blundering) {
				flight2Traj = trajSynth.getDeadReckoningTrajectory(ft);
			}
			// If the flight is conforming, synthesize a route trajectory for it,
			// assuming it's current track is its route track
			else {
				flight2Traj = trajSynth.getDeadReckoningTrajectory(ft);
			}
		}
	}
}
