package muse.algo;

import muse.bc.ClassHierarchy;
import muse.mem.CodeOutOfIndexException;
import muse.mem.Frame;
import muse.mem.Heap;
import muse.mem.Objekt;
import muse.mem.Reference;
import muse.mem.Simplex;
import muse.mem.State;

/**
 * Performs cast check for checkcast and instanceof bytecodes, and 
 * updates a state according to the corresponding semantics.
 * 
 * @author pietro
 */
public class CastChecker {
    private State runningState;
    private Reference tmpValue;
    
    /**
     * Constructor.
     * 
     * @param env the <code>SEEnv</code> of the symbolic execution.
     * @param runningState the current <code>State</code>.
     */
    public CastChecker(State runningState) {
    	this.runningState = runningState;
    }
    
    public void checkCheckcast() {
		checkType(
				new PostCheckStrategy() {
					public void doPostCheck(boolean isSubclass) {
						Frame frame = CastChecker.this.runningState.getStack().currentFrame();
						
				        if (!isSubclass) {
				        	Reference myExceRef = Util.loadInstance(runningState, "java/lang/ClassCastException");
				    		frame.push(myExceRef);
				    		CastChecker.this.runningState.doThrow(myExceRef);
				        	return;
				        }
				        
				        //pushes back the value on the operand stack
			            frame.push(tmpValue);
				        
				        try {
							CastChecker.this.runningState.incPC(3);
						} catch (CodeOutOfIndexException e) {
							Reference myExceRef = Util.loadInstance(CastChecker.this.runningState, "java/lang/VerifyError");
							frame.push(myExceRef);
							CastChecker.this.runningState.doThrow(myExceRef);
						}
					}
				});
	}
	
	public void checkInstanceof() {
		checkType(
				new PostCheckStrategy() {
					public void doPostCheck(boolean isSubclass) {
						Frame frame = CastChecker.this.runningState.getStack().currentFrame();
				        if (isSubclass) { 
				        	frame.push(new Simplex(1));
				        } else { 
							frame.push(new Simplex(0));
				        }
				        
				        try {
							CastChecker.this.runningState.incPC(3);
						} catch (CodeOutOfIndexException e) {
							Reference myExceRef = Util.loadInstance(CastChecker.this.runningState, "java/lang/VerifyError");
							frame.push(myExceRef);
							CastChecker.this.runningState.doThrow(myExceRef);
						}
					}
				});
	}
	
	private interface PostCheckStrategy {
		void doPostCheck(boolean isSubclass);
	}
	
	private void checkType(PostCheckStrategy c) {
        Frame frame = this.runningState.getStack().currentFrame();
        Heap heap = this.runningState.getHeap();
        int index;
        try {
	        byte tmp1 = this.runningState.getInstruction(1);
	        byte tmp2 = this.runningState.getInstruction(2);
	        index = Util.byteCat(tmp1,tmp2);
        } catch (CodeOutOfIndexException e) {
			Reference myExceRef = Util.loadInstance(this.runningState, "java/lang/VerifyError");
			frame.push(myExceRef);
			this.runningState.doThrow(myExceRef);
			return;
		}

        //gets in the current class constant pool the name 
        //of the class
        String classToOpen = frame.getName().getClassName();	
        String classT = this.runningState.getClassFileInterface().getClassFile(classToOpen).getClassSignature(index);

        //gets from the operand stack the reference to the 
        //object to be checked
        this.tmpValue = (Reference) frame.pop();
        
        //checks whether the object's class is a subclass 
        //of the class name from the constant pool
        boolean isSubclass;
        if (tmpValue.isNull()) {
        	isSubclass = true;  //the null value belongs to all classes
        } else {
	        Objekt objS = heap.getObject(tmpValue);
	        String classS = objS.getType();
	        ClassHierarchy hier = new ClassHierarchy(this.runningState.getClassFileInterface());
	        isSubclass = hier.isSubclass(classS, classT);
        }
        
        //performs the post check state updates
        c.doPostCheck(isSubclass);
	}
}
