package muse.algo;

import muse.bc.AttributeNotFoundException;
import muse.bc.ClassFile;
import muse.bc.FieldNotFoundException;
import muse.bc.Signature;
import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;
import muse.mem.ConstantPoolString;
import muse.mem.Klass;
import muse.mem.Objekt;
import muse.mem.Reference;
import muse.mem.State;
import muse.mem.Value;

/**
 * Strategy for initializing the static members of a given class.
 * 
 * @author pietro
 */
public class ClassInitializer {
	/**
	 * Loads a symbolic <code>Klass</code> object in the static area 
	 * of a <code>State</code>.
	 * 
	 * @param s the current <code>State</code>.
	 * @param className the name of the class to be loaded.
	 * @return the created <code>Klass</code>.
	 * @throws DecisionException 
	 */
	public static Klass loadSymbolicKlass(State s, String className, DecisionProcedure mgr) throws DecisionException {
		ClassFile classInterface = s.getClassFileInterface().getClassFile(className);
		Signature[] sgnFields = classInterface.getFieldsStatic();
		Klass k = new Klass(sgnFields, "[" + className + "]");
		k.setCreationEpoch(Objekt.Epoch.EPOCH_BEFORE_START);
		Util.initWithSymbolicValues(s, k);
		initConstantFields(s, className, k, mgr);
		return k;
	}
	
	/**
	 * Initializes a class by creating its <code>Klass</code> 
	 * and running its class initializer <code>&lt;clinit&gt;</code>. 
	 * If the class has already been initialized, then it does nothing.
	 * 
	 * @param s the current <code>State</code>.
	 * @param className the class to be initialized.
	 * @param mgr 
	 * @throws DecisionException 
	 */
	public static void initializeClass(State s, String className, DecisionProcedure mgr) throws DecisionException {
		createdFrames = 0;
		initializeClassRecursive(s, className, mgr);
	}

	/**
	 * 
	 */
	private static int createdFrames;
	
	/**
	 * Implements <code>initializeClass</code> by recursion.
	 * 
	 * @param s the current <code>State</code>.
	 * @param className the class to be initialized.
	 * @param mgr 
	 * @return <code>true</code> iff the initialization of 
	 *         <code>className</code> or of one of its superclasses 
	 *         fails for some reason.
	 * @throws DecisionException 
	 */
	private static boolean initializeClassRecursive(State s, String className, DecisionProcedure mgr) throws DecisionException {
		//if className is already initialized, then does nothing;
		//note that, if the class has been initialized before 
		//execution start, its Klass is not materialized now (this 
		//eases implementation, and the Klass will be materialized 
		//upon first access anyways)
		if (s.initialized(className)) {
			return false;
		}

		//creates the Klass and puts it in the method area
		ClassFile classInterface = s.getClassFileInterface().getClassFile(className);
		Signature[] sgnFields = classInterface.getFieldsStatic();
		Klass k = new Klass(sgnFields, "[" + className + "]");
		k.setCreationEpoch(Objekt.Epoch.EPOCH_AFTER_START);
		s.setKlass(k, className);
		
		//initializes the Klass'constant fields
		initConstantFields(s, className, k, mgr);
		
		//if className denotes a class rather than an interface, then 
		//recursively initializes its superclass if necessary
		boolean failed = false;
		if (!classInterface.isInterface()) {
			String superName = classInterface.getSuperClassName();
			failed = initializeClassRecursive(s, superName, mgr);
		}
		
		//if the initialization of the superclass didn't succeed, aborts 
		//initialization
		if (failed) {
			return true;
		}
		
		//in the case a <clinit> method exists, loads a Frame for it
		failed = false;
		String myExce = null;
		try {
			Signature sigClinit = new Signature(className, "()V", "<clinit>");
			if (classInterface.hasMethodDeclaration(sigClinit)) {
				Util.loadFrameStatic(s, sigClinit, null, 0);
				createdFrames++;
			}
		/* TODO Here I am in doubt about how I should manage exceptional
		 * situations. The JVM spec v2 (4.6, access_flags field discussion)
		 * states that the access flags of <clinit> should be ignored except for 
		 * strictfp. But it also says that if a method is either native 
		 * or abstract (from its access_flags field) it must have no code.
		 * What if a <clinit> is marked to be abstract or native? In such 
		 * case it should have no code. However, this shall not happen for 
		 * <clinit> methods. I will assume that in this case a verification 
		 * error should be raised.
		 */
		} catch (JavaReifyException e) {
			/* TODO By now, I will just reify the exception, but I should 
			 * analyze what to do in each case. Possible exceptions to reify
			 * (at least, the ones I detected until now):
			 * "java/lang/IncompatibleClassChangeError": determined if the 
			 *      method is not marked static; perhaps should plainly 
			 *      ignore as soon as there is code.
			 * "java/lang/VerifyError": determined if, when the pc or return 
			 *      address is set (to 0), a code out of range exception is 
			 *      raised - a really hairy case, which perhaps never happens.
			 */
			failed = true;
	    	myExce = e.subject();
		} catch (PleaseDoNativeException e) {
			failed = true;
	    	myExce = "java/lang/VerifyError";
		} finally {
			if (failed) {
				//pops all the frames created by the recursive calls
				for (int i = 1; i <= createdFrames; i++) {
					s.getStack().pop();
				}
		    	Reference myExceRef = Util.loadInstance(s, myExce);
				s.getStack().currentFrame().push(myExceRef);
		    	s.doThrow(myExceRef);
				return true; //returns failure
			}
		}
		
		//returns success
		return false;
	}
	
	/**
	 * Initializes the constant fields of a <code>Klass</code> with the 
	 * constant values specified in the classfile, if present.
	 * This is not necessary, as <code>SEGetstatic</code> must 
	 * check static final fields to prevent initialization, so 
	 * constant fields are not get from the <code>Klass</code>. 
	 * However, this keeps the <code>Klass</code> coherent with its 
	 * definition.
	 * 
	 * @param s the current <code>State</code>.
	 * @param className the class to be initialized.
	 * @param k the corresponding <code>Klass</code> object.
	 * @throws DecisionException 
	 */
	private static void initConstantFields(State s, String className, Klass k, DecisionProcedure mgr) throws DecisionException {
		ClassFile cf = s.getClassFileInterface().getClassFile(className);
		Signature[] flds = cf.getFieldsStatic();
		for (Signature sig : flds) {
			try {
				if (cf.isFieldFinal(sig) && cf.hasFieldConstantValue(sig)) {
					Value v;
					v = cf.fieldConstantValue(sig);
					if (v instanceof ConstantPoolString) {
						v = Util.loadStringLiteral(s, v.toString(), mgr);
					}
					k.setFieldValue(sig, v);
					//signatures are extracted directly from the classfile, 
					//so no resolution is necessary for using sig
				}
			} catch (FieldNotFoundException e) {
				//does nothing, this should never happen
			} catch (AttributeNotFoundException e) {
				//does nothing, this should never happen
			}
		}
	}
}