package muse.algo;

import muse.bc.Signature;
import muse.mem.CodeOutOfIndexException;
import muse.mem.Frame;
import muse.mem.FunctionApplication;
import muse.mem.Primitive;
import muse.mem.Reference;
import muse.mem.StackLastFrameException;
import muse.mem.State;
import muse.mem.Type;
import muse.mem.Value;
import muse.mem.ValueDoesNotSupportNativeException;

/**
 * Implements native method invocation by assuming that the invoked 
 * method is pure, i.e., it does not produce any effect other than 
 * returning a value on the operand stack according to the method's 
 * signature. More precisely:
 * <ul>
 * <li>If the method's return type is <code>void</code>, then the 
 * method invocation has no effect;</li>
 * <li>If the method's return type is primitive, and all its parameters 
 * have primitive type, then a <code>SEFunctionApplication</code> value 
 * reflecting the method's invocation is pushed on the operand stack; in 
 * the case the type of any parameter is not, a TODO 
 * exception is raised;</li>
 * <li>If the method's return type is a reference one (instance or array), 
 * then a fresh <code>SESymbolicReference</code> is pushed on the operand 
 * stack.</li>  
 * </ul>
 * 
 * @author pietro
 */
public class NativeInvokerPure implements NativeInvoker {
	private int pcOffset;
	
	public NativeInvokerPure(int pcOffset) {
		this.pcOffset = pcOffset;
	}

	public void doInvokeNative(State state, Signature methodSignatureResolved, Value[] args) throws ValueDoesNotSupportNativeException {		
		//determines the return value
		String returnType = Type.splitReturnValueDescriptor(methodSignatureResolved.getDescriptor());
		Value retVal;
		if (Type.isVoid(returnType)) {
			retVal = null;
		} else if (Type.isPrimitive(returnType)) {
			//requires all arguments are primitive
			Primitive[] argsPrim = new Primitive[args.length];
			for (int i = 0; i < args.length; i++) {
				if (args[i] instanceof Primitive) {
					argsPrim[i] = (Primitive) args[i];
				} else { 
					throw new ValueDoesNotSupportNativeException();
				}
			}
			retVal = new FunctionApplication(returnType, methodSignatureResolved.getName(), argsPrim);
		} else {
			retVal = state.createSymbol(returnType, "__NATIVE[" + state.getIdentifier() + "[" + state.getSequenceNumber() + "]");
		}
		
		//pushes the return value (if present) on the operand stack, 
		//or sets the state to stuck if no current frame exists
		try {
			Frame currentFrame = state.getStack().currentFrame();
			if (retVal != null) {
				currentFrame.push(retVal);
			}
		} catch (StackLastFrameException e) {
			state.setStuckReturn(retVal);
			return;
		}		

		//increments the program counter
		try {
			state.incPC(this.pcOffset);
		} catch (CodeOutOfIndexException e) {
			Reference myExceRef = Util.loadInstance(state, "java/lang/VerifyError");
			state.getStack().currentFrame().push(myExceRef);
			state.doThrow(myExceRef);
		}
	}
}
