package muse.algo;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import muse.bc.Signature;
import muse.mem.CodeOutOfIndexException;
import muse.mem.Frame;
import muse.mem.Reference;
import muse.mem.Simplex;
import muse.mem.StackLastFrameException;
import muse.mem.State;
import muse.mem.Value;


/**
 * Implements native method invocation by reifying the state 
 * of the interpreter, invoking an implementation of the method 
 * which must be available through Java reflection of the 
 * meta-level virtual machine, and reflecting back the resulting 
 * output. As a precondition, all the 
 * 
 * @author pietro
 *
 */
public class NativeInvokerReflect implements NativeInvoker {
	private int pcOffset;
	
	public NativeInvokerReflect(int pcOffset) {
		this.pcOffset = pcOffset;
	}
	
	public void doInvokeNative(State state,
			Signature methodSignatureResolved, Value[] args)
			throws CannotInvokeNativeException {
		try {
			//converts the arguments
			Object[] argsRefl = new Object[args.length];
			Class<?>[] argsClass = new Class[args.length];
			for (int i = 0; i < args.length; i++) {
				argsRefl[i] = args[i].getValueForNative();
				argsClass[i] = argsRefl[i].getClass();
			}
			
			//gets the method and invokes it
			Class<?> c = Class.forName(methodSignatureResolved.getClassName());
			Method m = c.getMethod(methodSignatureResolved.getName(), argsClass);
			Object retValRefl;
			if (Modifier.isStatic(m.getModifiers())) {
				retValRefl = m.invoke(null, argsRefl);
			} else {
				Object argThis = argsRefl[0]; //TODO reify argsRefl[0
				Object[] argsReflOthers = new Object[argsRefl.length - 1];
				for (int i = 0; i < argsReflOthers.length; i++) {
					argsReflOthers[i] = argsRefl[i + 1];
				}
				retValRefl = m.invoke(argThis, argsReflOthers);
			}
			
			//reifies the return value
			Value retVal;
			if (retValRefl.getClass().getName() == "byte") {
				retVal = new Simplex(((Byte) retValRefl).byteValue());
			} else if (retValRefl.getClass().getName() == "short") {
				retVal = new Simplex(((Short) retValRefl).shortValue());
			} else if (retValRefl.getClass().getName() == "int") {
				retVal = new Simplex(((Integer) retValRefl).intValue());
			} else if (retValRefl.getClass().getName() == "long") {
				retVal = new Simplex(((Long) retValRefl).longValue());
			} else if (retValRefl.getClass().getName() == "float") {
				retVal = new Simplex(((Float) retValRefl).floatValue());
			} else if (retValRefl.getClass().getName() == "double") {
				retVal = new Simplex(((Double) retValRefl).doubleValue());
			} else if (retValRefl.getClass().getName() == "char") {
				retVal = new Simplex(((Character) retValRefl).charValue());
			} else if (retValRefl.getClass().getName() == "boolean") {
				retVal = new Simplex(((Boolean) retValRefl).booleanValue());
			} else { 
				//TODO do something more relevant!
				return;
			}
			
			//pushes the return value on the operand stack
			try {
				Frame currentFrame = state.getStack().currentFrame();
				currentFrame.push(retVal);
			} catch (StackLastFrameException e) {
				state.setStuckReturn(retVal);
			}
		} catch (ClassNotFoundException e) {
			//TODO invent some relevant exception
			throw new CannotInvokeNativeException();
		} catch (SecurityException e) {
			//TODO invent some relevant exception
			throw new CannotInvokeNativeException();
		} catch (NoSuchMethodException e) {
			//TODO invent some relevant exception
			throw new CannotInvokeNativeException();
		} catch (IllegalArgumentException e) {
			//TODO invent some relevant exception
			throw new CannotInvokeNativeException();
		} catch (IllegalAccessException e) {
			//TODO invent some relevant exception
			throw new CannotInvokeNativeException();
		} catch (InvocationTargetException e) {
			//TODO invent some relevant exception
			throw new CannotInvokeNativeException();
		}

		//increments the program counter
		try {
			state.incPC(this.pcOffset);
		} catch (CodeOutOfIndexException e) {
			Reference myExceRef = Util.loadInstance(state, "java/lang/VerifyError");
			state.getStack().currentFrame().push(myExceRef);
			state.doThrow(myExceRef);
		}
	}
}
