package muse.algo;


import java.util.SortedSet;

import muse.dec.DecisionException;
import muse.mem.*;
import muse.tree.StateTree;

/**
 * Class for completing the semantics of bytecodes which require 
 * refinement of symbolic values.
 * 
 * @author Pietro Braione
 * @author unknown
 */
abstract class StateGenerator {
	protected final State runningState;
	protected final StateTree stateTree;

	/**
	 * Strategy for deciding the states following the current one.
	 * 
	 * @author Pietro Braione
	 * 
	 * @author pietro
	 *
	 * @param <R> the type of the possible outcomes of the decision.
	 */
	protected interface DecisionStrategy<R> {
		/**
		 * Decides the states following the current one.
		 * 
		 * @param result a (possibly empty) <code>SortedSet&lt;R&gt;</code>, 
		 *        which the method will update by storing in it an 
		 *        instance of <code>R</code> for each possible outcome of 
		 *        the decision. 
		 * @return <code>true</code> iff some decision outcomes 
		 *         require refinement of the produced state. 
		 * @throws DecisionException
		 * @throws CannotManageStateException
		 */
		public boolean decide(SortedSet<R> result) throws DecisionException, CannotManageStateException;
	}

	/**
	 * Strategy for refining a state.
	 * 
	 * @author pietro
	 *
	 * @param <R> the type of the possible outcomes of a previous decision.
	 */
	protected interface StateRefinementStrategy<R> {
		/**
		 * Refines a state.
		 * 
		 * @param s the <code>State</code> to be refined.
		 * @param r the outcome of a previous decision, which establishes 
		 *          the criterion on which the state shall be refined.
		 * @throws DecisionException
		 * @throws CannotManageStateException
		 * @throws JavaReifyException 
		 */
		public void refine(State s, R r) throws DecisionException, CannotManageStateException;
	}
	
	/**
	 * Strategy for updating a state, i.e., for completing its 
	 * bytecode semantics after a refinement.
	 * 
	 * @author pietro
	 *
	 * @param <R> the type of the possible outcome of a previous decision.
	 */
	protected interface StateUpdateStrategy<R> {
		/**
		 * Updates a state, i.e., completes its current bytecode's 
		 * semantics after a refinement.
		 * 
		 * @param s the <code>State</code> to be updated.
		 * @param r the outcome of a previous decision, which establishes 
		 *          the criterion on which the state shall be updated.
		 * @throws DecisionException
		 * @throws CannotManageStateException
		 */
		public void update(State s, R r) throws DecisionException, CannotManageStateException;
	}
	
	protected StateGenerator(State runningState, StateTree stateTree) {
		this.runningState = runningState;
		this.stateTree = stateTree;
	}
	
	/**
	 * Completes the bytecode semantics of a state by performing the following 
	 * actions:
	 * 
	 * <ul>
	 * <li>Performing a decision procedures and gathering its possible outcomes;</li>
	 * <li>Creating one state for each outcome, and perform on each state a refinement 
	 * procedure, if it is the case, to account for the outcome;</li>
	 * <li>Complete the bytecode semantics for each state, and add all the states 
	 * to the generator's <code>Initializer</code>, creating a backtrack point iff 
	 * more than one state is added.</li>
	 * </ul>
	 * 
	 * @param <R> the type of the possible outcomes of a decision.
	 * @param empty a <code>SortedSet&lt;R&gt;</code> which provides the 
	 *              assumed ordering for &lt;R&gt;'s and the default 
	 *              outcomes (usually none). After the invocation, 
	 *              <code>empty</code> contains the outcomes of the 
	 *              application of <code>aa</code>, i.e., as from 
	 *              <code>aa.decide(empty)</code>.
	 * @param aa a <code>DecisionStrategy&lt;R&gt;</code>, the decision 
	 *           procedure which is applied to calculate which are the 
	 *           next states.
	 * @param cc a <code>StateRefinementStrategy&lt;R&gt;</code>, the 
	 *           procedure which is applied to refine each of the produced 
	 *           states.
	 * @param bb a <code>StateUpdateStrategy&lt;R&gt;</code>, the
	 *           procedure which is applied to each state to complete the 
	 *           bytecode semantics. 
	 * @throws DecisionException
	 * @throws CannotManageStateException
	 * @throws JavaReifyException 
	 */
	protected <R> void generateStates(SortedSet<R> empty, DecisionStrategy<R> aa, StateRefinementStrategy<R> cc, StateUpdateStrategy<R> bb) throws DecisionException, CannotManageStateException {
		//decides the expansion of the symbolic values
		SortedSet<R> compResult = empty; 
		boolean shouldRefine = aa.decide(compResult);

		//adds a branch to the tree if the expansion yields
		//more than one outcome
		int cur = 1, tot = compResult.size();
		if (tot > 1) {
			this.stateTree.addBranch();
		}

		//generates the next states
		for (R r : compResult) {
			State s = (cur < tot ? (State) this.runningState.clone() : this.runningState);

			//possibly refines the state
			if (shouldRefine) {
				cc.refine(s, r);
			} else {
				s.resetLastPathConditionClause();
			}
			
			//completes the bytecode semantics
			bb.update(s, r);

            //adds the created states to the tree
			if (tot > 1) {
				this.stateTree.addState(s);
			}
			
			cur++;
		}
	}
}