package muse.algo;

import java.util.List;
import java.util.Set;
import java.util.SortedSet;

import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;
import muse.mem.Array;
import muse.mem.ArrayConstraint;
import muse.mem.CodeOutOfIndexException;
import muse.mem.Expression;
import muse.mem.Frame;
import muse.mem.Heap;
import muse.mem.Objekt;
import muse.mem.Primitive;
import muse.mem.Reference;
import muse.mem.ReferenceConcrete;
import muse.mem.ReferenceSymbolic;
import muse.mem.Simplex;
import muse.mem.State;
import muse.mem.Type;
import muse.mem.Value;
import muse.mem.Objekt.Epoch;
import muse.tree.DecisionAlternativeAload;
import muse.tree.DecisionAlternativeAloadArrayConstraint;
import muse.tree.DecisionAlternativeAloadOutbound;
import muse.tree.DecisionAlternativeAloadReference;
import muse.tree.DecisionAlternativeAloadReferenceClass;
import muse.tree.DecisionAlternativeAloadReferenceInstance;
import muse.tree.DecisionAlternativeAloadReferenceNull;
import muse.tree.DecisionAlternativeAloadValue;
import muse.tree.DecisionAlternativeAloadVisitor;
import muse.tree.DecisionAlternativeAnew;
import muse.tree.DecisionAlternativeAnewInbound;
import muse.tree.DecisionAlternativeAnewOutbound;
import muse.tree.DecisionAlternativeAnewVisitor;
import muse.tree.DecisionAlternativeAstore;
import muse.tree.StateTree;


/**
 * Class for completing the semantics of the *aload and *astore 
 * bytecodes.
 * 
 * @author pietro
 */
class StateGeneratorArray extends StateGenerator {
	private static final Simplex ZERO = new Simplex(0);
	private final DecisionProcedure mgr;
	private boolean partialReferenceResolution;
	private String notInitializedRefType;

	public StateGeneratorArray(State runningState, StateTree initializer, DecisionProcedure mgr) {
		super(runningState, initializer);
		this.mgr = mgr;
		this.partialReferenceResolution = false;
	}

	/**
	 * DecisionAlternativeAnewVisitor for update procedures; 
	 * reduces the exceptions thrown by the visitor methods. 
	 * 
	 * @author pietro
	 */
	private interface DecisionAlternativeAnewVisitorUpdate extends DecisionAlternativeAnewVisitor {
		void visitDecisionAlternativeAnewInbound(DecisionAlternativeAnewInbound dai) throws DecisionException, CannotManageStateException;
		void visitDecisionAlternativeAnewOutbound(DecisionAlternativeAnewOutbound dao);
	}

	public void generateStatesAnew(final String arrayType, final Primitive[] length, final int pcOffset) throws DecisionException, CannotManageStateException, PartialReferenceResolutionException {
		this.generateStates(this.mgr.getComparators().createEmptyDecisionAlternativeAnew(),  
		new DecisionStrategy<DecisionAlternativeAnew>() {
			public boolean decide(SortedSet<DecisionAlternativeAnew> result) throws DecisionException, CannotManageStateException {
				//builds the Primitive clause stating that all 
				//the dimension lengths are nonnegative, and
				//calculates the number of layers that can be 
				//directly created by the instruction, i.e.,
				//the number of concrete members before the first
				//symbolic member in length, plus one
				int layersToCreateNow = 0;
				Primitive lengthsNonNegative = null;
				{
					boolean first = true;
					boolean increment = true;
					for (Primitive l : length) {
						if (increment) {
							layersToCreateNow++;
						}
						increment = (l instanceof Simplex);

						if (first) {
							lengthsNonNegative = l.ge(ZERO);
							first = false;
						} else {
							lengthsNonNegative = lengthsNonNegative.and(l.ge(ZERO));
						}
					}
				}

				//determines whether the lengths can be (non)negative
				Expression nonNegative, negative;
				boolean canBeNonNegative, canBeNegative;
				if (lengthsNonNegative instanceof Simplex) {
					nonNegative = null;
					negative = null;
					canBeNonNegative = (Boolean) ((Simplex) lengthsNonNegative).getActualValue();
					canBeNegative = !canBeNonNegative;
				} else {
					nonNegative = (Expression) lengthsNonNegative; 
					negative = (Expression) nonNegative.not(); 
					canBeNonNegative = mgr.isSat(nonNegative);
					canBeNegative = mgr.isSat(negative);
				}
				if (canBeNonNegative) {
					result.add(new DecisionAlternativeAnewInbound(nonNegative, layersToCreateNow));
				}
				if (canBeNegative) {
					result.add(new DecisionAlternativeAnewOutbound(negative));
				}

				return (result.size() > 1);
			}
		},
		new StateRefinementStrategy<DecisionAlternativeAnew>() {
			public void refine(State s, DecisionAlternativeAnew r) throws DecisionException, CannotManageStateException {
				s.addClauseToPathCondition(r.exp);
			} 
		}, 
		new StateUpdateStrategy<DecisionAlternativeAnew>() {
			State s;
			Heap heap;
			Frame frame;
			
			DecisionAlternativeAnewVisitorUpdate visitorUpdate = 
				new DecisionAlternativeAnewVisitorUpdate() {
					public void visitDecisionAlternativeAnewInbound(DecisionAlternativeAnewInbound dai)
					throws DecisionException, CannotManageStateException {
						int layersToCreateNow = dai.toCreate;

						//calculates the number of dimensions as declared in arraySignature
						//and checks that the length parameter does not exceed this number
						int dimDecl;
						for (dimDecl = 0; dimDecl < arrayType.length() && arrayType.charAt(dimDecl) == Type.ARRAYOF; dimDecl++) {
							;
						}
						if (dimDecl < length.length) {
							Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
							frame.push(myExceRef);
							s.doThrow(myExceRef);
							return;
						}

						//determines the signature of the array member
						//String arrayMemberSignature = arrayType.substring(dimDecl);
						//TODO check that arrayMemberSignature refers to an existing class (when should this be done???)

						//all checks passed: creates the multidimensional array as layers of 
						//monodimensional arrays

						//builds the initialization value for the arrays in the last layer, i.e., 
						//the default array element value if the whole multidimensional array 
						//can be created now, otherwise a suitable ArrayConstraint allowing to 
						//lazily build the layers that cannot be built now when the array will
						//be accessed
						Value initValue = null;
						if (layersToCreateNow < length.length) {
							Primitive[] lConstr = new Primitive[length.length - layersToCreateNow];
							for (int k = layersToCreateNow; k < length.length; k++) {
								lConstr[k - layersToCreateNow] = length[k];
							}
							initValue = new ArrayConstraint(lConstr);
						}

						//the reference to be pushed on the operand stack at the end of the
						//creation; note that it is initialized to null, but this is just 
						//to make the compiler happy. It will be initialized during the loop, 
						//at the end of which it must be nonnull 
						ReferenceConcrete toPush = null;
						
						{
							//array of references to the Arrays created at the previous iteration
							Reference[] prev = null;

							//the number of arrays to be created in the current layer
							int toCreateInCurrentLayer = 1;

							//creates all the layers of monodimensional arrays that it can create now, 
							//initializes them, puts them in the heap, and sets toPush with a reference
							//to the topmost array
							for (int currentLayer = 0; currentLayer < layersToCreateNow; currentLayer++) {
								//caches the length of the arrays in the current layer, if it is concrete, 
								//and determines if it is (concretely or provably) 0 to early stop the 
								//creation process
								int lengthCurrentLayer;						
								boolean zeroBreak;
								if (length[currentLayer] instanceof Simplex) {
									lengthCurrentLayer = ((Integer) ((Simplex) length[currentLayer]).getActualValue());
									zeroBreak = (lengthCurrentLayer == 0); 
								} else {
									lengthCurrentLayer = 0;  //just to avoid compiler complaints, in such case this variable is not used
									Expression lengthCurrentLayerEqualToZero = (Expression) length[currentLayer].eq(ZERO); 
									Expression lengthCurrentLayerNotEqualToZero = (Expression) lengthCurrentLayerEqualToZero.not(); 
									zeroBreak = mgr.isSat(lengthCurrentLayerEqualToZero); 
									zeroBreak = zeroBreak && !mgr.isSat(lengthCurrentLayerNotEqualToZero);
								}

								//creates the i-th layer of arrays
								Reference[] next = new Reference[toCreateInCurrentLayer];
								for (int k = 0; k < toCreateInCurrentLayer; k++) {
									//creates the k-th array in the layer
									Array a = null;
									{
										String subarrayType = arrayType.substring(currentLayer);
										try {
											a = new Array(false, initValue, length[currentLayer], subarrayType, null);
										} catch (JavaReifyException e) {
											ReferenceConcrete myExceRef = Util.loadInstance(s, e.subject());
											frame.push(myExceRef);
											s.doThrow(myExceRef);
											return;
										}
									}

									//sets its epoch
									a.setCreationEpoch(Epoch.EPOCH_AFTER_START);

									//stores the array in the heap
									ReferenceConcrete ref = new ReferenceConcrete(heap.addNew(a));

									//stores the reference to the created array
									if (currentLayer == 0) { //topmost reference
										toPush = ref;
									} else {				
										//stores the reference in one of the arrays created 
										//at (i-1)-th iteration
										int prevArraySize = toCreateInCurrentLayer / prev.length; //size of arrays created at iteration i-1
										((Array) heap.getObject(prev[k /prevArraySize])).setElement(new Simplex(k % prevArraySize), ref, mgr);
									}

									//saves the created reference for initialization  
									//at (i+1)-th iteration
									next[k] = ref;

								}
								prev = next;
								if (length[currentLayer] instanceof Simplex) {
									toCreateInCurrentLayer *= lengthCurrentLayer;
								}

								//exits if the length of the arrays in the current layer is zero
								if (zeroBreak) {
									break;
								}
							}
						}

						//pushes the reference
						frame.push(toPush);

						//increments the program counter
						try {
							s.incPC(pcOffset);
						} catch (CodeOutOfIndexException e) {
							Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
							frame.push(myExceRef);
							s.doThrow(myExceRef);
						}
					}

					public void visitDecisionAlternativeAnewOutbound(DecisionAlternativeAnewOutbound dao) {
						Reference myExceRef = Util.loadInstance(s, "java/lang/NegativeArraySizeException");
						frame.push(myExceRef);
						s.doThrow(myExceRef);
					}				
			};
				
			public void update(State s, DecisionAlternativeAnew r) throws DecisionException, CannotManageStateException {
				this.s = s;
				this.heap = s.getHeap();
				this.frame = s.getStack().currentFrame();
				
				try {
					r.accept(this.visitorUpdate);
				} catch (DecisionException e) {
					throw e;
				} catch (CannotManageStateException e) {
					throw e;
				} catch (Throwable e) {
					//TODO unreachable code
				}
			} 
		});
		if (this.partialReferenceResolution) {
			throw new PartialReferenceResolutionException(this.notInitializedRefType);
		}
	}

	public void generateStatesAstore(final Reference myObjectRef, final Primitive index, final Value value) throws DecisionException, CannotManageStateException, PartialReferenceResolutionException {
		this.generateStates(this.mgr.getComparators().createEmptyDecisionAlternativeAstore(), 
		new DecisionStrategy<DecisionAlternativeAstore>() {
			public boolean decide(SortedSet<DecisionAlternativeAstore> result) throws DecisionException {
				Array arrayObj = (Array) runningState.getHeap().getObject(myObjectRef);
				if (arrayObj.mayBeInRange(index, mgr)) {
					result.add(DecisionAlternativeAstore.IN);
				}
				if (arrayObj.mayBeOutOfRange(index, mgr)) {
					result.add(DecisionAlternativeAstore.OUT);
				}
				return (result.size() > 1);
			} 
		}, 
		new StateRefinementStrategy<DecisionAlternativeAstore>() {
			public void refine(State s, DecisionAlternativeAstore r) {
				Array arrayObj = (Array) s.getHeap().getObject(myObjectRef);
				switch (r) {
				case IN:
					s.addClauseToPathCondition((Expression) arrayObj.inRange(index));
					break;
				case OUT:
					s.addClauseToPathCondition((Expression) arrayObj.outOfRange(index));
					break;
				}
				//note that, in the case refine is invoked, then both "index in range" 
				//and "index out of range" are satisfiable, thus both arrayObj.inRange(index)
				//and arrayObj.outOfRange(index) must return an Expression
			} 
		}, 
		new StateUpdateStrategy<DecisionAlternativeAstore>() {
			public void update(State s, DecisionAlternativeAstore r) throws DecisionException {
				switch (r) {
				case IN:
					Array arrayObj = (Array) s.getHeap().getObject(myObjectRef);
					arrayObj.setElement(index, value, StateGeneratorArray.this.mgr);
					try {
						s.incPC();
					} catch (CodeOutOfIndexException e) {
						Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
						s.getStack().currentFrame().push(myExceRef);
						s.doThrow(myExceRef);
					}
					break;
				case OUT:
					Reference myExceRef = Util.loadInstance(s, "java/lang/ArrayIndexOutOfBoundsException");
					s.getStack().currentFrame().push(myExceRef);
					s.doThrow(myExceRef);
					break;
				}
			} 
		}
		);
		if (this.partialReferenceResolution) {
			throw new PartialReferenceResolutionException(this.notInitializedRefType);
		}
	}

	/**
	 * DecisionAlternativeAloadVisitor for refinement procedures; 
	 * reduces the exceptions thrown by the visitor methods. 
	 * 
	 * @author pietro
	 */
	private interface DecisionAlternativeAloadVisitorRefine extends DecisionAlternativeAloadVisitor {
		void visitDecisionAlternativeAloadReferenceClass(DecisionAlternativeAloadReferenceClass dac) throws DecisionException, CannotManageStateException;
		void visitDecisionAlternativeAloadReferenceInstance(DecisionAlternativeAloadReferenceInstance dai) throws DecisionException;
		void visitDecisionAlternativeAloadReferenceNull(DecisionAlternativeAloadReferenceNull dan) throws DecisionException;
		void visitDecisionAlternativeAloadOutbound(DecisionAlternativeAloadOutbound dao);
		void visitDecisionAlternativeAloadValue(DecisionAlternativeAloadValue dav) throws DecisionException;
		void visitDecisionAlternativeAloadArrayConstraint(DecisionAlternativeAloadArrayConstraint daa);
	}
	
	/**
	 * DecisionAlternativeAloadVisitor for update procedures; 
	 * reduces the exceptions thrown by the visitor methods. 
	 * 
	 * @author pietro
	 */
	private interface DecisionAlternativeAloadVisitorUpdate extends DecisionAlternativeAloadVisitor {
		void visitDecisionAlternativeAloadReferenceClass(DecisionAlternativeAloadReferenceClass dac);
		void visitDecisionAlternativeAloadReferenceInstance(DecisionAlternativeAloadReferenceInstance dai);
		void visitDecisionAlternativeAloadReferenceNull(DecisionAlternativeAloadReferenceNull dan);
		void visitDecisionAlternativeAloadOutbound(DecisionAlternativeAloadOutbound dao);
		void visitDecisionAlternativeAloadValue(DecisionAlternativeAloadValue dav);
		void visitDecisionAlternativeAloadArrayConstraint(DecisionAlternativeAloadArrayConstraint daa) throws JavaReifyException, DecisionException;
	}
	
	public void generateStatesAload(final Reference myObjectRef, final Primitive index) throws DecisionException, CannotManageStateException, PartialReferenceResolutionException {
		this.generateStates(this.mgr.getComparators().createEmptyDecisionAlternativeAload(), 
		new DecisionStrategy<DecisionAlternativeAload>() {
			public boolean decide(SortedSet<DecisionAlternativeAload> result) throws DecisionException, CannotManageStateException {
				boolean retVal = false;
				Array arrayObj = (Array) runningState.getHeap().getObject(myObjectRef);
				Set<Array.ArrayEntry> entries = arrayObj.getElement(index, mgr);
				for (Array.ArrayEntry e : entries) {
					Expression exp = e.getExpression();
					boolean fresh = false;
					
					//takes the value of the current entry
					Value val;
					if (e instanceof Array.ArrayEntryValue) {
						val = ((Array.ArrayEntryValue) e).getValue();
						if (val == null) {
							fresh = true;
							val = runningState.createSymbol(arrayObj.getType().substring(1), arrayObj.getOrigin() + "[" + index + "]");
						}
					} else { //e instanceof Array.ArrayEntryOutOfBound
						val = null;
					}

					//production of a fresh target always requires a 
					//refinement action
					if (fresh) {
						retVal = true;
					}

					//builds result
					if (val == null) {
						result.add(new DecisionAlternativeAloadOutbound(exp));
					} else if (Type.isUninitializedReference(val)) {
						//TODO this (highly sensible) strategy code is common with StateGeneratorReference(State, Initializer, Simplifier, int).DecisionStrategy$0.decide(SortedSet<DecisionAlternativeReference>)!!!! Refactor!!!!
						ReferenceSymbolic referenceToRefine = (ReferenceSymbolic) val; 
						StateGeneratorArray.this.notInitializedRefType = referenceToRefine.getTypeNotInitialized();			        		
						String notInitializedRefOrigin = referenceToRefine.getOrigin();

						//adds all compatible objects in the state's heap
						List<Integer> compatibleObjectIndex = runningState.getObjectsCompatible(StateGeneratorArray.this.notInitializedRefType);
						for (int i : compatibleObjectIndex) {
							result.add(new DecisionAlternativeAloadReferenceInstance(exp, referenceToRefine, i));
						}

						//adds symbolic instances for compatible classes
						Set<String> compatibleClassNames = runningState.getTypesCompatible(StateGeneratorArray.this.notInitializedRefType, notInitializedRefOrigin);
						for (String s : compatibleClassNames) {
							result.add(new DecisionAlternativeAloadReferenceClass(exp, referenceToRefine, s));
						}
						StateGeneratorArray.this.partialReferenceResolution = compatibleClassNames.isEmpty();

						//adds null
						result.add(new DecisionAlternativeAloadReferenceNull(exp, referenceToRefine));

						//uninitialized symbolic references always require a refinement action
						retVal = true; 
					} else if (val instanceof ArrayConstraint) {
						result.add(new DecisionAlternativeAloadArrayConstraint(exp, (ArrayConstraint) val));
					} else {  //is a (concrete or symbolic initialized) reference, or a primitive
						result.add(new DecisionAlternativeAloadValue(exp, val, fresh));
					}
				}

				//also the size of the result matters to whether refine or not 
				retVal = retVal || (result.size() > 1);

				return retVal;
			}
		}, 
		new StateRefinementStrategy<DecisionAlternativeAload>() {
			State s;
			DecisionAlternativeAloadVisitorRefine visitorRefine = 
				new DecisionAlternativeAloadVisitorRefine() {
				public void visitDecisionAlternativeAloadOutbound(DecisionAlternativeAloadOutbound dao) {
					s.addClauseToPathCondition(dao.exp);
				}

				public void visitDecisionAlternativeAloadReferenceClass(DecisionAlternativeAloadReferenceClass dac)
				throws DecisionException, CannotManageStateException {
					//generates a symbolic instance and refines the reference with 
					//the position of the symbolic instance
					if (Type.isArray(dac.className)) {
						Util.loadSymbolicInstance(s, dac.className, dac.refinedRef, null);
					} else {
						Util.loadSymbolicArray(s, dac.className, dac.refinedRef, null, mgr);
					}

					//augments the path condition (a clone of the symbolic instance is 
					//created at the purpose) and updates the state's reference
					Objekt o = (Objekt) s.getHeap().getObject(dac.refinedRef).clone();
					s.addClauseToPathCondition((ReferenceSymbolic) dac.refinedRef.clone(), o);
					Array a = (Array) s.getHeap().getObject(myObjectRef);
					a.setElement(index, dac.refinedRef, mgr);
				}

				public void visitDecisionAlternativeAloadReferenceInstance(DecisionAlternativeAloadReferenceInstance dai)
				throws DecisionException {
					//generates the refined reference (setting it to the instance's position)
					dai.refinedRef.initialize(dai.instancePosition);

					//augments the path condition and updates the state's reference
					s.addClauseToPathCondition((ReferenceSymbolic) dai.refinedRef.clone(), null);
					Array a = (Array) s.getHeap().getObject(myObjectRef);
					a.setElement(index, dai.refinedRef, mgr);
				}

				public void visitDecisionAlternativeAloadReferenceNull(DecisionAlternativeAloadReferenceNull dan)
				throws DecisionException {
					//generates the refined reference (setting it to null)
					dan.refinedRef.initializeNull();

					//augments the path condition and updates the state's reference; 
					//note that the order is relevant: since the last added clause 
					//is sent to the simplifier, it must be dan.exp
					s.addClauseToPathCondition(dan.refinedRef.clone(), null);
					s.addClauseToPathCondition(dan.exp);
					Array a = (Array) s.getHeap().getObject(myObjectRef);
					a.setElement(index, dan.refinedRef, mgr);
				}

				public void visitDecisionAlternativeAloadValue(DecisionAlternativeAloadValue dav)
				throws DecisionException {
					s.addClauseToPathCondition(dav.exp);
					//if the value is fresh, then it writes it in the array
					if (dav.fresh) {
						Array a = (Array) s.getHeap().getObject(myObjectRef);
						a.setElement(index, dav.val, mgr);
					}
				}

				public void visitDecisionAlternativeAloadArrayConstraint(DecisionAlternativeAloadArrayConstraint daa) {
					s.addClauseToPathCondition(daa.exp);
				}
			};

			public void refine(State s, DecisionAlternativeAload r) throws DecisionException,CannotManageStateException{
				this.s = s;
				try {
					r.accept(this.visitorRefine);
				} catch (DecisionException e) {
					throw e;
				} catch (CannotManageStateException e) {
					throw e;
				} catch (Throwable e) {
					//TODO unreachable code
				}
			}
		}, 
		new StateUpdateStrategy<DecisionAlternativeAload>() {
			State s;

			DecisionAlternativeAloadVisitorUpdate visitorUpdate = 
				new DecisionAlternativeAloadVisitorUpdate() {
				public void visitDecisionAlternativeAloadOutbound(DecisionAlternativeAloadOutbound dao) {
					Reference myExceRef = Util.loadInstance(s, "java/lang/ArrayIndexOutOfBoundsException");
					s.getStack().currentFrame().push(myExceRef);
					s.doThrow(myExceRef);
				}

				private void visitDecisionAlternativeAloadReference(DecisionAlternativeAloadReference dar) {
					//pushes the refined reference on the 
					//operand stack
					Frame frame = s.getStack().currentFrame(); 
					frame.push(dar.refinedRef);

					//updates the program counter
					try {
						s.incPC(1);
					} catch (CodeOutOfIndexException e) {
						Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
						frame.push(myExceRef);
						s.doThrow(myExceRef);
					}
				}
				
				public void visitDecisionAlternativeAloadReferenceClass(DecisionAlternativeAloadReferenceClass dac) {
					this.visitDecisionAlternativeAloadReference(dac);
				}

				public void visitDecisionAlternativeAloadReferenceInstance(DecisionAlternativeAloadReferenceInstance dai) {
					this.visitDecisionAlternativeAloadReference(dai);
				}

				public void visitDecisionAlternativeAloadReferenceNull(DecisionAlternativeAloadReferenceNull dan) {
					this.visitDecisionAlternativeAloadReference(dan);
				}

				public void visitDecisionAlternativeAloadValue(DecisionAlternativeAloadValue dav) {
					//pushes the refined reference on the 
					//operand stack
					Frame frame = s.getStack().currentFrame(); 
					frame.push(dav.val);

					//updates the program counter
					try {
						s.incPC(1);
					} catch (CodeOutOfIndexException e) {
						Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
						frame.push(myExceRef);
						s.doThrow(myExceRef);
					}
				}

				public void visitDecisionAlternativeAloadArrayConstraint(DecisionAlternativeAloadArrayConstraint daa)
				throws JavaReifyException, DecisionException {
					//lazily creates the array from the constraint
					Array a = (Array) s.getHeap().getObject(myObjectRef);
					ReferenceConcrete ref = new ReferenceConcrete(s.getHeap().addNew(new Array(false, daa.constr.next(), daa.constr.getLength(), a.getType().substring(1), null)));
					a.setElement(index, ref, mgr);

					//pushes the reference to the fresh array on the 
					//operand stack
					Frame frame = s.getStack().currentFrame(); 
					frame.push(ref);

					//updates the program counter
					try {
						s.incPC(1);
					} catch (CodeOutOfIndexException e) {
						Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
						frame.push(myExceRef);
						s.doThrow(myExceRef);
					}
				}
			};

			public void update(State s, DecisionAlternativeAload r) throws DecisionException, CannotManageStateException {
				this.s = s;
				try {
					r.accept(this.visitorUpdate);
				} catch (Throwable e) {
					//TODO unreachable code
				}
			}
		}
		);
		if (this.partialReferenceResolution) {
			throw new PartialReferenceResolutionException(this.notInitializedRefType);
		}
	}
}
