package muse.algo;

import java.util.SortedSet;

import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;
import muse.mem.Any;
import muse.mem.CodeOutOfIndexException;
import muse.mem.Expression;
import muse.mem.Operator;
import muse.mem.Primitive;
import muse.mem.Reference;
import muse.mem.Simplex;
import muse.mem.State;
import muse.tree.DecisionAlternativeBoolean;
import muse.tree.DecisionAlternativeComparison;
import muse.tree.DecisionAlternativeSwitch;
import muse.tree.StateTree;


/**
 * Class for completing the semantics of the if*, *cmp* and *switch 
 * bytecodes, which require the refinement of a symbolic primitive.
 * 
 * @author pietro
 */
class StateGeneratorPrimitive extends StateGenerator {
	private final DecisionProcedure mgr;
	
	public StateGeneratorPrimitive(State runningState, StateTree initializer, DecisionProcedure mgr) {
		super(runningState, initializer);
		this.mgr = mgr;
	}
	
	public void generateStatesIfcond(final Primitive val1, final String type, final Primitive val2, final int branchOffset) throws DecisionException {
		try {
			this.generateStates(
					this.mgr.getComparators().createEmptyDecisionAlternativeBoolean(),
					new DecisionStrategy<DecisionAlternativeBoolean>() {
						public boolean decide(SortedSet<DecisionAlternativeBoolean> result) throws DecisionException {
							boolean shouldRefine;
							
							if ((val1 instanceof Any) || (val2 instanceof Any)) {
								//1 - condition involving "don't care" values
								result.add(DecisionAlternativeBoolean.TT);
								result.add(DecisionAlternativeBoolean.FF);
								shouldRefine = false; //"don't care" does not require refinement
							} else {
								Primitive condition = new Expression(val1, type, val2).normalize();
								if (condition instanceof Simplex) {
									//2 - condition not involving symbolic values
									boolean conditionValue = (Boolean) ((Simplex) condition).getActualValue();
									result.add(DecisionAlternativeBoolean.to(conditionValue));
								} else {
									//3 - condition involving symbolic values
									StateGeneratorPrimitive.this.mgr.simplifyDecision(new Expression(val1, type, val2), result);
								}
								shouldRefine = (result.size() > 1);
							}
							return shouldRefine;
						}
					}, 
					new StateRefinementStrategy<DecisionAlternativeBoolean>() {
						public void refine(State s, DecisionAlternativeBoolean r) {
							Expression condTrue; 
							if (r.value()) {
								condTrue = new Expression(val1, type, val2);
							} else {
								condTrue = (Expression) ((new Expression(val1, type, val2)).not());
							}
							s.addClauseToPathCondition(condTrue);
						}
					},
					new StateUpdateStrategy<DecisionAlternativeBoolean>() {
						public void update(State s, DecisionAlternativeBoolean r) {
							try {
								if (r.value()) {
									s.incPC(branchOffset);
								} else {
									s.incPC(3);
								}
							} catch (CodeOutOfIndexException e) {
								Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
								s.getStack().currentFrame().push(myExceRef);
								s.doThrow(myExceRef);
							}
						}
					}
			);
		} catch (CannotManageStateException e) {
			//this should never happen
		}
	}

	public void generateStatesNcmp(final Primitive val1, final Primitive val2) throws DecisionException {
		try {
			this.generateStates(
					this.mgr.getComparators().createEmptyDecisionAlternativeComparison(),
					new DecisionStrategy<DecisionAlternativeComparison>() {
						public boolean decide(SortedSet<DecisionAlternativeComparison> result) throws DecisionException {
							boolean shouldRefine;
							if ((val1 instanceof Any) || (val2 instanceof Any)) {
								//1 - condition involving "don't care" values
								result.add(DecisionAlternativeComparison.GT);
								result.add(DecisionAlternativeComparison.EQ);
								result.add(DecisionAlternativeComparison.LT);
								shouldRefine = false;
							} else {
								Primitive conditionEq = val1.eq(val2);
								Primitive conditionGt = val1.gt(val2);
								if ((conditionEq instanceof Simplex) && (conditionGt instanceof Simplex)) {
									//2 - condition not involving symbolic values
									boolean conditionEqValue = (Boolean) ((Simplex) conditionEq).getActualValue();
									boolean conditionGtValue = (Boolean) ((Simplex) conditionGt).getActualValue();
									if (conditionEqValue) {
										result.add(DecisionAlternativeComparison.EQ);
									} else if (conditionGtValue) {
										result.add(DecisionAlternativeComparison.GT);
									} else {
										result.add(DecisionAlternativeComparison.LT);
									}
								} else {
									//3 - condition involving symbolic values
									mgr.simplifyComparison(val1, val2, result);
								}
								shouldRefine = (result.size() > 1);
							}
					        return shouldRefine;
						}
					}, 
					new StateRefinementStrategy<DecisionAlternativeComparison>(){
						public void refine(State s, DecisionAlternativeComparison r) {
			            	s.addClauseToPathCondition((Expression) new Expression(val1, r.toOperator(), val2).normalize());
						}
					},
					new StateUpdateStrategy<DecisionAlternativeComparison>() {
						public void update(State s, DecisionAlternativeComparison r) {
				            s.getStack().currentFrame().push(new Simplex(r.value()));
							try {
								s.incPC();
							} catch (CodeOutOfIndexException e) {
								Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
								s.getStack().currentFrame().push(myExceRef);
								s.doThrow(myExceRef);
							}
						}
					} 
			);
		} catch (CannotManageStateException e) {
			//this should never happen
		}		
	}
	
	public void generateStatesSwitch(final Primitive op, final boolean isTableSwitch, final int deflt, final int low, final int high) throws DecisionException {
		final int entrySizeInBytes = (isTableSwitch ? 4 : 8);
		final int tableStart = this.runningState.getPC();
		final int tableEnd = tableStart + (high - low + 1) * entrySizeInBytes;
		final byte[] code = this.runningState.getCode();

		try {
			this.generateStates(
					this.mgr.getComparators().createEmptyDecisionAlternativeSwitch(), 
					new DecisionStrategy<DecisionAlternativeSwitch>() {
						public boolean decide(SortedSet<DecisionAlternativeSwitch> result) throws DecisionException {
							boolean shouldRefine;
							if (op instanceof Any) {
								//1 - condition involving "don't care" values
								if (isTableSwitch) {
									for (int i = low; i <= high; i++) {
										result.add(new DecisionAlternativeSwitch(i));
									}
								} else {
									byte[] t = new byte[entrySizeInBytes];
									for (int k = tableStart; k < tableEnd; k += entrySizeInBytes) {
										for (int i = 0; i < entrySizeInBytes; i++) {
											t[i] = code[k + i];
										}
										int match = Util.byteCat(t[0], t[1], t[2], t[3]);
										result.add(new DecisionAlternativeSwitch(match));
									}
								}
								result.add(DecisionAlternativeSwitch.alternativeDefault());
								shouldRefine = false;
							} else if (op instanceof Simplex) {
								//2 - condition not involving symbolic values
								int opValue = (Integer) ((Simplex) op).getActualValue();
								result.add(new DecisionAlternativeSwitch(opValue));
								shouldRefine = false;
							} else {
						       	//3 - condition involving symbolic values
								StateGeneratorPrimitive.this.mgr.simplifySwitch(op, isTableSwitch, low, high, tableStart, tableEnd, code, result);
					            shouldRefine = (result.size() > 1);
					        }
							return shouldRefine;
						}
					}, 
					new StateRefinementStrategy<DecisionAlternativeSwitch>() {
						public void refine(State s, DecisionAlternativeSwitch r) {
							int index = r.value();
							boolean valueOption = !r.isDefault();
							byte[] t = new byte[entrySizeInBytes];

							Expression branchCondition = null;
							if (valueOption) {
								branchCondition = (Expression) op.eq(new Simplex(index));
							} else if (isTableSwitch) {
								branchCondition = (Expression) op.le(new Simplex(low)).or(op.gt(new Simplex(high)));
							} else { //default and lookupswitch
								//scans all entries and adds a clause for each
								for (int k = tableStart; k < tableEnd; k += entrySizeInBytes) {
									for (int i = 0; i < entrySizeInBytes; i++) {
										t[i] = code[k + i];
									}
									int match = Util.byteCat(t[0], t[1], t[2], t[3]);
						            if (branchCondition == null) {
						            	branchCondition = new Expression(op, Operator.NE, new Simplex(match));
						            } else {
										branchCondition = new Expression(branchCondition, Operator.AND, new Expression(op, Operator.NE, new Simplex(match)));
						            }
								}
						    }
							s.addClauseToPathCondition(branchCondition);
						}
					},
					new StateUpdateStrategy<DecisionAlternativeSwitch>() {
						public void update(State s, DecisionAlternativeSwitch r) {
							int index = r.value();
							boolean def = r.isDefault();
							byte[] t = new byte[entrySizeInBytes];
					
							//calculates the jump offset
							int jumpOffset;
							if (def) {
								jumpOffset = deflt;
							} else if (isTableSwitch) {
								if (index >= low && index <= high) {
									for (int i = 0; i < entrySizeInBytes; i++) {
										t[i] = code[tableStart + (index - low) * entrySizeInBytes + i];
									}
									jumpOffset = Util.byteCat(t[0], t[1], t[2], t[3]);
								} else {
									jumpOffset = deflt;
								}
							} else { 
								//search of entry (unoptimized linear search)
								jumpOffset = deflt;
								for (int k = tableStart; k < tableEnd; k += entrySizeInBytes) {
									for (int i = 0; i < entrySizeInBytes; i++) {
										t[i] = code[k + i];
									}
									int match = Util.byteCat(t[0], t[1], t[2], t[3]);
									if (index == match) {
										jumpOffset = Util.byteCat(t[4], t[5], t[6], t[7]);
										break; //slight acceleration
									} else if (index < match) {
										break; //slight acceleration
									}
								}
							}
							
							//jumps
							try {
								s.incPC(jumpOffset);
							} catch (CodeOutOfIndexException e) {
								Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
								s.getStack().currentFrame().push(myExceRef);
								s.doThrow(myExceRef);
							}
						}
					} 
			);
		} catch (CannotManageStateException e) {
			//this should never happen
		}
	}	
}
