package muse.algo;

import java.util.List;
import java.util.Set;
import java.util.SortedSet;

import muse.bc.Signature;
import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;
import muse.mem.CodeOutOfIndexException;
import muse.mem.Frame;
import muse.mem.Instance;
import muse.mem.Klass;
import muse.mem.LocalVariablesArea;
import muse.mem.Objekt;
import muse.mem.Reference;
import muse.mem.ReferenceSymbolic;
import muse.mem.State;
import muse.mem.Type;
import muse.tree.DecisionAlternativeReference;
import muse.tree.DecisionAlternativeReferenceClass;
import muse.tree.DecisionAlternativeReferenceNull;
import muse.tree.DecisionAlternativeReferenceObject;
import muse.tree.DecisionAlternativeReferenceVisitor;
import muse.tree.StateTree;


/**
 * Class for completing the semantics of the getfield, getstatic and *load* 
 * bytecodes, which require the refinement of a symbolic reference 
 * ("lazy initialization").
 * 
 * @author Pietro Braione
 * @author unknown
 */
class StateGeneratorReference extends StateGenerator {
	/**
	 * Strategy for refining the source of the symbolic 
	 * reference subject to refinement. 
	 * 
	 * @author Pietro Braione
	 */
	private interface SourceReferenceRefinementStrategy {
		void refine(State s, ReferenceSymbolic newRef);
	}

	/**
	 * DecisionAlternativeReferenceVisitor for refinement procedures; 
	 * reduces the number exceptions thrown by the visitor methods. 
	 * 
	 * @author pietro
	 */
	private interface DecisionAlternativeReferenceVisitorRefine extends DecisionAlternativeReferenceVisitor {
		void visitDecisionAlternativeReferenceClass(DecisionAlternativeReferenceClass drc) throws DecisionException, CannotManageStateException;
		void visitDecisionAlternativeReferenceNull(DecisionAlternativeReferenceNull drn) throws DecisionException, CannotManageStateException;
		void visitDecisionAlternativeReferenceObject(DecisionAlternativeReferenceObject dro) throws DecisionException, CannotManageStateException;
	}
	
	private final DecisionProcedure mgr;
	private final DecisionStrategy<DecisionAlternativeReference> aa; 		//same for all
	private final StateRefinementStrategy<DecisionAlternativeReference> cc; //depends on DecisionAlternative
	private final StateUpdateStrategy<DecisionAlternativeReference> bb;     //depends on DecisionAlternative
	private SourceReferenceRefinementStrategy f;							//depends on bytecode
	private ReferenceSymbolic notInitializedRef;
	private boolean partialReferenceResolution;
	
	/**
	 * Constructor.
	 * 
	 * @param runningState
	 *        the <code>State</code> currently under execution.
	 * @param initializer
	 *        the <code>Initializer</code> where the created 
	 *        <code>State</code>s will be put.
	 * @param pcOffset
	 *        the offset of the next bytecode.
	 */
	public StateGeneratorReference(State runningState, StateTree initializer, final DecisionProcedure mgr, final int pcOffset) {
		super(runningState, initializer);
		this.mgr = mgr;
		this.aa =
			new DecisionStrategy<DecisionAlternativeReference>() {
			public boolean decide(SortedSet<DecisionAlternativeReference> result) {
				//TODO this (highly sensible) strategy code is common with SEStateGeneratorArray.generateStatesAload(SEReference, Primitive).DecisionStrategy$0.decide(SortedSet<DecisionAlternativeArray>)!!!! Refactor!!!!
				//adds all compatible objects in the state's heap
				List<Integer> compatibleObjectIndex = StateGeneratorReference.this.runningState.getObjectsCompatible(StateGeneratorReference.this.notInitializedRef.getTypeNotInitialized());
				for (int i : compatibleObjectIndex) {
					ReferenceSymbolic referenceToRefine = StateGeneratorReference.this.notInitializedRef.clone();
					DecisionAlternativeReferenceObject d = new DecisionAlternativeReferenceObject(referenceToRefine, i);
					result.add(d);
				}

				//adds symbolic instances for compatible classes
				Set<String> compatibleClassNames = StateGeneratorReference.this.runningState.getTypesCompatible(StateGeneratorReference.this.notInitializedRef.getTypeNotInitialized(), StateGeneratorReference.this.notInitializedRef.getOrigin());
				StateGeneratorReference.this.partialReferenceResolution = compatibleClassNames.isEmpty();
				for (String className : compatibleClassNames) {
					ReferenceSymbolic referenceToRefine = StateGeneratorReference.this.notInitializedRef.clone();
					DecisionAlternativeReferenceClass c = new DecisionAlternativeReferenceClass(referenceToRefine, className);
					result.add(c);
				}

				//adds null
				DecisionAlternativeReferenceNull n = new DecisionAlternativeReferenceNull(StateGeneratorReference.this.notInitializedRef.clone());
				result.add(n);

				//references always require a refinement action
				return true;
			}
		};
		this.cc = 
			new StateRefinementStrategy<DecisionAlternativeReference>() {
				State s;
				DecisionAlternativeReferenceVisitor visitorRefine = 
					new DecisionAlternativeReferenceVisitorRefine() {
					public void visitDecisionAlternativeReferenceClass(DecisionAlternativeReferenceClass drc) 
					throws DecisionException, CannotManageStateException {
						//generates a symbolic instance and refines the reference with 
						//the position of the symbolic instance
						if (Type.isArray(drc.objectName)) {
							Util.loadSymbolicArray(s, drc.objectName, drc.refinedRef, null, mgr);
						} else { //(is the name of a class)
							Util.loadSymbolicInstance(s, drc.objectName, drc.refinedRef, null);
						}

						//augments the path condition (a clone of the symbolic instance is 
						//created at the purpose) and updates the state's reference
						Objekt o = (Objekt) s.getHeap().getObject(drc.refinedRef).clone();
						s.addClauseToPathCondition((ReferenceSymbolic) drc.refinedRef.clone(), o);
						StateGeneratorReference.this.f.refine(s, drc.refinedRef);
					}

					public void visitDecisionAlternativeReferenceNull(DecisionAlternativeReferenceNull drn) 
					throws DecisionException, CannotManageStateException {
						//generates the refined reference (setting it to null)
						drn.refinedRef.initializeNull();

						//augments the path condition and updates the state's reference
						s.addClauseToPathCondition((ReferenceSymbolic) drn.refinedRef.clone(), null);
						StateGeneratorReference.this.f.refine(s, drn.refinedRef);
					}

					public void visitDecisionAlternativeReferenceObject(DecisionAlternativeReferenceObject dro) 
					throws DecisionException, CannotManageStateException {
						//generates the refined reference (setting it to the instance's position)
						dro.refinedRef.initialize(dro.heapPosition);

						//augments the path condition and updates the state's reference
						s.addClauseToPathCondition((ReferenceSymbolic) dro.refinedRef.clone(), null);
						StateGeneratorReference.this.f.refine(s, dro.refinedRef);
					}
				};
			
			public void refine(State s, DecisionAlternativeReference r) throws DecisionException, CannotManageStateException {
				this.s = s;
				try {
					r.accept(visitorRefine);
				} catch (DecisionException e) {
					throw e;
				} catch (CannotManageStateException e) {
					throw e;
				} catch (Throwable e) {
					//TODO unreachable code
				}
				//r.refine(s);
			}
		};
		this.bb = 
			new StateUpdateStrategy<DecisionAlternativeReference>() {

			public void update(State s, DecisionAlternativeReference r) {
				//pushes the refined reference on the 
				//operand stack
				Frame frame = s.getStack().currentFrame(); 
				frame.push(r.refinedRef);

				//updates the program counter
				try {
					s.incPC(pcOffset);
				} catch (CodeOutOfIndexException e) {
					Reference myExceRef = Util.loadInstance(s, "java/lang/VerifyError");
					frame.push(myExceRef);
					s.doThrow(myExceRef);
				}
				//r.update(s);
			}			
		};
		this.partialReferenceResolution = false;
	}

	/**
	 * Generates states from <code>runningState</code> by expanding 
	 * an uninitialized reference, where <code>runningState</code> 
	 * is provided by the constructor. 
	 * The concrete references produced by the expansion replace the 
	 * corresponding symbolic ones in their respective states, and are 
	 * pushed on the top of their respective operand stacks. 
	 * The produced states are stored in the <code>initializer</code> 
	 * provided by the constructor.
	 * 
	 * @param notInitializedRef 
	 *        the (uninitialized) <code>ReferenceSymbolic</code> to be expanded.
	 * @param index
	 *        an integer value identifying the source of  
	 *        <code>notInitializedRef</code>; it is interpreted as the  
	 *        slot number in the local variable area of 
	 *        <code>runningState</code>'s current frame, where 
	 *        <code>runningState</code> is provided by the constructor. 
	 * @throws CannotManageStateException 
	 * @throws DecisionException 
	 * @throws PartialReferenceResolutionException 
	 */
	public void generateStatesLoad(final ReferenceSymbolic notInitializedRef, final int index) throws DecisionException, CannotManageStateException, PartialReferenceResolutionException {
		this.notInitializedRef = notInitializedRef;
		this.f = 
			new SourceReferenceRefinementStrategy() { 
			public void refine(State s, ReferenceSymbolic newRef) {
				LocalVariablesArea lvt = s.getStack().currentFrame().getLocalVariableTable();
				lvt.set(index, newRef);
			}
		};
		this.generateStates(this.mgr.getComparators().createEmptyDecisionAlternativeReference(), this.aa, this.cc, this.bb);
		if (this.partialReferenceResolution) {
			throw new PartialReferenceResolutionException(this.notInitializedRef.getTypeNotInitialized());
		}
	}

	/**
	 * Generates states from <code>runningState</code> by expanding 
	 * an uninitialized reference, where <code>runningState</code> 
	 * is provided by the constructor. 
	 * The concrete references produced by the expansion replace the 
	 * corresponding symbolic ones in their respective states, and are 
	 * pushed on the top of their respective operand stacks. 
	 * The produced states are stored in the <code>Initializer</code> 
	 * provided by the constructor.
	 * 
	 * @param notInitializedRef 
	 *        the (uninitialized) <code>ReferenceSymbolic</code> to be expanded.
	 * @param index
	 *        an integer value identifying the source of  
	 *        <code>notInitializedRef</code>; it is interpreted as the  
	 *        row number in the constant pool of some <code>Instance</code>'s 
	 *        classfile, containing the signature of one of its fields. 
	 * @param src 
	 *        the <code>Reference</code> in <code>runningState</code>'s 
	 *        heap, of some <code>Instance</code>. 
	 * @throws CannotManageStateException 
	 * @throws DecisionException 
	 * @throws PartialReferenceResolutionException 
	 */
	public void generateStatesGetfield(final ReferenceSymbolic notInitializedRef, final int index, final Reference src) throws DecisionException, CannotManageStateException, PartialReferenceResolutionException {
		this.notInitializedRef = notInitializedRef;
		this.f =
			new SourceReferenceRefinementStrategy() { 
			public void refine(State s, ReferenceSymbolic newRef) {
				//fetches the instance from s.heap
				Instance myObject = (Instance) s.getHeap().getObject(src);

				//gets the signature of the field from the current class
				Frame frame = s.getStack().currentFrame();
				String classToOpen = frame.getName().getClassName();
				Signature fieldSignature = s.getClassFileInterface().getClassFile(classToOpen).getFieldSignature(index);
				//TODO verify whether this is correct!!!! Is the target object on the operand stack top???

				//resolves and sets the instance's field
				boolean failed = false;
				String myExce = null;
				try {
					Signature fieldSignatureResolved = Util.resolveField(s.getClassFileInterface(), fieldSignature);
					myObject.setFieldValue(fieldSignatureResolved, newRef);
				} catch (JavaReifyException e) {
					failed = true;
					myExce = e.subject();
				} finally {
					if (failed) {
						Reference myExceRef = Util.loadInstance(s, myExce);
						frame.push(myExceRef);
						s.doThrow(myExceRef);
						return;
					}
				}		
			}
		};
		this.generateStates(this.mgr.getComparators().createEmptyDecisionAlternativeReference(), this.aa, this.cc, this.bb);
		if (this.partialReferenceResolution) {
			throw new PartialReferenceResolutionException(this.notInitializedRef.getTypeNotInitialized());
		}
	}

	/**
	 * Generates states from <code>runningState</code> by expanding 
	 * an uninitialized reference, where <code>runningState</code> 
	 * is provided by the constructor. 
	 * The concrete references produced by the expansion replace the 
	 * corresponding symbolic ones in their respective states, and are 
	 * pushed on the top of their respective operand stacks. 
	 * The produced states are stored in the <code>initializer</code> 
	 * provided by the constructor.
	 * 
	 * @param notInitializedRef 
	 *        the (uninitialized) <code>ReferenceSymbolic</code> to be expanded.
	 * @param index
	 *        an integer value identifying the source of  
	 *        <code>notInitializedRef</code>; it is interpreted as the  
	 *        row number in the current class' constant pool containing 
	 *        the signature of a static field of the current class.
	 * @throws CannotManageStateException 
	 * @throws DecisionException 
	 * @throws PartialReferenceResolutionException 
	 */
	public void generateStatesGetstatic(final ReferenceSymbolic notInitializedRef, final int index) throws DecisionException, CannotManageStateException, PartialReferenceResolutionException {
		this.notInitializedRef = notInitializedRef;
		this.f = 
			new SourceReferenceRefinementStrategy() { 
			public void refine(State s, ReferenceSymbolic newRef) {
				//gets the signature of the static field from the current class
				Frame frame = s.getStack().currentFrame();
				String currentClassName = frame.getName().getClassName();
				Signature fieldSignature = s.getClassFileInterface().getClassFile(currentClassName).getFieldSignature(index);

				//fetches the klass of the field from s.staticStore
				Klass myObject = s.getKlass(fieldSignature.getClassName());

				//resolves and sets the klass' field
				boolean failed = false;
				String myExce = null;
				try {
					Signature fieldSignatureResolved = Util.resolveField(s.getClassFileInterface(), fieldSignature);
					myObject.setFieldValue(fieldSignatureResolved, newRef);
				} catch (JavaReifyException e) {
					failed = true;
					myExce = e.subject();
				} finally {
					if (failed) {
						Reference myExceRef = Util.loadInstance(s, myExce);
						frame.push(myExceRef);
						s.doThrow(myExceRef);
						return;
					}
				}		
			}
		};
		this.generateStates(this.mgr.getComparators().createEmptyDecisionAlternativeReference(), this.aa, this.cc, this.bb);
		if (this.partialReferenceResolution) {
			throw new PartialReferenceResolutionException(this.notInitializedRef.getTypeNotInitialized());
		}
	}	
}
