package muse.algo;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;

import muse.bc.*;
import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;
import muse.jvm.Environment;
import muse.mem.*;
import muse.mem.Objekt.Epoch;


/**
 * Class that supply utility method for algorithm
 */
public class Util {
	/**
	 * Given four bytes, returns the int resulting by their 
	 * bitwise concatenation
	 */
	public static int byteCat(byte first, byte second, byte third, byte fourth) {
		int first_32 = first;
		int second_32 = second;
		int third_32 = third;
		int fourth_32 = fourth;

		first_32 = first_32 << 24;
		second_32 = (second_32 << 16) & 0xFF0000;  
		third_32 = (third_32 << 8) & 0xFF00;  
		fourth_32 = fourth_32 & 0xFF;  

		return (first_32 | second_32 | third_32 | fourth_32);
	}

	/**
	 * Given two bytes, returns the int resulting by their 
	 * bitwise concatenation.
	 */
	public static int byteCat(byte first, byte second) {
		return byteCat((byte) 0, (byte) 0, first, second);
	}

	/**
	 * Given two bytes, returns the short resulting from their 
	 * bitwise concatenation.
	 */
	public static short byteCatShort(byte first, byte second) {
		//narrowing just discards the higher order bits, 
		//which however are zeros 
		return (short) byteCat(first, second);  
	}

	/**
     * Creates a new frame for a nonstatic method invocation and pushes it 
     * on a state's stack. The actual parameters of the invocation are 
	 * initialized with symbolic values.
     *  
     * @param state 
     *        the <code>State</code> to which the frame is added.
     * @param methodSignature 
     *        the <code>Signature</code> of the method for which the 
     *        frame is built.
     * @param thisRef
     *        a <code>Reference</code> for an object on the heap, 
     *        which is the subject (<code>this</code>) of the method 
     *        invocation. The <code>methodSignature</code> 
     *        must indicate a method in the class of the object.
     * @boolean isRoot
     *        <code>true</code> iff this is a root frame (i.e., with no invoker).
     * @param pcOffset 
     *        the offset from the current program counter of 
     *        the return program counter. Meaningful iff <code>!isRoot</code>.
	 * @throws JavaReifyException when frame creation does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
	 * @throws PleaseDoNativeException when the method is native. 
     */
    public static void loadSymbolicFrame(State state, Signature methodSignature, Reference thisRef, boolean isRoot, int pcOffset) throws JavaReifyException, PleaseDoNativeException {
    	//gets the method's signature
        String[] arrayDescriptor = Type.splitParametersDescriptors(methodSignature.getDescriptor());
		int count = arrayDescriptor.length + 1;
		
		//produces the args as symbolic values from the method's signature
		Value[] args = new Value[count];
		args[0] = thisRef;
		for (int i = 1; i < count; i++) {
			//gets the field signature and name
			String tmpDescriptor = arrayDescriptor[i - 1];

            //builds a symbolic value from signature and name
           	args[i] = state.createSymbol(tmpDescriptor, "");
		}
		
		//creates and initializes the frame, and pushes on the state's 
		//stack frame
		Util.createFrame(false, state, methodSignature, false, args, isRoot, pcOffset);
		
		//sets the symbol's origin according to parameter's names
		//TODO make this block a LocalVariablesArea method
		String methodName = methodSignature.getName();
		LocalVariablesArea lvt = state.getStack().currentFrame().getLocalVariableTable();
		int slot = 0;
		for (Value arg: args) {
			if (arg instanceof Symbolic) {
				Symbolic sy = (Symbolic) arg;
				Variable v = lvt.buildVar(slot, 0);
				sy.setOrigin(methodName + "[" + v.getName() + "]");
			}
			
			//next slot
			slot++;
			if (!Type.isCat_1(arg.getType())) {
				slot++;
			}
		}
    }

    /**
     * Creates a new frame for a static method invocation and pushes it 
     * on a state's stack. The actual parameters of the invocation are 
	 * initialized with symbolic values.
     * 
     * @param state
     *        the <code>State</code> to which the frame is added.
     * @param methodSignature
     *        the <code>Signature</code> of the method for which the 
     *        frame is built.
     * @param isRoot
     *        <code>true</code> iff this is a root frame (i.e., with no invoker).
     * @param pcOffset
     *        the offset from the current program counter of 
     *        the return program counter. Meaningful iff <code>!isRoot</code>.
	 * @throws JavaReifyException when frame creation does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
	 * @throws PleaseDoNativeException when the method is native. 
     */
    public static void loadSymbolicFrameStatic(State state, Signature methodSignature, boolean isRoot, int pcOffset) throws JavaReifyException, PleaseDoNativeException {
    	//TODO collapse with loadSymbolicFrame
    	//gets the method's signature
        String[] arrayDescriptor = Type.splitParametersDescriptors(methodSignature.getDescriptor());
		int count = arrayDescriptor.length;
		
		//produces the args as symbolic values from the method's signature
		Value[] args = new Value[count];
		for (int i = 0; i < count; i++) {
			//gets the field signature and name
			String tmpDescriptor = arrayDescriptor[i];

            //builds a symbolic value from signature and name
           	args[i] = state.createSymbol(tmpDescriptor, "");
		}
		
		//creates and initializes the frame, and pushes on the state's 
		//stack frame
		Util.createFrame(false, state, methodSignature, true, args, isRoot, pcOffset);
		
		//sets the symbol's origin according to parameter's names
		//TODO make this block a LocalVariablesArea method
		String methodName = methodSignature.getName();
		LocalVariablesArea lvt = state.getStack().currentFrame().getLocalVariableTable();
		int slot = 0;
		for (Value arg: args) {
			if (arg instanceof Symbolic) {
				Symbolic sy = (Symbolic) arg;
				Variable v = lvt.buildVar(slot, 0);
				sy.setOrigin(methodName + "[" + v.getName() + "]");
			}
			
			//next slot
			slot++;
			if (!Type.isCat_1(arg.getType())) {
				slot++;
			}
		}
    }

    /**
	 * Creates a new frame for a nonstatic method invocation and pushes it 
	 * on a state's stack. The actual parameters of the invocation are 
	 * initialized with values from the invoking frame's operand stack.
	 * 
	 * @param isSpecial
	 *        <code>true</code> if INVOKESPECIAL method code lookup rules 
	 *        must be applied, <code>false</code> if INVOKEVIRTUAL/INVOKEINTERFACE 
	 *        code lookup rules must be applied. 
     * @param state 
     *        the <code>State</code> to which the frame is added.
     * @param methodSignature
     *        the <code>Signature</code> of the method for which the 
     *        frame is built. If <code>isSpecial == true</code> it must 
     *        be the <code>Signature</code> of the resolved method, 
     *        otherwise the class indicated in <code>methodSignature</code>
     *        is ignored.
     * @param argArray
     *        the array of method call arguments.
     * @param pcOffset 
     *        the offset from the current program counter of 
     *        the return program counter.
	 * @throws JavaReifyException when frame creation does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
	 * @throws PleaseDoNativeException when the method is native. 
     */
	public static void loadFrame(boolean isSpecial, State state, Signature methodSignature, Value[] argArray, int pcOffset) throws JavaReifyException, PleaseDoNativeException {
		//ensures the receiver of the invocation is not null
		if (((Reference) argArray[0]).isNull()) {
			throw new JavaReifyException("java/lang/NullPointerException");
		}

		//creates and initializes the frame, and pushes it on the state's 
		//frame stack
		Util.createFrame(isSpecial, state, methodSignature, false, argArray, false, pcOffset);
	}

	/**
	 * Parses the signature of a method, and pops from the operand stack 
	 * a number of values consistent with the number of method parameters.
	 * @param state 
	 *        the <code>State</code> from which the values are popped. 
	 * @param methodSignature
	 *        the <code>SESignature</code> of a method.
	 * @param isStatic
	 *        <code>true</code> iff the method is static.
	 * @return a <code>Value[]</code> of all the parameters, where the i-th 
	 *         array element is the value for the i-th parameter in the 
	 *         <code>methodSignature</code>. 
	 */
	public static Value[] popMethodCallArgs(State state, Signature methodSignature, boolean isStatic) {
		Frame frame = state.getStack().currentFrame();
		String[] arrayDescriptor = Type.splitParametersDescriptors(methodSignature.getDescriptor());
		int count = (isStatic ? arrayDescriptor.length : arrayDescriptor.length + 1);
		Value[] argArray = new Value[count];
		for (int i = 1; i <= count; i++) { 
			argArray[count - i] = frame.pop();
		}
		return argArray;
	}
	
	/**
	 * Creates a new frame for a static method invocation and pushes it 
	 * on a state's stack. The actual parameters of the invocation are 
	 * initialized with values from the invoking frame's operand stack.
	 * 
	 * @param state
     *        the <code>State</code> to which the frame is added.
	 * @param methodSignature
     *        the <code>SESignature</code> of the method for which the 
     *        frame is built.
     * @param argArray
     *        the array of method call arguments.
     * @param pcOffset 
     *        the offset from the current program counter of 
     *        the return program counter.
	 * @throws JavaReifyException when frame creation does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
	 * @throws PleaseDoNativeException when the method is native. 
	 */
	public static void loadFrameStatic(State state, Signature methodSignature, Value[] argArray, int pcOffset) throws JavaReifyException, PleaseDoNativeException {
		Util.createFrame(false, state, methodSignature, true, argArray, false, pcOffset);
	}

    /**
	 * Creates a new frame for a method invocation and pushes it on a 
	 * state's frame stack.
	 * 
	 * @param isSpecial
	 *        <code>true</code> if INVOKESPECIAL method invocation rules 
	 *        must be applied, <code>false</code> if INVOKEVIRTUAL/INVOKEINTERFACE 
	 *        method invocation rules must be applied. It is ignored if 
	 *        <code>isStatic == true</code>.
     * @param state 
     *        the <code>State</code> to which the frame is added.
     * @param methodSignatureResolved
     *        the <code>SESignature</code> of the resolved method for 
     *        which the frame must be built.
     * @param isStatic
     *        <code>true</code> iff INVOKESTATIC method invocation rules 
     *        must be applied.
     * @param args
     *        an array of <code>Value</code>s which will be used to
     *        initialize the local variables, up to exhaustion of the 
     *        array. The other local variables will be initialized with 
     *        the default values for their respective types. If  
     *        <code>args[i] == null</code>, the i-th local variable will be 
     *        initialized to its default value. Counting of variables 
     *        does *not* follow Java's local variable table convention 
     *        of variables with size two - i.e., if <code>args[i]</code> is the 
     *        initialization value of local variable k, and this variable 
     *        has type double, then the initialization value of the next 
     *        local variable k + 2 is contained in <code>args[i + 1]</code>. 
     *        If <code>args == null || args.length == 0</code> all the 
     *        local variables will be initialized to their default values.
     * @boolean isRoot
     *        <code>true</code> iff this is a root frame (i.e., with no invoker).
     * @param pcOffset 
     *        the offset from the current program counter of 
     *        the return program counter. Meaningful iff <code>!isRoot</code>.
	 * @throws JavaReifyException when frame creation does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
     * @throws ValueDoesNotSupportNativeException 
     * @throws PleaseDoNativeException 
     */
	private static void createFrame(boolean isSpecial, State state, Signature methodSignatureResolved, boolean isStatic, Value[] args, boolean isRoot, int pcOffset) throws JavaReifyException, PleaseDoNativeException {
		//gets the class from which method code lookup is started in iterative
		//lookup procedures, i.e., the current class for INVOKESPECIAL and 
		//the class of the invocation receiver for INVOKEVIRTUAL and 
		//INVOKEINTERFACE
		ClassFile luStartClass = null;
		if (!isStatic) {
			String luStartClassName = null;
			if (isSpecial) {
				Frame currentFrame = state.getStack().currentFrame();
				luStartClassName = currentFrame.getName().getClassName();
			} else {
				luStartClassName = state.getHeap().getObject((Reference) args[0]).getType();
			}
			luStartClass = state.getClassFileInterface().getClassFile(luStartClassName);
		}
		
		//performs method code lookup
		ClassFile classMethodImpl = null;
		if (isStatic) {               //INVOKESTATIC
			classMethodImpl = Util.lookupMethodImplStatic(state.getClassFileInterface(), methodSignatureResolved);
		} else if (isSpecial) {       //INVOKESPECIAL
		    classMethodImpl = Util.lookupMethodImplSpecial(luStartClass, state.getClassFileInterface(), methodSignatureResolved);
		} else {                      //INVOKEVIRTUAL, INVOKEINTERFACE 
			classMethodImpl = Util.lookupMethodImplVirtualInterface(luStartClass, state.getClassFileInterface(), methodSignatureResolved);
		}
		
		//exits if the method is native
		if (classMethodImpl.isMethodNative(methodSignatureResolved)) {
			throw new PleaseDoNativeException();
		}
		
        //gets the method's local variable table information
		LocalVariableTable lvtMethod;
		lvtMethod = classMethodImpl.getLocalVariableTable(methodSignatureResolved);
		
		//builds the frame's local variable area
		LocalVariablesArea lv = new LocalVariablesArea(lvtMethod, args);

		//builds an empty operand stack
		OperandStack op = new OperandStack();
		
		//gets the method's line number table 
		LineNumberTable lntMethod;
		lntMethod = classMethodImpl.getLineNumberTable(methodSignatureResolved);
		
		//saves the return program counter
		if (!isRoot) {
			try {
				state.setReturnPC(pcOffset);
			} catch (CodeOutOfIndexException e) {
				throw new JavaReifyException("java/lang/VerifyError");
			}
		}

		//builds the new frame and pushes it on the state's stack
		Signature methodSignature = new Signature(classMethodImpl.getClassName(), methodSignatureResolved.getDescriptor(), methodSignatureResolved.getName());
		state.getStack().push(new Frame(lntMethod, op, lv, methodSignature));

		//sets the method bytecode
		state.setCode(classMethodImpl.getMethodCodeBySignature(methodSignature));
	}
	
	/**
	 * Performs method lookup according to the semantics of the 
	 * INVOKEVIRTUAL bytecode.
	 * 
	 * @param receiverClass the <code>SEClassFile</code> of the 
	 *        method invocation's receiver.
	 * @param methodSignature the signature of the resolved method 
	 *        which must be looked up.
	 * @return the <code>SEClassFile</code> of the superclass of
	 *         <code>methodSignature.getClassName()</code> which 
	 *         contains the method implementation.
	 * @throws JavaReifyException when method lookup does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
	 */
	public static ClassFile lookupMethodImplVirtualInterface(ClassFile receiverClass, ClassFileInterface cfi, Signature methodSignature) throws JavaReifyException {
		ClassHierarchy hier = new ClassHierarchy(cfi);
		for (ClassFile f : hier.superclasses(receiverClass.getClassName())) {
			if (f.hasMethodImplementation(methodSignature)) {
				return f;
			}
		}
		throw new JavaReifyException("java/lang/AbstractMethodError");
	}
	
	/**
	 * Performs method lookup according to the semantics of the 
	 * INVOKESPECIAL bytecode.
	 * 
	 * @param currentClass the class of the invoker
	 * @param methodSignature the signature of the resolved method 
	 *        which must be looked up.
	 * @return the <code>SEClassFile</code> of the class which 
	 *         contains the method implementation.
	 * @throws JavaReifyException when method lookup does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
	 */
	public static ClassFile lookupMethodImplSpecial(ClassFile currentClass, ClassFileInterface cfi, Signature methodSignature) throws JavaReifyException {
		ClassFile retVal = null;
		boolean useNonVirtual = true;

		//determines whether should look for the implementation in 
		//the current class superclasses (virtual semantics) or in 
		//the method's signature class (nonvirtual semantics) and 
		//at the same time calculates the superclass for virtual semantics
		if (!methodSignature.getName().equals("<init>") && 
			currentClass.isSuperInvoke()) {
			boolean skippedFirst = false;
			ClassHierarchy hier = new ClassHierarchy(cfi);
			
			for (ClassFile f : hier.superclasses(currentClass.getClassName())) {
				if (skippedFirst) { 
					if (f.hasMethodImplementation(methodSignature))
						retVal = f;
					if (f.getClassName().equals(methodSignature.getClassName())) {
						useNonVirtual = false;
						break;
					}
				}
				skippedFirst = true;
			}
		}
		
		//nonvirtual (i.e., <init>) case: gets the method implementation
		//from the resolved method's class
		if (useNonVirtual) {
			ClassFile resolvedClass = cfi.getClassFile(methodSignature.getClassName());
			if (resolvedClass.hasMethodImplementation(methodSignature))
				retVal = resolvedClass;
			else
				retVal = null;
		}
		
		if (retVal == null)
			throw new JavaReifyException("java/lang/AbstractMethodError");

		return retVal;
	}
	
	/**
	 * Performs method lookup according to the semantics of the 
	 * INVOKESTATIC bytecode.
	 * 	 
	 * @param methodSignature the signature of the resolved method 
	 *        which must be looked up.
	 * @return the <code>SEClassFile</code> of the class which 
	 *         contains the method implementation.
	 * @throws JavaReifyException when method lookup does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.
	 */
	public static ClassFile lookupMethodImplStatic(ClassFileInterface cfi, Signature methodSignature) throws JavaReifyException {
		ClassFile retVal = cfi.getClassFile(methodSignature.getClassName());		
		if (!retVal.isMethodStatic(methodSignature)) {
			throw new JavaReifyException("java/lang/IncompatibleClassChangeError");
		}
		return retVal;
	}

	/**
	 * Performs field resolution (see JVM Specification, sec. 5.4.3.2).
	 * 
	 * @param cfi the <code>SEClassFileInterface</code> of the execution.
	 * @param fieldSignature the <code>SESignature</code> of the field to be resolved.
	 * @return the <code>SESignature</code> of the declaration of the resolved field.
	 * @throws JavaReifyException when method resolution does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.    
	 */
	public static Signature resolveField(ClassFileInterface cfi, Signature fieldSignature) throws JavaReifyException {
		//lookup starts from the field's signature class
		ClassFile classFile = cfi.getClassFile(fieldSignature.getClassName());

		//searches in classFile a declaration for the field
		Signature fieldSignatureResolved = null;
		if (classFile.hasFieldDeclaration(fieldSignature)) {
			fieldSignatureResolved = fieldSignature;
		}

		//if nothing has been found, searches in the superinterfaces
        ClassHierarchy hier = new ClassHierarchy(cfi);
		if (fieldSignatureResolved == null) {
			for (ClassFile cf : hier.superinterfaces(fieldSignature.getClassName())) {
        		if (cf.hasFieldDeclaration(fieldSignature)) {
        			fieldSignatureResolved = 
        				new Signature(cf.getClassName(), fieldSignature.getDescriptor(), fieldSignature.getName());
        			break;
        		}
			}
		}
        
		//if still nothing has been found, searches in the superclasses
		if (fieldSignatureResolved == null) {
	       	for (ClassFile cf : hier.superclasses(fieldSignature.getClassName())) {
	       		if (cf.hasFieldDeclaration(fieldSignature)) {
	       			fieldSignatureResolved = 
	       				new Signature(cf.getClassName(), fieldSignature.getDescriptor(), fieldSignature.getName());
	       			break;
	       		}
	       	}
		}
		
        //if still nothing has been found, raises an exception
		if (fieldSignatureResolved == null) {
			throw new JavaReifyException("java/lang/NoSuchFieldError");
		}
					
		//TODO checks accessibility of the field, and throws IllegalAccessError in case

		//everything went ok
		return fieldSignatureResolved;
	}
	
	/**
	 * Performs both method and interface method resolution (see JVM Specification, sec. 5.4.3.3 and 5.4.3.4).
	 * 
	 * @param isInterface <code>true</code> iff the method to be resolved is an interface method.
	 * @param methodSignature the <code>SESignature</code> of the method to be resolved.
	 * @return the <code>SESignature</code> of the declaration of the resolved method.
	 * @throws JavaReifyException when method resolution does not succeed; the exception 
	 *         stores the name of the Java exception that the virtual machine must raise.    
	 */
	public static Signature resolveMethod(ClassFileInterface cfi, boolean isInterface, Signature methodSignature) throws JavaReifyException {
		//checks if the symbolic reference to the method class 
		//is an interface
		ClassFile classFile = cfi.getClassFile(methodSignature.getClassName());
		if ((isInterface && !classFile.isInterface()) || (!isInterface && classFile.isInterface())) {
			throw new JavaReifyException("java/lang/IncompatibleClassChangeError");
		}

		//attempts to find a superclass or superinterface containing 
		//a declaration for the method
		Signature methodSignatureResolved = null;
		ClassFile classFileResolve = null;

		//searches for the method declaration in the superclasses (only method resolution)
        ClassHierarchy hier = new ClassHierarchy(cfi);
        if (!isInterface) {
        	for (ClassFile cf : hier.superclasses(methodSignature.getClassName())) {
        		if (cf.hasMethodDeclaration(methodSignature)) {
        			classFileResolve = cf;
        			methodSignatureResolved = 
        				new Signature(cf.getClassName(), methodSignature.getDescriptor(), methodSignature.getName());
        			break;
        		}
        	}
        }
				
        //if nothing has been found, searches in the superinterfaces
		if (methodSignatureResolved == null) {
			for (ClassFile cf : hier.superinterfaces(methodSignature.getClassName())) {
        		if (cf.hasMethodDeclaration(methodSignature)) {
        			classFileResolve = cf;
        			methodSignatureResolved = 
        				new Signature(cf.getClassName(), methodSignature.getDescriptor(), methodSignature.getName());
        			break;
        		}
			}
		}
        
        //if still nothing has been found, searches in the java.lang.Object class 
        //(only interface method resolution)
		if (methodSignatureResolved == null && isInterface) {
			ClassFile cf = cfi.getClassFile("java/lang/Object");
    		if (cf.hasMethodDeclaration(methodSignature)) {
    			classFileResolve = cf;
    			methodSignatureResolved = 
    				new Signature(cf.getClassName(), methodSignature.getDescriptor(), methodSignature.getName());
    		}
		}
        
        //if still nothing has been found, raises an exception
		if (methodSignatureResolved == null)
			throw new JavaReifyException("java/lang/NoSuchMethodError");
		
		//additional checks for noninterface methods
		if (!isInterface) {
			//checks the resolved method is not abstract in the case 
			//the class where we started is not an abstract class
			if (!classFile.isAbstract() && classFileResolve.isMethodAbstract(methodSignatureResolved))
				throw new JavaReifyException("java/lang/AbstractMethodError");
	
			//TODO checks accessibility of the method, and throws IllegalAccessError in case
		}
			
		//everything went ok
		return methodSignatureResolved;
	}
	
	/**
	 * Initializes some <code>Instance</code> with symbolic values.
	 * 
	 * @param s a <code>State</code>.
	 * @param myObj an <code>Instance</code> which will be initialized with 
	 *              symbolic values created by <code>s</code>.
	 */
	public static void initWithSymbolicValues(State s, Instance myObj) {
        Signature[] mySgnArray = myObj.getFieldSignatures();
        for (int j = 0; j < myObj.getLength(); j++) {
            //gets the field signature and name
            Signature myActualSignature = mySgnArray[j];
            String tmpDescriptor = myActualSignature.getDescriptor();
            String tmpName = myActualSignature.getName();

            //builds a symbolic value from signature and name 
            //and assigns it to the field
            myObj.setFieldValue(myActualSignature, 
            		s.createSymbol(tmpDescriptor, myObj.getOrigin() + "." + tmpName));
        }
	}
	
    /**
     * Creates a new <code>Instance</code> of a given class in the heap of 
	 * a state. The <code>Instance</code>'s fields are initialized with symbolic 
	 * values.
	 *  
     * @param state 
     *        the <code>State</code> whose heap will receive 
	 *        the new object.
     * @param className 
     *        the name of the class of the new object.
     * @param ref
     *        a previously created <code>SESymbolicReference</code> to be updated; if 
     *        it is <code>null</code> a new <code>SEConcreteReference</code> will be created. 
     * @param thisName
     *        a <code>String</code> representing the name of the object; used in the 
     *        case <code>ref == null</code> to give a meaningful name to the 
     *        created object, otherwise the name is calculated from <code>ref</code>.
     * @return a <code>SEReference</code> to the newly created object, either 
     *         <code>ref</code> or a new <code>SEConcreteReference</code>.
     */
    public static Reference loadSymbolicInstance(State state, String className, ReferenceSymbolic ref, String thisName) {
    	//determines the object's origin
    	String origin;
    	if (ref == null) {
    		origin = thisName;
    	} else {
    		origin = ref.getOrigin();
    	}
    	
    	//creates the instance
    	Instance myObj = Util.createNewInstance(className, state.getClassFileInterface(), origin);

    	//sets the epoch
        myObj.setCreationEpoch(Objekt.Epoch.EPOCH_BEFORE_START);
        
        //initializes the instance
        initWithSymbolicValues(state, myObj);
        
        //puts the instance in the heap and returns a reference to it
        int pos = state.getHeap().addNew(myObj);
        if (ref == null) {
        	return new ReferenceConcrete(pos);
        } else {
        	ref.initialize(pos);
        	return ref;
        }
    }

	/**
	 * Creates a new <code>Instance</code> of a given class in the 
	 * heap of a state. The <code>Instance</code>'s fields are initialized 
	 * with the default values for each field's type.
	 * 
	 * @param state the <code>State</code> whose heap will receive 
	 *              the new object.
	 * @param className the name of the class of the new object.
	 * @return a <code>SEConcreteReference</code> to the newly created object.
	 */
	public static ReferenceConcrete loadInstance(State state, String className) {
        Instance myObj = Util.createNewInstance(className, state.getClassFileInterface(), null);
        myObj.setCreationEpoch(Objekt.Epoch.EPOCH_AFTER_START);
		return new ReferenceConcrete(state.getHeap().addNew(myObj));
	}
	
	/**
	 * Creates a new <code>Instance</code> of class String in a state's heap 
	 * and loads it in the state's literal area, in the case such literal 
	 * does not exist yet.
	 * 
	 * @param state the <code>State</code> where the string will be loaded.
	 * @param value a <code>String</code> representing the value stored in the 
	 *              created <code>Instance</code>.
	 * @return a <code>SEConcreteReference</code> to the created <code>Instance</code>, 
	 *         or to a previously created <code>Instance</code>. 
	 * @throws DecisionException 
	 */
	public static ReferenceConcrete loadStringLiteral(State state, String value, DecisionProcedure mgr) throws DecisionException {
		//tries to fetch the value from the state's literals
		ReferenceConcrete retVal = state.getStringLiteral(value);
		
		//if it does not find an instance for the literal, 
		//it creates one
		if (retVal == null) {
			//creates the instance and puts it in the state's literals
			retVal = loadInstance(state, "java/lang/String");
			state.setStringLiteral(value, retVal);
			
			//directly initializes the instance's fields
			Instance i = (Instance) state.getHeap().getObject(retVal);
			i.setFieldValue(new Signature("java/lang/String", "" + Type.INT, "hash"), new Simplex(0));
			i.setFieldValue(new Signature("java/lang/String", "" + Type.INT, "offset"), new Simplex(0));
			i.setFieldValue(new Signature("java/lang/String", "" + Type.INT, "count"), new Simplex(value.length()));
			int[] dims = new int[1]; dims[0] = value.length();
			ReferenceConcrete fldValue = null;
			try {
				//fldValue = loadArray(state, "" + Type.ARRAYOF + Type.CHAR, dims, mgr);
				fldValue = new ReferenceConcrete(state.getHeap().addNew(new Array(false, null, new Simplex(value.length()), "" + Type.ARRAYOF + Type.CHAR, null)));
			} catch (JavaReifyException e) {
				//TODO this should never happen, handle somehow
			}
			i.setFieldValue(new Signature("java/lang/String", "" + Type.ARRAYOF + Type.CHAR, "value"), fldValue);
			
			//initializes the instance's field "value" by copying 
			//the String value parameter in it
			Array a = (Array) state.getHeap().getObject(fldValue);
			for (int k = 0; k < value.length(); k++) {
				char c = value.charAt(k);
				a.setElement(new Simplex(k), new Simplex(c), mgr);
			}
		}
		
		return retVal;
	}

	/**
	 * Creates an <code>Instance</code> of a given class.
	 * 
	 * @param className the name of the class as a <code>String</code>.
	 * @param cfi the <code>SEClassFileInterface</code> of the class.
	 * @param origin the origin of the <code>Instance</code>, or 
	 *               <code>null</code> iff the instance is not created 
	 *               by lazy initialization.
	 * @return the created <code>Instance</code>.
	 */
	private static Instance createNewInstance(String className, ClassFileInterface cfi, String origin) {
        //accumulates all the field signatures of the class and of 
        //its superclasses 
        Vector<Signature[]> myVector = new Vector<Signature[]>(0);
        int nfields = 0;
        ClassHierarchy hier = new ClassHierarchy(cfi);
        for (ClassFile c : hier.superclasses(className)) {
            Signature[] sgnFields = c.getFieldsNonStatic();
            myVector.add(sgnFields);
            nfields += sgnFields.length;
        }
        
        //flattens myVector into mySgnArray
        Signature[] mySgnArray = new Signature[nfields];
        {   //block to make the definition of j temporary
            int j = 0;
            for (Signature[] sig : myVector) {
                for (int k = 0; k < sig.length; k++) { 
                    mySgnArray[j] = sig[k];
                    j++;
                }
            }
        }
		
        //creates the new instance of the object and returns it
        return new Instance(mySgnArray, className, origin);		
	}

	/**
     * Creates a new <code>Array</code> of a given class in the heap of 
	 * a state. The <code>Array</code>is initialized with symbolic 
	 * values.
	 * 
     * @param s
     *        the <code>State</code> whose heap will receive 
	 *        the new object.
     * @param arraySignature 
     *        the type of the array.
     * @param ref
     *        a previously created <code>SESymbolicReference</code> to be updated; if 
     *        it is <code>null</code> a new <code>SEConcreteReference</code> will be created. 
     * @param name
     *        a <code>String</code> representing the name of the object; used in the 
     *        case <code>ref == null</code> to give a meaningful name to the 
     *        created object, otherwise the name is calculated from <code>ref</code>.
	 * @param mgr
	 *        a <code>SESimplifier</code> set with the current path condition.
     * @return a <code>SEReference</code> to the newly created object, either 
     *         <code>ref</code> or a new <code>SEConcreteReference</code>.
	 * @throws DecisionException
	 * @throws CannotManageStateException
	 */
	public static Reference loadSymbolicArray(State s, String arraySignature, ReferenceSymbolic ref, String name, DecisionProcedure mgr) throws DecisionException, CannotManageStateException {
		//determines the object's origin
    	String origin;
    	if (ref == null) {
    		origin = name;
    	} else {
    		origin = ref.getOrigin();
    	}

		//creates the symbolic array length and assumes it positive
		Primitive length = (Primitive) s.createSymbol("" + Type.INT, ref.getOrigin());
		s.addClauseToPathCondition((Expression) length.ge(new Simplex(0)));
		
		//creates the array
		Array a = null;
		try {
			a = new Array(true, null, length, arraySignature, origin);
		} catch (JavaReifyException e) {
			//TODO this should never happen
		}
		
		//sets the epoch
		a.setCreationEpoch(Epoch.EPOCH_BEFORE_START);
		
		//puts the array in the heap and returns a reference to it
		int pos = s.getHeap().addNew(a);
		if (ref == null) {
			return new ReferenceConcrete(pos);
		} else {
			ref.initialize(pos);
			return ref;
		}
	}
	
	public static boolean isFieldAccessible(Signature accessor, Signature fld, Environment env, ClassFileInterface cfi) {
		return true; //TODO implement visibility rules
	}
	
	public static boolean isMethodAccessible(Signature accessor, Signature meth, Environment env, ClassFileInterface cfi) {
		return true; //TODO implement visibility rules
	}
	
	public static String getSrcFileRow(String className, String[] srcPath, String pathSep, int row) {
		if (row != -1 && !srcPath.equals("")) {
			for (String s: srcPath) {
				String fileName = (s.endsWith(pathSep) ? s : s.concat(pathSep)) + className.split("\\$")[0] + ".java"; 
				BufferedReader fr = null;
				try {
					fr = new BufferedReader(new FileReader(fileName));
					String l;
					for (int i = 1; ((l = fr.readLine()) != null) ; i++) {
						if (i == row) {
							return l;
						}
					}
				} catch (FileNotFoundException e) {
					//does nothing
				} catch (IOException e) {
					//does nothing
				} finally {
					try {
						if (fr != null) fr.close();
					} catch (IOException e) {
						//does nothing
					}
				}
			}
		}
		return null;
	}	
}