package muse.apps;

import muse.algo.CannotManageStateException;
import muse.algo.PartialReferenceResolutionException;
import muse.dec.DecisionBacktrackException;
import muse.dec.DecisionRunningException;
import muse.jvm.CannotBacktrackException;
import muse.jvm.Engine;
import muse.jvm.EngineStuckException;
import muse.tree.StateTree.BranchPoint;

/**
 * Class implementing a full method running algorithm.
 * 
 * @author pietro
 */
public class Runner {
	public static class Actions {
		public boolean atTraceStart() { return false; }
		
		public boolean atPreStep(boolean isRootBranch) { return false; }
		
		public boolean atPostStep(BranchPoint bp) { return false; }
		
		public boolean atTraceEnd(boolean isRootBranch) { return false;	}

		public boolean atBacktrack() { return false; }
		
		public void atEnd() { }		

		public boolean atDecisionRunningException(DecisionRunningException e) 
		throws DecisionRunningException { return true; }

		public boolean atEngineStuckException(EngineStuckException e) 
		throws EngineStuckException { return true; }

		public boolean atCannotManageStateException(CannotManageStateException e) 
		throws CannotManageStateException { return true; }

		public boolean atPartialReferenceResolutionException(PartialReferenceResolutionException e) 
		throws PartialReferenceResolutionException { return true; }
		
		public boolean atDecisionBacktrackException(DecisionBacktrackException e) 
		throws DecisionBacktrackException { return true; }
		
		public boolean atCannotBacktrackException(CannotBacktrackException e) 
		throws CannotBacktrackException { return true;	}		
	}
	
	/** The symbolic execution engine used by the <code>Runner</code>. */
	private Engine engine;

	/** 
	 * <code>true</code> iff the <code>Runner</code> must step Java instructions, 
	 * <code>false</code> iff the <code>Runner</code> must step bytecodes.
	 */
	private boolean stepNext;
	
	/** The <code>Actions</code> to be performed while visiting the execution tree. */
	private Actions actions;

	/**
	 * Constructor. 
	 */
	public Runner(Engine engine, boolean stepNext) {
		this.engine = engine;
		this.stepNext = stepNext;
		this.actions = new Actions();
	}
	
	public Runner(Engine engine, boolean stepNext, Actions actions) {
		this(engine, stepNext);
		this.actions = actions;
	}

	/**
	 * Executes the method.
	 * 
	 * @throws PartialReferenceResolutionException 
	 * @throws DecisionBacktrackException 
	 * @throws CannotBacktrackException 
	 * @throws DecisionRunningException 
	 * @throws EngineStuckException 
	 * @throws CannotManageStateException 
	 */
	public void run() throws PartialReferenceResolutionException, DecisionBacktrackException, CannotBacktrackException, DecisionRunningException, EngineStuckException, CannotManageStateException  {
		//runs the method		
		boolean isRootBranch;
		boolean stop;

		//performs the symbolic execution loop
		while (true) {
			stop = this.actions.atTraceStart();
			if (stop) {
				return;
			}

			//steps until the last state of the trace is reached
			isRootBranch = true;
			while (this.engine.canStep()) {
				stop = this.actions.atPreStep(isRootBranch);
				if (stop) {
					return;
				}

				//performs a step and prints a warning in the case some reference has been 
				//partially resolved
				BranchPoint bp = null;
				try {
					bp = (stepNext ? this.engine.stepNext() : this.engine.step());
				} catch (DecisionRunningException e) {
					stop = this.actions.atDecisionRunningException(e);
					if (stop) {
						return;
					}
				} catch (EngineStuckException e) {
					stop = this.actions.atEngineStuckException(e);
					if (stop) {
						return;
					}
				} catch (CannotManageStateException e) {
					stop = this.actions.atCannotManageStateException(e);
					if (stop) {
						return;
					}
				} catch (PartialReferenceResolutionException e) {
					stop = this.actions.atPartialReferenceResolutionException(e);
					if (stop) {
						return;
					}
				}

				stop = this.actions.atPostStep(bp);
				if (stop) {
					return;
				}

				if (bp == null) { 
					isRootBranch = false;
				} else {
					isRootBranch = true;
				}
			}

			//stuck state reached 
			stop = this.actions.atTraceEnd(isRootBranch);
			if (stop) {
				return;
			}

			//backtracks
			if (this.engine.canBacktrack()) {
				stop = this.actions.atBacktrack();
				if (stop) {
					return;
				}
				try {
					this.engine.backtrack();
				} catch (DecisionBacktrackException e) {
					stop = this.actions.atDecisionBacktrackException(e);
					if (stop) {
						return;
					}
				} catch (CannotBacktrackException e) {
					stop = this.actions.atCannotBacktrackException(e);
					if (stop) {
						return;
					}
				}
			} else {
				this.actions.atEnd();
				return;
			}
		}
	}
}
