package muse.apps;

import java.util.HashSet;
import java.util.Map;

import muse.algo.Util;
import muse.mem.Array;
import muse.mem.Expression;
import muse.mem.Frame;
import muse.mem.FunctionApplication;
import muse.mem.Heap;
import muse.mem.Instance;
import muse.mem.Klass;
import muse.mem.LocalVariablesArea;
import muse.mem.Objekt;
import muse.mem.OperandStack;
import muse.mem.PathCondition;
import muse.mem.Primitive;
import muse.mem.PrimitiveSymbolic;
import muse.mem.ReferenceSymbolic;
import muse.mem.Simplex;
import muse.mem.State;
import muse.mem.StaticMethodArea;
import muse.mem.ThreadStack;
import muse.mem.Type;
import muse.mem.Value;
import muse.mem.Variable;


public abstract class StateFormatterText implements StateFormatter {
	protected final String lineSep = System.getProperty("line.separator");
	protected final String pathSep = System.getProperty("file.separator");
	protected final String indent = "\t";
	protected final BytecodeFormatter bfmt = new BytecodeFormatter();

	protected String[] srcPath;
	protected String indentTxt = "";
	
	protected String formatOutput = "";
	
	public StateFormatterText(String[] srcPath) {
		this.srcPath = srcPath;
	}
	
	public void format(State s) {		
        if (s.isStuck()) {
        	this.formatOutput += "<Leaf state";
        	if (s.getStuckException() != null) {
        		this.formatOutput += ", raised exception: " + s.getStuckException().toString() + ">" + lineSep;
        	} else if (s.getStuckReturn() != null) {
        		this.formatOutput += ", returned value: " + s.getStuckReturn().toString() + ">" + lineSep;
        	} else {
        		this.formatOutput += ">" + lineSep;
        	}
        } else {
        	Frame currentFrame = s.getStack().currentFrame();
        	this.formatOutput += "Method signature: " + currentFrame.getName() + lineSep;
	        if (this.srcPath == null) {
	        	this.formatOutput += "Program counter: " + s.getPC() + lineSep;
	        	this.formatOutput += "Next bytecode: " + this.bfmt.select(s.getInstruction()).format(currentFrame) + lineSep; 
	        } else {
	        	this.formatOutput += "Source line: " + this.sourceLine(currentFrame) + lineSep;
	        }
        }
        this.formatOutput += "Path condition: " + this.formatPathCondition(s.getPathCondition()) + lineSep;
        this.formatOutput += "Static store: {" + lineSep + this.formatStaticStore(s.getStaticStore()) + lineSep + "}" + lineSep;
        this.formatOutput += "Heap: {" + lineSep + this.formatHeap(s.getHeap()) + lineSep + "}" + lineSep;
        if (!s.isStuck()) {
        	this.formatOutput += "Stack: {" + lineSep + this.formatStack(s.getStack()) + lineSep + "}";
        }
	}

	/**
	 * Must be invoked at the end of a <code>emit()</code> implementation.
	 */
	protected void cleanup() {
		this.formatOutput = "";
	}
	
	/*private String formatBytecode(byte[] byteCode) {
    	String tmpCode = "";
        for (int i = 0; i < byteCode.length; i++) {
            tmpCode = tmpCode + " " + byteCode[i];
        }
        return tmpCode;
	}*/

	protected String formatStack(ThreadStack s) {
    	this.increaseIndent();
        String tmpRet = this.indentTxt;
        int j = 0;
        for (Frame f : s.frames()) {
            tmpRet += "Frame[" + j + "]: {" + lineSep + this.formatFrame(f) + lineSep;
            tmpRet += this.indentTxt + "}";
            if (j < s.frames().size() - 1) 
            	tmpRet += lineSep + this.indentTxt;
            j++;
        }
        this.decreaseIndent();
        return(tmpRet);
	}

	protected String formatFrame(Frame f) {
		this.increaseIndent();
		String tmp = "";
        tmp += this.indentTxt + "Method signature: " + f.getName().toString() + lineSep;
        if (this.srcPath == null) {
	        tmp += this.indentTxt + "Program counter: " + f.getProgramCounter() + lineSep;
	        tmp += this.indentTxt + "Program counter after return: "; 
	        tmp += ((f.getReturnProgramCounter() == Frame.UNKNOWN_PC) ? "<UNKNOWN>" : f.getReturnProgramCounter()) + lineSep;
	        tmp += this.indentTxt + "Next bytecode: " + this.bfmt.select(f.getInstruction()).format(f) + lineSep; 
        } else {
            tmp += this.indentTxt + "Source line: " + this.sourceLine(f) + lineSep;
        }
        tmp += this.indentTxt + "Operand Stack: {"+ lineSep + this.formatOperandStack(f.getOperandStack()) + lineSep + this.indentTxt + "}" + lineSep;
        tmp += this.indentTxt + "Local Variables: {" + lineSep + this.formatLocalVariables(f.getLocalVariableTable(), f.getProgramCounter()) + lineSep + this.indentTxt + "}";
        this.decreaseIndent();
        return tmp;
	}
	
	protected String formatOperandStack(OperandStack s) {
		this.increaseIndent();
        String tmp = this.indentTxt;
        int i = 0;
        for (Value v : s.values()) {
            tmp += "Operand[" + i +"]: " + this.formatValue(v);
            if (i < s.values().size() - 1)  {
            	tmp += lineSep + this.indentTxt;
            }
            i++;
        }
        this.decreaseIndent();
        return(tmp);
	}
	
	protected String formatLocalVariables(LocalVariablesArea l, int programCounter) {
		this.increaseIndent();
        String tmp = this.indentTxt;
        boolean isFirst = true;
       	for (int i : l.slots()) {
            if (!isFirst) {
            	tmp += lineSep + this.indentTxt;
            }
            tmp += "Variable[" + i + "]: "+ this.formatVariable(l.buildVar(i, programCounter));
            isFirst = false;
        }
        this.decreaseIndent();
        return tmp;
	}
	
	protected String formatPathCondition(PathCondition pc) {
		String expression = "";
		String where = "";
		this.increaseIndent();
    	boolean doneFirstExpression = false;
    	boolean doneFirstWhere = false;
		HashSet<String> doneSymbols = new HashSet<String>();
		for (PathCondition.Clause c : pc.getClauses()) {
   			expression += (doneFirstExpression ? (" &&" + lineSep) : "") + this.indentTxt;
   			doneFirstExpression = true;
    		if (c.hasCondition()) {
    			expression += this.formatValue(c.getCondition());
    			String expressionFormatted = this.formatExpressionForPathCondition(c.getCondition(), doneSymbols);
        		if (expressionFormatted.equals("")) {
        			//does nothing
        		} else {
        			where += (doneFirstWhere ? (" &&" + lineSep) : "") + this.indentTxt + expressionFormatted;
        			doneFirstWhere = true;
        		}
    		} else if (c.hasRefinement()) {
    			ReferenceSymbolic ref = c.getRefinement(); 
    			expression += ref.toString() + " == ";
    			if (ref.isNull()) {
    				expression += "null";
    			} else {
    				expression += "Object[" + ref.getHeapPosition() + "]";
    			}
    			String referenceFormatted = this.formatReferenceForPathCondition(ref, doneSymbols); 
        		if (referenceFormatted.equals("")) {
        			//does nothing
        		} else {
        			where += (doneFirstWhere ? (" &&" + lineSep ) : "") + this.indentTxt + referenceFormatted;
        			doneFirstWhere = true;
        		}
    		} else {
    			expression += "access(" + c.getClassName() +")";
    		}
		
		}
		this.decreaseIndent();
		return (expression.equals("") ? "" : (lineSep + this.indentTxt + expression)) + (where.equals("") ? "" : (lineSep + this.indentTxt + "where:" + lineSep + this.indentTxt + where));
	}
	
	protected String formatReferenceForPathCondition(ReferenceSymbolic r, HashSet<String> done) {
		if (done.contains(r.toString())) {
			return "";
		} else {
			return r.toString() + " == " + r.getOrigin();
		}
	}
	
	protected String formatExpressionForPathCondition(Expression e, HashSet<String> done) {
		Primitive firstOp = e.getFirstOperand();
		Primitive secondOp = e.getSecondOperand();
		String retVal = "";
        if (firstOp != null) {
        	retVal = this.formatPrimitiveForPathCondition(firstOp, done);
        }
        String secondOpFormatted = this.formatPrimitiveForPathCondition(secondOp, done);
        if (retVal.equals("") || secondOpFormatted.equals("")) {
    		//does nothing
        } else {
       		retVal += " &&" + lineSep + this.indentTxt;
        }
        retVal += secondOpFormatted;
        return retVal;
	}
	
	protected String formatPrimitiveForPathCondition(Primitive p, HashSet<String> done) {
		if (p instanceof Expression) {
			return this.formatExpressionForPathCondition((Expression) p, done);
		} else if (p instanceof PrimitiveSymbolic) {
			if (done.contains(p.toString())) {
				return "";
			} else {
				done.add(p.toString());
				return p.toString() + " == " + ((PrimitiveSymbolic) p).getOrigin();
			}
		} else if (p instanceof FunctionApplication) {
			return this.formatFunctionApplicationForPathCondition((FunctionApplication) p, done);
		} else { //(p instanceof SEAny || p instanceof Simplex)
			return "";
		}
	}

	public String formatFunctionApplicationForPathCondition(FunctionApplication a, HashSet<String> done) {
		String retVal = "";
		boolean first = true;
		for (Primitive p : a.getArgs()) {
			String argFormatted = this.formatPrimitiveForPathCondition(p, done);
			if (argFormatted.equals("")) {
				//does nothing
			} else { 
				retVal += (first ? "" : " &&" + lineSep + this.indentTxt) + argFormatted;
				first = false;
			}
		}
		retVal += ")";
		return retVal;
	}

	public String formatHeap(Heap h) {
		this.increaseIndent();
        String tmpRet = this.indentTxt;
        int j = 0;
        for (Map.Entry<Integer, Objekt> e : h.getObjects().entrySet()) {
        	Objekt o = e.getValue();
            tmpRet += "Object[" + e.getKey() + "]: " + "{" + lineSep;
            tmpRet += this.formatObject(o) + lineSep;
            tmpRet += this.indentTxt + "}";
            if (j < h.getObjects().size() - 1) {
            	tmpRet += lineSep + this.indentTxt;
            }
            j++;
        }
        this.decreaseIndent();
        return(tmpRet);
	}

	public String formatStaticStore(StaticMethodArea a) {
		this.increaseIndent();
        String tmpRet = this.indentTxt;
        int j = 0;
        for (Map.Entry<String, Klass> ee : a.getObjects().entrySet()) {
        	Klass k = ee.getValue();
        	String c = ee.getKey();
            tmpRet += "Class[" + c + "]: " + "{" + lineSep;
            tmpRet += this.formatObject(k) + lineSep;
            tmpRet += this.indentTxt + "}";
            if (j < a.getObjects().entrySet().size() - 1) {
            	tmpRet += lineSep + this.indentTxt;
            }
            j++;
        }
        this.decreaseIndent();
        return(tmpRet);
	}

	public String formatObject(Objekt o) {
		//explicit dispatch on type
		if (o instanceof Array) {
			Array a = (Array) o;
			this.increaseIndent();
			String str = this.indentTxt + "Type: " + a.getType();
			str += lineSep + this.indentTxt + "Length: " + a.getLength(); 
			str += lineSep + this.indentTxt + "Items: {";
			if (!a.isNormalized()) {
				this.increaseIndent();
			}
			boolean isChar = (a.getContentType() == Type.CHAR);
			boolean printAsString = isChar;
			if (printAsString) {
				for (Array.ArrayEntryValue e : a.values()) {
					printAsString = printAsString && (e.getValue() instanceof Simplex);
				}
			}
			if (printAsString) {
				str += "\"";
			}
			boolean skipComma = true;
			for (Array.ArrayEntryValue e : a.values()) {
				if (a.isNormalized()) {
					str += (skipComma ? "" : ", ") + this.formatArrayEntry(e, false);
					if (!printAsString) {
						skipComma = false;
					}
				} else {
					str += lineSep + this.indentTxt + this.formatArrayEntry(e, true);
				}
			}
			if (printAsString) {
				str += "\"";
			}
			if (!a.isNormalized()) {
				str += lineSep;
				this.decreaseIndent();
				str += this.indentTxt;
			}
			str += "}";
	        this.decreaseIndent();
			return str;
		} else if (o instanceof Instance) {
			Instance i = (Instance) o;
			this.increaseIndent();
	        String tmp = this.indentTxt + "Class: " + i.getType();
	        int z = 0;
	        for (Map.Entry<String, Variable> e : i.fields().entrySet()) {
	            tmp +=  lineSep + this.indentTxt + "Field[" + z + "]: " + formatVariable(e.getValue());
	            z++;
	        }
	        this.decreaseIndent();
	        return tmp;
		} else {
			return o.toString();
		}
	}
	
	public String formatArrayEntry(Array.ArrayEntryValue e, boolean showExpression) {
		String exp;
		String val;
		if (e.getExpression() == null) {
			exp = "true";
		} else {
			exp = this.formatValue(e.getExpression());
		}
		if (e.getValue() == null) {
			val = "<UNKNOWN>";
		} else {
			val = this.formatValue(e.getValue());
		}
		return (showExpression ? (exp + " -> ") : "") + val;	
	}
	
	
	public String formatVariable(Variable v) {
        String tmp;
        Value val = v.getValue(); 
        if (val == null) {
            tmp = "ERROR: no value has been assigned to this variable.";
        } else {
            tmp = this.formatValue(val);
        }
        return ("Name: " + v.getName() + ", Type: " + v.getType() + ", Value: " + tmp);
	}
	
	public String formatValue(Value val) {
        String tmp = val.toString();
        if (val instanceof ReferenceSymbolic) {
        	ReferenceSymbolic ref = (ReferenceSymbolic) val;
        	if (ref.isInitialized()) {
        		if (ref.isNull()) {
        			tmp += " == null";
        		} else {
        			tmp += " == Object[" + ref.getHeapPosition() + "]";
        		}
        	}
        }
        return tmp;
	}

	private void increaseIndent() {
		this.indentTxt += this.indent;
	}
	
	private void decreaseIndent() {
		this.indentTxt = this.indentTxt.substring(1);
	}	
	
	private String sourceLine(Frame f) {
    	String retVal = "";
    	int sourceRow = f.getSourceRow();
    	if (sourceRow == -1) { 
    		retVal += "<UNKNOWN>";
    	} else { 
    		retVal += "(" + sourceRow + "): ";
    		String row = Util.getSrcFileRow(f.getName().getClassName(), this.srcPath, this.pathSep, sourceRow);
    		retVal += (row == null ? "<UNKNOWN>" : row);
    	}
    	return retVal;
	}
}
