package muse.apps.mc;

import java.util.*;

public class BA {
	HashSet<State> states;
    HashMap<String, State> stateNames;  //associates every state name with its state
    State initialState;
    HashSet<State> acceptingStates;
    HashSet<State> currentStates;
    Binding binds;

    // Stack allows us to save current states of BA and thus supports backtracking.
    Stack<HashSet<State>> stateStack = new Stack<HashSet<State>>(); 
    Stack<Binding> bindingStack = new Stack<Binding>(); 

/** Constructor - creates an empty Buchi automaton. */
protected BA() {
    states = new HashSet<State>();
    stateNames = new HashMap<String, State>();
    initialState = null;
    acceptingStates = new HashSet<State>();
    currentStates = new HashSet<State>();
    binds = new Binding();
} // BA()

/** Allows us to build an automaton incrementaly by adding particular states. */
protected void addState(State s) {
	if (s.isInit()) {
		if (initialState != null) {
			states.remove(initialState);
		}
		initialState = s;
	}
	if (s.isAccept()) {
		acceptingStates.add(s);
	}
	states.add(s);
	for (StateName name : s.name) {
		stateNames.put(name.toString(), s);
	}
} // addState

/**
 * Initialize the automaton - set the initial, accepting and current states.
 */
public void initialize() {
	currentStates.add(initialState);
}

public boolean isInitialized() {
	return !(currentStates.isEmpty());
}

/**
 * @return The set of all atomic propositions used in this Buchi automaton.
 */
public HashSet<String> getAtomicPropositions() {
	HashSet<String> propositions = new HashSet<String>(); 
    for (State s : states)
    	s.getAtomicPropositions(propositions);
	return propositions;
}

/**
 * Perform a step of the automaton under current binding.
 */
public boolean doStep(){
    boolean status = false;
    // Check all current states.
    HashSet<StateName> newStateNames = new HashSet<StateName>();
    for (State s : currentStates) {
        status = s.doStep(binds, newStateNames) || status;
    }

    HashSet<State> newStates = new HashSet<State>();
    for (StateName sn : newStateNames) {
    	newStates.add(getState(sn));
    }
    	
    currentStates = newStates;
    return status;
}

public State getState(StateName sn) {
	return stateNames.get(sn.toString());
}

/**
 * Return true if the set of current states is empty.
 */
public boolean emptyCurrentStates(){
    return currentStates.isEmpty();
}

/**
 * Return true if the set of current states contains at least one accepting state.
 */
public boolean isInAcceptingState() {
	boolean accepts = false;
    Iterator<State> i = currentStates.iterator();
    // Current states.
    while ( i.hasNext() && !accepts ) {
    	accepts = accepts || acceptingStates.contains(i.next());
    }
	return accepts;
}

/**
 * Safes current states and bindings for future backtrack.
 */
public void pushState() {
	HashSet<State> clone = new HashSet<State>(currentStates);
	stateStack.push(clone);
	bindingStack.push(binds.clone());
	return;
}

/**
 * Removes states from the top of the stack and set them as current states.
 */
public void popState() {
	if ( !stateStack.isEmpty() ) {
        currentStates = stateStack.pop();
	} else {
		System.err.println("BA.java: popState() is not possible because stateStack is empty.");
	}
	if ( !bindingStack.isEmpty() ) {
		binds = bindingStack.pop();
	} else {
		System.err.println("BA.java: popState() is not possible because bindingStack is empty.");
	}
	return;
}

/**
 * Sets states from the top of the stack as current states.
 */
public void peekState() {
	if ( !stateStack.isEmpty() ) {
		currentStates = new HashSet<State>(stateStack.peek());
	} else {
		System.err.println("BA.java: peekState() is not possible because stateStack is empty.");
	}
	if ( !bindingStack.isEmpty() ) {
		binds = bindingStack.peek().clone();
	} else {
		System.err.println("BA.java: peekState() is not possible because bindingStack is empty.");
	}
	return;
}

/**
 * Returns this automaton in the never claim view.
 */
public String toString() {
    StringBuffer sb = new StringBuffer();
    sb.append("States and transitions:");
    for ( State s : states ) {
       sb.append("\n" + s.toString());
       for ( Rule r : s.rules )
           sb.append("\n    "+ "( "+r.condition.toString()+" ) -> "+getState(r.nextState).toString());
    }
                                                               // Initial state.
    if (initialState!=null)
        sb.append("\nInitial state:     "+initialState.name.toString());
                                                            // Accepting states.
    if ( acceptingStates.size() >0 ) {
        sb.append("\nAccepting states:  ");
        for ( State s : acceptingStates )
            sb.append(s.name.toString()+"  ");
    }
                                                              // Current states.
    if ( currentStates.size() >0 ) {
        sb.append("\nCurrent states:    ");
        for ( State s : currentStates )
            sb.append(s.name.toString()+"  ");
    }
    sb.append("\n");
    return sb.toString();
}

/**
 * Returns string representation of current states.
 */ 
public String states2String() {
    StringBuffer sb = new StringBuffer("Current states ("+currentStates.size()+"):  ");
    Iterator<?> i = currentStates.iterator();
                                                              // Current states.
    while ( i.hasNext() ) {
        sb.append(i.next().toString());
        sb.append("  ");
    }
    return sb.toString();
}

} // class BA
