package muse.apps.mc;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

import muse.MUSE_INFO;
import muse.algo.CannotManageStateException;
import muse.algo.PartialReferenceResolutionException;
import muse.bc.ClassFileFactoryJavassist;
import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;
import muse.dec.DecisionProcedureSicstus;
import muse.dec.DecisionQuitException;
import muse.jvm.CannotBacktrackException;
import muse.jvm.Engine;
import muse.jvm.EngineStuckException;
import muse.jvm.ExecutionObserver;
import muse.jvm.InitializationException;
import muse.jvm.NonexistingObservedVariablesException;
import muse.mem.CannotRefineException;
import muse.mem.Simplex;
import muse.mem.State;
import muse.mem.Value;
import muse.tree.StateTree.BranchPoint;


public class Check {
	private static final String pathSep = System.getProperty("file.separator");
	private static final String classpathSep = ";";
	private static final String signatureSep = ":";
	private static final String substMapSep = "=";
	private static final String originSep = "@";

	public static class Parameters {
		int heapScope = 100; // maximum size of heap
		String methodSignature = null;
		String defaultPath = ".";
		String automatonClassName = null;
		String substitutionMapFname = null;
		String classInitFname = null;
		String sicstusServerAddress = "127.0.0.1";
		int sicstusServerPort = 8888;
		String executionClasspath = "." + pathSep;
		int verbLevel = 0;
		String outName = null;
		String errName = null;
	}

	private enum TVL { UNKNOWN, TRUE, FALSE };

	private Parameters p = null;
	private PrintStream out = System.out;
	private PrintStream err = System.err;
	private Engine engine = null;
	private TVL[] atomicVariables;
	private String[] atomicPropositions;
	private int atomicVariablesSize = 0;
	private Stack<TVL[]> apStack = new Stack<TVL[]>();
	private Stack<BranchPoint> btStack = new Stack<BranchPoint>();

	/** 
	 * Creates a Buchi automaton from a class given by the <code>className</code>. 
	 * 
	 * @param className determine a class describing a Buchi automaton construction.
	 * @return Buchi automaton
	 */
	private BA createBA() {
		BA object = null;
		final String automatonPath = this.p.defaultPath;
		try {
			ClassLoader l = new ClassLoader() {
				@Override
				public Class<?> findClass(String name) {
					String fullPathName = automatonPath + name + ".class";
					File f = new File(fullPathName);
					FileInputStream fs = null;
					try {
						fs = new FileInputStream(f);
						byte[] b = new byte[(int) f.length()];
						fs.read(b);
						return defineClass(name, b, 0, b.length);
					} catch (ClassFormatError e) {
						Check.this.err.println(e);
						return null;
					} catch (IndexOutOfBoundsException e) {
						Check.this.err.println(e);
						return null;
					} catch (SecurityException e) {
						Check.this.err.println(e);
						return null;
					} catch (FileNotFoundException e) {
						Check.this.err.println(e);
						return null;
					} catch (IOException e) {
						Check.this.err.println(e);
						return null;
					} finally {
						try {
							if (fs != null) {
								fs.close();
							}
						} catch (IOException e) {
							//does nothing
						}
					}
				}
			};
			Class<?> classDefinition = l.loadClass(p.automatonClassName); 
			object = (BA) classDefinition.newInstance();
		} catch (InstantiationException e) {
			out.println(e);
		} catch (IllegalAccessException e) {
			out.println(e);
		} catch (ClassNotFoundException e) {
			out.println(e);
		}
		return object;
	}

	/**
	 * Prepares the symbolic execution observers.
	 * 
	 * @param propositions the atomic propositions for the observers.
	 */
	private void setObservers (HashSet<String> propositions) {
		// Creates a field of boolean variables corresponding to atomic propositions.
		atomicVariablesSize = propositions.size();
		atomicVariables = new TVL[atomicVariablesSize];
		atomicPropositions = new String[atomicVariablesSize];
		String[] varNames = new String[atomicVariablesSize];
		ExecutionObserver[] obs = new ExecutionObserver[atomicVariablesSize];
		int i = 0;
		for (String name : propositions) {
			final int fi = i;
			atomicVariables[i] = TVL.UNKNOWN;
			atomicPropositions[i] = name;
			varNames[i] = "_" + name.toUpperCase();
			obs[i] = new ExecutionObserver() { public void update() { refreshAP(fi); } };
			i++;
		}

		// Sets the observers.
		engine.setObservers(varNames, obs);
	}

	/**
	 * Actualize a value of a variable when a corresponding atomic proposition 
	 * is changed.
	 *  
	 * @param ap index of atomic proposition/boolean variable.
	 */
	private void refreshAP(int ap) {
		Value v  = this.engine.getObservedVariableValue(ap);
		if (v instanceof Simplex) {
			Object o = ((Simplex) v).getActualValue();
			if (o instanceof Integer) {
				this.atomicVariables[ap] = ((((Integer) o).intValue() == 1) ? TVL.TRUE : TVL.FALSE);
			}
		}
	}

	/**
	 * Constructor.
	 * 
	 * @param p the <code>Parameters</code> of the model check.
	 */
	public Check(Parameters p) {
		this.p = p;
	}

	/**
	 * Creates the expression simplifier, the symbolic execution engine, 
	 * the Buchi automaton observer, and perform checking. 
	 */
	public int check() {
		if (this.p.verbLevel >= 1) {
			this.out.println("========= Welcome to the MUSE v. " + MUSE_INFO.CURRENT_VERSION + " symbolic model checker. =========\n");
		}

		// Checks presence of compulsory parameters and arguments
		if (this.p.methodSignature == null) {
			this.out.println("Check: Please specify the name of a method to check.");
			return 1;
		}
		if (this.p.automatonClassName == null) {
			this.out.println("Check: Please specify the name of a class with observing Buchi automaton.");
			return 1;
		}

		// Creates the output file
		if (this.p.outName != null) {
			try {
				File f = new File(p.defaultPath + p.outName);
				this.out = new PrintStream(f);
			} catch (Exception exc) {
				this.err.println("Check: Error during creation of output file. Continuing with default.");
				this.out = System.out; //just for safety
			}
		}

		// Creates an observing Buchi automaton.
		BA ba = createBA();
		if (ba == null) {
			this.out.println("Check: Creation of an observing Buchi automaton failed.");
			return 1;
		}

		// Prints the Buchi automaton.
		if (this.p.verbLevel >= 2) {
			this.out.println("Automaton for the property '" + this.p.automatonClassName + "':\n" + ba.toString());
		}

		// Creates a simplifier used by the engine.
		DecisionProcedure simplifier = null;
		try {
			simplifier = new DecisionProcedureSicstus(p.sicstusServerAddress, p.sicstusServerPort, null);
		} catch (DecisionException e) {
			this.err.println("Check: Sicstus simplifier not found.");
			return 1;
		}

		// Creates a symbolic execution engine.
		this.engine = new Engine(new ClassFileFactoryJavassist(), simplifier);

		// Sets the engine's class path, method signature and observers.
		this.engine.addClassPath(this.p.executionClasspath.split(classpathSep));
		String[] sig = p.methodSignature.split(signatureSep);
		if (sig.length != 3) {
			this.err.println("Check: Method signature incorrect.");
			return 1;
		}
		this.engine.setMethodName(sig[0], sig[1], sig[2]);
		this.setObservers(ba.getAtomicPropositions());

		// Sets the engine's class init.
		if (this.p.classInitFname != null) {
			BufferedReader fr = null;
			try {
				fr = new BufferedReader(new FileReader(this.p.defaultPath + this.p.classInitFname));
				String l;
				while ((l = fr.readLine()) != null) {
					engine.addNotInitializedClass(l);
				}
			} catch (FileNotFoundException e) {
				//does nothing
			} catch (IOException e) {
				//does nothing
			} finally {
				try {
					if (fr != null) {
						fr.close();
					}
				} catch (IOException e) {
					//does nothing
				}
			}
		}

		// Sets the engine's substitution map.
		{
			List<String[]> substitutionMap = new LinkedList<String[]>();

			if (this.p.substitutionMapFname != null) {
				BufferedReader fr = null;
				try {
					fr = new BufferedReader(new FileReader(this.p.defaultPath + this.p.substitutionMapFname));
					String l;
					while ((l = fr.readLine()) != null) {
						String seps = substMapSep + originSep;
						String strNonSep = "[^" + seps + "]+";
						String pattern = strNonSep + "(" + originSep + strNonSep + ")?" + substMapSep + strNonSep; 
						if (l.matches(pattern)) {
							substitutionMap.add(l.split("[" + seps + "]"));
						}
					}
				} catch (FileNotFoundException e) {
					//does nothing
				} catch (IOException e) {
					//does nothing
				} finally {
					try {
						if (fr != null) {
							fr.close();
						}
					} catch (IOException e) {
						//does nothing
					}
				}
			}
			for (String[] s : substitutionMap) {
				if (s.length == 3) {
					this.engine.setClassSubst(s[0], s[1], s[2]);
				} else { //length must be 2
					this.engine.setClassSubst(s[0], null, s[1]);
				}
			}
		}

		// Initializes the symbolic execution.
		BranchPoint btPoint = null;
		try { 
			this.engine.init();
		} catch (DecisionException e) {
			this.err.println("Check: Simplifier initialization failed.");
			return 1;
		} catch (InitializationException e) {
			this.err.println("Check: Initializer initialization failed.");
			return 1;
		} catch (NonexistingObservedVariablesException e) {
			this.err.print("Check: Engine cannot observe atomic proposition" + (e.getVariableIndices().size() > 1 ? "s " : " "));
			boolean first = true;
			for (int i : e.getVariableIndices()) {
				if (first) {
					first = false;
				} else {
					this.err.print(", ");
				}
				this.err.print(this.atomicPropositions[i]);
			}
			this.err.print(".");
			return 1;
		}
		if (this.p.verbLevel >= 2) {
			this.out.println("Initialization completed, starting verification.");
			this.out.println();
		}

		// Stores a copy of the initial state for later displaying counterexample.
		State initialState = (State) this.engine.getCurrentState().clone();

		// Performs verification.
		long startTime = System.currentTimeMillis();
		boolean counterexampleFound = false;
		boolean someTraceNotFullyExplored = false;
		String notFullyExploredBranches = "";
		try {
			while (!counterexampleFound && (this.engine.canStep() || this.engine.canBacktrack())) {
				if (this.engine.canStep()) { // we are not at the and of the trace
					// We perform a step of the symbolic executor.
					try {
						btPoint = this.engine.step();
					} catch (EngineStuckException e) {
						//does nothing, it should never happen
					} catch (CannotManageStateException e) {
						// If the engine is unable to step to the next state, we stop the current trace
						if (someTraceNotFullyExplored) { 
							notFullyExploredBranches += "\n"; 
						}
						someTraceNotFullyExplored = true;
						notFullyExploredBranches += this.engine.getCurrentState().getIdentifier() + "[" + engine.getCurrentState().getSequenceNumber() + "] (cause: engine cannot handle state)";
						this.engine.stopCurrentTrace();
					} catch (PartialReferenceResolutionException e) {
						//TODO implement some reporting; by now just prints a warning
						this.out.println("WARNING: a symbolic reference has not been fully resolved. This usually yields an underspecified analysis.");
					}

					// If we are passing a backtrack point we store values to stacks.
					if (btPoint != null) {
						if (this.p.verbLevel >= 2) {
							this.out.println("BacktrackPoint "+ btPoint + " stored into the stack.");
						}
						this.btStack.push(btPoint);
						this.apStack.push(this.atomicVariables.clone());
						ba.pushState();
					}

					// Determines whether the observing Buchi automaton must perform the initial step
					// and in the case performs it. 
					if (!ba.isInitialized()) {
						boolean automatonCanStart = true;
						for (TVL t : atomicVariables) {
							if (t == TVL.UNKNOWN) {
								automatonCanStart = false; 
								break;
							}
						}
						if (automatonCanStart) {
							ba.initialize();
						}
					}

					// Performs the step of the observing Buchi automaton.
					if (ba.isInitialized()) {
						ba.binds.clearEvaluation();
						for (int i = 0; i < atomicVariablesSize; i++) {
							ba.binds.setValue(atomicPropositions[i], ((atomicVariables[i] == TVL.TRUE) ? true : false));
						}
						if (ba.doStep()) {
							// If the Buchi automaton reaches an accept_all state we stop the exploration 
							// of current trace and return the counterexample. 
							this.engine.stopCurrentTrace();
							counterexampleFound = true;
						} else if (ba.emptyCurrentStates()) {
							// If set of current states of the Buchi automaton is empty we stop the  
							// exploration of the current trace (the automaton does not recognize 
							// a counterexample prefix). 
							this.engine.stopCurrentTrace();
						}
					}

					// If we exceed the analysis scope, we stop the current trace.
					if (this.engine.getCurrentState().getHeap().getSize() > this.p.heapScope) {
						if (someTraceNotFullyExplored) {
							notFullyExploredBranches += "\n"; 
						}
						someTraceNotFullyExplored = true;
						notFullyExploredBranches += this.engine.getCurrentState().getIdentifier() + "[" + engine.getCurrentState().getSequenceNumber() + "] (cause: scope exhausted)";
						this.engine.stopCurrentTrace();
					}

					// We produce the step-by-step feedback.
					if (p.verbLevel >= 2) {
						for (int i = 0; i < atomicVariablesSize; i++) {
							out.print(atomicPropositions[i] + " == " + atomicVariables[i] + "  ");
						}
						out.print("  BA " + ba.states2String());
						out.println("  Program current state: " + engine.getCurrentState().getIdentifier() + "[" + engine.getCurrentState().getSequenceNumber() + "]");
					}
				} else { // we are at the end of the trace
					if (ba.isInAcceptingState()) {
						// If observer is in an accepting state the property does not hold.
						counterexampleFound = true;
					} else {
						// We return to the next not explored branch (backtracking).
						try {
							btPoint = this.engine.backtrack();
						} catch (CannotBacktrackException e) {
							//does nothing, this should never happen
						} 
						// We remove obsolete values from the stacks. 
						while (!btPoint.equals(this.btStack.peek()) ) {
							if (this.p.verbLevel >= 2) {
								this.out.println("Backtrack point " + this.btStack.peek() + " removed from the stack.");
							}
							this.btStack.pop();
							this.apStack.pop();
							ba.popState();
						}
						// We restore values from the stack valid for current backtrack point.
						if (p.verbLevel >= 2) {
							out.println("Returned to backtrack point " + btStack.peek() + ".");
						}
						this.atomicVariables = this.apStack.peek().clone();
						ba.peekState();
					}
				}
			}

			// Prints the verification result.
			if (p.verbLevel >= 1) {
				if (p.verbLevel >= 2) {
					this.out.println();
				}
				this.out.print("Analysis complete, analyzed " + this.engine.getAnalyzedStates() + " symbolic states, elapsed time: ");
				long elapsedTime = System.currentTimeMillis() - startTime;
				if (elapsedTime < 1000) {
					this.out.print(elapsedTime + " msec ");
				} else if (elapsedTime < 60000) {
					this.out.print( ( elapsedTime/1000 ) + " sec " + ( elapsedTime % 1000 ) + " msec ");
				} else if (elapsedTime < 3600000){
					this.out.print( ( elapsedTime/60000 ) + " min " + (( elapsedTime % 60000 ) / 1000) + " sec " + (( elapsedTime % 60000 ) % 1000) + " msec ");
				} else {
					this.out.print( ( elapsedTime/3600000 ) + " hr " + (( elapsedTime % 3600000 ) / 60000) + " min " + ((( elapsedTime % 3600000) % 60000 ) / 1000) + " sec " + (( ( elapsedTime % 3600000) % 60000 ) % 1000) + " msec ");
				}
				long elapsedTimeSimplifier = this.engine.getTimeSimplifier();
				double tmp = elapsedTimeSimplifier;
				double timeSimplifierRatio = (tmp / elapsedTime);
				String timeSimplifierPercent = new DecimalFormat("##.##%").format(timeSimplifierRatio);
				this.out.print("(" + timeSimplifierPercent + " in simplifier), ");
				this.out.println("average speed: " + this.engine.getAnalyzedStates() * 1000 / elapsedTime + " states/sec.");
				this.out.println();
			}

			if (this.p.verbLevel >= 1) {
				this.out.print("The property '" + this.p.automatonClassName + "' ");
			}

			if (counterexampleFound) {
				if (p.verbLevel >= 1) {
					// TODO Print the whole trace which violates the property.
					out.println("does not hold; the violation witness follows.");
					State finalState = engine.getCurrentState();
					try {
						initialState.refine(finalState.getPathCondition());
						this.out.println();
						this.out.println("Initial state:");
						this.out.println(initialState);
					} catch (CannotRefineException e) {
						this.err.println("Check: unexpected error while refining initial state.");
					}
					this.out.println();
					this.out.println("Final state:");
					this.out.println(finalState);
				} else {
					this.out.println("No");
				}
			} else if (someTraceNotFullyExplored) {
				if (this.p.verbLevel >= 1) {
					this.out.println("cannot be decided along branches:");
					this.out.println(notFullyExploredBranches);
					this.out.println("No violating final state found along other branches.");
				} else {
					this.out.println("Inconclusive");
				}
			} else {
				if (this.p.verbLevel >= 1) {
					this.out.println("holds.");
				} else {
					this.out.println("Yes");
				}
			}

			//quits
		} catch (DecisionException e) {
			this.err.println("Check: Simplifier error.");
			return 1;
		} finally {
			try {
				this.engine.quit();
			} catch (DecisionQuitException e) {
				this.err.println("Check: Simplifier error during quit.");
				return 1;
			}				
		}
		return 0;
	}

	//Here starts the command line interface code.
	private static final String SWITCH_CLASSPATH = "cp";
	private static final String SWITCH_HELP = "h";
	private static final String SWITCH_HELP_2 = "?";
	private static final String SWITCH_OUTPUT = "o";
	private static final String SWITCH_DEFAULT_PATH = "fp";
	private static final String SWITCH_PROPERTY = "p";
	private static final String SWITCH_SCOPE = "sc";
	private static final String SWITCH_SERVER = "se";
	private static final String SWITCH_SUBST_MAP = "sm";
	private static final String SWITCH_CLASS_INIT = "ci";
	private static final String SWITCH_VERBOSE = "v";
	private static final String SWITCH_RESERVED = "R";
	private static final String SWITCH_RESERVED_NAME = SWITCH_RESERVED + "n";
	private static final char networkAddressSep = ':';
	private static final char commandLineSwitch = '/';

	//parameters exclusively used by the CLI
	private static boolean help = false;
	private static String reservedName = "java " + Check.class.getCanonicalName();
	private static boolean defaultName = true;

	public static void main(String[] args) {
		Parameters p = parseParameters(args);
		if (help) {
			System.out.print(help());
			System.exit(0);
		}
		Check c = new Check(p);
		System.exit(c.check());
	}

	private static String help() {
		String retVal = "" +
		//12345678901234567890123456789012345678901234567890123456789012345678901234567890
		"Usage: \n\n" +
		"  " + reservedName + " " + commandLineSwitch + SWITCH_PROPERTY + " <property>\n" + 
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_DEFAULT_PATH + " <pathname> " + "] \n" + 
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_CLASSPATH + " <classpath> " + "] \n" + 
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_OUTPUT + " <fnameo> " + "] \n" + 
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_SCOPE + " <mscope> " + "] \n" + 
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_SUBST_MAP + " <fnamem> " + "] \n" + 
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_CLASS_INIT + " <fnamec> " + "] \n" + 
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_SERVER + " <host>" + networkAddressSep + "<port> " + "] \n" +
		repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_VERBOSE + " <vlevel>" + "] \n" +
		repeat(' ', reservedName.length() + 3) + "<className>" + signatureSep + "<paramSignature>" + signatureSep + "<methodName>\n\n" +
		"  " + reservedName + " " + commandLineSwitch + SWITCH_HELP + " | " + commandLineSwitch + SWITCH_HELP_2 + "\n\n" + 
		"Where:\n\n" +
		"<property>       is the name of the Java class implementing the property as a\n" +
		"                 Buchi automaton;\n" +
		"<pathname>       is the default path for <property>, <fnamem>, <fnamec> and\n" +
		"                 <fnameo> (default: .);\n" +
		"<classpath>      is a list of paths separated by the character '" + classpathSep +"', which\n" +
		"                 will be used as the classpath of the symbolic execution\n" + 
		"                 (default: .);\n" +
		"<fnameo>         is the name of the output file (default: stdout);\n" +
		"<mscope>         is an integer value indicating the maximum scope of the\n" + 
		"                 execution, i.e., the maximum allowed heap size, reached\n" + 
		"                 which a branch is abandoned (default: 100);\n" +
		"<fnamem>         is the name of the file containing the substitution map\n" + 
		"                 specifying how a symbolic reference of a given type\n" +
		"                 is initialized with fresh concrete objects\n" +
		"                 (default: no substitution map, all symbolic references\n" +
		"                 will be initialized by instances of their types, or no fresh\n" +
		"                 instances if the type is abstract);\n" +
		"<fnamec>         is the name of the file containing the names of the classes\n" +
		"                 that are not to be considered initialized at the start of the\n" +
		"                 symbolic execution\n" +
		"                 (default: no file, all classes are considered initialized);\n" +
		"<host>" + networkAddressSep + "<port>    is the network address of the Sicstus server\n" +
		"                 (default: localhost" + networkAddressSep + "8888);\n" +
		"<vlevel>         is the level of verbosity (from 0 to 2, default: 0);\n" + 
		"<className>      is the name of the class containing the method to be checked;\n" +
		"<paramSignature> describes the types of the method's parameters and return\n" + 
		"                 values;\n" +
		"<methodName>     is the name of the method to be checked.\n";
		if (defaultName)
			retVal += "\n" +
			//12345678901234567890123456789012345678901234567890123456789012345678901234567890
			"Example: we assume that the model checker is deployed in c:\\mc\\mc.jar, the\n" +
			"property classfile is c:\\prop\\Safe.class, the standard Java libraries are in\n" + 
			"c:\\jre\\bin\\rt.jar, the Sicstus server is listening at port 8888 of server\n" + 
			"131.151.20.44, the method to be verified is\n\n" + 
			"  boolean myPackage.MyClass.myMethod(Object o)\n\n" + 
			"the software under verification is deployed in c:\\proj\\proj.jar and that\n" + 
			"we want to abort search as the current state's heap size exceeds 500 objects.\n" + 
			"Then, the analysis is run by invoking:\n\n" + 
			"  java -cp c:\\mc\\mc.jar " + Check.class.getCanonicalName() + " "+ commandLineSwitch + SWITCH_DEFAULT_PATH + " c:\\prop\n" +  
			"       " + commandLineSwitch + SWITCH_PROPERTY + " Safe " + commandLineSwitch + SWITCH_OUTPUT + " out.txt " + commandLineSwitch + SWITCH_SERVER + " 131.151.20.44" + 
			networkAddressSep + "8888\n" + 
			"       " + commandLineSwitch + SWITCH_CLASSPATH + " c:\\proj\\proj.jar" + classpathSep + 
			"c:\\jre\\bin\\rt.jar " + commandLineSwitch + SWITCH_SCOPE + " 500\n" + 
			"       myPackage/MyClass" + signatureSep + "(Ljava/lang/Object;)Z" + signatureSep + "myMethod\n";
		return retVal;
	}

	/**
	 * Analyzes command line arguments and creates a <code>Parameters</code> 
	 * instance.
	 * 
	 * @param args command line arguments.
	 * @return a <code>Parameters</code> instance set accordingly with 
	 * the command line arguments, or <code>null</code> if the "help" 
	 * switch is present. 
	 */
	private static Parameters parseParameters(String[] args) {
		Parameters retVal = new Parameters();
		boolean classpath = false;
		boolean output = false;
		boolean property = false;
		boolean property_path = false;
		boolean scope = false;
		boolean server = false;
		boolean reserved_name = false;
		boolean subst_map = false;
		boolean class_init = false;
		boolean verbose = false;

		for (String s : args) {
			if (classpath) {
				retVal.executionClasspath = s;
				classpath = false;
			} else if (output) {
				retVal.outName = s;
				output = false;
			} else if (property) {
				retVal.automatonClassName = s;
				property = false;
			} else if (property_path) {
				if (s.length() < pathSep.length() || (!pathSep.equals(s.substring(s.length() - pathSep.length()))))
					s += pathSep;
				retVal.defaultPath = s;
				property_path = false;
			} else if (scope) {
				try {
					retVal.heapScope = Integer.parseInt(s);
				} catch (NumberFormatException exc) {
					//does nothing
				}
				scope = false;
			} else if (server) {
				try {
					String[] ss = s.split("" + networkAddressSep);
					retVal.sicstusServerAddress = ss[0];
					retVal.sicstusServerPort = Integer.parseInt(ss[1]);
				} catch (NumberFormatException exc) {
					//does nothing
				} catch (IndexOutOfBoundsException exc) {
					//does nothing
				}
				server = false;
			} else if (subst_map) {
				retVal.substitutionMapFname = s;
				subst_map = false;
			} else if (class_init) {
				retVal.classInitFname = s;
				class_init = false;
			} else if (verbose) {
				retVal.verbLevel = Integer.parseInt(s);
				verbose = false;
			} else if (reserved_name) {
				reservedName = s;
				defaultName = false;
				reserved_name = false;
			} else if (s.equals(commandLineSwitch + SWITCH_CLASSPATH)) {
				classpath = true;
			} else if (s.equals(commandLineSwitch + SWITCH_HELP) ||
					s.equals(commandLineSwitch + SWITCH_HELP_2)) {
				help = true;
			} else if (s.equals(commandLineSwitch + SWITCH_OUTPUT)) {
				output = true;
			} else if (s.equals(commandLineSwitch + SWITCH_PROPERTY)) {
				property = true;
			} else if (s.equals(commandLineSwitch + SWITCH_DEFAULT_PATH)) {
				property_path = true;
			} else if (s.equals(commandLineSwitch + SWITCH_SCOPE)) {
				scope = true;
			} else if (s.equals(commandLineSwitch + SWITCH_SERVER)) {
				server = true;
			} else if (s.equals(commandLineSwitch + SWITCH_SUBST_MAP)) {
				subst_map = true;
			} else if (s.equals(commandLineSwitch + SWITCH_CLASS_INIT)) {
				class_init = true;
			} else if (s.equals(commandLineSwitch + SWITCH_VERBOSE)) {
				verbose = true;
			} else if (s.equals(commandLineSwitch + SWITCH_RESERVED_NAME)) {
				reserved_name = true;
			} else {
				retVal.methodSignature = s;
			}
		}
		return retVal;
	}

	/**
	 * Builds a string by repeating a character. 
	 * @param c the <code>char</code> to be repeated.
	 * @param ln the length of the string.
	 * @return
	 */
	private static String repeat(char c, int ln) {
		char[] cc = new char[ln];
		for (int i = 0; i < ln; i++) {
			cc[i] = c;
		}
		return new String(cc);
	}
}
