package muse.apps.run;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.text.DecimalFormat;
import java.util.LinkedList;
import java.util.List;

import muse.MUSE_INFO;
import muse.algo.CannotInvokeNativeException;
import muse.algo.PartialReferenceResolutionException;
import muse.algo.UndefInstructionException;
import muse.apps.Runner;
import muse.apps.StateFormatter;
import muse.apps.StateFormatterGraphviz;
import muse.apps.StateFormatterText;
import muse.apps.StateFormatterTextTrace;
import muse.bc.ClassFileFactoryJavassist;
import muse.dec.DecisionBacktrackException;
import muse.dec.DecisionException;
import muse.dec.DecisionInitException;
import muse.dec.DecisionProcedure;
import muse.dec.DecisionProcedureAlwSat;
import muse.dec.DecisionProcedureCVC3;
import muse.dec.DecisionProcedureConsole;
import muse.dec.DecisionProcedureSicstus;
import muse.dec.DecisionQuitException;
import muse.dec.DecisionRunningException;
import muse.jvm.Engine;
import muse.jvm.EngineStuckException;
import muse.jvm.ExecutionObserver;
import muse.jvm.InitializationException;
import muse.jvm.NonexistingObservedVariablesException;
import muse.mem.CannotRefineException;
import muse.mem.State;
import muse.tree.StateTree.BranchPoint;


/**
 * The <code>Run</code> class implements an interactive runner, 
 * which allows to symbolically execute a method. At each step 
 * of the symbolic execution <code>Run</code> prints: 
 * 
 * <ul>
 * <li> the identifier of the backtrack point created by the 
 * step, whenever such a backtrack point is created;</li> 
 * <li> the current execution state in textual form.</li>
 * </ul>
 * 
 * <code>Run</code> executes a trace until it reaches a stuck 
 * state; then, if it can backtrack to some state, 
 * it asks the user via console whether it should backtrack. 
 * <code>Run</code> tries to establish a connection with a 
 * decision procedure to prune unfeasible branches; in the case 
 * it fails to connect to a valid decision procedure it queries 
 * via console the user whenever a branch point is encountered.
 * A <code>Run</code> session can be automatically dumped to file 
 * for offline inspection.  
 * 
 * @author Pietro Braione
 */
public class Run {
    /** Constant for default (localhost) IP address. */
    private static final String LOCALHOST_IP = "127.0.0.1";

    /** Constant for default port. */
    private static final int DEFAULT_PORT = 8888;
    
    /** Constant for path separator. */
	private static final String PATH_SEP = System.getProperty("file.separator");
	
	/** Constant for path list separator. */
	private static final String MULTI_PATH_SEP = ";";
	
	/** Constant for signature separator. */
	private static final String SIGNATURE_SEP = ":";
	
	/** Constant for substitution map file separator. */
	private static final String SUBST_MAP_SEP = "=";

	/** Constant for origin separator. */
	private static final String ORIGIN_SEP = "@";
	
	/**
     * Class encapsulating the protocol of the runner's 
     * parameters. These are:
     * 
     * <ul>
     * <li>The output file, where the output produced to 
     * stdout and stderr is copied (by default, no output 
     * file is created and the output is just sent to 
     * stdout and stderr);</li>
     * <li>The text mode, indicating how the rows on the 
     * output file must be terminated, whether in a Windows 
     * (i.e., with characters CR LF) or in a UNIX (i.e., 
     * just with LF) style, or in the default style of the 
     * current platform; by default, the platform's text 
     * mode is used. This parameter is relevant only when 
     * an output file is specified;</li>
     * <li>The interaction mode, indicating the amount of 
     * required user interaction; the possible alternatives 
     * are:
     * <ul>
     * <li>prompts user at all computation steps (just pause), 
     * at backtrack points (queries whether it should backtrack 
     * or not) and in case the Sicstus simplifier is unavailable, 
     * also at undecidable branches (queries which 
     * branches should be explored);</li>
     * <li>prompts at backtrack points and undecidable branches;
     * this is the default behaviour;</li> 
     * <li>prompts only at undecidable branches and backtracks 
     * at all backtrack points;</li>
     * <li>never prompts the user and explores all branches of the 
     * conditions it is unable to decide.</li>
     * </ul>
     * </li>
     * <li>The temporal granularity mode, indicating the amount 
     * of information about the symbolic execution which is actually 
     * displayed; the possible alternatives are:
     * <ul>
     * <li>Displays all the traversed states; this is the default 
     * behaviour;</li>
     * <li>Displays only the initial (root) state, the decision 
     * (branches) states, and the stuck (leaves) states;</li>
     * <li>Displays all the pairs initial/stuck state, where each 
     * initial state is refined with the path condition of the 
     * corresponding stuck state;</li>
     * <li>Displays only the stuck (leaves) states.</li>
     * </ul>
     * </li>
     * <li>The output format mode used for displaying states;
     * the currently available format modes are a plaintext 
     * prettyprint of the state (default) and one where the 
     * heap is rendered as a graphviz/dot string;</li>
     * <li>The choice of a simplifier for deciding arithmetic 
     * expressions; currently Sicstus and CVC3 are supported;
     * <li>The host and the port of the Sicstus server; if 
     * none is specified, the address 127.0.0.1:8888 is attempted. 
     * If no server is available at the prescribed host:port, the 
     * user is prompted at each branch, or all branches are 
     * explored, depending on the interaction mode;</li>
     * <li>The path of the cvc3 executable;</li>
     * <li>The visualization status of the interactions 
     * between the simplifier and the runner; by default, 
     * nothing is showed, but it is possible to dump a high-level 
     * description of the exchanged messages on the output;</li> 
     * <li>The semicolumn-separated list of paths where the
     * source code must be found, in which case, if the source 
     * code for some class is found in the path, the code will 
     * be executed from a source row to the next one, skipping 
     * visualization of the intermediate states.
     * This is another way to control the granularity at which 
     * the symbolic execution is displayed. 
     * By default, the source code information is ignored and 
     * all states at the bytecode granularity are shown.</li>
     * <li>The initial state of the symbolic execution, or 
     * alternatively, a semicolon-separated classpath, the 
     * name of a substitution map file, and the signature of 
     * a method whose symbolic execution must be performed.</li>
     * <li>Three different bounds (scopes) for limiting the 
     * exploration of the symbolic execution state space.</li>
     * <li>The signature of a variable which is assumed initially 
     * volatile.</li>
     * </ul> 
     * 
     * @author pietro
     *
     */
	public static class Parameters {
		/** The name of the output file. */
		String dumpFileName = null;
		
		/** The text mode. */
        TextMode textMode = null;
        
        /** The interaction mode. */
        InteractionMode interactionMode = InteractionMode.PROMPT_BACKTRACK;
        
        /** The temporal granularity mode. */
        TemporalGranularityMode granularityMode = TemporalGranularityMode.ALL;

        /** The format mode. */
        OutputFormatMode outputFormatMode = OutputFormatMode.PLAINTEXT;
        
        /** The network address of the Sicstus server. */
        String sicstusServerAddress = LOCALHOST_IP;

        /** The port of the Sicstus server. */
        int sicstusServerPort = DEFAULT_PORT;
        
        /** 
         * <code>true</code> iff the interactions between the 
         * runner and the sicstus server must be logged to the output. 
         */
        boolean showSimplifierInteraction = false;
        
        /**  
         * The source code path, or <code>null</code> iff no path is specified, 
         * which is the default. 
         */
		String srcPath = null;
		
		/**
         * The step mode, i.e., the temporal granularity of the steps.
		 */
		StepMode stepMode = StepMode.BYTECODE;

        /** 
         * The initial <code>State</code> of the symbolic execution, or
         * <code>null</code> iff an initial state for a method invocation 
         * must be created by the runner; by default it is <code>null</code>.
         */
        State initialState = null;
        
        /**  
         * The classpath; overridden by <code>initialState</code>'s classpath
         * when <code>initialState != null</code>.
         */
		String classPath = "." + PATH_SEP;

        /**  
         * The name of the file containing the class substitution map, or <code>null</code>
         * for no file; overridden by <code>initialState</code>'s substitution map when 
         * <code>initialState != null</code>.
         * 
         */
		String substitutionMapFname = null;
		
        /**  
         * The name of the file containing the class init status, or <code>null</code>
         * for no file; overridden by <code>initialState</code>'s substitution map when 
         * <code>initialState != null</code>.
         */
		String classInitFname = null;

        /**  
         * The signature of the method to be executed; overridden by <code>initialState</code>'s 
         * current method when <code>initialState != null</code>.
         */
		String methodSignature = null;

		/**
		 * The simplifier to be used during the symbolic execution for deciding the 
		 * arithmetic conditions.
		 */
		SimplifierType simplifierType = SimplifierType.SICSTUS;

		/** The pathname of cvc3. */
		String cvc3Path = "." + PATH_SEP;
		
		/** The heap scope. */
		int heapScope = 0;

		/** The depth scope. */
		private int depthScope = 0;

		/** The count scope. */
		private int countScope = 0;
		
		/** The initially volatile variable. */
		private String varVolatile = null;
		
		/** The names of the variables observed by <code>this.observers</code>. */
		private String[] observedVarNames = null;
		
		/** The <code>ExecutionObserver</code>s. */
		private ExecutionObserver[] observers = null;
		
		/**
		 * Sets the name of the output file.
		 * @param s A <code>String</code> representing the pathname of a 
		 *          file where the console output (stdout and stderr) will 
		 *          be copied.
		 */
		public void setOutputFileName(String s) { this.dumpFileName = s; }
		
		/**
		 * Instructs not to copy the console output to any file, cancelling
		 * any previous invocation of the <code>setOutputFileName</code>
		 * method. This is the default behaviour. 
		 */
		public void setOutputFileNone() { this.dumpFileName = null; }
		
		/**
		 * Sets the line separator to CR LF.
		 */
		public void setTextModeWindows() { this.textMode = TextMode.WINDOWS; }
		
		/**
		 * Sets the line separator to LF.
		 */
		public void setTextModeUNIX() { this.textMode = TextMode.UNIX; }
		
		/**
		 * Resets the line separator to the platform default. This is 
		 * the default behaviour.
		 */
		public void setTextModePlatform() { this.textMode = null; }
		
		/**
		 * Sets the degree of user interaction to a mode where the
		 * user is prompted at all computation steps (just pause), 
		 * at backtrack points (queries whether it should backtrack 
		 * or not) and in case the Sicstus simplifier is unavailable, 
		 * also at undecidable branches (queries which 
		 * branches should be explored).
		 */
		public void setInteractionModeStepByStep() { this.interactionMode = InteractionMode.STEP_BY_STEP; }
		
		/**
		 * Sets the degree of user interaction to a mode where the
		 * user is prompted at backtrack points and undecidable branches;
		 * this is the default behaviour.
		 */
		public void setInteractionModePromptBacktrack() { this.interactionMode = InteractionMode.PROMPT_BACKTRACK; }
		
		/**
		 * Sets the degree of user interaction to a mode where the
		 * user is prompted only at undecidable branches and the runner 
		 * backtracks at all backtrack points.
		 */
		public void setInteractionModeOnlyBranchDecision() { this.interactionMode = InteractionMode.ONLY_BRANCH_DECISION; }
		
		/**
		 * Sets the degree of user interaction to a mode where the
		 * user is never prompted and the runner explores all branches of 
		 * the conditions it is unable to decide.
		 */
		public void setInteractionModeNoInteraction() { this.interactionMode = InteractionMode.NO_INTERACTION; }
		
		/**
		 * Sets the temporal granularity of execution display to a mode 
		 * where the runner displays all the traversed states; this is 
		 * the default behaviour.
		 */
		public void setTemporalGranularityModeAll() { this.granularityMode = TemporalGranularityMode.ALL; }

		/**
		 * Sets the temporal granularity of execution display to a mode 
		 * where the runner displays only the initial (root) state, the 
		 * decision (branches) states, and the stuck (leaves) states.
		 */
		public void setTemporalGranularityModeRootBranchesLeaves() { this.granularityMode = TemporalGranularityMode.ROOT_BRANCHES_LEAVES; }
		
		/**
		 * Sets the temporal granularity of execution display to a mode 
		 * where the runner displays all the pairs initial/stuck state, 
		 * where each initial state is refined with the path condition 
		 * of the corresponding stuck state.
		 */
		public void setTemporalGranularityModeRootLeavesRefined() { this.granularityMode = TemporalGranularityMode.ROOT_LEAVES_REFINED; }
		
		/**
		 * Sets the temporal granularity of execution display to a mode 
		 * where the runner displays only the stuck (leaves) states.
		 */
		public void setTemporalGranularityModeLeaves() { this.granularityMode = TemporalGranularityMode.LEAVES; }
		
		/**
		 * Sets the temporal granularity of execution display to a mode 
		 * where the runner does not display any state.
		 */
		public void setTemporalGranularityModeNone() { this.granularityMode = TemporalGranularityMode.NONE; }

		/**
		 * Sets the output format mode to produce a plaintext
		 * prettyprint representation of states; this is the 
		 * default behaviour. 
		 */
		public void setOutputFormatModePlaintext() { this.outputFormatMode = OutputFormatMode.PLAINTEXT; }
		
		/**
		 * Sets the output format mode to produce a textual
		 * trace of the covered statements in CSV format. 
		 */
		public void setOutputFormatModeTrace() { this.outputFormatMode = OutputFormatMode.TRACE; }
		
		/**
		 * Sets the output format mode to produce a representation 
		 * of states where the heap is formatted to a graphviz 
		 * string.
		 */
		public void setOutputFormatModeGraphviz() { this.outputFormatMode = OutputFormatMode.GRAPHVIZ; }
		
		/**
		 * Sets the path of the source files.
		 * 
		 * @param s a <code>String</code> representing a semicolon-separated 
		 * list of paths, where the source files will be searched.
		 */
		public void setSourcePath(String s) { this.srcPath = s; }
		
		/**
		 * Instructs the runner to step at every method call, displaying 
		 * only the initial and the final state of the methods execution. 
		 */
		public void setStepMethod() { this.stepMode = StepMode.METHOD; }
		
		/**
		 * Instructs the runner to step at every source code line. 
		 * In the case the source code of a class cannot be found,
		 * or the source path has not been set with a  
		 * call to <code>setSourcePath(String)</code>, the runner behaves 
		 * as for <code>setStepMethod()</code>.
		 */
		public void setStepSource() { this.stepMode = StepMode.SOURCE; }
		
		/**
		 * Instructs the engine to step at every bytecode;
		 * this is the default mode.
		 */
		public void setStepBytecode() { this.stepMode = StepMode.BYTECODE; }
		
		/**
		 * Sets the hostname or IP address where the Sicstus server 
		 * resides. By default the host is searched at 127.0.0.1.
		 * 
		 * @param s a <code>String</code> representing a hostname or 
		 *          IP address.
		 */
		public void setSicstusServerHost(String s) { this.sicstusServerAddress = s; }
		
		/**
		 * Sets the port number where the Sicstus server resides.
		 * By default the port number is assumed to be 8888.
		 * 
		 * @param i a <code>int</code> representing a port numer.
		 */
		public void setSicstusServerPort(int i) { this.sicstusServerPort = i; }
		
		/**
		 * Sets the pathname of the CVC3 executable.
		 * 
		 * @param s a <code>String</code> containing a valid pathname for
		 *          the CVC3 executable.
		 */
		public void setCVC3Path(String s) { this.cvc3Path = s; }
		
		/**
		 * Instructs the engine for using the Sicstus simplifier; this is the 
		 * default behaviour. 
		 */
		public void setSimplifierSicstus() { this.simplifierType = SimplifierType.SICSTUS; }
		
		/**
		 * Instructs the engine for using the CVC3 simplifier. 
		 */
		public void setSimplifierCVC3() { this.simplifierType = SimplifierType.CVC3; }

		/**
		 * Instructs the engine not to attempt looking for a simplifier; 
		 * the user will be queried according to the interaction 
		 * mode for resolving decisions.
		 */
		public void setSimplifierNone() { this.simplifierType = null; }
		
		/**
		 * Instructs the engine to print on the output all the interaction
		 * between the runner and the simplifier.
		 */
		public void setSimplifierShowInteraction() { this.showSimplifierInteraction = true; }

		/**
		 * Instructs the engine not to show the interaction with the 
		 * simplifier; this is the default mode.
		 */
		public void setSimplifierDontShowInteraction() { this.showSimplifierInteraction = false; }
		
		/** 
		 * Sets the <code>ExecutionObservers</code> to perform additional
		 * actions as a specific state is reached.
		 */ 
		public void setObservers(String[] observedVarNames, ExecutionObserver[] observers) {
			this.observedVarNames = observedVarNames;
			this.observers = observers;
		}

		/**
		 * Sets the engine's classpath, and cancels the effect of any 
		 * previous call to <code>setInitialState(State)</code>; the 
		 * default classpath is <code>"."</code>.
		 * 
		 * @param s a <code>String</code> containing a semicolon-separated 
		 *        classpath, where paths are expressed in Java format. 
		 */
		public void setClassPath(String s) { this.initialState = null; this.classPath = s; }
		
		/**
		 * Instructs the runner about the position of the substitution  
		 * map file, and cancels the effect of any previous call to 
		 * <code>setInitialState(State)</code>; by default, no 
		 * substitution map file is used.
		 * 
		 * @param s a <code>String</code> containing the pathname of 
		 *        a text file in Java format. Each row of the file must 
		 *        be in the format
		 *        &lt;className&gt;[<code>@root/</code>&lt;memberRegExp&gt;]
		 *        <code>=</code>&lt;className&gt;,
		 *        where &lt;className&gt; is the name of a class and 
		 *        &lt;memberRegExp&gt; is a regular expression matching
		 *        the attributes of the root object. As an example, if 
		 *        the root object is a linked list, <code>Node@root/head(/next)*=ConcreteNode</code>
		 *        specifies that the lazy initialization procedure must initialize 
		 *        with object of class <code>ConcreteNode</code> all the symbolic 
		 *        references to objects of class <code>Node</code> accessible 
		 *        as <code>root.head</code>, <code>root.head.next</code>, 
		 *        <code>root.head.next.next</code>, etc.
		 */
		public void setSubstitutionMapFileName(String s) { this.initialState = null; this.substitutionMapFname = s; }
		
		/**
		 * Instructs the runner about the position of the class   
		 * initialization file, and cancels the effect of any previous 
		 * call to <code>setInitialState(State)</code>; by default, no 
		 * class initialization file is used.
		 * 
		 * @param s a <code>String</code> containing the pathname of 
		 *        a text file in Java format. Each row of the file must
		 *        contain the name of a class which must be considered 
		 *        not already initialized when the symbolic execution begins. 
		 */
		public void setClassInitFileName(String s) { this.initialState = null; this.classInitFname = s; }
		
		/**
		 * Sets the signature of the method which must be symbolically executed, 
		 * and cancels the effect of any previous 
		 * call to <code>setInitialState(State)</code>.
		 * 
		 * @param s a <code>String</code> representing the signature of a method 
		 *        in the format 
		 *        &lt;className&gt;<code>:</code>&lt;parameters&gt;<code>:</code>&lt;methodName&gt;, 
		 *        e.g., <code>pkg/MainClass:([Ljava/lang/String;)V:main</code>. 
		 */
		public void setMethodSignature(String s) { this.initialState = null; this.methodSignature = s; }
		
		/** 
		 * Sets the signature of the variable which must be initially volatile
		 * and cancels the effect of any previous 
		 * call to <code>setInitialState(State)</code>.
		 */ 
		public void setVolatile(String varVolatile) { this.initialState = null; this.varVolatile = varVolatile; }
		
		/**
		 * Sets the initial state of the symbolic execution, and cancels the 
		 * effect of any previous call to <code>setClassPath(String)</code>,
		 * <code>setSubstitutionMapFileName(String)</code>, 
		 * <code>setClassInitFileName(String)</code>, <code>setMethodSignature(String)</code>
		 * and <code>setVolatile(String)</code>.
		 *  
		 * @param s a <code>State</code>.
		 */
		public void setInitialState(State s) { this.initialState = s; }

		/**
		 * Sets a limited heap scope. 
		 * The heap scope is the maximum size of the heap. If a state
		 * has a greater heap than the heap scope the exploration of the 
		 * branch it belongs is interrupted.
		 */
		public void setHeapScope(int heapScope) { this.heapScope = heapScope; }
		
		/**
		 * Sets an unlimited heap scope; this is the default behaviour. 
		 * The heap scope is the maximum size of the heap. If a state
		 * has a greater heap than the heap scope the exploration of the 
		 * branch it belongs is interrupted.
		 */
		public void setHeapScopeUnlimited() { this.heapScope = 0; }
		
		/**
		 * Sets a limited depth scope. 
		 * The depth of a state is the number of branches above it. If 
		 * a state has a depth greater than the depth scope the exploration 
		 * of the branch it belongs is interrupted.
		 */
		public void setDepthScope(int depthScope) { this.depthScope = depthScope; }
		
		/**
		 * Sets an unlimited depth scope; this is the default behaviour. 
		 * The depth of a state is the number of branches above it. If 
		 * a state has a depth greater than the depth scope the exploration 
		 * of the branch it belongs is interrupted.
		 */
		public void setDepthScopeUnlimited() { this.depthScope = 0; }
		
		/**
		 * Sets a limited count scope. 
		 * If a state has a number of predecessor states greater than the 
		 * count scope the exploration of the branch it belongs is interrupted.
		 */
		public void setCountScope(int countScope) { this.countScope = countScope; }
		
		/**
		 * Sets an unlimited count scope; this is the default behaviour.
		 * If a state has a number of predecessor states greater than the 
		 * count scope the exploration of the branch it belongs is interrupted.
		 */
		public void setCountScopeUnlimited() { this.countScope = 0; }
	}
	
	/**
	 * Enumeration of the possible text file line termination 
	 * conventions.
	 * @author Pietro Braione
	 */
	private enum TextMode {
		/** Separates the lines of text with the sequence of characters "\r\n". */ 
		WINDOWS, 
		
		/** Separates the lines of text with the character '\n'. */ 
		UNIX 
	};
	
	private enum StepMode {
		/** Steps at method invocations. */
		METHOD,
		
		/** Steps at source code lines. */
		SOURCE,
		
		/** Steps at bytecodes. */
		BYTECODE
	}

	/**
	 * Enumeration of the possible interaction mode of a runner 
	 * session.
	 * @author Pietro Braione
	 */
	private enum InteractionMode { 
		/** 
		 * Prompts the user at every step of the symbolic computation, 
		 * before each backtrack point and at branching points which
		 * is unable to decide.
		 */
		STEP_BY_STEP, 

		/** 
		 * Prompts the user before each backtrack point and at 
		 * branching points which is unable to decide.
		 */
		PROMPT_BACKTRACK, 
		
		/** 
		 * Prompts the user at branching points which is unable 
		 * to decide. 
		 */
		ONLY_BRANCH_DECISION, 

		/** 
		 * Does not prompt the user, exploring all paths in the 
		 * case it is unable to decide a branch. 
		 */
		NO_INTERACTION 
	}
	
	/**
	 * Enumeration of the possible temporal granularity of output, determining 
	 * which states are displayed and which are not during the execution.
	 * 
	 * @author pietro
	 */
	private enum TemporalGranularityMode {
		/** Displays all the states. */
		ALL,
		
		/** 
		 * Displays the root state, all the branch states and all the
		 * leaves, i.e., the stuck states.
		 */
		ROOT_BRANCHES_LEAVES,
		
		/**
		 * Displays each leave together with its root state refined 
		 * on the leave's path condition.
		 */
		ROOT_LEAVES_REFINED,

		/** Displays only the leaves, i.e., the stuck states. */
		LEAVES,
		
		/** Displays nothing. */
		NONE
	}
	
	/**
	 * Enumeration of the possible output display formats.
	 * 
	 * @author pietro
	 */
	private enum OutputFormatMode {
		/** Displays the states textually (with indentation). */ 
		PLAINTEXT,
		
		/** Displays just a trace of the traversed statements in CSV format. */ 
		TRACE,
		
		/** Displays the states as DOT graphs. */
		GRAPHVIZ
	}
	
	/**
	 * Enumeration of the possible simplifiers.
	 * 
	 * @author pietro
	 */
	private enum SimplifierType {
		/** Uses Sicstus CLP via network server. */
		SICSTUS, 
		
		/** Uses cvc3. */
		CVC3
	}

	/** The <code>Parameters</code> of the symbolic execution. */
	private Parameters parameters;
	
    /** <code>PrintStream</code>s for the output. */
    private PrintStream[] out;
    
    /** <code>PrintStream</code> for the error. */
    private PrintStream[] err;
    
	/** The <code>StateFormatter</code> to output states at branches. */
	private StateFormatter formatterBranches;
	
	/** The <code>StateFormatter</code> to output states not at branches. */
	private StateFormatter formatterOthers;
	
    /** The <code>Engine</code> used to run the method. */
    private Engine engine;
    
    /**
     * Constructor. 
     */
    public Run(Parameters parameters) {
        this.parameters = parameters;
    }
    
    private class ActionsRun extends Runner.Actions {
    	private State initialState;
    	private State currentState;
    	private String scopeExhaustionMessage;
    	
    	@Override
    	public boolean atTraceStart() {
    		//updates the current state
            this.currentState = Run.this.engine.getCurrentState();
            
            //saves the initial state
            if (this.initialState == null) {
            	this.initialState = currentState.clone();
            }
            
            //scope not yet exhausted
            this.scopeExhaustionMessage = null;
            
            return false;
    	}
    	
    	@Override
    	public boolean atPreStep(boolean isRootBranch) {
    		boolean stop = false;
    		
        	//displays the current state
        	if (Run.this.parameters.granularityMode == TemporalGranularityMode.ALL ||
        	    (Run.this.parameters.granularityMode == TemporalGranularityMode.ROOT_BRANCHES_LEAVES && isRootBranch)) {
        		Run.this.printState(this.currentState, isRootBranch);
        	}
            
            //prompts the user for next step in case of interactive mode
            if (Run.this.parameters.interactionMode == InteractionMode.STEP_BY_STEP) {
            	String ans = IO.readln(Run.this.out, PRE_NAME + PROMPT_SHOULD_STEP);
                if (ans.equals("x")) {
                    stop = true;
                }
            }
            
            return stop;
    	}
    	
    	@Override
    	public boolean atPartialReferenceResolutionException(PartialReferenceResolutionException e) {
    		//prints a warning
            IO.println(Run.this.out, PRE_NAME + WARNING_PARTIAL_REFERENCE_RESOLUTION_1 + 
            		   e.getReferenceStaticType() +  WARNING_PARTIAL_REFERENCE_RESOLUTION_2); 
    		return true;
    	}
    	
    	@Override
    	public boolean atPostStep(BranchPoint bp) {
    		//updates the current state
            this.currentState = Run.this.engine.getCurrentState();
            
            //if the scope has been exhausted, interrupts the exploration 
            //of the current trace
            if (Run.this.parameters.heapScope > 0 &&
            	this.currentState.getHeap().getSize() > Run.this.parameters.heapScope) {
            	Run.this.engine.stopCurrentTrace();
            	this.scopeExhaustionMessage = WARNING_SCOPE_EXHAUSTED_HEAP;
            }
            if (Run.this.parameters.depthScope > 0 &&
            	this.currentState.getDepth() > Run.this.parameters.depthScope) {
            	Run.this.engine.stopCurrentTrace();
            	this.scopeExhaustionMessage = WARNING_SCOPE_EXHAUSTED_DEPTH;
            }
            if (Run.this.parameters.countScope > 0 &&
            		this.currentState.getSequenceNumber() > Run.this.parameters.countScope) {
            	Run.this.engine.stopCurrentTrace();
            	this.scopeExhaustionMessage = WARNING_SCOPE_EXHAUSTED_COUNT;
            }
            
            return false;
    	}
    	
    	@Override
		public boolean atTraceEnd(boolean isRootBranch) {
    		if (Run.this.parameters.granularityMode == TemporalGranularityMode.NONE) {
    			;
    		} else {
    			//displays the leaf (stuck) state
    			if (Run.this.parameters.granularityMode == TemporalGranularityMode.ROOT_LEAVES_REFINED) {
    				State initialRefined = this.initialState.clone();
    				try {
    					initialRefined.refine(this.currentState.getPathCondition());
    				} catch (CannotRefineException e) {
    					IO.println(Run.this.out, PRE_NAME + ERROR_CANNOT_REFINE);
    					initialRefined = this.initialState;
    				}
    				Run.this.printState(initialRefined, true);
    			}
    			Run.this.printState(this.currentState, (Run.this.parameters.granularityMode == TemporalGranularityMode.LEAVES || 
    					Run.this.parameters.granularityMode == TemporalGranularityMode.ROOT_LEAVES_REFINED ||
    					isRootBranch));
    			if (this.scopeExhaustionMessage != null) {
					IO.println(Run.this.out, PRE_NAME + this.scopeExhaustionMessage);
    			}
    		}
            return false;
		}
    	
    	@Override
    	public boolean atBacktrack() {
    		boolean stop = false;
    		
        	//prompts the user for backtrack in the case of interactive mode
        	if (Run.this.parameters.interactionMode == InteractionMode.STEP_BY_STEP ||
        			Run.this.parameters.interactionMode == InteractionMode.PROMPT_BACKTRACK) {	 
        		String ans = IO.readln(Run.this.out, PRE_NAME + PROMPT_SHOULD_BACKTRACK);
                if (ans.equals("x")) {
                	stop = true;
                }
        	}
        	
        	return stop;
    	}
    	
    	@Override
    	public void atEnd() {
    		//prints a message
            IO.println(Run.this.out, PRE_NAME + MSG_NO_BACKTRACK);                        
    	}

    }

    /**
     * Executes the method.
     */
    public int run()  {
    	//sets up engine
        int retVal = processParametersAndSetUpEngine();
        if (retVal == 0) {
	        //runs the method		
			try {
	        	//starts the engine
        		//TODO this is for future implementation of variables observation inside the runner
				try {
					this.engine.init();
				} catch (NonexistingObservedVariablesException e) {
		        	for (int i : e.getVariableIndices()) {
		    			IO.println(this.err, PRE_NAME + 
		    					WARNING_PARAMETERS_UNRECOGNIZABLE_VARIABLE + 
		    					i + "-th.");
		        	}
				}
	
	        	//prepares for some statistics
	    		long startTime = System.currentTimeMillis();
	    		
	    		//runs
	    		Runner r = new Runner(this.engine, this.parameters.stepMode != StepMode.BYTECODE, new ActionsRun());
	    		r.run();
	
	    		//calculates statistics
	    		long elapsedTime = System.currentTimeMillis() - startTime;
	    		long elapsedTimeSimplifier = this.engine.getTimeSimplifier();
	    		double tmp = elapsedTimeSimplifier;
	    		double timeSimplifierRatio = (tmp / elapsedTime);
	    		String timeSimplifierPercent = new DecimalFormat("##.##%").format(timeSimplifierRatio);
	    		    		
	        	//prints statistics
		        IO.println(this.out, banner(null));
		        IO.println(this.out, PRE_NAME + MSG_END);                        
		        IO.print(this.out, PRE_NAME + MSG_END_STATES + this.engine.getAnalyzedStates() + ", ");
		        IO.print(this.out, MSG_END_ELAPSED + formatTime(elapsedTime) + " (" + timeSimplifierPercent + " in simplifier), " ); 
		        IO.println(this.out, MSG_END_SPEED + this.engine.getAnalyzedStates() * 1000 / elapsedTime + " states/sec." ); 
			} catch (DecisionInitException e) {
				IO.println(this.err, PRE_NAME + ERROR_ENGINE_INIT_SIMPLIFIER);
				IO.printException(this.err, e);
				retVal = 1;
			} catch (DecisionBacktrackException e) {
				IO.println(this.err, PRE_NAME + ERROR_ENGINE_SIMPLIFIER);
				IO.printException(this.err, e);
				retVal = 1;
			} catch (DecisionRunningException e) {
				IO.println(this.err, PRE_NAME + ERROR_ENGINE_SIMPLIFIER);
				IO.printException(this.err, e);
				retVal = 1;
			} catch (InitializationException e) {
				IO.println(this.err, PRE_NAME + ERROR_ENGINE_INIT_INITIALIZER);
				IO.printException(this.err, e);
				retVal = 1;
			} catch (UndefInstructionException e) {
		        IO.println(this.err, PRE_NAME + ERROR_UNDEF_BYTECODE);                        
				IO.printException(this.err, e);
				retVal = 1;
			} catch (CannotInvokeNativeException e) {
		        IO.println(this.err, PRE_NAME + ERROR_CANNOT_INVOKE_NATIVE);                        
				IO.printException(this.err, e);
				retVal = 1;
			} catch (EngineStuckException e) {
		        IO.println(this.err, PRE_NAME + ERROR_ENGINE_STUCK);
				IO.printException(this.err, e);
				retVal = 1;
			} catch (Throwable e) {
		        IO.println(this.err, PRE_NAME + ERROR_ENGINE_UNEXPECTED);                        
				IO.printException(this.err, e);
				retVal = 1;
			} finally {
				//quits the engine
				try {
					this.engine.quit();
		        } catch (DecisionQuitException e) {
					IO.println(this.err, PRE_NAME + ERROR_ENGINE_QUIT_SIMPLIFIER);
					IO.printException(this.err, e);
					retVal = 1;
		        }
		        
		        //closes all the output streams with the exception of 
		        //stdout/err 
				for (PrintStream p : this.out) {
					if (p != null && p != System.out) { 
						p.close();
					}
				}
				for (PrintStream p : this.err) {
					if (p != null && p != System.err) { 
						p.close();
					}
				}
			}
        }
        
		//returns the error code
		return retVal;
    }
    
    /**
     * Processes the provided <code>Parameters</code> and 
     * sets up the <code>SEEngine</code> which will be used 
     * by the runner to perform the symbolic execution.
     * 
     * @return a <code>int</code> value representing an error 
     * code.
     */
    private int processParametersAndSetUpEngine() {
        //TODO lots of controls on parameters
        
    	//sets the output and error streams
   		{
   			//first are to standard
   	    	this.out = new PrintStream[2];
   	    	this.err = new PrintStream[2];
   	    	this.out[0] = System.out;
   	   		this.err[0] = System.err;

   	   		//tries to open the dump file
	    	PrintStream fs;
	    	if (this.parameters.dumpFileName == null) {
	    		fs = null;
	    	} else {
		        try {
			        File f = new File(this.parameters.dumpFileName);
			        fs = new PrintStream(f);
		
		        	//sets line separator style
		        	if (this.parameters.textMode == TextMode.WINDOWS)
		        		System.setProperty("line.separator", "\r\n");
		        	else if (this.parameters.textMode == TextMode.UNIX)
		        		System.setProperty("line.separator", "\n");
		        } catch (Exception e) {
		        	IO.println(this.err, PRE_NAME + ERROR_DUMP_FILE_OPEN);
		        	fs = null;
		        }
	    	}
	    	
			this.out[1] = this.err[1] = fs;
   		}
		
        //prints a welcome message
		IO.println(this.out, PRE_NAME + MSG_WELCOME_TXT);

        //creates the output formatter
    	{
	    	String[] srcPath;
	    	if (this.parameters.stepMode == StepMode.BYTECODE) {
	    		srcPath = null;
	    	} else if (this.parameters.stepMode == StepMode.SOURCE && this.parameters.srcPath != null) {
	    		srcPath = this.parameters.srcPath.split(MULTI_PATH_SEP);
	    	} else {
	    		srcPath = new String[0];
	    	}
	    	
	        if (this.parameters.outputFormatMode == OutputFormatMode.PLAINTEXT) {
	        	this.formatterBranches = new StateFormatterText(srcPath) {
	        		@Override
	        		public void format(State s) {
	        			this.formatOutput += lineSep; //gutter
	        			this.formatOutput += banner(s.getIdentifier()) + " [" + s.getSequenceNumber() + "]";
	        			this.formatOutput += lineSep; //gutter
	        			super.format(s);
	        		}
	        		public void emit() {
						IO.print(Run.this.out, this.formatOutput);
						this.cleanup();
					}
	        	};        	
	        	this.formatterOthers = new StateFormatterText(srcPath) {
	        		@Override
	        		public void format(State s) {
	        			this.formatOutput += lineSep; //gutter
	        			this.formatOutput += banner(null) + " [" + s.getSequenceNumber() + "]";
	        			this.formatOutput += lineSep; //gutter
	        			super.format(s);
	        		}
	        		public void emit() {
	        			IO.print(Run.this.out, this.formatOutput);
	        			this.cleanup();
	        		}
	        	};        	
	        } 
	        else if (this.parameters.outputFormatMode == OutputFormatMode.TRACE) {
	        	this.formatterBranches = new StateFormatterTextTrace() {
	        		public void emit() {
						IO.print(Run.this.out, this.formatOutput);
						this.cleanup();
					}
	        	};        	
		        this.formatterOthers = new StateFormatterTextTrace() {
	        		public void emit() {
	        			IO.print(Run.this.out, this.formatOutput);
	        			this.cleanup();
	        		}
	        	};        	
	        } else { // outputFormatMode == OutputFormatMode.GRAPHVIZ
	        	this.formatterBranches = new StateFormatterGraphviz(srcPath) {
	        		@Override
	        		public void format(State s) {
	        			this.formatOutput += lineSep; //gutter
	        			this.formatOutput += banner(s.getIdentifier()) + " [" + s.getSequenceNumber() + "]";
	        			this.formatOutput += lineSep; //gutter
	        			super.format(s);
	        		}
	        		public void emit() {
						IO.print(Run.this.out, this.formatOutput);
						this.cleanup();
					}
	        	};
	        	this.formatterOthers = new StateFormatterGraphviz(srcPath) {
	        		@Override
	        		public void format(State s) {
	        			this.formatOutput += lineSep; //gutter
	        			this.formatOutput += banner(null) + " [" + s.getSequenceNumber() + "]";
	        			this.formatOutput += lineSep; //gutter
	        			super.format(s);
	        		}
	        		public void emit() {
						IO.print(Run.this.out, this.formatOutput);
						this.cleanup();
					}
	        	};
	        }
    	}

		//creates the decision procedure and the engine
		{
			//tries to create decision procedure
			DecisionProcedure dec = null;
			if (this.parameters.simplifierType == SimplifierType.SICSTUS) {
				if (this.parameters.sicstusServerAddress != null) {
					IO.println(this.out, PRE_NAME + MSG_TRY_SICSTUS + this.parameters.sicstusServerAddress + ":" + this.parameters.sicstusServerPort + " ...");
					try {
						dec =
							new DecisionProcedureSicstus(this.parameters.sicstusServerAddress, 
									this.parameters.sicstusServerPort, 
									(this.parameters.showSimplifierInteraction ? this.out : null));
					} catch (DecisionException e) {
						IO.println(this.err, PRE_NAME + ERROR_SICSTUS_FAILED); 
					}
				}
			} else if (this.parameters.simplifierType == SimplifierType.CVC3) {
				IO.println(this.out, PRE_NAME + MSG_TRY_CVC3 + this.parameters.cvc3Path + " ...");
				try {
					dec =
						new DecisionProcedureCVC3(this.parameters.cvc3Path, 
								(this.parameters.showSimplifierInteraction ? this.out : null));
				} catch (DecisionException e) {
					IO.println(this.err, PRE_NAME + ERROR_CVC3_FAILED); 
				}
			}

			//if decision procedure creation fails, or if none is required, 
			//creates a fallback decision procedure
			if (dec == null) {
				if (this.parameters.interactionMode == InteractionMode.NO_INTERACTION) {
					dec = new DecisionProcedureAlwSat();
					IO.println(this.out, PRE_NAME + MSG_SIMPLIFIER_FALLBACK_BASIC);
				} else {
					dec = new DecisionProcedureConsole(this.out);
					IO.println(this.out, PRE_NAME + MSG_SIMPLIFIER_FALLBACK_INTERACTIVE);
				}
			}

			//sets the engine
			this.engine = new Engine(
					new ClassFileFactoryJavassist(), 
					//new ClassFileFactoryJaba("C:/"), 
					dec);
		}

		//sets the engine's parameters; first all the parameters determining the initial state...
        if (this.parameters.initialState == null) {
        	//specifies the environment of the symbolic execution:
        	//first, the classpath...
        	this.engine.addClassPath(this.parameters.classPath.split(MULTI_PATH_SEP));
        	
        	//...then, the substitution map...
    		List<String[]> substitutionMap = new LinkedList<String[]>();
    		if (this.parameters.substitutionMapFname != null) {
    			BufferedReader fr = null;
    			try {
    				fr = new BufferedReader(new FileReader(this.parameters.substitutionMapFname));
    				String l;
    				while ((l = fr.readLine()) != null) {
						String seps = SUBST_MAP_SEP + ORIGIN_SEP;
						String strNonSep = "[^" + seps + "]+";
						String pattern = strNonSep + "(" + ORIGIN_SEP + strNonSep + ")?" + SUBST_MAP_SEP + strNonSep; 
						if (l.matches(pattern)) {
							substitutionMap.add(l.split("[" + seps + "]"));
						}
    				}
    			} catch (FileNotFoundException e) {
    				//does nothing
    			} catch (IOException e) {
    				//does nothing
    			} finally {
    				try {
    					if (fr != null) { 
    						fr.close();
    					}
    				} catch (IOException e) {
    					//does nothing
    				}
    			}
    		}
    		for (String[] su : substitutionMap) {
				if (su.length == 3) {
					engine.setClassSubst(su[0], su[1], su[2]);
				} else { //length must be 2
					engine.setClassSubst(su[0], null, su[1]);
				}
    		}
    		
    		//...next, the class init...
    		if (this.parameters.classInitFname != null) {
    			BufferedReader fr = null;
    			try {
    				fr = new BufferedReader(new FileReader(this.parameters.classInitFname));
    				String l;
    				while ((l = fr.readLine()) != null) {
						engine.addNotInitializedClass(l);
    				}
    			} catch (FileNotFoundException e) {
    				//does nothing
    			} catch (IOException e) {
    				//does nothing
    			} finally {
    				try {
    					if (fr != null) {
    						fr.close();
    					}
    				} catch (IOException e) {
    					//does nothing
    				}
    			}
    		}

    		//...then, the method to be executed...
    		String[] sig = this.parameters.methodSignature.split(SIGNATURE_SEP);
    		if (sig.length != 3) {
    			IO.println(this.err, PRE_NAME + ERROR_PARAMETERS_UNRECOGNIZABLE_SIGNATURE);
    			return 1;
    		}
    		this.engine.setMethodName(sig[0], sig[1], sig[2]);
    		
    		//...finally, the initially volatile variable
    		if (this.parameters.varVolatile != null) {
    			this.engine.setVolatile(this.parameters.varVolatile);
    		}
        } else {
        	this.engine.setInitialState(this.parameters.initialState);
        }
        //...then the observers
		this.engine.setObservers(this.parameters.observedVarNames, this.parameters.observers);
				
        return 0;
    }

    /**
     * Displays a state on <code>this.out</code>.
     * @param s the state to be displayed.
     * @param isRootBranch <code>true</code> iff the state is at a branch point
     */
    private void printState(State s, boolean isRootBranch) {
    	StateFormatter f;
    	if (isRootBranch) {
    		f = this.formatterBranches;
    	} else {
    		f = this.formatterOthers;
    	}
        f.format(s);
        f.emit();
    }
    
    //Static members.
    
    /** Length of separator between text areas. */
    private static final int BANNER_LENGTH = 35;

    /** Char for separator between text areas. */
    private static final char BANNER_CHAR = '.';
    
    /** Prefix: name of the application. */
    private static final String PRE_NAME = Run.class.getCanonicalName() + ": ";

    /** Message: welcome. */
    private static final String MSG_WELCOME_TXT = "========= Welcome to the MUSE v. " + MUSE_INFO.CURRENT_VERSION + " symbolic runner. =========";

    /** Message: trying to connect to the Sicstus simplifier. */
    private static final String MSG_TRY_SICSTUS = "Trying to connect to the Sicstus simplifier at network address ";

    /** Message: trying to connect to the CVC3 simplifier. */
    private static final String MSG_TRY_CVC3 = "Trying to connect to the CVC3 simplifier at ";

    /** Message: no more pending backtrack points. */
	private static final String MSG_NO_BACKTRACK = "Trace finished, no pending backtrack point.";

    /** Message: will interact via console for clause simplification. */
	private static final String MSG_SIMPLIFIER_FALLBACK_INTERACTIVE = "Will work in interactive clause simplification mode.";

    /** Message: will not perform clause simplification. */
	private static final String MSG_SIMPLIFIER_FALLBACK_BASIC = "Will work in noninteractive full execution tree exploration mode.";

	/** Message: end of symbolic execution. */
    private static final String MSG_END = "The symbolic execution ends here.";

    /** Message: analyzed states. */
    private static final String MSG_END_STATES = "Analyzed states: ";
    
	/** Message: elapsed time. */
    private static final String MSG_END_ELAPSED = "Elapsed time: ";

	/** Message: average speed. */
    private static final String MSG_END_SPEED = "Average speed: ";

    /** Warning message: unrecognizable signature. */
    private static final String WARNING_PARAMETERS_UNRECOGNIZABLE_VARIABLE = "Unrecognizable variable will not be observed: "; 
    
    /** Warning message: partial reference resolution (part 1). */
    private static final String WARNING_PARTIAL_REFERENCE_RESOLUTION_1 = "Partial resolution of reference with type ";

    /** Warning message: partial reference resolution (part 2). */
    private static final String WARNING_PARTIAL_REFERENCE_RESOLUTION_2 = ". Subsequent reasoning on this subtree will be limited to the " + 
                                                                         "known portion of the heap. Perhaps you should improve your class substitution specification."; 

    /** Warning message: exhausted scope. */
    private static final String WARNING_SCOPE_EXHAUSTED = "scope exhausted along this branch.";

    /** Warning message: exhausted heap scope. */
    private static final String WARNING_SCOPE_EXHAUSTED_HEAP = "Heap " + WARNING_SCOPE_EXHAUSTED;

    /** Warning message: exhausted depth scope. */
    private static final String WARNING_SCOPE_EXHAUSTED_DEPTH = "Depth " + WARNING_SCOPE_EXHAUSTED;

    /** Warning message: exhausted count scope. */
    private static final String WARNING_SCOPE_EXHAUSTED_COUNT = "Count " + WARNING_SCOPE_EXHAUSTED;

    /** Error message: unrecognizable signature. */
    private static final String ERROR_PARAMETERS_UNRECOGNIZABLE_SIGNATURE = "Unrecognizable method signature."; 
    
    /** Error message: unable to open dump file. */
    private static final String ERROR_DUMP_FILE_OPEN = "Could not open the dump file. The session will be displayed on console only.";

    /** Error message: unable to connect with the Sicstus simplifier. */
    private static final String ERROR_SICSTUS_FAILED = "Failed connection to Sicstus simplifier."; 
    
    /** Error message: unable to connect with the CVC3 simplifier. */
    private static final String ERROR_CVC3_FAILED = "Failed connection to CVC3 simplifier."; 
    
    /** Error message: unable to initialize engine, because unable to initialize Sicstus simplifier. */
    private static final String ERROR_ENGINE_INIT_SIMPLIFIER = "Unexpected failure during the initialization of the execution engine (cause: simplifier)."; 
    
    /** Error message: unable to initialize engine, because unable to initialize the <code>Initializer</code>. */
    private static final String ERROR_ENGINE_INIT_INITIALIZER = "Unexpected failure during the initialization of the execution engine (cause: initializer)."; 
        
    /** Error message: Sicstus simplifier failure. */
    private static final String ERROR_ENGINE_SIMPLIFIER = "Error while interacting with the Sicstus simplifier."; 
        
    /** Error message: unable to quit engine, because unable to quit the Sisctus simplifier. */
    private static final String ERROR_ENGINE_QUIT_SIMPLIFIER = "Unexpected failure while quitting the engine (cause: simplifier)."; 
        
    /** Error message: cannot manage a bytecode. */
    private static final String ERROR_UNDEF_BYTECODE = "Met an unmanageable bytecode."; 
        
    /** Error message: cannot manage a native method invocation. */
    private static final String ERROR_CANNOT_INVOKE_NATIVE = "Met an unmanageable native method invocation."; 
        
    /** Error message: unexpected failure (stepping while engine stuck). */
    private static final String ERROR_ENGINE_STUCK = "Unexpected failure while running (attempted step while in a stuck state)."; 
        
    /** Error message: cannot refine state. */
    private static final String ERROR_CANNOT_REFINE = "Unexpected failure while refining initial state with final state path condition. The nonrefined initial state will be printed.";
    
    /** Error message: unexpected failure. */
    private static final String ERROR_ENGINE_UNEXPECTED = "Unexpected failure of unidentifiable source."; 
        
    /** Prompt: ask user whether should continue execution by stepping. */
	private static final String PROMPT_SHOULD_STEP = "Proceed with next step? (x: abort, any other: step): ";

    /** Prompt: ask user whether should continue execution by backtracking. */
    private static final String PROMPT_SHOULD_BACKTRACK = "Trace finished, pending backtrack points; proceed with backtrack? (x: abort, any other: backtrack): ";

    /**
     * Returns a banner around a <code>String</code>.
     * @param s a <code>String</code> around which the banner is built, 
     *          or <code>null</code> for a clean banner.
     * @return the banner.
     */
    private static String banner(String s) {
    	String retVal = ((s == null) ? "" : (".:: " + s + " ::. "));    	
        for (int i = retVal.length() + 1; i <= BANNER_LENGTH; i++) {
        	retVal += BANNER_CHAR;
        }
        return retVal;
    }
    
	//Here starts the command line interface code.
	private static final String SWITCH_HELP = "h";
	private static final String SWITCH_HELP_2 = "?";
	private static final String SWITCH_CLASSPATH = "cp";
	private static final String SWITCH_OUTPUT = "o";
	private static final String SWITCH_SOURCE_PATH = "fp";
	private static final String SWITCH_CVC3 = "cvc3";
	private static final String SWITCH_SICSTUS_SERVER = "se";
	private static final String SWITCH_SHOW_SIMPLIFIER = "ss";
	private static final String SWITCH_SUBST_MAP = "sm";
	private static final String SWITCH_CLASS_INIT = "ci";
	private static final String SWITCH_TEXT_MODE = "t";
	private static final String SWITCH_INTERACTION_MODE = "i";
	private static final String SWITCH_TEMPORAL_GRANULARITY = "g";
	private static final String SWITCH_OUTPUT_FORMAT = "fmt";
	private static final String SWITCH_RESERVED = "R";
	private static final String SWITCH_RESERVED_NAME = SWITCH_RESERVED + "n";
	private static final char networkAddressSep = ':';
	private static final char commandLineSwitch = '/';

	//parameters exclusively used by the CLI
    private static boolean help = false;
	private static String reservedName = "java " + Run.class.getCanonicalName();
	private static boolean defaultName = true;
    
    public static void main(String[] args) {
    	Parameters p = parseParameters(args);
    	if (help) {
			System.out.print(help());
			System.exit(0);
		}
    	Run inst = new Run(p);
		System.exit(inst.run());
    }
    
    private static Parameters parseParameters(String[] args) {
		Parameters retVal = new Parameters();
		boolean classpath = false;
		boolean output = false;
		boolean source_path = false;
		boolean cvc3 = false;
		boolean server = false;
		boolean subst_map = false;
		boolean class_init = false;
		boolean text_mode = false;
		boolean interaction_mode = false;
		boolean gr_mode = false;
		boolean out_fmt = false;
		boolean reserved_name = false;
		
		for (String s : args) {
			if (classpath) {
				retVal.setClassPath(s);
				classpath = false;
			} else if (output) {
				retVal.setOutputFileName(s);
				output = false;
			} else if (source_path) {
				retVal.setSourcePath(s);
				source_path = false;
			} else if (cvc3) {
				retVal.setSimplifierCVC3();
				retVal.setCVC3Path(s);
				cvc3 = false;
			} else if (server) {
				try {
					retVal.setSimplifierSicstus();
					String[] ss = s.split("" + networkAddressSep);
					retVal.setSicstusServerHost(ss[0]);
					retVal.setSicstusServerPort(Integer.parseInt(ss[1]));
				} catch (NumberFormatException exc) {
					//does nothing
				} catch (IndexOutOfBoundsException exc) {
					//does nothing
				}
				server = false;
			} else if (subst_map) {
				retVal.setSubstitutionMapFileName(s);
				subst_map = false;
			} else if (class_init) {
				retVal.setClassInitFileName(s);
				class_init = false;
			} else if (text_mode) {
				try {
					int i = Integer.parseInt(s);
					for (TextMode m : TextMode.values()) {
						if (i == m.ordinal()) {
							retVal.textMode = m;
							break;
						}
					}
				} catch (NumberFormatException exc) {
					//does nothing
				}
				text_mode = false;
			} else if (interaction_mode) {
				try {
					//TODO this impl is ugly
					int i = Integer.parseInt(s);
					for (InteractionMode m : InteractionMode.values()) {
						if (i == m.ordinal()) {
							retVal.interactionMode = m;
							break;
						}
					}
				} catch (NumberFormatException exc) {
					//does nothing
				}
				interaction_mode = false;
			} else if (gr_mode) {
				try {
					//TODO this impl is ugly
					int i = Integer.parseInt(s);
					for (TemporalGranularityMode m : TemporalGranularityMode.values()) {
						if (i == m.ordinal()) {
							retVal.granularityMode = m;
							break;
						}
					}
				} catch (NumberFormatException exc) {
					//does nothing
				}
				gr_mode = false;
			} else if (out_fmt) {
				try {
					//TODO this impl is ugly
					int i = Integer.parseInt(s);
					for (OutputFormatMode m : OutputFormatMode.values()) {
						if (i == m.ordinal()) {
							retVal.outputFormatMode = m;
							break;
						}
					}
				} catch (NumberFormatException exc) {
					//does nothing
				}
				out_fmt = false;
			} else if (reserved_name) {
				reservedName = s;
				defaultName = false;
				reserved_name = false;
			} else if (s.equals(commandLineSwitch + SWITCH_HELP) ||
					 s.equals(commandLineSwitch + SWITCH_HELP_2)) {
				help = true;
			} else if (s.equals(commandLineSwitch + SWITCH_CLASSPATH)) {
				classpath = true;
			} else if (s.equals(commandLineSwitch + SWITCH_OUTPUT)) {
				output = true;
			} else if (s.equals(commandLineSwitch + SWITCH_SOURCE_PATH)) {
				source_path = true;
			} else if (s.equals(commandLineSwitch + SWITCH_CVC3)) {
				cvc3 = true;
			} else if (s.equals(commandLineSwitch + SWITCH_SICSTUS_SERVER)) {
				server = true;
			} else if (s.equals(commandLineSwitch + SWITCH_SHOW_SIMPLIFIER)) {
				retVal.showSimplifierInteraction = true;
			} else if (s.equals(commandLineSwitch + SWITCH_SUBST_MAP)) {
				subst_map = true;
			} else if (s.equals(commandLineSwitch + SWITCH_CLASS_INIT)) {
				class_init = true;
			} else if (s.equals(commandLineSwitch + SWITCH_TEXT_MODE)) {
				text_mode = true;
			} else if (s.equals(commandLineSwitch + SWITCH_INTERACTION_MODE)) {
				interaction_mode = true;
			} else if (s.equals(commandLineSwitch + SWITCH_TEMPORAL_GRANULARITY)) {
				gr_mode = true;
			} else if (s.equals(commandLineSwitch + SWITCH_OUTPUT_FORMAT)) {
				out_fmt = true;
			} else if (s.equals(commandLineSwitch + SWITCH_RESERVED_NAME)) {
				reserved_name = true;
			} else {
				retVal.methodSignature = s;
			}
		}
		return retVal;
    }
    
    private static String help() {
    	final String lineSep = System.getProperty("line.separator");
		String retVal = "" +
		//12345678901234567890123456789012345678901234567890123456789012345678901234567890
		  "Usage:" + lineSep + lineSep +
		  "  " + reservedName + lineSep + 
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_CLASSPATH + " <classpath> " + "]" + lineSep + 
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_SOURCE_PATH + " <srcpath> " + "]" + lineSep + 
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_OUTPUT + " <fnameo> " + "]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_SUBST_MAP + " <fnamem> " + "]" + lineSep + 
		  repeat(' ', reservedName.length() + 3) + "[["+ commandLineSwitch + SWITCH_CLASS_INIT + " <fnamec> " + "] |" + lineSep +
		  repeat(' ', reservedName.length() + 3) + " ["+ commandLineSwitch + SWITCH_CVC3 + " <fnamecvc>" + "]]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_SICSTUS_SERVER + " <host>" + networkAddressSep + "<port> " + "]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_SHOW_SIMPLIFIER + "]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_TEXT_MODE + " <textm> " + "]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_INTERACTION_MODE + " <intm> " + "]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_TEMPORAL_GRANULARITY + " <gr> " + "]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "["+ commandLineSwitch + SWITCH_OUTPUT_FORMAT + " <fmt> " + "]" + lineSep +
		  repeat(' ', reservedName.length() + 3) + "<className>" + SIGNATURE_SEP + "<paramSignature>" + SIGNATURE_SEP + "<methodName>" + lineSep + lineSep +
		  "  " + reservedName + " " + commandLineSwitch + SWITCH_HELP + " | " + commandLineSwitch + SWITCH_HELP_2 + lineSep + lineSep +
		  "Where:" + lineSep + lineSep +
		  "<classpath>      is a list of paths separated by the character '" + MULTI_PATH_SEP +"', which" + lineSep +
		  "                 will be used as the classpath of the symbolic execution" + lineSep + 
		  "                 (default: .);" + lineSep +
		  "<srcpath>        is a list of paths separated by the character '" + MULTI_PATH_SEP +"', which" + lineSep + 
		  "                 contains the source code for better visualization" + lineSep + 
		  "                 (default: no source path, steps performed at bytecode" + lineSep +
		  "                 granularity);" + lineSep +
		  "<fnameo>         is the name of the file where stdout will be copied" + lineSep +
		  "                 (default: no dump file, only stdout);" + lineSep +
		  "<fnamem>         is the name of the file containing the substitution map," + lineSep +
		  "                 specifying how a symbolic reference of a given type" + lineSep +
		  "                 is initialized with fresh concrete objects" + lineSep +
		  "                 (default: no substitution map, all symbolic references" + lineSep +
		  "                 will be initialized by instances of their types, or no fresh" + lineSep +
		  "                 instances if the type is abstract);" + lineSep +
		  "<fnamec>         is the name of the file containing the names of the classes" + lineSep +
		  "                 that are not to be considered initialized at the start of the" + lineSep +
		  "                 symbolic execution\n" +
		  "                 (default: no file, all classes are considered initialized);" + lineSep +
		  "<fnamecvc>       is the pathname of the CVC3 executable;" + lineSep +
		  commandLineSwitch + SWITCH_SHOW_SIMPLIFIER + "              activates the visualization of the interactions between" + lineSep +
		  "                 the runner and the simplifier;" + lineSep +
		  "<host>" + networkAddressSep + "<port>    is the network address of the Sicstus server" + lineSep +
		  "                 (default: localhost" + networkAddressSep + "8888);" + lineSep +
		  "<textm>          is the text line termination standard (0 for Windows, 1 for" + lineSep + 
		  "                 UNIX, default: the current platform's standard);" + lineSep +
		  "<intm>           is the interaction mode (0 prompts all steps, backtrack" + lineSep + 
		  "                 points and undecidable branches, 1 prompts backtrack points" + lineSep +
		  "                 and undecidable branches, 2 prompts only undecidable branches" + lineSep +
		  "                 and backtracks at all backtrack points, 3 does not prompt the" + lineSep + 
		  "                 user and explores all paths in the case it is unable to" + lineSep + 
		  "                 decide a branch, default: 1);" + lineSep +
		  "<gr>             is the temporal granularity of the output display (0 displays" + lineSep + 
		  "                 all the states of the execution, 1 displays only the starting" + lineSep + 
		  "                 state, all the states at branches and all the stuck states," + lineSep + 
		  "                 2 displays the obtained pairs starting state - stuck state," + lineSep + 
		  "                 where the starting state is refined with the path condition" + lineSep + 
		  "                 of the stuck state, 3 displays only the stuck state," + lineSep +
		  "                 4 does not display any state, default: 0);" + lineSep +
		  "<fmt>            is the output format (0 for plaintext, 1 for graphviz," + lineSep +
		  "                 default: 0);" + lineSep + 
		  "<className>      is the name of the class containing the method to be checked;" + lineSep +
		  "<paramSignature> describes the types of the method's parameters and return" + lineSep + 
		  "                 values;" + lineSep +
		  "<methodName>     is the name of the method to be checked." + lineSep;
		if (defaultName)
			retVal +=  lineSep +
		//12345678901234567890123456789012345678901234567890123456789012345678901234567890
		  "Example: we assume that the runner is deployed in c:\\run\\run.jar, the" + lineSep + 
		  "standard Java libraries are in c:\\jre\\bin\\rt.jar, the Sicstus server is" + lineSep +
		  "listening at port 8888 of server 131.151.20.44, the method to be verified is" + lineSep + lineSep + 
		  "  boolean myPackage.MyClass.myMethod(Object o)" + lineSep + lineSep + 
		  "the software under verification is deployed in c:\\proj\\proj.jar, that" + lineSep + 
		  "the source code is part in c:\\foo and part in c:\\baz, that we" + lineSep +
		  "want to display each refined starting state with the corresponding" + lineSep +
		  "stuck state, and that we want a copy of the output on the file d:\\out.txt." + lineSep +
		  "Then, the analysis is run by invoking:" + lineSep + lineSep + 
		  "  java -cp c:\\run\\run.jar " + Run.class.getCanonicalName() +  lineSep + 
		  "       " + commandLineSwitch +  SWITCH_SOURCE_PATH + " c:\\foo" + MULTI_PATH_SEP + "c:\\baz" + lineSep +  
		  "       " + commandLineSwitch + SWITCH_OUTPUT + " out.txt " + commandLineSwitch + SWITCH_SICSTUS_SERVER + " 131.151.20.44" + 
		  networkAddressSep + "8888" + lineSep + 
		  "       " + commandLineSwitch + SWITCH_CLASSPATH + " c:\\proj\\proj.jar" + MULTI_PATH_SEP + 
		  "c:\\jre\\bin\\rt.jar " + commandLineSwitch + SWITCH_TEMPORAL_GRANULARITY + " 2" + lineSep + 
		  "       myPackage/MyClass" + SIGNATURE_SEP + "(Ljava/lang/Object;)Z" + SIGNATURE_SEP + "myMethod" + lineSep;
		return retVal;
    }
	
	/**
	 * Builds a string by repeating a character.
	 *  
	 * @param c the <code>char</code> to be repeated.
	 * @param ln the length of the <code>String</code>; it 
	 *           must be nonnegative.
	 * @return the <code>String</code> obtained by repeating 
	 *         <code>c</code> </code>ln</code> times.
	 */
	private static String repeat(char c, int ln) {
		char[] cc = new char[ln];
		for (int i = 0; i < ln; i++) {
			cc[i] = c;
		}
		return new String(cc);
	}
	
	/**
	 * Formats a <code>long</code> value denoting a time interval 
	 * into a more readable <code>String</code>.
	 * 
	 * @param millis a <code>long</code> denoting a time interval 
	 *        in milliseconds.
	 * @return a <code>String</code> describing the interval in 
	 *         terms of hours, minutes, seconds and milliseconds.
	 */
	private static String formatTime(long millis) {
		if (millis < 1000) {
			return millis + " msec";
		} else if (millis < 60000) {
			return ( millis/1000 ) + " sec " + ( millis % 1000 ) + " msec";
		} else if (millis < 3600000){
			return ( millis/60000 ) + " min " + (( millis % 60000 ) / 1000) + " sec " + (( millis % 60000 ) % 1000) + " msec";
		} else {
			return ( millis/3600000 ) + " hr " + (( millis % 3600000 ) / 60000) + " min " + ((( millis % 3600000) % 60000 ) / 1000) + " sec " + (( ( millis % 3600000) % 60000 ) % 1000) + " msec";
		}
	}
}