package muse.bc;

import java.util.HashMap;
import java.util.Map;

/**
 * A <code>BytecodeDispatcher&lt;R&gt;</code> associates bytecodes with 
 * objects of a given class <code>R</code>. Associations can be defined by 
 * registering, for each bytecode, a <code>DispatchStrategy</code> which is invoked to generate 
 * the output object.  
 *  
 * @author pietro
 *
 * @param <R> the class of the object which is associated to bytecodes.  
 */
public abstract class BytecodeDispatcher<R> {
	/**
	 * A Strategy returning an object with class <code>RR</code>. 
	 * 
	 * @author pietro
	 *
	 * @param <RR> the class of the returned object.
	 */
	public interface DispatchStrategy<RR> { RR doIt() throws Throwable; }

	private DispatchStrategy<? extends R> dispatchNonexistent = 
		new DispatchStrategy<R>() { 
			public R doIt() { 
				return null; 
			} 
		};
	private Map<Byte, DispatchStrategy<? extends R>> dispatchTable = 
		new HashMap<Byte, DispatchStrategy<? extends R>>();
	
	/**
	 * Sets the <code>DispatchStrategy</code> for a given bytecode. If the bytecode
	 * was previously associated with another <code>DispatchStrategy</code>, the 
	 * former association is discarded and replaced with the new one.
	 * 
	 * @param instruction the bytecode.
	 * @param s the <code>DispatchStrategy</code> associated to <code>instruction</code>.
	 *          In the case <code>s == null</code> the method has no effect.
	 * @return <code>this</code> (allows to chain invocations). 
	 */
	public BytecodeDispatcher<R> setDispatchStrategy(byte instruction, DispatchStrategy<? extends R> s) {
		if (s != null) { 
			this.dispatchTable.put(instruction, s);
		}
		return this;
	}
	
	/**
	 * Sets the <code>DispatchStrategy</code> to be invoked by default 
	 * for nonregistered bytecodes. 
	 * If the <code>DispatchStrategy</code> was previously set, the 
	 * former association is discarded and replaced with the new one.
	 * 
	 * @param s the <code>DispatchStrategy</code> to be used by default 
	 *          with unregistered bytecodes.
	 *          In the case <code>s == null</code> the method has no effect.
	 * @return <code>this</code> (allows to chain invocations).
	 */
	public BytecodeDispatcher<R> setDispatchNonexistentStrategy(DispatchStrategy<? extends R> s) {
		if (s != null) {
			this.dispatchNonexistent = s;
		}
		return this;
	}
	
	/**
	 * Executes the <code>DispatchStrategy&lt;R&gt;</code> associated with a 
	 * bytecode, or the registered default strategy in case such association 
	 * does not exist, or does nothing if neither a registered Strategy for 
	 * the bytecode nor a default one has been previously set.
	 * 
	 * @param bytecode the bytecode.
	 * @return the <code>R</code> produced by the invoked 
	 *         <code>DispatchStrategy&lt;R&gt;</code>, or <code>null</code> 
	 *         if neither a registered Strategy for the bytecode nor a 
	 *         default one has been previously set.
	 */
	public R select(byte bytecode) throws Throwable {
		DispatchStrategy<? extends R> d;
		if (this.dispatchTable.containsKey(bytecode)) 
			d = this.dispatchTable.get(bytecode);
		else
			d = this.dispatchNonexistent;
		return d.doIt();
	}
}
