package muse.bc;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import muse.mem.Type;
import muse.mem.Value;


/**
 * Abstract class for managing the information on a single 
 * class file.
 * 
 */
public abstract class ClassFile {
	protected static final String classFileExtension = ".class";
	protected static final String jarFileExtension = ".jar";

	/**
     * Test whether the class is an interface.
     * 
     * @return <code>true</code> if the class is an interface,
     * <code>false</code> otherwise.
     */
    public abstract boolean isInterface();
    
    /**
     * Test whether the class is abstract.
     * 
     * @return <code>true</code> if the class is an interface,
     * <code>false</code> otherwise.
     */
    public abstract boolean isAbstract();

    /**
     * Tests the class visibility.
     * 
     * @return <code>true</code> if the class is public,
     * <code>false</code> otherwise.
     */
    public abstract boolean isPublic();
    
    /**
     * Tests the required <code>invokespecial</code> bytecode 
     * semantics.
     *  
     * @return <code>true</code> if the ACC_SUPER flag of the class 
     * is set, i.e., if the class requires modern semantics,
     * <code>false</code> if the class requires backward-compatible 
     * semantics.  
     */
    public abstract boolean isSuperInvoke();
    
    /**
     * Given the signature of a method, return the max length of the operand stack 
     * for that method.
     * @param methodSignature The signature for the searched method
     * @return int The max length of the operand stack
     *
     public int getMaxStackForMethod(SESignature methodSignature) throws RuntimeException
     {
     int index=getMethodIndex(methodSignature);
     MethodInfo metInf=jcr.getMethodInfo(index);
     for(int i=0;i<metInf.getAttributeCount();i++)
     {
     AttributeInfo attribute=metInf.getAttribute(i);
     if(cp.GetStringForUtf8Constant(attribute.getAttributeNameIndex()).equals("Code"))
     return (((CodeAttribute)attribute).getMaxStack());
     }
     throw new RuntimeException("Can't retrieve length of stack for this method");
     }*/
    
    /**
     * Test for the equality of class file structure, the test is on the name of 
     * the class.
     */
    @Override
    public boolean equals(Object o) {
    	if (!(o instanceof ClassFile)) {
    		return false; 
    	}
		return this.getClassName().equals(((ClassFile) o).getClassName());
    }

    /**
     * Tests whether the class has an implementation (i.e., a declaration 
     * plus bytecode) for a method with a given signature. 
     * 
     * @param methodSignature the signature of the method to be searched. 
     *                        Only the name of the method and its descriptor 
     *                        are considered (the signature's class name 
     *                        is ignored).
     * @return <code>true</code> if the class has an implementation for 
     *         the method with signature <code>MethodSignature</code>, 
     *         <code>false</code> otherwise.
     */
    public abstract boolean hasMethodImplementation(Signature methodSignature);
    
    /**
     * Tests whether the class has a declaration for a method with a given 
     * signature. 
     * 
     * @param methodSignature the signature of the method to be searched. 
     *                        Only the name of the method and its descriptor 
     *                        are considered (the signature's class name 
     *                        is ignored).
     * @return <code>true</code> if the class has an implementation for 
     *         the method with signature <code>methodSignature</code>, 
     *         <code>false</code> otherwise.
     */
    public abstract boolean hasMethodDeclaration(Signature methodSignature);
    
    /**
     * Tests whether a method in the class is declared abstract.
     * 
     * @param methodSignature the signature of the method to be checked.
     * @return <code>true</code> if the method is abstract, <code>false</code> 
     *         if the method is concrete.
     * @throws MethodNotFoundException iff <code>hasMethodDeclaration(methodSignature) == false</code>.
     */
    public abstract boolean isMethodAbstract(Signature methodSignature) throws MethodNotFoundException;
    
    /**
     * Tests whether a method in the class is declared static.
     * 
     * @param methodSignature the signature of the method to be checked.
     * @return <code>true</code> if the method is static, <code>false</code> 
     *         if the method is not static.
     * @throws MethodNotFoundException iff <code>hasMethodDeclaration(methodSignature) == false</code>.
     */
    public abstract boolean isMethodStatic(Signature methodSignature) throws MethodNotFoundException;
    
    /**
     * Tests whether a method in the class is declared native.
     * 
     * @param methodSignature the signature of the method to be checked.
     * @return <code>true</code> if the method is native, <code>false</code> 
     *         if the method is not native.
     * @throws MethodNotFoundException iff <code>!hasMethodDeclaration(methodSignature)</code>.
     */
    public abstract boolean isMethodNative(Signature methodSignature) throws MethodNotFoundException;
    
    /**
     * Given the signature of a method, returns its exception table.
     * 
     * @param MethodSignature the <code>SESignature</code> of a method.
     * @return the <code>SEExceptionTable</code> of the method (empty in the case 
     *         the method has no exception handler).
     * @throws MethodNotFoundException iff <code>!hasMethodDeclaration(methodSignature)</code>.
     * @throws MethodCodeNotFoundException iff the method has no code.
     */
    public abstract ExceptionTable getExceptionTable(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException;
    
    /**
     * Given the signature of a method, returns a local variable table for that method.
     * 
     * @param methodSignature the structure that contains the signature of a method.
     * @return a <code>SELocalVariableTable</code> for the method.
     * @throws MethodNotFoundException iff <code>!hasMethodDeclaration(methodSignature)</code>.
     * @throws MethodCodeNotFoundException iff the method has no code.
     * @throws LocalVariableInfoNotFoundException if it cannot build a local variable table 
     * from the information contained in the class file. 
     */
    public abstract LocalVariableTable getLocalVariableTable(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException, LocalVariableInfoNotFoundException ;    

    /**
     * Returns a method's <code>SELineNumberTable</code>. 
     * @param methodSignature a <code>SESignature</code>.
     * @return the method's <code>SELineNumberTable</code>.
     * @throws MethodNotFoundException iff <code>!hasMethodDeclaration(methodSignature)</code>.
     * @throws MethodCodeNotFoundException iff the method has no code.
     * @throws LocalVariableInfoNotFoundException if it cannot build a line number table 
     * from the information contained in the class file. 
     */
    public abstract LineNumberTable getLineNumberTable(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException, LocalVariableInfoNotFoundException;    

    /**
     * Returns a value for the passed index of constant pool.
     */
    public abstract Value getValueFromConstantPool(int index) throws InvalidIndexException;

    /**
     * Given a Methodref index of the constant pool, returns the array 
     * of byte code; it is equivalent to 
     * <code>getMethodCodeBySignature(getMethodSignature(methodRef))</code>
     * @param methodRef the CONSTANT_Methodref of searched method
     * @return a <code>byte[]</code> containing the byte code.
     * @throws MethodNotFoundException iff 
     * <code>!hasMethodDeclaration(getMethodSignature(methodRef))</code>
     * @throws MethodCodeNotFoundException iff the method has no bytecode.
     */
    public byte[] getMethodCodeByMethodRef(int methodRef) throws MethodNotFoundException, MethodCodeNotFoundException {
        return (getMethodCodeBySignature(getMethodSignature(methodRef)));
    }
    
    /**
     * Given the signature of a method, returns the bytecode of 
     * the method. 
     * 
     * @param methodSignature the <code>SESignature</code> of a method.
     * @return a <code>byte[]</code> containing the byte code.
     */
    public abstract byte[] getMethodCodeBySignature(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException;
    
    /**
     * Checks whether the class declares a field.
     * 
     * @param fieldSignature the <code>SESignature</code> of a method. The 
     * class of the signature is ignored.
     * @return <code>true</code> iff the class declares a field with  
     * <code>fieldSignature</code>'s name and type.
     */
    public abstract boolean hasFieldDeclaration(Signature fieldSignature);
    
    public abstract boolean isFieldFinal(Signature fieldSignature) throws FieldNotFoundException;
    
    public abstract boolean hasFieldConstantValue(Signature fieldSignature) throws FieldNotFoundException;

    public abstract Value fieldConstantValue(Signature fieldSignature) throws FieldNotFoundException, AttributeNotFoundException;

    /**
     * Gets all the nonstatic (instance) fields declared by this class 
     * (not by its superclasses).
     *  
     * @return an array of <class>SESignature</class>, one for each 
     *         nonstatic field declared in the class.
     */
    public abstract Signature[] getFieldsNonStatic();
    
    /**
     * Gets all the static (class) fields declared by this class 
     * (not by its superclasses).
     *  
     * @return an array of <class>SESignature</class>, one for each 
     *         static field declared in the class.
     */
    public abstract Signature[] getFieldsStatic();
    
    /**
     *Given an index of the constant pool of CONSTANT_FieldRef type, returns the signature of the Field
     *@param fieldRef Is the CONSTANT_Fieldref of searched field
     *@return Signature Is the structure that contains the signature of a field.
     */
    public abstract Signature getFieldSignature(int fieldRef) throws InvalidIndexException ;
    
    /**
     *Given an index of the constant pool of CONSTANT_Methodref type, returns the signature of the Method
     *@param methodRef Is the CONSTANT_Methodref of searched field
     *@return Signature Is the structure that contains the signature of a method.
     */
    public abstract Signature getMethodSignature(int methodRef) throws InvalidIndexException ;
    
    /**
     *Given an index of the constant pool of CONSTANT_InterfaceMethodref type, returns the signature of the Method
     *@param methodRef Is the CONSTANT_InferfaceMethodref of searched field
     *@return Signature Is the structure that contains the signature of a method.
     */
    public abstract Signature getInterfaceMethodSignature(int methodRef) throws InvalidIndexException ;
    
    /**
     * Given an index of the constant table of CONSTANT_Class type, returns the signature 
     * of the class.
     * 
     * @param classRef the CONSTANT_Classref of searched class.
     * @return the <code>String</code> that contains the signature of a class, for a 
     * class the signature is only the name.
     */
    public abstract String getClassSignature(int classRef) throws InvalidIndexException;
    
    /**
     * Returns the name of the superclass.
     * 
     * @return the name of the superclass as a <code>String</code>.
     */
    public abstract String getSuperClassName();
    
    /**
     * Returns the list of the names of the superinterfaces.
     * 
     * @return a <code>List&lt;String&gt;</code> containing all the 
     *         names of this class' superinterfaces.
     */
    public abstract List<String> getSuperInterfaceNames();
    
    /**
     * Returns the name of the class.
     * 
     * @return the name of the class.
     */
    public abstract String getClassName();

    /**
     * Returns the length of the local variable table of a method.
     * 
     * @param methodSignature a <code>SESignature</code>.
     * @return a <code>int</code> representing the length in bytes of the 
     *         local variable table of the method with signature 
     *         <code>methodSignature</code>, in the case such method exists 
     *         in the class and has code.
     * @throws MethodNotFoundException iff the method does not exist in the class.
     * @throws MethodCodeNotFoundException iff the method has not the 
     *         Code attribute.
     */
    public abstract int getLocalVariableLength(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException;
    
    /**
     * Returns the length of the bytecode of a method.
     * 
     * @param methodSignature a <code>SESignature</code>.
     * @return a <code>int</code> representing the length in bytes of the 
     *         bytecode of the method with signature 
     *         <code>methodSignature</code>, in the case such method exists 
     *         in the class and has code.
     * @throws MethodNotFoundException iff the method does not exist in the class.
     * @throws MethodCodeNotFoundException iff the method has not the 
     *         Code attribute.
     */
    public abstract int getCodeLength(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException;

    /**
     * Opens a classfile contained in a .jar archive. 
     * 
     * @param jarFileName a <code>String</code> containing the pathname 
     *                    of a jar file.
     * @param className a <code>String</code> containing the name of a
     *                  class in internal format.
     * @return an <code>InputStream</code> to the class file.
     * @throws IOException iff it is unable to open the stream for any 
     *                     reason.
     */
    protected InputStream getClassFileFromJar(String jarFileName, String className) throws IOException {
    	JarFile jf = new JarFile(jarFileName);	
		String classFilePathNameJar = className + classFileExtension;
		JarEntry e = jf.getJarEntry(classFilePathNameJar);
		if (e != null) 
			return jf.getInputStream(e);
		else
			throw new IOException();
    }
    
    /**
     * Creates a default local variable table from a method's signature.
     * 
     * @param methodSignature the <code>SESignature</code> of a method.
     * @return a <code>SELocalVariableTable</code> containing entries 
     *         inferred from the method's parameters.
     */
    protected LocalVariableTable defaultLocalVariableTable(Signature methodSignature) {
        //if no LocalVariableTable attribute is found, tries to create the local 
        //variable table from information on the method's signature
    	boolean isStatic = this.isMethodStatic(methodSignature);
    	String[] parDescList = Type.splitParametersDescriptors(methodSignature.getDescriptor());
    	LocalVariableTable LVT = new LocalVariableTable((isStatic ? parDescList.length : (parDescList.length + 1)), this.getLocalVariableLength(methodSignature));
    	int i = 0;
    	short slot = 0;
    	if (!isStatic) {
        	LVT.setEntry(slot, Type.REFERENCE + this.getClassName() + Type.TYPEEND, 
        			     "this", 0, this.getCodeLength(methodSignature));
    		i++; slot++;
    	}
    	for (String descriptor : parDescList) {
    		LVT.setEntry(slot, descriptor, 
    				     "__PARAM[" + i + "]", 0, this.getCodeLength(methodSignature));
    		i++; slot++;
    		if (!Type.isCat_1(descriptor.charAt(0))) {
    			slot++;
    		}
    	}
    	return LVT;
    }
    
    /**
     * Creates the default line number table to be returned in 
     * the case a method has not the LineNumberTable attribute.
     * 
     * @return the default (empty) <code>SELineNumberTable</code>.
     */
    protected LineNumberTable defaultLineNumberTable() {
    	return new LineNumberTable(0);
    }
}