package muse.bc;

import java.util.LinkedList;
import java.util.List;

import muse.mem.ConstantPoolString;
import muse.mem.Simplex;
import muse.mem.Value;


import javassist.*;
import javassist.bytecode.AccessFlag;
import javassist.bytecode.CodeAttribute;
import javassist.bytecode.ConstPool;
import javassist.bytecode.LineNumberAttribute;
import javassist.bytecode.LocalVariableAttribute;
//also uses javassist.bytecode.ExceptionTable, not imported to avoid name clash

public class ClassFileJavassist extends ClassFile {
	private CtClass cls;
	private ConstPool cp;
	
	public ClassFileJavassist(String className, ClassPool cpool) {
		try {
			this.cls = cpool.get(className.replace("/", "."));
			this.cp = this.cls.getClassFile().getConstPool();
		} catch (NotFoundException e) {
			throw new ClassFileNotFoundException();
		}
	}

	@Override
	public String getClassName() {
		return this.cls.getName().replace(".", "/");
	}

	@Override
	public String getClassSignature(int classRef) throws InvalidIndexException {
		if (classRef < 1 || classRef > this.cp.getSize() ||
				this.cp.getTag(classRef) != ConstPool.CONST_Class)
			throw new InvalidIndexException();
		return this.cp.getClassInfo(classRef).replace(".", "/");
	}

	@Override
	public boolean isPublic() {
		return Modifier.isPublic(this.cls.getModifiers());
	}

	@Override
	public boolean isSuperInvoke() {
		//note that we use getClassFile().getAccessFlag() because 
		//getModifiers() does not provide the ACC_SUPER flag
		return ((this.cls.getClassFile().getAccessFlags() & AccessFlag.SUPER) != 0);	
	}

	@Override
	public Signature getFieldSignature(int fieldRef) throws InvalidIndexException {
		if (fieldRef < 1 || fieldRef > this.cp.getSize() ||
				this.cp.getTag(fieldRef) != ConstPool.CONST_Fieldref) {
			throw new InvalidIndexException();
		}
        String containerClass = this.cp.getFieldrefClassName(fieldRef).replace('.', '/');
        String descriptor = this.cp.getFieldrefType(fieldRef);
        String name = this.cp.getFieldrefName(fieldRef);
        return new Signature(containerClass, descriptor, name);
	}

	private List<Signature> getFields(boolean areStatic) {
    	List<Signature> fields = new LinkedList<Signature>();
    	CtField[] fieldsJA = this.cls.getDeclaredFields();
    	for (CtField fld : fieldsJA) {
    		if (Modifier.isStatic(fld.getModifiers()) == areStatic) {
    			Signature sig = new Signature(this.getClassName(), fld.getSignature(), fld.getName());
    			fields.add(sig);
    		}
    	}
    	return fields;
	}
	
	@Override
	public Signature[] getFieldsNonStatic() {
    	List<Signature> fieldsList = this.getFields(false);
    	Signature[] retval = new Signature[fieldsList.size()];
    	fieldsList.<Signature>toArray(retval);
    	return retval;
	}

	@Override
	public Signature[] getFieldsStatic() {
    	List<Signature> fieldsList = this.getFields(true);
    	Signature[] retval = new Signature[fieldsList.size()];
    	fieldsList.<Signature>toArray(retval);
    	return retval;
	}

	@Override
	public Signature getInterfaceMethodSignature(int methodRef) throws InvalidIndexException {
		if (methodRef < 1 || methodRef > this.cp.getSize() ||
				this.cp.getTag(methodRef) != ConstPool.CONST_InterfaceMethodref)
			throw new InvalidIndexException();
		return new Signature(this.cp.getInterfaceMethodrefClassName(methodRef).replace(".", "/"), 
				this.cp.getInterfaceMethodrefType(methodRef), 
				this.cp.getInterfaceMethodrefName(methodRef));
	}


	/**
	 * Finds a method declaration in the classfile.
	 * 
	 * @param methodSignature a <code>SESignature</code>
	 * @return <code>null</code> if no method with <code>methodSignature</code> 
	 *         signature is declared in <code>this</code>, otherwise the 
	 *         <code>CtMethod</code> for it; the class name in <code>methodSignature</code>
	 *         is ignored
	 */
	private CtBehavior findMethod(Signature methodSignature) {
		CtConstructor cc = this.cls.getClassInitializer();
		if (methodSignature.getName().equals("<clinit>"))
			return cc;
		
		CtBehavior[] bs = this.cls.getDeclaredBehaviors();
		for (CtBehavior b : bs) {
			String internalName = 
				(((b instanceof CtConstructor) && (!((CtConstructor) b).isClassInitializer())) ? 
						"<init>" : 
						b.getName());
			if (internalName.equals(methodSignature.getName()) &&
				b.getSignature().equals(methodSignature.getDescriptor()))
				return b;
		}
		return null;
	}
	
	private CodeAttribute getMethodCodeAttribute(Signature methodSignature) 
	 throws MethodNotFoundException, MethodCodeNotFoundException {
		CtBehavior b = this.findMethod(methodSignature);
		if (b == null) throw new MethodNotFoundException();
		CodeAttribute ca = b.getMethodInfo().getCodeAttribute();
		if (ca == null) throw new MethodCodeNotFoundException(); 
		return ca;
	}
	

	@Override
	public ExceptionTable getExceptionTable(Signature methodSignature)
	    throws MethodNotFoundException, MethodCodeNotFoundException {
		javassist.bytecode.ExceptionTable et = this.getMethodCodeAttribute(methodSignature).getExceptionTable();

		ExceptionTable retVal = new ExceptionTable(et.size());
		for (int i = 0; i < et.size(); i++) {
	        ExceptionTableEntry exEntry = new ExceptionTableEntry();
            exEntry.setStartPC(et.startPc(i));
            exEntry.setEndPC(et.endPc(i));
            exEntry.setPCHandle(et.handlerPc(i));
            int exType = et.catchType(i);
            if (exType == 0) { //any exception
            	exEntry.setType("java/lang/Throwable");
            } else {
            	exEntry.setType(this.getClassSignature(exType));
            }
            retVal.addEntry(exEntry);
		}
		return retVal;
	}
	
	@Override
	public int getLocalVariableLength(Signature methodSignature)
	 throws MethodNotFoundException, MethodCodeNotFoundException {
		return this.getMethodCodeAttribute(methodSignature).getMaxLocals();
	}

	@Override
	public int getCodeLength(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException {
		return this.getMethodCodeAttribute(methodSignature).getCodeLength();
	}

	@Override
	public LocalVariableTable getLocalVariableTable(Signature methodSignature) 
	 throws MethodNotFoundException, MethodCodeNotFoundException, LocalVariableInfoNotFoundException  {
		CodeAttribute ca = this.getMethodCodeAttribute(methodSignature);
		LocalVariableAttribute lvtJA = (LocalVariableAttribute) ca.getAttribute("LocalVariableTable");
		
        if (lvtJA == null)
        	return this.defaultLocalVariableTable(methodSignature);
        	
        //builds the local variable table from the LocalVariableTable attribute 
    	//information; this has always success
        LocalVariableTable LVT = new LocalVariableTable(lvtJA.tableLength(), ca.getMaxLocals());
        for (int i = 0; i < lvtJA.tableLength(); i++) {
        	LVT.setEntry(lvtJA.index(i), lvtJA.descriptor(i), 
        			     lvtJA.variableName(i), lvtJA.startPc(i),  lvtJA.codeLength(i));
            }
        return LVT;
	}

	@Override
	public byte[] getMethodCodeBySignature(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException {
		return this.getMethodCodeAttribute(methodSignature).getCode();
	}

	@Override
	public Signature getMethodSignature(int methodRef) throws InvalidIndexException {
		if (methodRef < 1 || methodRef > this.cp.getSize() ||
				this.cp.getTag(methodRef) != ConstPool.CONST_Methodref)
			throw new InvalidIndexException();
		return new Signature(this.cp.getMethodrefClassName(methodRef).replace(".", "/"), 
				this.cp.getMethodrefType(methodRef), 
				this.cp.getMethodrefName(methodRef));
	}

	@Override
	public String getSuperClassName() {
		String name = this.cls.getClassFile().getSuperclass();
		if (name != null) {
			name = name.replace(".", "/");
		}
		return name;
	}

	@Override
	public List<String> getSuperInterfaceNames() {
		LinkedList<String> retVal = new LinkedList<String>();
		String[] ifs = this.cls.getClassFile().getInterfaces();
		
		for (String s : ifs) {
			retVal.add(s.replace(".", "/"));
		}
		return retVal;
	}

	@Override
	public Value getValueFromConstantPool(int index) throws InvalidIndexException {
		if (index < 1 || index > this.cp.getSize()) {
	        throw new InvalidIndexException();
		}
		int tag = this.cp.getTag(index);
		switch (tag) {
        case(ConstPool.CONST_Integer):
            return(new Simplex(this.cp.getIntegerInfo(index)));
        case(ConstPool.CONST_Float):
            return(new Simplex(this.cp.getFloatInfo(index)));
        case(ConstPool.CONST_Long):
            return(new Simplex(this.cp.getLongInfo(index)));
        case(ConstPool.CONST_Double):
            return(new Simplex(this.cp.getDoubleInfo(index)));
        case(ConstPool.CONST_String):
            return(new ConstantPoolString(this.cp.getStringInfo(index)));
		}
        throw new InvalidIndexException();
	}
	
	@Override
	public boolean hasMethodDeclaration(Signature methodSignature) {
		return (this.findMethod(methodSignature) != null);
	}

	@Override
	public boolean hasMethodImplementation(Signature methodSignature) {
		CtBehavior b = this.findMethod(methodSignature);
		return (b != null && (b.getMethodInfo().getCodeAttribute() != null || Modifier.isNative(b.getModifiers())));
	}

	@Override
	public boolean isAbstract() {
		return Modifier.isAbstract(this.cls.getModifiers());
	}

	@Override
	public boolean isInterface() {
		return this.cls.isInterface();
	}

	@Override
	public boolean isMethodAbstract(Signature methodSignature) throws MethodNotFoundException {
		CtBehavior b = this.findMethod(methodSignature);
		if (b == null) throw new MethodNotFoundException();
		return Modifier.isAbstract(b.getModifiers());
	}

	@Override
	public boolean isMethodNative(Signature methodSignature) throws MethodNotFoundException {
		CtBehavior b = this.findMethod(methodSignature);
		if (b == null) throw new MethodNotFoundException();
		return Modifier.isNative(b.getModifiers());
	}

	@Override
	public boolean isMethodStatic(Signature methodSignature) throws MethodNotFoundException {
		CtBehavior b = this.findMethod(methodSignature);
		if (b == null) {
			throw new MethodNotFoundException();
		}
		return Modifier.isStatic(b.getModifiers());
	}

	@Override
	public boolean hasFieldDeclaration(Signature fieldSignature) {
		return (this.findField(fieldSignature) != null);
	}

	@Override
	public LineNumberTable getLineNumberTable(Signature methodSignature) throws MethodNotFoundException, MethodCodeNotFoundException, LocalVariableInfoNotFoundException {
		CodeAttribute ca = this.getMethodCodeAttribute(methodSignature);
		LineNumberAttribute lnJA = (LineNumberAttribute) ca.getAttribute("LineNumberTable");
		
		if (lnJA == null)
			return this.defaultLineNumberTable();
		LineNumberTable LN = new LineNumberTable(lnJA.tableLength());
		for (int i = 0; i < lnJA.tableLength(); i++) {
			LN.addRow(lnJA.startPc(i), lnJA.lineNumber(i));
        }
        return LN;
	}

	@Override
	public Value fieldConstantValue(Signature fieldSignature) throws FieldNotFoundException, AttributeNotFoundException {
		CtField fld = this.findField(fieldSignature);
		if (fld == null) {
			throw new FieldNotFoundException();
		}
		int cpVal = fld.getFieldInfo().getConstantValue();
		if (cpVal == 0) {
			throw new AttributeNotFoundException();
		}
		return this.getValueFromConstantPool(cpVal);
	}

	@Override
	public boolean hasFieldConstantValue(Signature fieldSignature) throws FieldNotFoundException {
		CtField fld = this.findField(fieldSignature);
		if (fld == null) {
			throw new FieldNotFoundException();
		}
		return (fld.getConstantValue() != null);
	}

	@Override
	public boolean isFieldFinal(Signature fieldSignature) throws FieldNotFoundException {
		CtField fld = this.findField(fieldSignature);
		if (fld == null) {
			throw new FieldNotFoundException();
		}
		return Modifier.isFinal(fld.getModifiers());
	}
	
	private CtField findField(Signature fieldSignature) {
    	CtField[] fieldsJA = this.cls.getDeclaredFields();
    	for (CtField fld : fieldsJA) {
    		if (fld.getSignature().equals(fieldSignature.getDescriptor()) && 
    				fld.getName().equals(fieldSignature.getName())) {
    			return fld;
    		}
    	}
    	return null;
	}
}