package muse.bc;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;

/**
 * Class allowing to handle the hierarchy of Java classes as specified 
 * by some <code>SEEnv</code>. 
 *  
 * @author pietro
 *
 */
public class ClassHierarchy {
	private ClassFileInterface cfi;

	/**
	 * Constructor.
	 * 
	 * @param env a <code>SEEnv</code> specifying the execution environment
	 *            conditions for finding Java classes.  
	 */
	public ClassHierarchy(ClassFileInterface cfi) {
		this.cfi = cfi;
	}

	/**
	 * Produces all the superclasses of a given class.
	 * 
	 * @param startClassName the name of the class whose superclasses 
	 *                       are returned.
	 * @return an <code>Iterable&lt;SEClassFile&gt;</code> containing 
	 *         all the superclasses of <code>startClassName</code> (included).
	 */
	public Iterable<ClassFile> superclasses(String startClassName) {
		return new ListSuperclasses(startClassName);
	}

	/**
	 * Produces all the superclasses of a given class.
	 * 
	 * @param startClassName the name of the class whose superclasses 
	 *                       are returned.
	 * @return an <code>Iterable&lt;SEClassFile&gt;</code> containing 
	 *         all the superclasses of <code>startClassName</code> (included).
	 */
	public Iterable<ClassFile> superinterfaces(String startClassName) {
		return new ListSuperinterfaces(startClassName);
	}

	/**
	 * Checks whether a class is a subclass of another one. Does not 
	 * work for interfaces.
	 * 
	 * @param sub a class name
	 * @param sup another class name
	 * @return <code>true</code> if <code>sub.equals(sup)</code> of <code>sub</code> 
	 *         is a subclass of <code>sup</code>, <code>false</code> otherwise.
	 */
	public boolean isSubclass(String sub, String sup) {
		//TODO verify subclass check has the JVM specification semantics.
		for (ClassFile f : superclasses(sub)) { 
			String s = f.getClassName();
			if (s.equals(sup)) {
				return true;
			}
		}
		for (ClassFile f : superinterfaces(sub)) {
			if (f.getClassName().equals(sup)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * <code>Iterable&lt;SEClassFile&gt;</code> for upwardly scanning 
	 * a class hierarchy.
	 *  
	 * @author pietro
	 */
	private class ListSuperclasses implements Iterable<ClassFile> {
		private String startClassName;

		/**
		 * Constructor.
		 * 
		 * @param startClassName 
		 *        The name of the class from where the iteration is started. 
		 *        Note that the iterator firstly visits this class: thus, 
		 *        the first call to <code>hasNext()</code> will return true 
		 *        iff <code>startClassName != null</code> and such class 
		 *        exists in the environment defined by <code>SEEnv.this.env</code>.
		 */
		public ListSuperclasses(String startClassName) {
			this.startClassName = startClassName;
		}

		public Iterator<ClassFile> iterator() {
			return new MyIterator(this.startClassName);
		}        

		/**
		 * <code>Iterator&lt;SEClassFile&gt;</code> for
		 * upwardly scanning a class hierarchy.
		 * 
		 * @author Pietro Braione
		 */
		private class MyIterator implements Iterator<ClassFile> {
			private String nextClassName;

			public MyIterator(String startClassName) {
				this.nextClassName = null;

				ClassFile cf = null;
				try {
					cf = ClassHierarchy.this.cfi.getClassFile(startClassName);
				} catch (ClassFileNotFoundException ce) {
					return;
				}
				this.nextClassName = cf.getClassName();
			}

			public boolean hasNext() {
				return (nextClassName != null);
			}

			public ClassFile next() {
				//ensures the method precondition
				if (!this.hasNext())
					throw new NoSuchElementException();

				//gets the next class' class file (should always be successful)
				ClassFile retval = null;
				try {
					retval = ClassHierarchy.this.cfi.getClassFile(nextClassName);
				} catch (ClassFileNotFoundException e) {
					throw new RuntimeException("Unexpected error while iterating class hierarchy");
				} 

				//prepares for the next invocation
				this.nextClassName = retval.getSuperClassName();

				//returns the result
				return retval;
			}

			public void remove() {
				throw new UnsupportedOperationException();
			}
		}
	}

	/**
	 * <code>Iterable&lt;SEClassFile&gt;</code> for upwardly scanning 
	 * an interface hierarchy.
	 *  
	 * @author pietro
	 */
	private class ListSuperinterfaces implements Iterable<ClassFile> {
		private String startClassName;

		/**
		 * Constructor.
		 * 
		 * @param startClassName 
		 *        The name of the class from where the iteration is started. 
		 *        Note that the iterator firstly visits this class iff it is 
		 *        an interface: thus, the first call to <code>hasNext()</code> 
		 *        will return true iff <code>startClassName != null</code> and 
		 *        <code>startClassName</code> exists in the environment 
		 *        defined by <code>SEEnv.this.env</code>, and it is an 
		 *        interface.
		 */
		public ListSuperinterfaces(String startClassName) {
			this.startClassName = startClassName;
		}

		public Iterator<ClassFile> iterator() {
			return new MyIterator(this.startClassName);
		}        

		/**
		 * <code>Iterator&lt;SEClassFile&gt;</code> for
		 * upwardly scanning an interface hierarchy. For 
		 * the sake of simplicity it scans in breadth-first 
		 * order. It does not visit a same interface twice. 
		 * 
		 * @author Pietro Braione
		 */
		private class MyIterator implements Iterator<ClassFile> {
			private LinkedList<String> nextClassNames;
			private HashSet<String> visited;

			public MyIterator(String startClassName) {
				this.nextClassNames = new LinkedList<String>();
				this.visited = new HashSet<String>();

				ClassFile cf = null;
				try {
					cf = ClassHierarchy.this.cfi.getClassFile(startClassName);
				} catch (ClassFileNotFoundException ce) {
					return;
				}
				if (cf.isInterface()) 
					this.nextClassNames.add(cf.getClassName());
				else
					this.nextClassNames.addAll(cf.getSuperInterfaceNames());
			}

			public boolean hasNext() {
				return !(this.nextClassNames.isEmpty());
			}

			public ClassFile next() {
				//ensures the method precondition
				if (!this.hasNext())
					throw new NoSuchElementException();

				//gets the next unvisited interface names
				String nextClassName = this.nextClassNames.getFirst(); 
				this.visited.add(nextClassName);

				//gets its class file (should always be successful)
				ClassFile retval = null;
				try {
					retval = ClassHierarchy.this.cfi.getClassFile(nextClassName);
				} catch (ClassFileNotFoundException e) {
					throw new RuntimeException("Unexpected error while iterating interface hierarchy");
				} 

				//prepares for the next invocation:
				//1 - adds all the superinterfaces of the visited one to the 
				//    yet-to-visit interfaces
				this.nextClassNames.addAll(retval.getSuperInterfaceNames());
				//2 - removes all the already-visited superinterfaces from 
				//    the yet-to-visit ones
				this.nextClassNames.removeAll(this.visited);

				//returns the result
				return retval;
			}

			public void remove() {
				throw new UnsupportedOperationException();
			}
		}
	}
}
