package muse.dec;

import java.util.SortedSet;

import muse.mem.Expression;
import muse.mem.Primitive;
import muse.tree.DecisionAlternativeBoolean;
import muse.tree.DecisionAlternativeComparison;
import muse.tree.DecisionAlternativeSetFactory;
import muse.tree.DecisionAlternativeSwitch;


/**
 * Interface for a decision procedure. A <code>DecisionProcedure</code> accumulates 
 * a satisfiable clause and allows to check whether another clause is satisfiable
 * under it.
 */
public abstract class DecisionProcedure {
	protected long time;
	private DecisionAlternativeSetFactory c;

    public DecisionProcedure() { 
    	this.time = 0L;
    	this.c = new DecisionAlternativeSetFactory();
    }
    
    public DecisionProcedure(DecisionAlternativeSetFactory c) {
    	this.time = 0L;
    	this.c = c;
    }
    
    public DecisionAlternativeSetFactory getComparators() { 
    	return this.c; 
    }
    
    /**
     * Adds a clause in the simplifier's clause accumulator.
     * 
     * @param c the clause to be added as a <code>Primitive</code>.
     * @throws DecisionException upon failure.
     */
    public abstract void pushClause(Primitive c) throws DecisionException;
    
    /**
     * Drops the simplifier's current accumulated clause.
     * 
     * @throws DecisionException upon failure.
     */
    public abstract void resetClause() throws DecisionException;

    /**
     * Drops the simplifier's current accumulated clause and 
     * replaces it with another one.  
     * 
     * @param c the clause which must replace the currently accumulated one.
     * @throws DecisionException upon failure.
     */
    public void setClause(Primitive c) throws DecisionException {
    	this.resetClause();
    	this.pushClause(c);
    }
    
    public abstract Primitive getAccumulatedClause() throws DecisionException;
    
    /**
     * Determines the satisfiability of an <code>Expression</code> under the
     * <code>DecisionProcedure</code>'s accumulated clause.
     * 
     * @param exp an <code>Expression</code>.
     * @return <code>true</code> iff <code>exp</code> is satisfiable under
     *         the <code>DecisionProcedure</code>'s accumulated clause.
     * @throws DecisionException upon failure.
     */
    public abstract boolean isSat(Expression exp) throws DecisionException;
    
    /**
     * Simplifies a decision for a branch statement.
     * 
     * @param exp an <code>Expression</code> representing a logical clause.
     * @param upd a <code>SortedSet&lt;DecisionAlternativeBoolean&gt;</code>, which the metod 
     *            will updated by adding to it <code>DecisionAlternativeBoolean.TT</code> 
     *            iff <code>exp</code> is satisfiable under the simplifier's 
     *            accumulated clause, and <code>DecisionAlternativeBoolean.FF</code> iff
     *            <code>exp.not()</code> is satisfiable under the simplifier's 
     *            accumulated clause. Note that the two conditions are not 
     *            mutually exclusive.
     * @throws DecisionException upon failure.
     */
    public abstract void simplifyDecision(Expression exp, SortedSet<DecisionAlternativeBoolean> upd) throws DecisionException;
    
    /**
     * Simplifies a comparison for a compare statement.
     * 
     * @param val1 a <code>Primitive</code>.
     * @param val2 another <code>Primitive</code>.
     * @param upd a <code>SortedSet&lt;ComparisonValue&gt;</code>, which the metod 
     *            will update by adding to it <code>ComparisonValue.GT</code>, 
     *            (respectively, <code>ComparisonValue.EQ</code>, <code>ComparisonValue.LT</code>)
     *            iff <code>val1.gt(val2)</code> (respectively, <code>val1.eq(val2)</code>, 
     *            <code>val1.lt(val2)</code>) is satisfiable under the simplifier's 
     *            accumulated clause. Note that the three conditions are not 
     *            mutually exclusive.
     * @throws DecisionException upon failure.
     */
    public abstract void simplifyComparison(Primitive val1, Primitive val2, SortedSet<DecisionAlternativeComparison> upd) throws DecisionException;

    /**
     * Simplifies a table or a range for a switch statement.
     * 
     * @param val a <code>Primitive</code>. 
     * @param isTableSwitch <code>true</code> iff the semantics is that of a tableswitch.
     * @param low the lower value for the switch table. 
     * @param high the higher value for the switch table.
     * @param tableStart the start of the table in <code>code</code>.
     * @param tableEnd the end of the table in <code>code</code>.
     * @param code the bytecode containing the switch table.
     * @param upd a <code>SortedSet&lt;DecisionAlternativeSwitch&gt;</code>, which the metod 
     *            will update by adding to it all the <code>DecisionAlternativeSwitch</code>s 
     *            such that the equality of <code>val</code> to the associated
     *            value in the switch table is satisfiable under the simplifier's 
     *            accumulated clause.
     * @throws DecisionException upon failure.
     */
    public abstract void simplifySwitch(Primitive val, boolean isTableSwitch, int low, int high, int tableStart, int tableEnd, byte[] code, SortedSet<DecisionAlternativeSwitch> upd) throws DecisionException;
    
    /**
     * Quits the simplifier. It must be invoked after the use of the 
     * simplifier to release resources.
     * 
     * @throws DecisionException upon failure.
     */
    public void quit() throws DecisionException { };
    
    /**
     * Gets the time spent by the simplifier's computations.
     * 
     * @return a <code>long</code> representing the spent time 
     *         in milliseconds.
     */
    public long getTime() {
    	return this.time;
    }
}