package muse.dec;

import java.io.PrintStream;
import java.util.SortedSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import muse.apps.run.IO;
import muse.mem.Expression;
import muse.mem.Primitive;
import muse.tree.DecisionAlternativeBoolean;
import muse.tree.DecisionAlternativeComparison;
import muse.tree.DecisionAlternativeSwitch;


public class DecisionProcedureConsole extends DecisionProcedure {
	private static final String TURNSTILE = " |- ";
	private static final String PROMPT = "? > ";
	private final DecisionProcedureAlwSat bs = new DecisionProcedureAlwSat();
	private final PrintStream[] ps;
	
	public DecisionProcedureConsole(PrintStream[] ps) {
		this.ps = ps;
	}
	
	@Override
	public void pushClause(Primitive c) {
		this.bs.pushClause(c);
	}

	@Override
	public void resetClause() {
		this.bs.resetClause();
	}

	@Override
	public Primitive getAccumulatedClause() {
		return this.bs.getAccumulatedClause();
	}

	@Override
	public boolean isSat(Expression exp) {
		long start = System.currentTimeMillis();
		boolean retVal;
		IO.println(this.ps, "SESimplifierInteractiveConsole: Please solve the following clause:");
		IO.println(this.ps, "SESimplifierInteractiveConsole: " +  this.getAccumulatedClause() + TURNSTILE + exp.toString());
		IO.println(this.ps, "SESimplifierInteractiveConsole: Possible results: unsatisfiable (u), satisfiable (any other).");
    	String ans = IO.readln(this.ps, PROMPT);

        if (ans.equals("u")) {
        	retVal = false;
        } else  {
        	retVal = true;
        }
        this.time += System.currentTimeMillis() - start;
        return retVal;
	}

	@Override
    public void simplifyDecision(Expression exp, SortedSet<DecisionAlternativeBoolean> upd) {
		long start = System.currentTimeMillis();
		IO.println(this.ps, "SESimplifierInteractiveConsole: Please solve the following condition:");
		IO.println(this.ps, "SESimplifierInteractiveConsole: " + this.getAccumulatedClause() + TURNSTILE + exp);
		IO.println(this.ps, "SESimplifierInteractiveConsole: Possible results: always true (t), always false (f), both (any other).");
        String ans = IO.readln(this.ps, PROMPT);
        
        if (ans.equals("t")) {
        	upd.add(DecisionAlternativeBoolean.TT);
        } else if (ans.equals("f")) {
        	upd.add(DecisionAlternativeBoolean.FF);
        } else {
        	upd.add(DecisionAlternativeBoolean.TT);
        	upd.add(DecisionAlternativeBoolean.FF);
        }
        this.time += System.currentTimeMillis() - start;
    }
    
    @Override
    public void simplifyComparison(Primitive val1, Primitive val2, SortedSet<DecisionAlternativeComparison> upd) {        
		long start = System.currentTimeMillis();
    	IO.println(this.ps, "SESimplifierInteractiveConsole: Please solve the following comparison:");
		IO.println(this.ps, "SESimplifierInteractiveConsole: " +  this.getAccumulatedClause() + TURNSTILE + val1.toString() + ">=<" + val2.toString());
    	IO.println(this.ps, "SESimplifierInteractiveConsole: Possible results: always greater (G), equals (E), always less (L), greater or equal (g), less or equal (e), different (d), all cases possible (any other).");
    	String ans = IO.readln(this.ps, PROMPT);

        if (ans.equals("G")) {
        	upd.add(DecisionAlternativeComparison.GT);
        } else if (ans.equals("E")) {
        	upd.add(DecisionAlternativeComparison.EQ);
        } else if (ans.equals("L")) {
        	upd.add(DecisionAlternativeComparison.LT);
        } else if (ans.equals("g")) {
        	upd.add(DecisionAlternativeComparison.GT);
        	upd.add(DecisionAlternativeComparison.EQ);
        } else if (ans.equals("l")) {
        	upd.add(DecisionAlternativeComparison.EQ);
        	upd.add(DecisionAlternativeComparison.LT);
        } else if (ans.equals("d")) {
        	upd.add(DecisionAlternativeComparison.GT);
        	upd.add(DecisionAlternativeComparison.LT);
        } else {
        	upd.add(DecisionAlternativeComparison.GT);
        	upd.add(DecisionAlternativeComparison.EQ);
        	upd.add(DecisionAlternativeComparison.LT);
        }
        this.time += System.currentTimeMillis() - start;
    }

	@Override
	public void simplifySwitch(Primitive val, boolean isTableSwitch, int low, int high, int tableStart, int tableEnd, byte[] code, SortedSet<DecisionAlternativeSwitch> upd) throws DecisionException {
		long start = System.currentTimeMillis();
		IO.println(this.ps, "SESimplifierInteractiveConsole: Please solve the following " + (isTableSwitch? "tableswitch" : "lookupswitch") + ":");
		if (isTableSwitch) {
			IO.println(this.ps, "SESimplifierInteractiveConsole: " + this.getAccumulatedClause() + TURNSTILE + tableStart + "..." + tableEnd);
		} else {
			IO.println(this.ps, "SESimplifierInteractiveConsole: " + this.getAccumulatedClause() + TURNSTILE + low + "..." + high);
		}
		IO.println(this.ps, "SESimplifierInteractiveConsole: Provide a blank-separated list of integer values plus d for default.");
		String ans = IO.readln(this.ps, PROMPT);

        Pattern p = Pattern.compile("(-?\\d+|d)");
        Matcher m = p.matcher(ans);

        while (m.find()) {
        	String s = m.group();
        	try {
            	int i = Integer.parseInt(s);
            	upd.add(new DecisionAlternativeSwitch(i));        		
        	} catch (NumberFormatException e) {
        		upd.add(DecisionAlternativeSwitch.alternativeDefault());
        	}
        }
        this.time += System.currentTimeMillis() - start;
	}
	
	@Override
	public long getTime() {
		return this.time + this.bs.getTime();
	}
}