package muse.dec;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.HashSet;
import java.util.Stack;

import muse.mem.Expression;
import muse.mem.ExpressionVisitor;
import muse.mem.FunctionApplication;
import muse.mem.Operator;
import muse.mem.Primitive;
import muse.mem.PrimitiveSymbolic;
import muse.mem.Simplex;
import muse.mem.Term;
import muse.mem.Type;


/**
 * Interface with the CVC3 SMT solver.
 * 
 * @author pietro
 */
public class DecisionProcedureExternalInterfaceCVC3 extends DecisionProcedureExternalInterface {
	private static final String lineSep = System.getProperty("line.separator");
	private static final String prompt = "CVC> ";
	private static final String prologue = 
		"J_byte : TYPE = SUBTYPE( LAMBDA (x : INT) : x >= -128 AND x <= 127 ); " +
		"J_short : TYPE = SUBTYPE( LAMBDA (x : INT) : x >= -32768 AND x <= 32767 ); " +
		"J_int : TYPE = SUBTYPE( LAMBDA (x : INT) : x >= -2147483648 AND x <= 2147483647 ); " +
		"J_long : TYPE = SUBTYPE ( LAMBDA (x : INT) : x >= -9223372036854775808 AND x <= 9223372036854775807 ); " +
		"J_float : TYPE = SUBTYPE ( LAMBDA (x : REAL) : x >= (14 * (10^(-46))) AND x <= (34028235 * (10^31)) ); " +
		"J_double : TYPE = SUBTYPE ( LAMBDA (x : REAL) : x >= (49 * (10^(-325))) AND x <= (17976931348623157 * (10^292)) ); " +
		"J_boolean : TYPE = SUBTYPE ( LAMBDA (x : INT) : x >= 0 AND x <= 1 ); " +
		"J_char : TYPE = SUBTYPE ( LAMBDA (x : INT) : x >= 0 AND x <= 65535 ); " + lineSep;
	private static final String push = "PUSH;" + lineSep;
	private static final String pop = "POP;" + lineSep;
	private static final Simplex tt = new Simplex(true);
	
	private boolean working;
	private Process cvc3;
	private BufferedReader cvc3In;
	private BufferedWriter cvc3Out;
	private Primitive cond;
	private Primitive context;

	public DecisionProcedureExternalInterfaceCVC3(String path) throws DecisionExternalProtocolException, IOException {
		this.working = true;
        this.cvc3 = Runtime.getRuntime().exec(path + " +int"); 
        this.cvc3In = new BufferedReader(new InputStreamReader(this.cvc3.getInputStream()));
        this.cvc3Out = new BufferedWriter(new OutputStreamWriter(this.cvc3.getOutputStream()));
        this.readPrompt();
        this.cvc3Out.write(prologue);
        this.cvc3Out.flush();
        this.readPrompt();
        this.clear();
	}
	
	private void readPrompt() throws IOException, DecisionExternalProtocolException {
        for (int c : prompt.toCharArray()) {
        	if (c != this.cvc3In.read()) {
        		this.working = false;
        		throw new DecisionExternalProtocolException();
        	}
        }
	}

	public boolean isWorking() {
		return this.working;
	}

	public void clear() throws DecisionExternalProtocolException, IOException {
		this.discardFormula();
		this.context = tt;
	}

	public void putCondition(Primitive cond) throws DecisionExternalProtocolException, IOException {
		this.cond = cond;
	}
	
	public void appendCondToAssumptionsAs(boolean value) throws DecisionExternalProtocolException, IOException {
		if (value) {
			this.context = this.context.and(this.cond);
		} else {
			this.context = this.context.and(this.cond.not());
		}
	}

	public void discardFormula() throws DecisionExternalProtocolException, IOException {
		this.cond = tt;
	}
	
	private static final String OTHER = "";
	
	private static boolean yieldsBooleanOperator(String operator) {
		return (operator.equals(Operator.LT))  ||
               (operator.equals(Operator.LE))  ||
               (operator.equals(Operator.EQ))  ||
               (operator.equals(Operator.NE))  ||
               (operator.equals(Operator.GE))  ||
               (operator.equals(Operator.GT))  ||
		       (operator.equals(Operator.AND)) ||
               (operator.equals(Operator.OR))  ||
               (operator.equals(Operator.NOT));
	}
	
	private static boolean isBooleanOperator(String operator) {
		return (operator.equals(Operator.AND)) ||
               (operator.equals(Operator.OR))  ||
               (operator.equals(Operator.NOT));
	}
	
	private static String CVC3Op(String operator) {
        if (operator.equals(Operator.ADD))      return "+";
        else if (operator.equals(Operator.SUB)) return "-";
        else if (operator.equals(Operator.MUL)) return "*";
        else if (operator.equals(Operator.DIV)) return "/";
        else if (operator.equals(Operator.NEG)) return "-";
        else if (operator.equals(Operator.LT))  return "<";
        else if (operator.equals(Operator.LE))  return "<=";
        else if (operator.equals(Operator.EQ))  return "=";
        else if (operator.equals(Operator.NE))  return "/=";
        else if (operator.equals(Operator.GE))  return ">=";
        else if (operator.equals(Operator.GT))  return ">";
        else if (operator.equals(Operator.AND)) return "AND";
        else if (operator.equals(Operator.OR))  return "OR";
        else if (operator.equals(Operator.NOT)) return "NOT";
        else return OTHER;
	}

	private static String CVC3PrimitiveType(char type) {
		if (type == Type.BYTE) {
			return "J_byte"; 
		} else if (type == Type.SHORT) {
			return "J_short"; 
		} else if (type == Type.INT) {
			return "J_int";
		} else if (type == Type.LONG) {
			return "J_long";
		} else if (type == Type.FLOAT) {
			return "J_float";
		} else if (type == Type.DOUBLE) {
			return "J_double";
		} else if (type == Type.BOOLEAN) {
			return "J_boolean";
		} else if (type == Type.CHAR) {
			return "J_char";
		} else {
			//this should be unreachable!
			return OTHER;
		}
	}

	private static String CVC3Double(String s) {
		String[] parts = s.split("E");
		String[] partsMantissa = parts[0].split("\\."); 
		long exponent;
		if (s.length() == 2) {
			exponent = Long.parseLong(parts[1]);
		} else {
			exponent = 0L;
		}
		if (partsMantissa.length == 2) {
			exponent -= partsMantissa[1].length();
		}
		String retVal = "(";
		retVal += partsMantissa[0] + (partsMantissa.length == 2 ? partsMantissa[1] : "");
		retVal += ((exponent == 0) ? "" : "* (10 ^ (" + exponent + "))");
		retVal += ")";
		return retVal;
	}

	private class CVC3ExpressionVisitor implements ExpressionVisitor {
		private boolean isBooleanExpression = true;
		private HashSet<String> declaredVars = new HashSet<String>();
		private StringBuffer decl = new StringBuffer("");
		private Stack<String> clauseStack = new Stack<String>();
		
		public String popClause() { return this.clauseStack.pop(); }
		public String getDecl() { return this.decl.toString(); }
		
		public CVC3ExpressionVisitor() { }
		public CVC3ExpressionVisitor(CVC3ExpressionVisitor v, boolean isBooleanExpression) {
			this.isBooleanExpression = isBooleanExpression;
			this.declaredVars = v.declaredVars;
			this.decl = v.decl;
			this.clauseStack = v.clauseStack;
		}

		public void visitFunctionApplication(FunctionApplication x) throws DecisionExternalProtocolException, IOException {
			if (working) {
				String fun = x.getOperator();

				if (fun.equals(Operator.POW)) {
					for (Primitive p : x.getArgs()) { //they are two! 
						p.accept(new CVC3ExpressionVisitor(this, false));
					}
					this.clauseStack.push("( (" + this.clauseStack.pop() + ") ^ (" + this.clauseStack.pop() + ") )");
				} else {
					//the whole application is treated as an uninterpreted 
					//symbolic value
	                char mytype = x.getType();
	                if (Type.isPrimitive(mytype)) {
						m.mangle(x).accept(this);
	                } else {
	                    throw new DecisionExternalProtocolException("Wrong function return type");
	                }
					return;
				}
			}
		}
		

		public void visitExpression(Expression e) throws DecisionExternalProtocolException, IOException {
			if (working) {
				String operation = e.getOperator();
				Primitive firstOp = e.getFirstOperand();
				Primitive secondOp = e.getSecondOperand();

				String op = CVC3Op(operation);
				boolean isBooleanOperator = isBooleanOperator(operation);

				if (yieldsBooleanOperator(operation) == this.isBooleanExpression) {
					//the expression is well-formed.

					//1 - The operator corresponds to a CVC3 operator:
					//    builds the expression
					if (!op.equals(OTHER)) { 
						int ops = 0;
						if (secondOp != null) {
							ops++;
							secondOp.accept(new CVC3ExpressionVisitor(this, isBooleanOperator));
						}
						if (firstOp != null)  {
							ops++;
							firstOp.accept(new CVC3ExpressionVisitor(this, isBooleanOperator));
						}
						if (ops == 2) {
							this.clauseStack.push("( (" + this.clauseStack.pop() + ")" + op + "(" + this.clauseStack.pop() + ") )");
						} else if (ops == 1) {
							this.clauseStack.push("( " + op + this.clauseStack.pop() + " )");
						} else {
							//TODO this should be unreachable
						}
						//2 - The operator does not correspond to a Sicstus operator, 
						//    but it can be implemented in terms of these: builds a 
						//    suitable expression and visits it
					} else if (operation.equals(Operator.REM)) {
						Primitive val = firstOp.sub(firstOp.div(secondOp).mul(secondOp));
						val.accept(this);
					} else if (operation.equals(Operator.SHL)) {
						Primitive[] args = new Primitive[2];
						args[0] = new Simplex(2);
						args[1] = secondOp;
						Primitive val = firstOp.mul(Primitive.apply("" + firstOp.getType(), Operator.POW, args));
						val.accept(this);
					} else if (operation.equals(Operator.SHR)) {
						Primitive[] args = new Primitive[2];
						args[0] = new Simplex(2);
						args[1] = secondOp;
						Primitive val = firstOp.div(Primitive.apply("" + firstOp.getType(), Operator.POW, args));
						val.accept(new CVC3ExpressionVisitor(this, false));
						//3 - the operator is not supported: mangles the subexpressions into 
						//    a symbolic value and sends it (unsupported operators are:
						//    Operator.USHR, Operator.ANDBW, Operator.ORBW, Operator.XOR)
					} else {
						m.mangle(e).accept(this);
					}
				} else {
					//the expression is ill-formed
					throw new RuntimeException("Met an ill-formed expression."); //TODO improve management of internal errors
				}
			}
		}

		public void visitPrimitiveSymbolic(PrimitiveSymbolic s) throws DecisionExternalProtocolException, IOException {
			this.putSymbol(s.getType(), s.toString());
		}
		
		public void visitTerm(Term x) throws DecisionExternalProtocolException, IOException {
			this.putSymbol(x.getType(), x.toString());
		}
		
		private void putSymbol(char type, String symbol) {
			String cvc3VarName = symbol.substring(1, symbol.length() - 1);
			if (this.declaredVars.contains(cvc3VarName)) {
				//does nothing
			} else {
				this.declaredVars.add(cvc3VarName);
				this.decl.append(cvc3VarName + " : " + CVC3PrimitiveType(type) + "; ");
			}
			this.clauseStack.push(cvc3VarName);
		}
		
		public void visitSimplex(Simplex x) throws DecisionExternalProtocolException, IOException {
			Object obj = x.getActualValue();
	        if (working) {
	            char mytype = x.getType();
	            if (mytype == Type.BYTE ||
	                mytype == Type.SHORT ||
	                mytype == Type.INT ||
	                mytype == Type.LONG ||
	                mytype == Type.CHAR) {
	                this.clauseStack.push(obj.toString());
	            } else if (mytype == Type.FLOAT ||
	                     mytype == Type.DOUBLE) {
	                this.clauseStack.push(CVC3Double(obj.toString()));
	            } else if (mytype == Type.BOOLEAN) {
	            	if ((Boolean) obj) {
	            		if (this.isBooleanExpression) {
	            			this.clauseStack.push("TRUE");
	            		} else {
	            			this.clauseStack.push("1");
	            		}
	            	} else {
	            		if (this.isBooleanExpression) {
	            			this.clauseStack.push("FALSE");
	            		} else {
	            			this.clauseStack.push("0");
	            		}
	            	}
	            }
	        }
		}		
	}

	public boolean curAssumptionsImplyCondIs(boolean value) throws DecisionExternalProtocolException, IOException {
		//visits the context and the current expression and
		//builds the query
		CVC3ExpressionVisitor v = new CVC3ExpressionVisitor();
		this.context.accept(v);
		if (value) {
			this.cond.accept(v);
		} else {
			this.cond.not().accept(v);
		}
		String decl = v.getDecl();
		String cnd = v.popClause();
		String ctx = v.popClause();
		String query = decl + "ASSERT " + ctx + "; QUERY " + cnd + ";" + lineSep;
		
		//queries CVC3
		this.cvc3Out.write(push);
		this.cvc3Out.flush();
		this.readPrompt();
		this.cvc3Out.write(query);
		this.cvc3Out.flush();
		String ans = this.cvc3In.readLine();
		this.readPrompt();
		this.cvc3Out.write(pop);
		this.cvc3Out.flush();
		this.readPrompt();

		//returns the result
		return ans.equals("Valid.");
	}

	public void fail() {
		this.cvc3.destroy();		
	}

	public void quit() throws DecisionExternalProtocolException, IOException {
		this.cvc3Out.close();
		try {
			if (this.cvc3.waitFor() != 0) {
				throw new DecisionExternalProtocolException();
			}
		} catch (InterruptedException e) {
			throw new DecisionExternalProtocolException();
		}
	}
}
