package muse.dec;

import java.util.ArrayList;
import java.util.HashMap;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.net.UnknownHostException;

import muse.mem.Expression;
import muse.mem.ExpressionVisitor;
import muse.mem.FunctionApplication;
import muse.mem.Operator;
import muse.mem.Primitive;
import muse.mem.PrimitiveSymbolic;
import muse.mem.Simplex;
import muse.mem.Term;
import muse.mem.Type;



/**
 * Interface with the Sicstus simplifier. Use the simplifier via TCP/IP.
 * 
 * Protocol (request/reply) for communicating with C++ SICSTUS Prolog.
 * 
 * Commands:
 *   I - Init. Initialize the state of Sicstus. The assumptions set used
 *   for proofs is initialized to the identically true assumption. Use
 *   immediately after connecting to the server. No other command will
 *   be accepted if a I command is not received.
 *   
 *   F - Formula. Prefix for all commands related to transmitting formulas
 *   to the Sicstus server. Each formula transfer starts with a F1
 *   command and ends with a F0 command.
 *   
 *   F1 - Begin formula communication (formula operands and operators
 *   in postfixed format are expected hereafter).  As a
 *   precondition, there must be no other (part of) formula
 *   currently stored.
 *   
 *   FO - Operator. After this, the Sicstus server will wait and parse
 *   a string representing an operator.  Possible operators are +,
 *   -, *, /, _ (that is, unary negation of a number), >, G
 *   (greater than or equals), = (equals), !  (not equals), <, L
 *   (less than or equals), A (and), O (or), N (not), I (implies)
 *   
 *    FI - Integer value. After this, the Sicstus server will wait and
 *    parse a string representing an integer value.
 *    
 *    FR - Real value. After this, the Sicstus server will wait and
 *    parse a string representing a real value.
 *    
 *    FB - Boolean value. After this, the Sicstus server will wait and
 *    parse for a string representing a boolean value, i.e., "true"
 *    or "false"
 *    
 *    FS - Symbolic value. After this, the Sicstus server will wait and
 *    parse two subsequent strings: first, a single character
 *    representing the type of the symbol: i.e., I (integer), R
 *    (real) or B (boolean); second, a string representing the
 *    symbol name.
 *    
 *    F0 - End formula communication.
 *    
 *    P - Prove. Prefix for commands that attempt to prove a formula, using
 *    the current assumption set as knowledge base. As a precondition,
 *    there must be a formula currently stored.
 *    
 *    PT - Attempt to prove that the formula is true.
 *    
 *    PF - Attempt to prove that the formula is false.
 *    
 *    A - Append. Add a formula to the current assumptions set and then
 *    discard it. As a precondition, there must be a formula currently
 *    stored.
 *    
 *    AT - Append the formula as it is.
 *    
 *    AF - Negate the formula and then append it.
 *    
 *    D - Discard. Discard a formula. As a precondition, there must be a
 *    formula currently stored.
 *    
 *    C - Clear. Deletes the whole assumption set.
 *    
 *    Q - Quit. End the Sicstus session.
 *    
 * Return codes are as follows:
 * 
 *    250 Requested simplifier action okay, completed
 *    
 *    354 Continue with formula input; end with the F0 command
 *    
 *    451 Action completed with negative answer: unable to prove a formula
 *    or empty command
 *    
 *    554 Unrecoverable error, the transaction failed and the server as been
 *    reset: syntax errors, parsing errors, wrong command sequences, etc.
 *     
 */
@SuppressWarnings("unused")
public class DecisionProcedureExternalInterfaceSicstus extends DecisionProcedureExternalInterface {
    // Constants: result codes 
    final private static String protoOK     = "250";
    final private static String protoCONT   = "354";
    final private static String protoNO     = "451";
	final private static String protoERR    = "554";
    
    private BufferedWriter sicstusWriter;
    private BufferedReader sicstusReader;    
    private boolean working = false;
    private SicstusEmitterExpressionVisitor v = new SicstusEmitterExpressionVisitor();
    
    /**
     * An ExpressionVisitor which builds a Sicstus string 
     * representing the expression. It is completely stateless, 
     * and it just redispatches to suitable methods of the container 
     * object.
     * 
     * @author pietro
     */
    private class SicstusEmitterExpressionVisitor implements ExpressionVisitor {
    	public void visitExpression(Expression e) throws DecisionExternalProtocolException, IOException {
    		DecisionProcedureExternalInterfaceSicstus.this.putExpression(e);
    	}

    	public void visitPrimitiveSymbolic(PrimitiveSymbolic s) throws IOException, DecisionExternalProtocolException {
    		DecisionProcedureExternalInterfaceSicstus.this.putPrimitiveSymbolic(s);
    	}

    	public void visitSimplex(Simplex x) throws IOException, DecisionExternalProtocolException {
    		DecisionProcedureExternalInterfaceSicstus.this.putSimplex(x);
    	}

		public void visitFunctionApplication(FunctionApplication x) throws DecisionExternalProtocolException, IOException {
			DecisionProcedureExternalInterfaceSicstus.this.putFunctionApplication(x);
		}
		
		public void visitTerm(Term x) throws DecisionExternalProtocolException, IOException {
			DecisionProcedureExternalInterfaceSicstus.this.putTerm(x);
		}
    }
    
	private static final char OTHER = '\0';
	
	private static char sicstusOp(String operator) {
        if (operator.equals(Operator.ADD))      return '+';
        else if (operator.equals(Operator.SUB)) return '-';
        else if (operator.equals(Operator.MUL)) return '*';
        else if (operator.equals(Operator.DIV)) return '/';
        else if (operator.equals(Operator.NEG)) return '_';
        else if (operator.equals(Operator.LT))  return '<';
        else if (operator.equals(Operator.LE))  return 'L';
        else if (operator.equals(Operator.EQ))  return '=';
        else if (operator.equals(Operator.NE))  return '!';
        else if (operator.equals(Operator.GE))  return 'G';
        else if (operator.equals(Operator.GT))  return '>';
        else if (operator.equals(Operator.AND)) return 'A';
        else if (operator.equals(Operator.OR))  return 'O';
        else if (operator.equals(Operator.NOT)) return 'N';
        else return OTHER;
	}

	private static char sicstusFun(String function) {
		if (function.equals(Operator.ABS))      return '|';
		else if (function.equals(Operator.SIN)) return 'S';
		else if (function.equals(Operator.COS)) return 'C';
		else if (function.equals(Operator.TAN)) return 'T';
		else if (function.equals(Operator.POW)) return 'P';
		//else if (function.equals(Operator.EXP)) return 'E';
		else if (function.equals(Operator.MIN)) return 'm';
		else if (function.equals(Operator.MAX)) return 'M';
		else return OTHER;
	}

    /**
     * Constructor. Tries to connect to the server specified 
     * by its parameters.
     * 
     * @throws DecisionExternalProtocolException
     * @throws IOException 
     */
    public DecisionProcedureExternalInterfaceSicstus(String serverAddress, int serverPort) throws DecisionExternalProtocolException, IOException  {
    	this.working = false;
        this.connect(serverAddress, serverPort);
    }

    private void connect(String serverAddress, int serverPort) throws DecisionExternalProtocolException, IOException {
        //establishing the connection
        Socket conn2CppSicstus = null;
        try {
            conn2CppSicstus = new Socket(serverAddress,serverPort);
        } catch(UnknownHostException e) {
            throw new UnknownHostException("No simplifier found at address: " + 
                    serverAddress + ':' + serverPort);
        }
        this.sicstusWriter = new BufferedWriter(new
                OutputStreamWriter(conn2CppSicstus.getOutputStream()));
        this.sicstusReader = new BufferedReader(new
                InputStreamReader(conn2CppSicstus.getInputStream()));
        this.sicstusWriter.write("I",0,1);
        this.sicstusWriter.flush();
        String reply = sicstusReader.readLine();
        if (!reply.startsWith(protoOK)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }
        
        this.working = true;
    }
    
    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#isWorking()
	 */
    public boolean isWorking() { 
    	return this.working; 
    } 
    
    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#putCondition(SEE.Memory.Primitive)
	 */
    public void putCondition(Primitive cond) throws
    DecisionExternalProtocolException, IOException {
        this.sicstusWriter.write("F1",0,2);
        this.sicstusWriter.flush();
        String reply = sicstusReader.readLine();
        if (!reply.startsWith(protoCONT)) {
            fail();
            throw new DecisionExternalProtocolException(reply);
        }
        cond.accept(v);
        this.sicstusWriter.write("F0",0,2);
        this.sicstusWriter.flush();
        reply = sicstusReader.readLine();
        if (!reply.startsWith(protoOK)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }               
    }
    
    private void putExpression(Expression e) throws IOException, DecisionExternalProtocolException {
        if (working) {
            String operation = e.getOperator();
            Primitive firstOp = e.getFirstOperand();
            Primitive secondOp = e.getSecondOperand();

            char op = sicstusOp(operation);

            //1 - The operator corresponds to a Sicstus operator:
            //    sends the expression
            if (op != OTHER) { 
            	if (secondOp != null) { 
            		secondOp.accept(v);
            	}
            	if (firstOp != null)  {
            		firstOp.accept(v);
            	}
            	putOperator(op);
            //2 - The operator does not correspond to a Sicstus operator, 
            //    but it can be implemented in terms of these: builds a 
            //    suitable expression and sends it
            } else if (operation.equals(Operator.REM)) {
            	Primitive val = firstOp.sub(firstOp.div(secondOp).mul(secondOp));
            	val.accept(v);
            } else if (operation.equals(Operator.SHL)) {
                Primitive[] args = new Primitive[2];
                args[0] = new Simplex(2);
                args[1] = secondOp;
                Primitive val = firstOp.mul(Primitive.apply("" + firstOp.getType(), Operator.POW, args));
                val.accept(v);
            } else if (operation.equals(Operator.SHR)) {
                Primitive[] args = new Primitive[2];
                args[0] = new Simplex(2);
                args[1] = secondOp;
                Primitive val = firstOp.div(Primitive.apply("" + firstOp.getType(), Operator.POW, args));
                val.accept(v);
            //3 - the operator is not supported: mangles the subexpressions into 
            //    a symbolic value and sends it (unsupported operators are:
           	//    Operator.USHR, Operator.ANDBW, Operator.ORBW, Operator.XOR)
            } else {
            	m.mangle(e).accept(v);
            }
        }
    }
    
	private void putPrimitiveSymbolic(PrimitiveSymbolic s) throws IOException, DecisionExternalProtocolException {
        if (working) {
            char mytype = s.getType();
            if (Type.isPrimitive(mytype)) {
            	this.putSymbol(mytype, s.getValue());
            } else {
                throw new DecisionExternalProtocolException("Wrong symbol type");
            }
        }
	}

	private void putSimplex(Simplex x) throws IOException, DecisionExternalProtocolException {
		Object obj = x.getActualValue();
        if (working) {
            char mytype = x.getType();
            if (mytype == Type.BYTE ||
                mytype == Type.SHORT ||
                mytype == Type.INT ||
                mytype == Type.LONG ||
                mytype == Type.CHAR) {
                this.putInteger(obj);
            } else if (mytype == Type.FLOAT ||
                     mytype == Type.DOUBLE) {
                this.putReal(obj);
            } else if (mytype == Type.BOOLEAN) {
                this.putBoolean(obj);
            }
        }
	}

	private void putFunctionApplication(FunctionApplication x) throws DecisionExternalProtocolException, IOException {
		if (working) {
			String fun = x.getOperator();
			char op = sicstusFun(fun);

			//TODO this disables interpretation, until support on the server side will be flawless
			op = OTHER; 

			if (op != OTHER) {
				for (Primitive p : x.getArgs()) { 
					p.accept(v);
				}
				this.putOperator(op);
			} else {
				//if the function cannot be interpreted by Sicstus, 
				//the whole application is treated as an uninterpreted 
				//symbolic value
                char mytype = x.getType();
                if (Type.isPrimitive(mytype)) {
                	m.mangle(x).accept(v);
                } else {
                    throw new DecisionExternalProtocolException("Wrong function return type");
                }
			}
		}
	}
    
    private void putInteger(Object i) throws DecisionExternalProtocolException, IOException {
        if (!(i instanceof Integer)&&!(i instanceof Short) &&
                !(i instanceof Long)&&!(i instanceof Byte) &&
                !(i instanceof Character)) {
            throw new ClassCastException("SIInterfaceSicstus.putInteger");
        }
        String i2string=i.toString();
        this.sicstusWriter.write("FI"+i2string,0,2+i2string.length());
        this.sicstusWriter.flush();
        String reply = this.sicstusReader.readLine();
        if (!reply.startsWith(protoCONT)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }
    
    private void putReal(Object r) throws
    DecisionExternalProtocolException, IOException {
        if(!(r instanceof Float)&&!(r instanceof Double)) { 
            throw new ClassCastException("SIInterfaceSicstus.putReal");
        }
        String r2string=r.toString();
        sicstusWriter.write("FR"+r2string,0,2+r2string.length());
        sicstusWriter.flush();
        String reply=sicstusReader.readLine();
        if (!reply.startsWith(protoCONT)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }
    
    private void putBoolean(Object b) throws
    DecisionExternalProtocolException, IOException {
        if (!(b instanceof Boolean)) {
        	throw new ClassCastException("SIInterfaceSicstus.putBoolean");
        }
        String b2string=b.toString();
        sicstusWriter.write("FB"+b2string,0,2+b2string.length());
        sicstusWriter.flush();
        String reply=sicstusReader.readLine();
        if(!reply.startsWith(protoCONT)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }
    
    private void putOperator(char op) throws
    DecisionExternalProtocolException, IOException {
        sicstusWriter.write("FO" + op, 0, 3);
        sicstusWriter.flush();
        String reply=sicstusReader.readLine();
        if (!reply.startsWith(protoCONT)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }
    
    private void putTerm(Term x) throws
    DecisionExternalProtocolException, IOException {
    	this.putSymbol(x.getType(), x.toString());
    }
    	
    private void putSymbol(char type, String name) throws
    DecisionExternalProtocolException, IOException{
    	char typeSicstus;
    	
    	if (type ==Type.BYTE ||
                type == Type.SHORT ||
                type == Type.INT ||
                type == Type.LONG ||
                type == Type.CHAR) {
    		typeSicstus = 'I';
    	} else if (type == Type.FLOAT ||
                type == Type.DOUBLE) {
    		typeSicstus = 'R';
    	} else { //type == Type.BOOLEAN
            typeSicstus = 'I'; //TODO it should be typeSicstus = 'B', but apparently a 'B' annoys the Sicstus server. Since treatment of booleans as integers is consistent with Java semantics we leave it this way.
    	}
        sicstusWriter.write("FS"+typeSicstus+name,0,3+name.length());
        sicstusWriter.flush();
        String reply=sicstusReader.readLine();
        if (!reply.startsWith(protoCONT)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }
    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#curAssumptionsImplyCondIs(boolean)
	 */
    public boolean curAssumptionsImplyCondIs(boolean value) throws
    DecisionExternalProtocolException, IOException{
        if (value) { 
        	sicstusWriter.write("PT",0,2);
        } else {
        	sicstusWriter.write("PF",0,2);
        }
        sicstusWriter.flush();
        String reply = sicstusReader.readLine();
        if (reply.startsWith(protoOK)) {
            return true;
        } else if(reply.startsWith(protoNO)) {
            return false;
        } else {
            fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }    
    
    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#appendCondToAssumptionsAs(boolean)
	 */
    public void appendCondToAssumptionsAs(boolean value) throws
    DecisionExternalProtocolException, IOException{
        if (value) {
        	sicstusWriter.write("AT",0,2);
        } else {
        	sicstusWriter.write("AF",0,2);
        }
        sicstusWriter.flush();
        String reply = sicstusReader.readLine();
        if(!reply.startsWith(protoOK)) {
            fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }
    
    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#discardFormula()
	 */
    public void discardFormula() throws DecisionExternalProtocolException, IOException{
        this.sicstusWriter.write("D",0,1);
        this.sicstusWriter.flush();
        String reply = this.sicstusReader.readLine();
        if (!reply.startsWith(protoOK)) {
        	fail();
            throw new DecisionExternalProtocolException(reply);
        }
    }
    
    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#clear()
	 */
    public void clear() throws DecisionExternalProtocolException, IOException {
        sicstusWriter.write("C",0,1);
        sicstusWriter.flush();
        String reply=sicstusReader.readLine();
        if (!reply.startsWith(protoOK)) {
            fail();
            throw new DecisionExternalProtocolException(reply);
        }
        //TODO here we could reset the mangler to free some resources
    }

    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#quit()
	 */
    public void quit() throws DecisionExternalProtocolException, IOException {
        sicstusWriter.write("Q",0,1);
        sicstusWriter.flush();
        String reply=sicstusReader.readLine();
   		try {
			sicstusWriter.close();
		} catch (IOException e) {
			//nothing to do
		}
   		try {
			sicstusReader.close();
		} catch (IOException e) {
			//nothing to do
		}
        if (!reply.startsWith(protoOK)) {
            throw new DecisionExternalProtocolException(reply);
        }
    }
    
    /* (non-Javadoc)
	 * @see SEE.SIInterface.SIInterface#fail()
	 */
    public void fail() {
    	working = false;
    }    
}