package muse.jvm;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import muse.algo.Algorithm;
import muse.algo.AlgorithmContext;
import muse.algo.BytecodeDispatcherAlgorithm;
import muse.algo.CannotManageStateException;
import muse.algo.PartialReferenceResolutionException;
import muse.algo.SEInit;
import muse.bc.ClassFileFactory;
import muse.bc.Opcodes;
import muse.bc.Signature;
import muse.dec.DecisionBacktrackException;
import muse.dec.DecisionException;
import muse.dec.DecisionInitException;
import muse.dec.DecisionProcedure;
import muse.dec.DecisionQuitException;
import muse.dec.DecisionRunningException;
import muse.mem.Any;
import muse.mem.Expression;
import muse.mem.Instance;
import muse.mem.Klass;
import muse.mem.PathCondition;
import muse.mem.State;
import muse.mem.Type;
import muse.mem.Value;
import muse.tree.StateTree;
import muse.tree.StateTree.BranchPoint;


/**
 * The <code>Engine</code> class is the core class of the symbolic executor.
 * An <code>Engine</code> implements a JVM able to symbolically execute the 
 * methods of an arbitrary Java class. 
 * An <code>Engine</code> can be integrated with verification and testing 
 * tools by means of the following features:
 * 
 * <ul>
 * <li>stepwise symbolic execution of public methods of Java classes;</li>
 * <li>predictable exploration of decision points where the path condition
 * allows more than one branch, and possibility to backtrack the execution to
 * unexplored branches;</li>
 * <li>support to volatile variables, whose value is never assumed, allowing 
 * the unconditional exploration of both branches of a decision point;</li>
 * <li>support to observation of changes in state variables based on
 * notification.</li>
 * </ul>
 * 
 * @author Pietro Braione
 * @author unknown
 */
public class Engine {

	/** The engine's <code>DecisionProcedure</code>. */
	private DecisionProcedure simplifier;
	
	/** The engine's <code>ClassFileFactory</code>. */
	private ClassFileFactory classFileFactory;
	
	/** The engine's <code>Environment</code>. */
	private Environment env;
	
	/** The engine's <code>StateTree</code>. */
	private StateTree stateTree;
	
	/** The <code>AlgorithmContext</code> for the <code>Algorithm</code>s. */
	private AlgorithmContext ctx; 

	/** The engine's <code>BytecodeDispatcherAlgorithm</code>. */
	private BytecodeDispatcherAlgorithm dispatcher;
	
	/**
	 * The name of the class containing the method which is symbolically
	 * executed by the engine when started.
	 */
	private String className;

	/**
	 * The <code>Signature</code> of the method which is symbolically executed by the engine
	 * when started.
	 */
	private Signature methodSignature;

	/** The initial JVM <code>State</code> of the symbolic execution. */
	private State initialState;

	/** The current JVM <code>State</code> of the symbolic execution. */
	private State currentState;

	/** The <code>Signature</code>s of the variables under observation. */
	private Signature[] varSigs;

	/** The <code>ExecutionObserver</code>s of the variables <code>varSigs</code>. */
	private ExecutionObserver[] obs;

	/** Cache for the current <code>Value</code>s of the variables under observation. */
	private Value[] values;

	/** Cache for the <code>Value</code>s of the variables under observation (for backtrack). */
	private Map<BranchPoint, Value[]> savedValues;
	
	/** The <code>Signature</code> of the variable initialized with the <code>Any</code> value. */
	private Signature varVolatile;
	
	/** <code>true</code> iff the <code>Engine</code> has been stopped. */
	private boolean engineStopped;
	
	/** The total number of <code>State</code>s analyzed by the <code>Engine</code>. */
	private long analyzedStates;

	//Costructors

	/**
	 * Constructor.
	 * 
	 * @param classFileFactory a <code>ClassFileFactory</code> which 
	 *        will be used by the engine to retrieve classfiles.
	 * @param simplifier a <code>DecisionProcedure</code> which will be 
	 *        used by the engine to simplify symbolic expression and 
	 *        decide branches.
	 */
	public Engine(ClassFileFactory classFileFactory, DecisionProcedure simplifier) {
		//prepares the architecture
		this.simplifier = simplifier;
		this.classFileFactory = classFileFactory;
		this.env = new Environment();
		this.stateTree = new StateTree();
		this.dispatcher = new BytecodeDispatcherAlgorithm();
	}

	// Public methods

	/**
	 * Adds a set of paths to the classpath.
	 * 
	 * @param paths a <code>String[]</code> representing a list 
	 * of paths.
	 */
	public void addClassPath(String[] paths) {
		for (String path : paths) {
			this.env.addClassPath(path);
		}
	}
	
    /**
     * Specifies a possible replacement for a class. Typically, an abstract 
     * class must have at least one or more concrete replacement classes for 
     * solving symbolic references with abstract type with concrete instances.
     * 
     * @param toReplace   the name of the class to be replaced. It must 
     *                    be <code>toReplace != null</code>.
     * @param originExp   an expression describing the origins of the 
     *                    symbolic references which match this replacement.
     *                    If <code>originExp == null</code>, all the symbolic 
     *                    references with static type <code>toReplace</code> 
     *                    will match. 
     * @param replacement the name of the class which is a possible 
     *                    replacement for <code>toReplace</code>. It must 
     *                    be <code>toReplace != null</code>. During the 
     *                    symbolic execution, every symbolic reference with 
     *                    class <code>toReplace</code> will be expanded, 
     *                    when necessary, to a symbolic object with class 
     *                    <code>replacement</code>.
     */
    public void setClassSubst(String toReplace, String originExp, String replacement) {
    	this.env.addClassSubst(toReplace, originExp, replacement);
    }
    
    /**
     * Adds a class name to the set of not initialized classes.
     * 
     * @param s a class name as a <code>String</code>.
     */
    public void addNotInitializedClass(String s) {
    	this.env.addNotInitializedClass(s);
    }

	/**
	 * Sets the signature of the method to be symbolically executed by the
	 * engine. The method <emph>must not</emph> be static.
	 * 
	 * @param className a class name as a <code>String</code>.
	 * @param paramsSig the signature of the method's parameters as a 
	 *                  <code>String</code>.
	 * @param methodName the name of the method as a <code>String</code>. 
	 */
	public void setMethodName(String className, String paramsSig, String methodName) {
		this.methodSignature = new Signature(className, paramsSig, methodName);
		this.className = className;
	}

	/**
	 * Sets an initial state for the symbolic execution.
	 * 
	 * @param s
	 *        <code>State</code> which will be used as the initial state 
	 *        (i.e., the state where the engine steps after a call to 
	 *        <code>init()</code>) of the symbolic execution. It must 
	 *        be <code>s != null</code>.
	 */
	public void setInitialState(State s) {
		this.initialState = s;
		this.methodSignature = s.getStack().currentFrame().getName();
		this.className = this.methodSignature.getClassName();
		this.env = s.getClassFileInterface().getEnvironment();
	}

	/**
	 * Instructs the engine to notify changes in values of a list of program
	 * variables to a corresponding list of observers. It MUST be invoked 
	 * after the <code>setMethodName<code> or <code>setInitialState</code>
	 * method.
	 * 
	 * @param varSigs
	 *        an array of <code>String</code>s, each representing the
	 *        name of a variable to be monitored; the method has no effect
	 *        if <code>varNames</code> is null, or has zero length.
	 *        Variables must be instance variables of the root object, i.e., 
	 *        of the object instance which is created to start the execution 
	 *        of the method specified by means of the <code>setMethodName</code> 
	 *        method, and must have type <code>boolean</code>.
	 * @param obs
	 *        an array of <code>ExecutionObserver</code>s which are notified
	 *        upon changes of the variables specified in
	 *        <code>varNames</code>; the method has no effect if
	 *        <code>obs</code> is <code>null</code>, or has a length different 
	 *        from that of <code>varNames</code>. Otherwise, the i-th observer
	 *        in <code>obs</code> will be notified upon changes in the
	 *        values of the i-th variable in <code>varNames</code>.
	 */
	public void setObservers(String[] varNames, ExecutionObserver[] obs) {
		if (varNames != null && obs != null && 
				varNames.length > 0 && varNames.length == obs.length) {
			this.varSigs = new Signature[varNames.length];
			for (int i = 0; i < varNames.length; i++) {
				this.varSigs[i] = new Signature(this.className, "" + Type.BOOLEAN, varNames[i]);
			}
			this.obs = obs;
			for (ExecutionObserver o : obs) {
				o.e = this;
			}
		}
	}

	/**
	 * Instructs the <code>Engine</code> to consider a boolean variable as initially 
	 * volatile. A variable is volatile when it has a different value each time is read.
	 * Note that the <code>Engine</code> will assign a special "any" value to 
	 * the variable upon start, but that the variable's content can be overwritten and copied 
	 * during the symbolic execution as any other variable. This means that a variable 
	 * is not "statically" volatile, but it is as long as it contains the "any" value. 
	 * If the variable's value is overwritten it will no longer be volatile, and if the 
	 * "any" value is copied into another variable, this will also become volatile. 
	 * This behaviour is a possible source of unsoundness. For this reason a volatile 
	 * variable should not be overwritten nor should the "any" value escape from it 
	 * into other variables. The meant use of the "any" value is for simulating nondeterministic 
	 * choice by putting the variable containing it as the condition of a branch.   
	 * 
	 * @param varVolatile 
	 *        a <code>String</code> containing a variable name. 
	 *        The variable must be an instance variable of the root object, i.e., 
	 *        of the object instance which is created to start the execution 
	 *        of the method specified by means of the <code>setMethodName</code> 
	 *        method. The variable must also be of primitive type <code>boolean</code>. 
	 */
	public void setVolatile(String varVolatile) {
		this.varVolatile = new Signature(this.className, "" + Type.BOOLEAN, varVolatile);
	}
	
	/**
	 * Starts the symbolic execution by stepping the engine in a suitable
	 * initial state. In such state the path condition is set to the
	 * precondition specified by the last call to <code>setPrecondition</code>,
	 * and the JVM state is configured as follows:
	 * 
	 * <ul>
	 * <li> the heap has one object, an instance of the class specified by the
	 * last call to <code>setClassName</code>: this object is called the 
	 * <emph>root object</emph>;</li>
	 * <li> the stack has one frame for an invocation of the method specified by
	 * the last call to <code>setMethodName</code>;</li>
	 * <li> all the instance variables of the root object and all the local variables 
	 * in the current frame are initialized with suitable symbolic values, with
	 * the exception of <code>this</code> which is set to the root object;</li>
	 * <li> the path condition is set to the expression provided by means of the 
	 * <code>setPrecondition</code> method.</li>
	 * </ul>
	 * @throws DecisionException  in case initialization of the simplifier fails. 
	 * @throws InitializationException in case initialization of the initializer fails.
	 * @throws NonexistingObservedVariablesException in case some of the provided 
	 *         observed variable names cannot be observed.  
	 */
	public void init() throws DecisionInitException, InitializationException, NonexistingObservedVariablesException {		
		//sets the state count to zero
		this.analyzedStates = 0L;
		
		//propagates the environment to the classfile factory
		this.classFileFactory.setEnv(this.env);
		
		//builds the context for the Algorithms
		this.ctx = new AlgorithmContext(this.env, this.classFileFactory, this.stateTree, this.simplifier);

		//creates the initial state
		SEInit algo = this.dispatcher.select();
		if (this.initialState == null) {
			algo.exec(this.methodSignature, ctx);
		} else {
			algo.exec(this.initialState, ctx);
		}
		
		//extracts the initial state from the tree and 
		//sets it as the current state; also saves a copy
		//of it for future reference
		this.currentState = this.stateTree.nextState();
		this.initialState = this.currentState.clone();
		
		//sets the initially volatile variable with Any 
		if (this.hasVolatile()) {
			this.getRootObject().setFieldValue(this.varVolatile,
					new Any(this.getRootObject().getType()));
		}
		
		//saves the values of the observed variables
		List<Integer> nonexistingVariables = new LinkedList<Integer>();
		if (this.hasObservers()) {
			this.values = new Value[this.numObservers()];
			for (int i = 0; i < this.numObservers(); i++) {
				this.values[i] = this.getObservedVariableValue(i);
				if (this.values[i] == null) {
					nonexistingVariables.add(i);
					this.obs[i] = null;
				}
			}
			this.savedValues = new HashMap<BranchPoint, Value[]>();
		}
		
		//if some of the observed variables does not exist, throws 
		//the related exception
		if (nonexistingVariables.size() > 0) {
			throw new NonexistingObservedVariablesException(nonexistingVariables);
		}
	}
	
	/**
	 * Adds an assumption to the current state. The operation has no 
	 * effect in the case the additional assumption is not satisfiable 
	 * under the state's path condition.
	 * 
	 * @param exp 
	 *        a <code>SEExpression</code> representing the additional 
	 *        assumption. 
	 * @throws DecisionException iff the Sicstus simplifier
	 *         fails for any reason.
	 * @throws ContradictionException iff <code>exp</code> is not 
	 *         satisfiable under the current state's path condition.  
	 */
	public void assume(Expression exp) throws DecisionException, ContradictionException {
		if (this.simplifier.isSat(exp)) {
			this.currentState.addClauseToPathCondition(exp);
			this.updateSimplifier();
		} else {
			throw new ContradictionException();
		}
	}
	
	/**
	 * Returns the engine's current JVM state.
	 * 
	 * @return a <code>State</code> object representing the current engine JVM
	 *         state.
	 */
	public State getCurrentState() {
		return this.currentState;
	}

	/**
	 * Checks whether the engine can step.
	 * 
	 * @return <code>true</code> iff the engine can step. The engine can step unless 
	 *         the current engine <code>State</code> is stuck, or no backtrack has 
	 *         been performed since the last invocation of the <code>stopCurrentTrace</code>
	 *         method.
	 */
	public boolean canStep() {
		return !(this.currentState.isStuck() || this.engineStopped);
	}

	/**
	 * Steps the execution by performing the current bytecode.
	 * 
	 * @return the <code>BactrackPoint</code> created after the execution of the 
	 *         current bytecode, allowing to resume the execution from the states 
	 *         produced by it, or <code>null</code> if the bytecode execution 
	 *         does not produce more than one possible next state.
	 * @throws EngineStuckException when the method is invoked in a state where 
	 *         <code>this.canStep() == false</code>.
	 * @throws DecisionRunningException iff the simplifier fails for any reason.
	 * @throws CannotManageStateException iff the engine is unable to calculate 
	 *         the next state because of some engine limitations.
	 * @throws PartialReferenceResolutionException iff some symbolic 
	 *         reference has been partially resolved (i.e., not 
	 *         resolved to a fresh symbolic object). 
	 */
	public BranchPoint step() throws EngineStuckException, DecisionRunningException, CannotManageStateException, PartialReferenceResolutionException {
		if (!this.canStep()) {
			throw new EngineStuckException();
		}

		try {
			//steps
			Algorithm algo = this.dispatcher.select(this.currentState.getInstruction());
			algo.exec(this.currentState, this.ctx);

			//updates current state, simplifier and return value
			BranchPoint retVal;
			if (this.stateTree.createdBranch()) {
				this.currentState = this.stateTree.nextState();
				retVal = this.stateTree.nextBranch();
				this.updateSimplifier();
			} else {
				this.currentState.incSequenceNumber();
				retVal = null;
			}

			//manages variable observation
			if (this.hasObservers()) {
				//stores old values of observed values in case 
				//backtrack may be performed 
				if (retVal != null) {
					Value[] toSave = new Value[this.values.length];
					for (int i = 0; i < this.values.length; i++) { 
						if (this.values[i] != null) {
							toSave[i] = (Value) this.values[i].clone();
						}
					}
					this.savedValues.put(retVal, toSave);
				}
				
				//updates values of observed variables and notifies 
				//observers
				for (int i = 0; i < this.numObservers(); i++) {
					if (this.obs[i] != null && 
						!this.getObservedVariableValue(i).equals(this.values[i])) {
						this.values[i] = (Value) this.getObservedVariableValue(i).clone();
						this.obs[i].update();
					}
				}
			}

			//step correctly performed: increments the state count
			this.analyzedStates++;

			//returns
			return retVal;
		} catch (DecisionException e) {
			throw new DecisionRunningException();
		}
	}
	
	/**
	 * Returns the number of analyzed symbolic states.
	 * 
	 * @return a <code>long</code> value, the number of times the 
	 *         <code>step()</code> method has been invoked.  
	 */
	public long getAnalyzedStates() {
		return this.analyzedStates;
	}

	/**
	 * Checks whether the current state has an increased path condition, and 
	 * sends the additional clause to the simplifier.
	 * 
	 * @throws DecisionException iff the simplifier fails for any reason.
	 */
	private void updateSimplifier() throws DecisionException {
		if (this.currentState.augmentedPathCondition()) {
			PathCondition.Clause c = this.currentState.getLastPathConditionClause();
			if (c.hasCondition()) {
				this.simplifier.pushClause(c.getCondition());
			}
		}
	}
	
	/**
	 * Steps the execution to the next source code row by repeatedly 
	 * invoking <code>step()</code> until the source code row contained 
	 * in the classfile's debug information changes, or the current 
	 * method changes (invocation or return), or the state becomes stuck 
	 * (i.e., <code>canStep() == false</code>) or a backtrack 
	 * point is created (i.e., a branch in the symbolic execution 
	 * tree is reached), or some exception is raised.
	 * 
	 * @return the <code>BacktrackPoint</code> created after the 
	 *         step or <code>null</code>, identically as for the 
	 *         <code>step()</code> method (at most one 
	 *         <code>BacktrackPoint</code> is created by an 
	 *         invocation of <code>stepNext()</code>).
	 * @throws EngineStuckException iff the method is invoked in a 
	 *         state where <code>this.canStep() == false</code>.
	 * @throws DecisionRunningException iff the Sicstus simplifier
	 *         fails for any reason. 
	 * @throws CannotManageStateException iff the engine is unable 
	 *         to calculate the next state because of some engine 
	 *         limitations.
	 * @throws PartialReferenceResolutionException iff some symbolic 
	 *         reference has been partially resolved (i.e., not 
	 *         resolved to a fresh symbolic object). 
	 */
	public BranchPoint stepNext() throws EngineStuckException, DecisionRunningException, CannotManageStateException, PartialReferenceResolutionException {
		if (!this.canStep()) {
			throw new EngineStuckException();
		}
		
		int startSourceRow = this.currentState.getStack().currentFrame().getSourceRow();
		BranchPoint bp = null;
		while (bp == null && 
			   this.canStep() &&
			   this.currentState.getStack().currentFrame().getSourceRow() == startSourceRow) {
			if (this.atFrameChange()) {
				return this.step(); //just steps in the new frame
			}
			bp = this.step();
		}
		return bp;
	}
	
	/**
	 * Checks whether the next instruction may change the current frame.
	 * 
	 * @return <code>true</code> iff the next instruction is one of invoke*, 
	 * athrow or *return.
	 */
	private boolean atFrameChange() {
		byte currentInstruction = this.currentState.getInstruction(); 
		return (
				currentInstruction== Opcodes.OP_INVOKEVIRTUAL||
				currentInstruction == Opcodes.OP_INVOKESPECIAL ||
				currentInstruction == Opcodes.OP_INVOKESTATIC ||
				currentInstruction == Opcodes.OP_INVOKEINTERFACE ||
				currentInstruction == Opcodes.OP_ATHROW ||
				currentInstruction == Opcodes.OP_IRETURN ||
				currentInstruction == Opcodes.OP_FRETURN ||
				currentInstruction == Opcodes.OP_DRETURN ||
				currentInstruction == Opcodes.OP_ARETURN ||
				currentInstruction == Opcodes.OP_RETURN ||
				currentInstruction == Opcodes.OP_ATHROW
		);
	}

	/** 
	 * Adds a backtrack point allowing to resume the symbolic 
	 * execution from the current state. 
	 * 
	 * @return the created <code>BacktrackPoint</code>.
	 */
	public BranchPoint addBacktrackPoint() {
		State s = (State) this.currentState.clone();
		this.stateTree.addBranch();
		this.stateTree.addState(s);
		BranchPoint retVal = this.stateTree.nextBranch();
		if (this.hasObservers()) {
			Value[] toSave = new Value[this.values.length];
			for (int i = 0; i < this.values.length; i++) {
				toSave[i] = (Value) this.values[i].clone();
			}
			this.savedValues.put(retVal, toSave);
		}
		return retVal;

	}
	
	/**
	 * Stops the execution along the current trace.
	 */
	public void stopCurrentTrace() {
		this.engineStopped = true;
	}

	/**
	 * Checks whether the engine can be backtracked.
	 * 
	 * @return true iff the engine has at least one pending backtrack point.
	 */
	public boolean canBacktrack() {
		return stateTree.hasStates();
	}

	/**
	 * Backtracks the execution to the closest backtrack point up in the current
	 * execution trace.
	 * 
	 * @return the <code>BacktrackPoint</code> associated to the target of the
	 *         backtrack operation.
	 * @throws CannotBacktrackException iff <code>!this.canBacktrack()</code> 
	 *         before <code>this.backtrack()</code> is invoked.
	 */
	public BranchPoint backtrack() throws CannotBacktrackException, DecisionBacktrackException {
		//TODO dubious correctness of this implementation
		if (!this.canBacktrack()) {
			throw new CannotBacktrackException();
		}
		
		BranchPoint bp = this.stateTree.nextBranch();
		this.currentState = this.stateTree.nextState();
		try {
			this.simplifier.setClause(this.currentState.getPathCondition().getCondition());
		} catch (DecisionException e) {
			throw new DecisionBacktrackException();
		}
		this.engineStopped = false;
		
		if (this.hasObservers()) {
			this.values = this.savedValues.get(bp);
		}

		return bp;
	}

	/**
	 * Cleans-up the engine after its usage.
	 * 
	 * @throws DecisionQuitException when clean-up of the simplifier fails.
	 */
	public void quit() throws DecisionQuitException {
		//quits the simplifier
		try {
			this.simplifier.quit();
		} catch (DecisionException e) {
			throw new DecisionQuitException();
		}
	}
	
	/**
     * Gets the time spent by the simplifier's computations.
	 * 
     * @return a <code>long</code> representing the spent time 
     *         in milliseconds.
	 */
	public long getTimeSimplifier() {
		return this.simplifier.getTime();
	}

	// Private methods

	/**
	 * Gets the root object, i.e., the first object created to execute
	 * symbolically the method set by means of a call to
	 * <code>setMethodName</code>.
	 * 
	 * @return The first object in the heap of the current state.
	 */
	private Instance getRootObject() {
		return ((Instance) this.currentState.getHeap().getObjects().get(0));
	}

	/**
	 * Gets the root <code>Klass</code>, i.e., the object containing
	 * the static fields of the root object.
	 * 
	 * @return The <code>Klass</code> of the root object.
	 */
	private Klass getRootKlass() {
		return this.currentState.getKlass(this.className);
	}

	/**
	 * Tests whether there are some variables under observation.
	 * 
	 * @return true iff some variable with their observers have been registered
	 */
	private boolean hasObservers() {
		return (this.varSigs != null);
	}

	/**
	 * Gets the number of registered observers.
	 * 
	 * @return the number of specified observers variables, independently on the
	 *         fact that they do indeed exist.
	 */
	private int numObservers() {
		return (this.obs.length);
	}

	/**
	 * Gets the current value of a variable under observation.
	 * 
	 * @param i the ordinal number of the observed variable as in
	 *            <code>this.varNames</code>.
	 * @return the current value of <code>this.varNames[i]</code>.
	 */
	public Value getObservedVariableValue(int i) {
		Value retVal = this.getRootObject().getFieldValue(this.varSigs[i]);
		if (retVal == null) {
			retVal = this.getRootKlass().getFieldValue(this.varSigs[i]);
		}
		return retVal;
	}

	/**
	 * Tests whether some variable has been set as initially volatile.
	 * 
	 * @return <code>true</code> iff a previous call to <code>setVolatile</code>
	 *         has been performed independently on the fact that the variable 
	 *         does indeed exist.
	 */
	private boolean hasVolatile() {
		return (this.varVolatile != null);
	}
}