package muse.jvm;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Class representing the code environment for the symbolic 
 * execution (classpath and substitution map for abstract 
 * classes).
 * 
 * @author pietro
 */
public class Environment {
	private List<String> classPath;
	private Map<String, Set<Entry>> classSubstMap = new HashMap<String, Set<Entry>>();
	private Set<String> notInitializedClasses = new HashSet<String>();

	public class Entry {
		public String originExp;
		public String substClass;
		Entry(String originExp, String substClass) {
			this.originExp = originExp;
			this.substClass = substClass;
		}
		
		public boolean matchOrigin(String origin) {
			boolean retVal;
        	if (this.originExp == null) {
        		retVal = true;
        	} else {
        		//TODO this is really ugly, but it works with the current implementation of origins as strings. Improve it later to a separate language.
        		String pattern = this.originExp.replace("/", "\\."); //replaces the separator
        		Pattern p = Pattern.compile(pattern);
        		Matcher m = p.matcher(origin);
        		retVal = m.matches();
        	}
			return retVal;
		}
	}

	/**
	 * Adds a possible replacement for a given class name.
	 * 
	 * @param abstractClass the name of a class which should be substituted. 
	 *                      It must be <code>abstractClass != null</code>.
	 * @param substClass    the name of the replacement class. It must 
	 *                      be <code>substClass != null</code>.
	 */
	public void addClassSubst(String abstractClass, String substClass) {
		Set<Entry> c = classSubstMap.get(abstractClass);
		if (c == null) {
			c = new HashSet<Entry>();
			classSubstMap.put(abstractClass, c);
		}
		c.add(new Entry(null, substClass));
	}
	
	public void addNotInitializedClass(String s) {
		this.notInitializedClasses.add(s);
	}
	
	public Set<String> getNotInitializedClasses() {
		return this.notInitializedClasses;
	}

	/**
	 * Adds a possible replacement for a given class name.
	 * 
	 * @param abstractClass the name of a class which should be substituted. 
	 *                      It must be <code>abstractClass != null</code>.
	 * @param substClass    the name of the replacement class. It must 
	 *                      be <code>substClass != null</code>.
	 */
	public void addClassSubst(String abstractClass, String originExp, String substClass) {
		Set<Entry> c = classSubstMap.get(abstractClass);
		if (c == null) {
			c = new HashSet<Entry>();
			classSubstMap.put(abstractClass, c);
		}
		c.add(new Entry(originExp, substClass));
	}

	/**
	 * Returns all the specified replacements for a given class name.
	 * @param abstractClass the name of a class. 
	 * @return a <code>Set&lt;Entry&gt;</code> containing all 
	 *         the <code>Entry<code> instances prescribing possible 
	 *         replacements for <code>abstractClass</code>, or an empty 
	 *         <code>Set&lt;Entry&gt;</code> in the case 
	 *         <code>abstractClass</code> has no specified substitution.
	 */
	public Set<Entry> classSubstSet(String abstractClass) {
		Set<Entry> retVal = classSubstMap.get(abstractClass);
		if (retVal == null) {
			retVal = new HashSet<Entry>();
		}
		return retVal;
	}

	/**
	 * Adds a path to the classpath.
	 * 
	 * @param path a <code>String</code> representing a path 
	 * in the executor's classpath.
	 */
	public void addClassPath(String path) {
		//lazily initializes classPath
		if (classPath == null)
			classPath = new LinkedList<String>();

		if (path.endsWith(".jar"))
			classPath.add(path);
		else if (path.charAt(path.length() - 1) == '/')
			classPath.add(path);
		else
			classPath.add(path + "/");
	}

	public Iterable<String> classPath() {
		return new Iterable<String>() {
			public Iterator<String> iterator() {
				return classPath.iterator();
			}
		};
	}
}
