package muse.mem;

import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import muse.algo.JavaReifyException;
import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;


/**
 * Class implementing an array. It is realized as a small theory 
 * associating <code>SEExpression</code>s on the array index to 
 * the outcome of an array access when the access index satisfies 
 * the <code>SEExpression</code>. Two outcomes are possible, i.e., 
 * a <code>Value</code> is returned or an <code>IndexOutOfBoundsException</code> 
 * is raised. Such outcomes are represented by suitable instances 
 * of the subclasses of the <code>Array.ArrayEntry</code> class.
 * Note that the path condition of the state containing the array may 
 * need to be augmented to reflect the fact that some <code>Primitive</code> 
 * value <code>index</code>has been used for access. The <code>Array</code> 
 * class does not alter the state's path condition, but provides the methods 
 * <code>inRange(index)</code> and <code>outOfRange(index)</code>, which 
 * return the clauses which shall be added to the state.
 *  
 * @author pietro
 */
public class Array extends Objekt {
	private static final int T_BOOLEAN = 4;
	private static final int T_CHAR    = 5;
	private static final int T_FLOAT   = 6;
	private static final int T_DOUBLE  = 7;
	private static final int T_BYTE    = 8;
	private static final int T_SHORT   = 9;
	private static final int T_INT     = 10;
	private static final int T_LONG    = 11;

	private static final Term INDEX = new Term(Type.INT, "{INDEX}");

	/*Fields*/

	/** The type of the element's value. */
	private char contentType;

	/** The length of the array. */
	private final Primitive length;

	/** A <code>SEExpression</code> stating that <code>INDEX</code> is in range. */
	private final Expression indexInRange;

	/** The length of the array. */
	private LinkedList<ArrayEntryValue> values;
	
	/** 
	 * Indicates whether the array is normalized, i.e., whether it has as many 
	 * entries as its length, each corresponding to all the possible values of the 
	 * index, and ordered by index. This is possible only if this.length is a Simplex.
	 */ 
	private boolean normalized;

	/**
	 * Class representing an array entry. An array entry essentially 
	 * is a pair (condition, outcome), where the condition expresses 
	 * some constraint over the array index, and outcome is the outcome 
	 * of an array access when the value for the index satisfies the 
	 * constraint.
	 * 
	 * @author pietro
	 */
	public abstract static class ArrayEntry { 
		protected Expression exp;

		/** 
		 * Constructor.
		 * 
		 * @param exp A <code>SEExpression</code> which denotes a 
		 *        constraint over the array index. When <code>null</code>
		 *        it denotes <code>true</code> (i.e., an unconstrained 
		 *        index value). 
		 */
		private ArrayEntry(Expression exp) { this.exp = exp; }

		public Expression getExpression() { return this.exp; }
	}

	/**
	 * Class representing an array entry whose outcome is a correct 
	 * access (i.e., the access index is in the range 0..array.length).
	 * 
	 * @author pietro
	 */
	public static class ArrayEntryValue extends ArrayEntry implements Cloneable {
		protected Value val;

		/**
		 * Constructor.
		 * @param exp A <code>SEExpression</code> which denotes a 
		 *        constraint over the array index. When <code>null</code>
		 *        it denotes <code>true</code> (i.e., an unconstrained 
		 *        index value).
		 * @param val A <code>Value</code> which denotes the value returned  
		 *        by an array access under the condition <code>exp</code>. It 
		 *        can be any <code>Value</code> according to the array's type, 
		 *        or a <code>ArrayConstraint</code>, which denotes a reference 
		 *        to another array not yet available in the state's heap.
		 */
		private ArrayEntryValue(Expression exp, Value val) {
			super(exp);
			this.val = val;
		}

		public Value getValue() { return this.val; }

		@Override
		public ArrayEntryValue clone() {
			return new ArrayEntryValue(this.exp, this.val);
		}
	}

	/**
	 * Class representing an array entry whose outcome is the  
	 * generation of an <code>IndexOutOfBoundsException</code> 
	 * (i.e., the access index is outside the range 0..array.length).
	 * 
	 * @author pietro
	 *
	 */
	public static class ArrayEntryOutOfBound extends ArrayEntry { 
		private ArrayEntryOutOfBound(Expression exp) { super(exp); }
	}

	public static String checkAndReturnArrayType(int type) {
		String retVal = "" + Type.ARRAYOF;
		
		switch (type) {
		case T_BOOLEAN:
			retVal += Type.BOOLEAN;
			break;
		case T_CHAR:
			retVal += Type.CHAR;
			break;
		case T_FLOAT:
			retVal += Type.FLOAT;
			break;
		case T_DOUBLE:
			retVal += Type.DOUBLE;
			break;
		case T_BYTE:
			retVal += Type.BYTE;
			break;
		case T_SHORT:
			retVal += Type.SHORT;
			break;
		case T_INT:
			retVal += Type.INT;
			break;
		case T_LONG:
			retVal += Type.LONG;
			break;
		default:
			retVal = null;
		}
		return retVal;
	}
	
	private Array(Primitive length, String origin) throws JavaReifyException {
		super(origin);
		this.length = length;
		Expression e1 = new Expression(INDEX, Operator.GE, new Simplex(0));
		Expression e2 = new Expression(INDEX, Operator.LT, length);
		this.indexInRange  = (Expression) e1.and(e2);		
		this.values = new LinkedList<ArrayEntryValue>();
		this.normalized = (length instanceof Simplex);
	}

	/**
	 * Constructor.
	 * 
	 * @param initSymbolic <code>true</code> iff the array must be initialized 
	 *        with symbolic values.
	 * @param initValue a <code>Value</code> for initializing the array (ignored
	 *        whenever <code>initSymbolic == true</code>); if <code>initValue == null</code>
	 *        the default value for the array member type is used for initialization.
	 * @param length the number of elements in the array.
	 * @param type   the type of the array.
	 * @param origin the origin of the array.
	 * @throws JavaReifyException iff the type is not reference.
	 */
	public Array(boolean initSymbolic, Value initValue, Primitive length, String type, String origin) throws JavaReifyException {
		this(length, origin);

		boolean illFormed = false;
		if (type == null || type.length() < 2) {
			illFormed = true;
		} else {
			switch (type.charAt(1)) {
			case Type.BOOLEAN:
			case Type.CHAR:
			case Type.FLOAT:
			case Type.DOUBLE:
			case Type.BYTE:
			case Type.SHORT:
			case Type.INT:
			case Type.LONG:
			case Type.REFERENCE:
			case Type.ARRAYOF:
				break;
			default:
				illFormed = true;
			}
		}
		if (illFormed) {
			throw new JavaReifyException("java/lang/VerifyError");
		}
		this.type = type;
		this.contentType = this.type.charAt(1);
		this.setEntriesInit(initSymbolic, initValue);
	}

	private void setEntriesInit(boolean initSymbolic, Value initValue) {
		Value entryValue;
		if (initSymbolic) {
			entryValue = null;
		} else if (initValue == null) {
			entryValue = Type.createDefault(this.contentType); 
		} else {
			entryValue = initValue;
		}
		
		//in the case length is not symbolic, creates an entry for each 
		//possible value in the range; the rationale is, it is better having 
		//more, restrictive entries than less, liberal entries, since the 
		//most workload is on the simplifier communication part, and with 
		//restrictive entries we may hope that normalization will succeed 
		//upon array access, thus reducing the calls to the simplifier.
		if (this.normalized) {
			int ln = (Integer) ((Simplex) this.length).getActualValue();
			for (int i = 0; i < ln; i++) {
				this.values.add(new ArrayEntryValue(new Expression(INDEX, Operator.EQ, new Simplex(i)), 
						entryValue));
			}			
		} else {
			this.values.add(new ArrayEntryValue(this.indexInRange, entryValue));
		}
	}
	
	/**
	 * Returns the type of the object
	 */
	public String getType() {
		return this.type;
	}
	
	//TODO perhaps this method should be eliminated
	public char getContentType() {
		return this.contentType;
	}

	/**
	 * Returns the number of elements of the array.
	 */
	public Primitive getLength() {
		return this.length;
	}
	
	/**
	 * Checks whether the array is in normal form.
	 * 
	 * @return <code>true</code> iff the array is in normal form.
	 */
	public boolean isNormalized() {
		return this.normalized;
	}
	
	/**
	 * Sets all the entries with a value. It is ok to invoke 
	 * this method only after 
	 * 
	 * @param v a <code>Value</code>.
	 */
	public void replaceEntries(Value v) {
		for (ArrayEntryValue e : this.values) {
			e.val = v;
		}
	}

	/**
	 * Returns an element of the array.
	 * 
	 * @param index the index of the element in the array.
	 * @param mgr a <code>SESimplifier</code> set with the current path condition.
	 * @return a <code>SortedSet&lt;ArrayEntry&gt;</code> containing all the 
	 *         <code>ArrayEntry</code>s in the array which are compatible, 
	 *         under the current path condition, with an access based on <code>index</code>.
	 * @throws DecisionException if the simplifier fails.
	 */
	public Set<ArrayEntry> getElement(Primitive index, DecisionProcedure mgr) throws DecisionException {
		HashSet<ArrayEntry> retVal = new HashSet<ArrayEntry>();
		Primitive getIndexIsOutOfRange = this.indexInRange.replace(INDEX, index).normalize().not();
		Expression getIndexIsOutOfRangeExp = (getIndexIsOutOfRange instanceof Expression) ? ((Expression) getIndexIsOutOfRange) : null;

		if (this.normalized && index instanceof Simplex) {
			//fast case: access this.values as an array
			int actualIndex = (Integer) ((Simplex) index).getActualValue();
			if (actualIndex >= 0 && actualIndex < (Integer) ((Simplex) this.length).getActualValue()) {
				ArrayEntryValue entry = this.values.get(actualIndex).clone();
				entry.exp = null;
				retVal.add(entry);
			} else {
				retVal.add(new ArrayEntryOutOfBound(getIndexIsOutOfRangeExp)); //here is surely getIndexIsOutOfRangeExp == null
			}
		} else {
			//adds all the entries which can be satisfied by the index parameter 
			for (ArrayEntryValue e : this.values) {
				Expression getIndexIsInRangeExp = e.exp.replace(INDEX, index);
				Value getIndexIsInRangeNorm = getIndexIsInRangeExp.normalize();
				boolean inRange;
				boolean normIsSimplex = (getIndexIsInRangeNorm instanceof Simplex); 
				if (normIsSimplex) {
					inRange = (Boolean) ((Simplex) getIndexIsInRangeNorm).getActualValue();
				} else {
					inRange = mgr.isSat((Expression) getIndexIsInRangeNorm); 
				}
				if (inRange) {
					retVal.add(new ArrayEntryValue((normIsSimplex ? null : (Expression) getIndexIsInRangeNorm), e.val));
				}
			}
	
			//in the case the out of bounds case can be satisfied by the 
			//index parameter, a suitable entry is also added
			boolean outOfRange;
			boolean isSimplex = (getIndexIsOutOfRange instanceof Simplex); 
			if (isSimplex) {
				outOfRange = (Boolean) ((Simplex) getIndexIsOutOfRange).getActualValue();
			} else {
				outOfRange = mgr.isSat(getIndexIsOutOfRangeExp); 
			}
			if (outOfRange) {
				retVal.add(new ArrayEntryOutOfBound(getIndexIsOutOfRangeExp));
			}
		}

		return retVal;
	}

	/**
	 * Sets an element of the array. The array is left unchanged 
	 * iff <code>!mayBeInRange(index, mgr)</code>.  
	 * 
	 * @param index the position of the array element to set.  
	 * @param item the new <code>Value</code> for the array element at position <code>index</code>.
	 * @param mgr a <code>SESimplifier</code> set with the current path condition.
	 * @throws DecisionException if the simplifier fails.
	 */
	public void setElement(Primitive index, Value item, DecisionProcedure mgr) throws DecisionException {
		Expression formalIndexIsSetIndex = new Expression(INDEX, Operator.EQ, index);

		if (this.normalized && index instanceof Simplex) {
			int actualIndex = (Integer) ((Simplex) index).getActualValue();
			if (actualIndex >= 0 && actualIndex < (Integer) ((Simplex) this.length).getActualValue()) {
				this.values().get(actualIndex).val = item;
			}
		} else if (this.mayBeInRange(index, mgr)) {
			this.normalized = false;

			//constrains all entries to be different from the set index,
			//and eliminates all the unsatisfiable ones
			for (ListIterator<ArrayEntryValue> it = this.values.listIterator(); it.hasNext(); ) {
				ArrayEntryValue e = it.next();

				//determines whether the new index may fall in the range 
				//specified by the entry
				Primitive indexInRange = e.exp.replace(INDEX, index).normalize();
				boolean indexMayFallInRange;
				if (indexInRange instanceof Simplex) {
					indexMayFallInRange = (Boolean) ((Simplex) indexInRange).getActualValue();
				} else {
					indexMayFallInRange = mgr.isSat((Expression) indexInRange);
				}
				
				//in such case, if the new value differs from the entry's value, 
				//the entry's range must be constrained
				if (indexMayFallInRange && (e.val == null || !e.val.equals(item))) {
					Expression rangeConstrained = (Expression) e.exp.and(formalIndexIsSetIndex.not());
					if (mgr.isSat(rangeConstrained)) {
						e.exp = rangeConstrained;
					} else {
						//in the case the constrained entry is not satisfiable, 
						//it removes it
						it.remove();
					}
				}
			}

			//finally, adds a new entry for the set index value
			this.values.add(new ArrayEntryValue(formalIndexIsSetIndex, item));
			/* TODO should we put the constraint on index in the entry? 
			 * I think no, as long as the caller increases the path condition.
			 * However it should be checked; this is how the previous row 
			 * should be corrected in the positive case:
			 * 
			 * SEExpression setElementPost = (SEExpression) this.indexInRange.and(formalIndexIsSetIndex);
			 * this.values.add(new ArrayEntryValue(setElementPost, item));
			 */
		}
	}
	
	/**
	 * Returns a <code>Value</code> denoting the fact that an index is in 
	 * range.
	 * 
	 * @param index a <code>Primitive</code>
	 * @return a <code>SEExpression</code> denoting the fact that 
	 * <code>index</code> is in range. In the case such fact can be 
	 * proved or disproved by normalization, a <code>Simplex</code> 
	 * denoting <code>true</code> or <code>false</code> respectively 
	 * is returned. 
	 */
	public Primitive inRange(Primitive index) {
		Primitive retVal = this.indexInRange.replace(INDEX, index).normalize();
		return retVal;
	}
	
	/**
	 * Returns a <code>Value</code> denoting the fact that an index is 
	 * out of range.
	 * 
	 * @param index a <code>Primitive</code>
	 * @return a <code>SEExpression</code> denoting the fact that 
	 * <code>index</code> is out of range. In the case such fact can be 
	 * proved or disproved by normalization, a <code>Simplex</code> 
	 * denoting <code>true</code> or <code>false</code> respectively 
	 * is returned. 
	 */
	public Primitive outOfRange(Primitive index)  {
		Primitive retVal = this.inRange(index).not();
		return retVal;
	}

	/**
	 * Checks whether some value may be in range when used as index for 
	 * an array access.
	 * 
	 * @param index a <code>Primitive</code>.
	 * @param mgr a <code>SESimplifier</code>.
	 * @return <code>true</code> iff the fact that <code>index</code> 
	 *         is greater or equal to <code>0</code> and less than 
	 *         </code>this.getLength()</code> is satisfiable under the 
	 *         current assumption stored by <code>mgr</code>. 
	 * @throws DecisionException
	 */
	public boolean mayBeInRange(Primitive index, DecisionProcedure mgr) throws DecisionException {
		Primitive in = this.inRange(index);
		if (in instanceof Simplex) {
			return (Boolean) ((Simplex) in).getActualValue();
		} else {
			return mgr.isSat((Expression) in);
		}
	}

	/**
	 * Checks whether some value may be out of range when used as index for 
	 * an array access.
	 * 
	 * @param index a <code>Primitive</code>.
	 * @param mgr a <code>SESimplifier</code>.
	 * @return <code>true</code> iff the fact that <code>index</code> 
	 *         is less than <code>0</code> or greater or equal to  
	 *         </code>this.getLength()</code> is satisfiable under the 
	 *         current assumption stored by <code>mgr</code>. 
	 * @throws DecisionException
	 */
	public boolean mayBeOutOfRange(Primitive index, DecisionProcedure mgr) throws DecisionException  {
		Primitive out = this.outOfRange(index);
		if (out instanceof Simplex) {
			return (Boolean) ((Simplex) out).getActualValue();
		} else {
			return mgr.isSat((Expression) out);
		}
	}

	/**
	 * Returns a list of <code>ArrayEntryValue</code> representing the 
	 * array's theory.
	 * 
	 * @return a <code>List&lt;ArrayEntryValue&gt;</code>.
	 */
	public List<ArrayEntryValue> values() {
		return this.values;
	}

	@Override
	public String toString() {
		String str = "[Type:" + this.type + ", Length:" + this.length.toString() + ", Elements: {";
		boolean firstEntryPassed = false;
		for (ArrayEntryValue e : this.values) {
			if (firstEntryPassed) {
				str += ", ";
			} else {
				firstEntryPassed = true;
			}
			str += e.exp.toString() + "->" + e.val.toString();
		}
		str += "}]";
		return str;
	}

	@Override
	public Array clone() {
		Array o = (Array) super.clone();
		if (o == null) {
			return o;
		}

		//TODO being Values immutable it should not be necessary to clone this.length and this.indexInRange, refinement shouldn't change the situation as both are primitive. However, should investigate correctness.

		o.values = new LinkedList<ArrayEntryValue>();
		for (ArrayEntryValue e : this.values) {
			o.values.add(e.clone());
		}

		return o;
	}

	@Override
	public void refine(Map<String, ReferenceSymbolic> refs) {
		for (ArrayEntryValue e : this.values) {
			if (e.exp != null) {
				e.exp.refine(refs);
			}
			if (e.val != null) {
				e.val.refine(refs);
			}
		}
	}

	@Override
	public Set<Integer> symIdsPrimitive() {
		HashSet<Integer> retVal = new HashSet<Integer>();

		for (ArrayEntryValue e : this.values) {
			retVal.addAll(e.exp.symIds());
			if (e.val instanceof PrimitiveSymbolic) {
				retVal.add(((PrimitiveSymbolic) e.val).getId());
			} else if (e.val instanceof Expression) {
				retVal.addAll(((Expression) e.val).symIds());
			} else if (e.val instanceof FunctionApplication) {
				retVal.addAll(((FunctionApplication) e.val).symIds());
			}
		}

		return retVal;
	}

	@Override
	public Set<Integer> symIdsReference() {
		HashSet<Integer> retVal = new HashSet<Integer>();

		for (ArrayEntryValue e : this.values) {
			if (e.val instanceof ReferenceSymbolic) {
				retVal.add(((ReferenceSymbolic) e.val).getId());
			}
		}

		return retVal;
	}
}
