package muse.mem;

import java.util.Map;

import muse.bc.LineNumberTable;
import muse.bc.Signature;



/**
 * Class representing the activation record of a single method
 */
public class Frame implements Cloneable {
	/** 
	 * The value for the return program counter of 
	 * the topmost (current) frame.
	 */
    public final static int UNKNOWN_PC = -1;
    
	/** The frame's method line number table. */
	private LineNumberTable lnt;
	
	/** The frame's operand stack. */ 
    private OperandStack operandStack;

	/** The frame's local variable area. */ 
    private LocalVariablesArea localVariable;

    /** The program counter for the frame's method. */
    private int programCounter;
    
    /** The program counter when the frame will be again the current one. */
    private int returnProgramCounter;
    
    /** The bytecode of the frame's method. */
    private byte[] code;
    
    /** The signature of the frame's method. */
    private Signature mySignature;
    
    /**
     * Constructor.
     * @param opse operand ThreadStack
     * @param lvse Local variable table
     * @param methodSignature  method's signature of frame
     */
    public Frame(LineNumberTable lnt, OperandStack opse, LocalVariablesArea lvse, Signature methodSignature){
    	this.lnt = lnt;
        this.operandStack = opse;
        this.localVariable = lvse;
        this.programCounter = 0;
        this.returnProgramCounter = UNKNOWN_PC;
        this.code = new byte[0];
        this.mySignature = methodSignature;
    }
    
    /**
     * Put the passed value on the top of operand stack.
     * @param item: Value to put on the top of operand stack.
     */
    public void push(Value item) {
        operandStack.push(item);
    }
    
    /**
     * Returns the local variable table.
     */
    public LocalVariablesArea getLocalVariableTable() {
        return this.localVariable;
    }
    
    /**
     * Return and delete the value from the top of the operand stack.
     * @return Value represent the data on the top of operand stack.
     */
    public Value pop() {
        return this.operandStack.pop();
    }
    
    /**
     * Returns the source code row corresponding to the 
     * frame's program counter.
     *  
     * @return the source code row corresponding to the 
     *         frame's program counter, or <code>-1</code> 
     *         iff no debug information is available. 
     */
	public int getSourceRow() {
    	int retVal = -1;
    	for (LineNumberTable.Row r : this.lnt) {
    		if (r.start > this.programCounter) {
    			break;
    		}
    		retVal = r.lineNumber;
    	}
    	return retVal;
    }
    
    /**
     *Save the program counter used before a method call.
     *@param programCounter: the value of program counter
     *@param codeArray: the bytecode array
     */
    public void saveFrameProgramCounter(int programCounter) throws CodeOutOfIndexException {
    	this.boundCheckPCValue(programCounter);
        this.programCounter = programCounter;
        this.returnProgramCounter = UNKNOWN_PC;
    }
    
    public void saveFrameReturn(int returnProgramCounterOffset) throws CodeOutOfIndexException {
    	this.boundCheckPCValue(this.programCounter + returnProgramCounterOffset);
        this.returnProgramCounter = this.programCounter + returnProgramCounterOffset;
    }
    
    public void useReturnProgramCounter() throws CodeOutOfIndexException {
    	this.saveFrameProgramCounter(this.returnProgramCounter);
    }

    private void boundCheckPCValue(int newPC) throws CodeOutOfIndexException {
    	if (newPC < 0 || newPC >= this.code.length)
            throw new CodeOutOfIndexException();
    }

    public void saveFrameCode(byte[] codeArray){
        this.code = codeArray;
    }

    /**
     *Return the bytecode array saved for a method call
     *@return code[]: the bytecode array
     */
    public byte[] getCodeArray(){
        return this.code;
    }
    
    /**
     * Returns the frame's bytecode instruction pointed by 
     * the frame's program counter.
     * 
     * @return a <code>byte</code> representing the 
     *         bytecode pointed by the frame's program counter.
     */
    public byte getInstruction() {
    	return this.code[this.programCounter];
    }
    
    /**
     * Returns the frame's bytecode instruction pointed by 
     * the frame's program counter.
     * 
     * @return a <code>byte</code> representing the 
     *         bytecode pointed by the frame's program counter
     *         plus <code>displ</code>.
     * @throws CodeOutOfIndexException iff the frame's program
     *         counter plus <code>displ</code> does not point to 
     *         a bytecode.
     */
    public byte getInstruction(int displ) throws CodeOutOfIndexException {
    	this.boundCheckPCValue(this.programCounter + displ);
    	return this.code[this.programCounter + displ];
    }
    
    /**
     *Return the program counter saved for a method call
     *@return int: the value of program counter
     */
    public int getProgramCounter(){
        return this.programCounter;
    }
    
    public int getReturnProgramCounter(){
        return this.returnProgramCounter;
    }
    
    /**
     * Returns the signature of the frame's method.
     * @return the method's <code>Signature</code>.
     */
    public Signature getName() {
        return this.mySignature;
    }
    
    public OperandStack getOperandStack() {
    	return this.operandStack;
    }
    
    /**
     * Returns the topmost element in the operand stack of the frame, 
     * without removing it from the stack
     */
    @Deprecated
    public Value peekOperandStack() {
        return operandStack.returnLast();
    }
    
	public void refine(Map<String, ReferenceSymbolic> refs) {
		this.operandStack.refine(refs);
		this.localVariable.refine(refs);
	}
    
    @Override
    public Object clone() {
        Frame o = null;
        try {
            o = (Frame)super.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        
        o.code = new byte[code.length];
        for (int i = 0; i < code.length; i++) 
            o.code[i] = code[i];
        
        o.operandStack = (OperandStack)o.operandStack.clone();
        o.localVariable = (LocalVariablesArea)o.localVariable.clone();
        
        return o;
    }

	@Override
	public String toString(){
    	String tmp = "[";
        tmp += "Method:" + mySignature.toString() + ", ";
        tmp += "ProgramCounter:" + programCounter + ", ";
        tmp += "ReturnProgramCounter:" + (returnProgramCounter == UNKNOWN_PC ? "UNKNOWN" : returnProgramCounter) + ", ";
        tmp += "OperandStack:" + operandStack.toString() +", ";
        tmp += "Locals:" + localVariable.toString() + "]";
        return tmp;
    }
}