package muse.mem;

import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * Class that offers the same services of the heap in the JVM's memory.
 */
public final class Heap implements Cloneable {
    private SortedMap<Integer, Objekt> objList;
    private int nextIndex;
    
    /**
     * Constructor of an Heap structure.
     */
    public Heap() {
        this.objList = new TreeMap<Integer, Objekt>();
        this.nextIndex = 0;
    }
    
    /**
     * Stores a new object into the heap.
     * 
     * @param item the <code>ObjDef</code> to be stored in 
     *             the heap.
     * @return the position in the heap  
     *         of <code>item</code> in the case the object 
     *         is successfully stored.
     */
    public int addNew(Objekt item) {
        objList.put(this.nextIndex, item);
        int retVal = this.nextIndex;
        while (objList.containsKey(this.nextIndex)) {
        	this.nextIndex++;
        }
        return retVal;
    }
    
    /**
     * Sets an object into the heap.
     * 
     * @param ref a <code>Reference</code>; if it is a 
     *            <code>ReferenceSymbolic</code> it must be 
     *            <code>ref.isInitialized()</code> or the method 
     *            has no effect.
     * @param item an <code>Objekt</code>, which becomes the instance 
     *             stored at <code>ref</code>.
     */
    public void set(Reference ref, Objekt item) {
    	if (ref instanceof ReferenceSymbolic) {
    		if (!((ReferenceSymbolic) ref).isInitialized()) {
    			return;
    		}
    	}
    	int pos = ref.getHeapPosition();
    	this.objList.put(pos, item);
        while (objList.containsKey(this.nextIndex)) {
        	this.nextIndex++;
        }
    }
    
    /**
     * Gets an object from the heap.
     * 
     * @param ref a <code>Reference</code>.
     * @return the <code>ObjDef</code> at <code>ref</code>, or 
     *         <code>null</code> if the location at <code>ref</code> 
     *         is not mapped to a value (note that this differs from 
     *         the fact that the null value is present at location 
     *         <code>ref</code>, in which case TODO complete description
     **/
    public Objekt getObject(Reference ref) {
        return this.objList.get(ref.getHeapPosition());
    }
    
    public int getSize() {
    	return this.objList.size();
    }
    
    /**
     * Returns String representation of Heap
     */
    @Override
    public String toString() {
        String tmpRet = "[";
        int j = 0;
        for (Map.Entry<Integer, Objekt> e : this.objList.entrySet()) {
            tmpRet += e.getKey() + ":";
            tmpRet += e.getValue().toString();
            if (j < this.objList.size() - 1) 
            	tmpRet += ", ";
            j++;
        }
        tmpRet += "]";
        return(tmpRet);
    }
    
    @Override
    public Heap clone() {
        Heap h = null;
        try {
            h = (Heap) super.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        
        SortedMap<Integer, Objekt> objListClone = new TreeMap<Integer, Objekt>();
        
        for (Map.Entry<Integer, Objekt> e : this.objList.entrySet()) {
        	Objekt val = e.getValue();
            objListClone.put(e.getKey(), val.clone());
        }
        h.objList = objListClone;
        return h;
    }
    
    public Map<Integer, Objekt> getObjects() {
        return this.objList;
    }
    
	public void refine(Map<String, ReferenceSymbolic> refs) {
		for (Objekt o : this.objList.values()) {
			o.refine(refs);
		}
	}
}