package muse.mem;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import muse.bc.Signature;


/**
 * Class that represent an instance of an object in memory.
 *
 * It contain a reference to the Heap
 */

public class Instance extends Objekt {
    //Fields
    private HashMap<String, Variable> fields;
    private Signature[] fieldSignatures;
    /**
     * Constructor.
     * @param fieldSignatures an array of field signatures, field[i][0]contains the type. field[i][1] contains the name.
     * @param objType the type of the object.
     */
    public Instance(Signature[] fieldSignatures, String objType, String origin) {
    	super(origin);
        this.type = objType;
        this.fields = new HashMap<String, Variable>();
        this.fieldSignatures = fieldSignatures;
        for (Signature s : this.fieldSignatures) {
            fields.put(s.toString(), new Variable(s.getDescriptor(), s.getName()));
        }
    }
    
    /**
     * Gets a field of the <code>Instance</code>. Throws a runtime 
     * exception if the <code>Instance</code> has no field with the 
     * provided name.
     * 
     * @param name the name of the field.
     * @return a <code>Variable</code> object which is the field of the 
     * <code>Instance</code>. 
     */
    public Variable getFieldVariable(String name) {
        Variable tmpVariable = (Variable) fields.get(name);
        return tmpVariable;
    }
    
    /**
     * Gets the value in a field of the <code>Instance</code>.
     * 
     * @param name the name of the field.
     * @return a <code>Value</code> object which is the value 
     * stored in the field of the <code>Instance</code>, or 
     * <code>null</code> if the <code>Instance</code> has no 
     * field with name <code>name</code>. 
     */
    public Value getFieldValue(Signature sig) {
        try {
            return ((Variable) fields.get(sig.toString())).getValue();
        } catch (Exception e) {
            return null;
        }
    }
    
    /**
     * Sets the value of a field. Throws a runtime exception 
     * in the case the field does not exist.
     * 
     * @param field the <code>SESignature</code> of the field.
     * @param item the new <code>Value</code> that must be assigned to
     *             the field.
     */
    public void setFieldValue(Signature field, Value item) {
        ((Variable) fields.get(field.toString())).setValue(item);
    }
    
    /**
     * Return the dynamic type of the object.
     */
    public String getType() {
        return(this.type);
    }
    
    /**
     *Return the number of field.
     */
    public int getLength() {
        return(fields.size());
    }
    
    public Map<String, Variable> fields() {
    	return this.fields;
    }
    
    /**
     *Returns String rapresentation of Instance
     */
    @Override
    public String toString() {
        String tmp = "[Class:" + this.type + ", Fields:{";
        int i = 0;
        for (Map.Entry<String, Variable> e : this.fields.entrySet()) {
            tmp += e.getValue().toString();
            if (i < this.fields.entrySet().size() - 1)
            	tmp += ", ";
            i++;
        }
        tmp += "}]";
        return tmp;
    }
    
    @Override
    public Instance clone() {
        Instance o = null;
        o = (Instance) super.clone();
        if (o == null) return null;
        
        HashMap<String, Variable> newPropertiesMap = new HashMap<String, Variable>();
        
        for (String key : fields.keySet()) {
            Variable value = fields.get(key);
            Variable valueCopy = (Variable) value.clone();
            newPropertiesMap.put(key, valueCopy);
        }
        
        o.fields = newPropertiesMap;
        return o;
    }
    
    public Signature[] getFieldSignatures() {
    	return this.fieldSignatures;
    }
    
    @Override
	public void refine(Map<String, ReferenceSymbolic> refs) {
		for (Variable v : this.fields.values()) {
			Value val = v.getValue();
			val.refine(refs);
		}
	}

	@Override
	public Set<Integer> symIdsPrimitive() {
		HashSet<Integer> retVal = new HashSet<Integer>();
		for (Map.Entry<String, Variable> e : this.fields.entrySet()) {
			Value v = e.getValue().getValue(); 
			if (v instanceof PrimitiveSymbolic) {
				retVal.add(((PrimitiveSymbolic) v).getId());
			} else if (v instanceof Expression) {
				retVal.addAll(((Expression) v).symIds());
			} else if (v instanceof FunctionApplication) {
				retVal.addAll(((FunctionApplication) v).symIds());
			}
		}
		return retVal;
	}

	@Override
	public Set<Integer> symIdsReference() {
		HashSet<Integer> retVal = new HashSet<Integer>();
		for (Map.Entry<String, Variable> e : this.fields.entrySet()) {
			Value v = e.getValue().getValue(); 
			if (v instanceof ReferenceSymbolic) {
				retVal.add(((ReferenceSymbolic) v).getId());
			}
		}
		return retVal;
	}
}