package muse.mem;

import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import muse.bc.LocalVariableTable;


/**
 * Class representing a local variable memory area.
 */
public class LocalVariablesArea implements Cloneable {
	/** Values in the memory area, accessible by slot. */
	private SortedMap<Integer, Value> values;

	/** The local variable table for the method. */
	private LocalVariableTable lvt;
	
    /**
     * Constructor of the local variable table.
     * 
     * @param args array of variable for inizialization of local variable table.
     */
    public LocalVariablesArea(LocalVariableTable lvt, Value[] args) {
    	this.values = new TreeMap<Integer, Value>();
    	this.lvt = lvt;
		
		//initializes all the local variables by using args
        //until exhaustion, then DefaultValue
		int nargs = (args == null ? 0 : args.length);
		int j = 0;
		int slot = 0;
		while (slot < this.lvt.getSlots()) {
			Value val = null;
            if (j < nargs) {
            	val = args[j];
            	j++;
            } else {
            	val = DefaultValue.getInstance();
            }
            this.set(slot, val);
            
            //next slot
            slot++;
            if (!Type.isCat_1(val.getType())) {
            	slot++;
            }
		}
    }

    /**
     * Stores a value into a specific slot of the local variable table array.
     * 
     * @param slot slot number of the local variable.
     * @param item the value to be stored.  
     */
    public void set(int slot, Value val) {
    	//bound-checks slot and if necessary nullifies value at slot + 1  
    	if (val.getType() == Type.DOUBLE || val.getType() == Type.LONG) {
    		if (slot < 0 || slot >= this.lvt.getSlots() - 1)
    			return; //TODO throw JavaReifyException(???) ???
    		this.values.remove(slot + 1);
    	}
   		if (slot < 0 || slot >= this.lvt.getSlots())
   			return; //TODO throw JavaReifyException(???) ???

   		//stores val at slot
        this.values.put(slot, val);
    }
    
    public Value get(int slot) {
        Value retVal = this.values.get(slot);

        /* TODO the next case should never happen in well-formed code, however 
         * the JVM specification does not explicitly forbid it. Should 
         * we throw JavaReifyException("java/lang/VerifyError"), throw some 
         * InsufficientTypeInformationException to be defined, or rather return 
         * the true default value, which can be inferred from the bytecode??? 
         * Note that implementing the latter case would require deep 
         * refactoring, and perhaps is best done by performing inside  
         * SELoad.exec(State) the conversion of DefaultValue's instance 
         * into a true default value. 
         * Here we chose to return DefaultValue, in order to better support 
         * prettyprinting. Externally no support to DefaultValue handling 
         * is implemented, but it shouldn't permeate with well-behaving 
         * code so this can be safely added to the todo list.
         */
        if (DefaultValue.getInstance().equals(retVal))
        	;
        
        return retVal;
    }
    
    public Set<Integer> slots() {
    	return this.values.keySet();
    }
    
    /**
     * Builds a local variable.
     * 
     * @param slot the number of the slot.
     * @param curPC the current program counter.
     * @return a <code>Variable</code> at position.
     */
    public Variable buildVar(int slot, int curPC) {
    	//gets the value
        Value val = this.get(slot);
                
        //finds static information
        Variable retVal = null;
        LocalVariableTable.Row r = this.lvt.row(slot, curPC);
        if (r == null) {
        	retVal = new Variable("" + Type.UNKNOWN, "__LOCAL[" + slot + "]", val);
        } else {
    		retVal = new Variable(r.descriptor, r.name, val);
        }
        return retVal;
    }
    
    @Override
    public Object clone() {
        LocalVariablesArea o = null;
        try {
            o = (LocalVariablesArea) super.clone();
        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }
        
        SortedMap<Integer, Value> valuesCopy = new TreeMap<Integer, Value>();
        for (Integer key : this.values.keySet()) {
            Value val = this.values.get(key);
            Value valCopy = (Value) val.clone();
            valuesCopy.put(key, valCopy);
        }
        
        o.values = valuesCopy;
        return o;
    }
    
    /**
     * Returns a <code>String</code> representation for the local variable area
     */
    @Override
    public String toString() {
        String tmp = "[";
        int j = 0;
       	for (Map.Entry<Integer, Value> e : values.entrySet()) {
            tmp += e.getKey() + ":"+ e.getValue();
            if (j < values.size() - 1)
            	tmp += ", ";
            j++;
        }
       	tmp += "]";
        return tmp;
    }

	public void refine(Map<String, ReferenceSymbolic> refs) {
		for (Value v : this.values.values())
			v.refine(refs);
	}    
}