package muse.mem;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

/**
 * Class that represent a path condition.
 */
public class PathCondition implements Cloneable {
	public static class Clause implements Cloneable {
		private Expression p = null;
		private ReferenceSymbolic r = null;
		private Objekt o = null;
		private Klass k = null;
		private String className = null;
		
		/**
		 * Constructor for a clause over primitive symbolic values.
		 * 
		 * @param p a <code>Primitive</code> 
		 */
		Clause(Expression p) { this.p = p; }
		Clause(ReferenceSymbolic r, Objekt o) { this.r = r; this.o = o; }
		Clause(Klass k, String className) { this.k = k; this.className = className; }
		
		public boolean hasCondition() { return this.p != null; }
		public boolean hasRefinement() { return this.r != null; }
		public boolean hasObject() { return this.o != null; }
		public boolean hasKlass() { return this.k != null; }
		public Expression getCondition() { return this.p; }
		public ReferenceSymbolic getRefinement() { return this.r; }
		public Objekt getObject() { return this.o; }
		public Klass getKlass() { return this.k; }
		public String getClassName() { return this.className; }
		
		@Override
		public Clause clone() {
			Clause o;
			try {
				o = (Clause) super.clone();
			} catch (CloneNotSupportedException e) {
				return null;
			}
			//TODO do we need to clone everything?
			if (this.p != null) o.p = this.p.clone();
			if (this.r != null) o.r = this.r.clone();
			if (this.o != null) o.o = this.o.clone();
			if (this.k != null) o.k = this.k.clone();
			
			return o;
		}
		
		@Override
		public boolean equals(Object o) {
			if (!(o instanceof Clause))
				return false;
			Clause c = (Clause) o;
			if (this.p == null) {
				if (c.p != null) return false;
			} else {
				if (!this.p.equals(c.p)) return false;
			}

			if (this.r == null) {
				if (c.r != null) return false;
			} else {
				if (!this.r.equals(c.r)) return false;
			}

			if (this.o == null) {
				if (c.o != null) return false;
			} else {
				if (!this.o.equals(c.o)) return false;
			}
			return true;
		}
		
		@Override
		public String toString() {
			String retVal = "";
    		if (this.hasCondition()) {
    			retVal += this.getCondition().toString();
    		} else if (this.hasRefinement()) {
    			ReferenceSymbolic ref = this.getRefinement();
    			retVal += ref.toString() + " == " + (ref.isNull() ? "null" : ("Object[" + ref.getHeapPosition() + "]"));
    		} else {
    			retVal += "access(" + this.getClassName() +")";
    		}
    		return retVal;
		}
	}
	
	/** List of all the clauses added to the path condition */
	private List<Clause> clauses;
	
	/** true iff this.PC is invalid */
	private boolean dirty;
	
	/** caches the path condition as an expression, put to null means true */
    private Expression PC;
    
    /**
     * Default constructor.
     */
    public PathCondition() {
    	this.clauses = new ArrayList<Clause>();
    	this.dirty = false;
    	this.PC = null;
    }
    
    /**
     * Constructor accepting an initial condition.
     * @param condition the initial condition as a <code>Primitive</code>.
     */
    public PathCondition(Expression condition) {
    	this.clauses.add(new Clause(condition));
    	this.dirty = false;
        this.PC = condition;
    }
    
    /**
     * Returns all the <code>Clause</code>s of the <code>PathCondition</code>. 
     * @return a <code>Collection&lt;Clause&gt;</code> containing all the <code>Clause</code>s.
     */
    public Collection<Clause> getClauses() {
    	return this.clauses;
    }
    
    /**
     * Adds a clause to the path condition. The clause is a condition 
     * over primitive values.
     * 
     * @param condition the additional condition as a <code>SEExpression</code>.
     */
    public void concatenate(Expression condition) {
    	if (condition != null) {
    		this.clauses.add(new Clause(condition));
    		this.dirty = true;
    	}
    }

    /**
     * Adds a clause to the path condition. The clause is a refinement 
     * of a symbolic reference by an instance in its initial state. 
     * 
     * @param r the <code>SESymbolicReference</code> which is refined. It 
     *          must be <code>r != null</code>, or the method has no effect. 
     * @param o the <code>ObjDef</code> with which <code>r</code> is refined, 
     *          iff <code>o</code> is a fresh (i.e., not reachable before) 
     *          symbolic object, or <code>null</code> iff <code>r</code> is 
     *          refined to <code>null</code> or to an already available 
     *          symbolic object.
     */
    public void concatenate(ReferenceSymbolic r, Objekt o) {
    	if (r != null) {
    		this.clauses.add(new Clause(r, o));
    		//this.dirty doesn't change
    	}
    }

    /**
     * Adds a clause to the path condition. The clause is the introduction 
     * of a symbolic <code>Klass</code> object.
     *   
     * @param k the <code>Klass</code> object.
     * @param className the corresponding concrete class name as a <code>String</code>.
     */
    public void concatenate(Klass k, String className) {
   		this.clauses.add(new Clause(k, className));
   		//this.dirty doesn't change
    }

    /**
     * Returns the current <code>Primitive</code> value of the path 
     * condition.
     * 
     * @return a <code>Primitive</code> representing the logical && 
     * concatenation of all the clauses cumulated in <code>this</code>.
     */
    public Primitive getCondition() {
    	if (this.dirty) {
    		this.recalculatePC();
    	}
    	if (this.PC == null) {
    		return new Simplex(true);
    	}
        return this.PC;
    }

    /**
     * Scans all the <code>Clause</code>s and recalculates the 
     * value of <code>this.PC</code>.
     */
    private void recalculatePC() {
    	this.PC = null;
    	for (Clause c : this.clauses) {
    		if (c.hasCondition()) {
	    		if (this.PC == null) {
	    			this.PC = c.getCondition();
	    		} else {
	    			this.PC = new Expression(this.PC, Operator.AND, c.getCondition());
	    		}
    		}
    	}
    }
    
    /**
     * Returns the last <code>Clause</code> that has been added to <code>this</code>.
     * 
     * @return a <code>Clause</code>, the same added with the last call to
     * <code>concatenate</code>.
     */
    public Clause getLastAddedClause() {
    	if (this.clauses.size() > 0) {
    		return this.clauses.get(this.clauses.size() - 1);
    	} else {
    		return null;
    	}
    }
    
    /**
     * Tests whether the path condition refines another one.
     * 
     * @param pc the <code>PathCondition</code> to be compared against.
     * @return <code>true</code> iff <code>this</code> refines <code>pc</code>.
     */
    public boolean refines(PathCondition pc) {
    	if (pc.clauses == null) {
    		return true;
    	}
    	
    	Iterator<Clause> i = this.clauses.iterator();
    	for (Clause c : pc.clauses) {
    		if (!i.hasNext()) {
    			return false;
    		}
    		Clause cc = i.next();
    		if (!cc.equals(c)) {
    			return false;
    		}
    	}
    	return true;
    }
    
    @Override
    public String toString() {
    	String retVal = "";
    	boolean doneFirst = false;
    	for (Clause c : this.clauses) {
    		if (doneFirst) {
    			retVal += " && ";
    		}
    		doneFirst = true;
    		retVal += c.toString();
    	}
    	
    	if (retVal.equals("")) {
    		retVal = "true";
    	}
    	return retVal;
    }
    
    @Override
    public PathCondition clone() {
        PathCondition o = null;
        try {
            o = (PathCondition) super.clone();
        } catch (CloneNotSupportedException e) {
        	return o;
        }
        
        //does a deep copy of this.clauses
        if (o.clauses != null) {
        	o.clauses = new ArrayList<Clause>();
        	for (Clause c : this.clauses) {
        		o.clauses.add(c.clone());
        	}
        }
        
        //sets dirty to true; no need to clone PC
        o.dirty = true;
        return o;
    }
}
