package muse.mem;

import java.io.IOException;

import muse.dec.DecisionExternalProtocolException;


/**
 * Class that represent a primitive type value.
 */
public abstract class Primitive extends Value implements Cloneable {
    public abstract String toString();
    public abstract void accept(ExpressionVisitor v) throws IOException, DecisionExternalProtocolException;

    /**
     * Calculates the sum of <code>this</code> and another value.
     * 
     * @param param the <code>Primitive</code> representing the value to 
     *              be added to <code>this</code>.
     * @return a <code>Primitive</code> representing the result of the 
     *         operation.
     */
    public Primitive add(Primitive param){
        Expression exp = new Expression(this,Operator.ADD,param);
        return(exp.normalize());
    }
    /**
     * Calculates the product of <code>this</code> and another value.
     * 
     * @param param the <code>Primitive</code> representing the value to 
     *              be multiplied to <code>this</code>.
     * @return a <code>Primitive</code> representing the result of the 
     *         operation.
     */
    public Primitive mul(Primitive param){
        Expression exp=new Expression(this,Operator.MUL,param);
        return(exp.normalize());
    }
    
    /**
     * Calculates the difference between <code>this</code> and another value.
     * 
     * @param param the <code>Primitive</code> representing the value to 
     *              be subtracted from <code>this</code>.
     * @return a <code>Primitive</code> representing the result of the 
     *         operation.
     */
    public Primitive sub(Primitive param){
        Expression exp=new Expression(this,Operator.SUB,param);
        return(exp.normalize());
    }
    
    /**
     * Calculates the result of the division of <code>this</code> by
     * another value.
     * 
     * @param param the <code>Primitive</code> representing the value to 
     *              which <code>this</code> must be divided.
     * @return a <code>Primitive</code> representing the result of the 
     *         operation.
     */
    public Primitive div(Primitive param)
    {
        Expression exp=new Expression(this,Operator.DIV,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the reminder of itself and the passed value
     *@param param: the value to compute
     */
    public Primitive rem(Primitive param) {
        Expression exp = new Expression(this,Operator.REM,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the arithmetic negation of itself
     */
    public Primitive neg()
    {
        Expression exp=new Expression(Operator.NEG,this);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the bitwise and of itself and the passed value
     *@param param: the value to compute
     */
    public Primitive andBitwise(Primitive param) {
        Expression exp=new Expression(this,Operator.ANDBW,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the bitwise or of itself and the passed value
     *@param param: the value to compute
     */
    public Primitive orBitwise(Primitive param) {
        Expression exp=new Expression(this,Operator.ORBW,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the and of itself and the passed value
     *@param param: the value to compute
     */
    public Primitive and(Primitive param)
    {
        Expression exp=new Expression(this,Operator.AND,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the or of itself and the passed value
     *@param param: the value to compute
     */
    public Primitive or(Primitive param)
    {
        Expression exp=new Expression(this,Operator.OR,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the not of itself
     */
    public Primitive not()
    {
        Expression exp = new Expression(Operator.NOT,this);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the xor of itself and the passed value
     *@param param: the value to compute
     */
    public Primitive xor(Primitive param)
    {
        Expression exp=new Expression(this,Operator.XOR,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the shift left of itself by passed bit
     *@param param: the value to compute
     */
    public Primitive shl(Primitive param)
    {
        Expression exp=new Expression(this,Operator.SHL,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the shift right of itself by passed bit
     *@param param: the value to compute
     */
    public Primitive shr(Primitive param)
    {
        Expression exp=new Expression(this,Operator.SHR,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return the unsigned shift right of itself by passed bit
     *@param param: the value to compute
     */
    public Primitive ushr(Primitive param)
    {
        Expression exp=new Expression(this,Operator.USHR,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return a Primitive_bool value result of comparison
     *@param param: the value to compute
     */
    public Primitive eq(Primitive param) {
        Expression exp=new Expression(this,Operator.EQ,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return a Primitive_bool value result of comparison
     *@param param: the value to compute
     */
    public Primitive ne(Primitive param)
    {
        Expression exp=new Expression(this,Operator.NE,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return a Primitive_bool value result of comparison
     *@param param: the value to compute
     */
    public Primitive le(Primitive param)
    {
        Expression exp=new Expression(this,Operator.LE,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return a Primitive_bool value result of comparison
     *@param param: the value to compute
     */
    public Primitive lt(Primitive param)
    {
        Expression exp=new Expression(this,Operator.LT,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return a Primitive_bool value result of comparison
     *@param param: the value to compute
     */
    public Primitive ge(Primitive param) {
        Expression exp=new Expression(this,Operator.GE,param);
        return(exp.normalize());
    }
    
    /**
     *Given a primitive type, return a Primitive_bool value result of comparison
     *@param param: the value to compute
     */
    public Primitive gt(Primitive param) {
        Expression exp = new Expression(this,Operator.GT,param);
        return(exp.normalize());
    }
    
    public static Primitive apply(String type, String operator, Primitive[] args) {
    	FunctionApplication app = new FunctionApplication(type, operator, args);
    	return(app.normalize());
    }
}