package muse.mem;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import muse.algo.Util;
import muse.bc.ClassFile;
import muse.bc.ClassFileInterface;
import muse.bc.ClassHierarchy;
import muse.bc.ExceptionTable;
import muse.bc.ExceptionTableEntry;
import muse.jvm.Environment;
import muse.jvm.Environment.Entry;
import muse.mem.Objekt.Epoch;


/**
 * Class that represents the state of execution.
 */
public class State implements Cloneable {
	public static final String IDENTIFIER_DEFAULT = ".1";

	/** 
	 * The identifier of the state in the execution tree. Only 
	 * branching points are identified.
	 */
	private String identifier;

	/** The depth of the state, i.e., the number of branch points over it. */
	private int depth;
	
	/** The sequence number of the state along an execution tree branch. */
	private int sequenceNumber;
	
	/** The class loading interface used by this state to fetch classes. */ 
	private ClassFileInterface classFileInterface;
	
	/** The string literals. */
	private HashMap<String, ReferenceConcrete> stringLiterals;
    
    /** The JVM stack of the current execution thread. */
    private ThreadStack stack;
    
    /** The JVM heap (instances). */
    private Heap heap;
    
    /** The JVM static store. */
    private StaticMethodArea staticStore;
    
    /** The path condition of the state in the execution tree. */
    private PathCondition pathCondition;

    /** <code>true</code> iff the path condition has been increased. */ 
    private boolean augmentedPathCondition;
    
    /** <code>true</code> iff the state is stuck. */
    private boolean stuck;
	
    /** 
     * The exception raised to the root method's caller, 
     * meaningful only when <code>this.stuck == true</code>. 
     */ 
    private Reference exc;
    
    /** 
     * The return value for the root method's caller, 
     * meaningful only when <code>this.stuck == true</code>. 
     */ 
    private Value val;

    /** The next available identifier for a new reference-typed symbolic value. */
	private int nextIdRefSym;

    /** The next available identifier for a new primitive-typed symbolic value. */
	private int nextIdPrimSym;

    /** The identifiers already assigned to reference-typed symbolic values. */
	private Set<Integer> consumedSymIdsReference;

    /** The identifiers already assigned to primitive-typed symbolic values. */
	private Set<Integer> consumedSymIdsPrimitive;
	
	/** The names of the classes which are not yet initialized. */
	private Set<String> notInitializedClasses;
    
    /**
     * Constructor of an empty State.
     */
    public State(ClassFileInterface cfi) {
    	this.classFileInterface = cfi;
    	this.identifier = IDENTIFIER_DEFAULT;
    	this.depth = 0;
    	this.sequenceNumber = 0;
    	this.stringLiterals = new HashMap<String, ReferenceConcrete>();
        this.stack = new ThreadStack();
        this.heap = new Heap();
        this.staticStore = new StaticMethodArea();
        this.pathCondition = new PathCondition();
        this.augmentedPathCondition = false;
        this.stuck = false;
        this.exc = null;
        this.val = null;
        this.nextIdRefSym = this.nextIdPrimSym = 0;
        this.consumedSymIdsReference = new HashSet<Integer>();
        this.consumedSymIdsPrimitive = new HashSet<Integer>();
        this.notInitializedClasses = new HashSet<String>();
    }
    
    /**
     * Return Stack Frame
     */
    public ThreadStack getStack() {
        return this.stack;
    }
    
    /**
     * Tests whether a class has a <code>Klass</code> object.
     * 
     * @param className the name of the class.
     * @return <code>true</code> if a <code>Klass</code> exists  
     *         for <code>className</code> in the <code>State</code>'s 
     *         static store, <code>false</code> otherwise.
     */
    public boolean hasKlass(String className) {
    	return this.staticStore.contains(className);
    }

    /**
     * Tests whether a class is initialized.  
     *  
     * @param className the name of the class.
     * @return <code>true</code> if the class has been initialized, 
     *         <code>false</code> otherwise.
     */
    public boolean initialized(String className) {
    	return !this.notInitializedClasses.contains(className);
    }
    
    /**
     * Tests whether a class is initialized, and its initialization 
     * epoch is before the symbolic execution start.
     * 
     * @param className the name of the class.
     * @return <code>true</code> if the class has been initialized 
     *         <emph>and</emph> its initialization epoch is before 
     *         the symbolic execution start, <code>false</code> 
     *         otherwise.
     */
    public boolean initializedBeforeStart(String className) {
    	if (!this.initialized(className)) {
    		return false;
    	}
    	Klass k = this.staticStore.get(className);
    	if (k == null) {
    		return true;
    	}
    	return (k.creationEpoch() == Epoch.EPOCH_BEFORE_START);
    }
    
    /**
     * Sets a class to not initialized, unless it has already 
     * been initialized.
     * 
     * @param className the name of the class.
     */
    public void setNotInitialized(String className) {
    	if (this.staticStore.contains(className)) {
    		return;
    	}
    	this.notInitializedClasses.add(className);
    }
    
    /**
     * Returns the state's <code>SEClassFileInterface</code>. 
     * @return the state's <code>SEClassFileInterface</code>.
     */
    public ClassFileInterface getClassFileInterface() {
    	return this.classFileInterface;
    }
    
    /**
     * Returns the <code>Klass</code> object corresponding to 
     * a given class name.
     * 
     * @param className the name of the class.
     * @return the <code>Klass</code> object corresponding to 
     *         the memory representation of the class 
     *         <code>className</code>, or <code>null</code> 
     *         if the class has not been initialized.
     */
    public Klass getKlass(String className) {
    	return this.staticStore.get(className);
    }
    
    /**
     * Associates a <code>Klass</code> object to a class name.
     * 
     * @param k a <code>Klass</code>. 
     * @param className a <code>String</code representing the 
     *        name of a class.
     */
    public void setKlass(Klass k, String className) {
    	this.notInitializedClasses.remove(className);
    	this.staticStore.set(k, className);
    }

    /**
     * Returns a reference to the instance corresponding to a 
     * string literal.
     * 
     * @param value a <code>String</code>.
     * @return a <code>SEConcreteReference</code> to an <code>Instance</code> 
     *         of class <code>String</code> which equals <code>value</code>, 
     *         if such instance exists in the state's heap, <code>null</code> 
     *         otherwise.
     */
    public ReferenceConcrete getStringLiteral(String value) {
    	return this.stringLiterals.get(value);
    }
    
    /**
     * Associates to a string literal an <code>Instance</code> in the 
     * state's heap.
     * 
     * @param value a <code>String</code>.
     * @param ref a <code>SEConcreteReference</code> to an <code>Instance</code> 
     *            in the state's heap, which is meant to be an instance of 
     *            class <code>String</code> representing <code>value</code>.
     */
    public void setStringLiteral(String value, ReferenceConcrete ref) {
    	this.stringLiterals.put(value, ref);
    }
    
    /**
     * Returns the state's <code>Heap</code>.
     * 
     * @return the state's <code>Heap</code>.
     */
    public Heap getHeap() {
        return this.heap;
    }
    
    /**
     * Returns the identifiers of all the heap objects compatible 
     * with a given type.
     * 
     * @param type a <code>String</code> representing an array type 
     *        (e.g., <code>"[C"</code>) or a reference type
     *        (e.g., <code>"Ljava/lang/String"</code>).  
     * @return a <code>List&lt;Integer&gt;</code>, where each value 
     *         represents the identifier in the heap of an object 
     *         which is compatible with <code>type</code>, or 
     *         <code>null</code> iff <code>type</code> does not 
     *         denote a reference type.
     */
    public List<Integer> getObjectsCompatible(String type) {
    	if (Type.isReference(type) || Type.isArray(type)) {
    		//finds all the heap objects compatibile with the reference
    		List<Integer> retVal = new Vector<Integer>();
    		for (Map.Entry<Integer, Objekt> e : this.heap.getObjects().entrySet()) {
    			Integer i = e.getKey();
    			Objekt o = e.getValue();
    			//adds o to the list of compatible objects if o's 
    			//creation epoch comes before that of the symbol...
    			if (o.creationEpoch() == Objekt.Epoch.EPOCH_UNKNOWN ||
    					o.creationEpoch() == Objekt.Epoch.EPOCH_BEFORE_START) {
    				//...and if o is type compatible with the reference
    				if (Type.isArray(type)) {
    					if (o.getType().equals(type)) {
    						retVal.add(i);
    					}
    				} else { //(Type.isReference(type))
    					ClassHierarchy hier = new ClassHierarchy(this.classFileInterface);
    					String className = Type.getClassFromReferenceType(type);
    					if (hier.isSubclass(o.getType(), className)) { //TODO should we invoke getTypesCompatible instead???
    						retVal.add(i);
    					}
    				}
    			}
    		}
    		return retVal;
    	} else {
    		return null;
    	}
    }

    /**
     * Returns all the types compatible, according to the class substitution 
     * set stored by the state, to a given type.
     *    
     * @param type a <code>String</code> representing a type.
     * @param origin a <code>String</code> representing the origin of a 
     *               symbolic value.
     * @return a <code>Set&lt;String&gt;</code> where each member represents 
     *         a reference type which is compatible, according to the class 
     *         substitution set, with <code>type</code>, or <code>null</code>
     *         iff <code>type</code> does not denote a reference or array type.
     */
    public Set<String> getTypesCompatible(String type, String origin) {
   		HashSet<String> retVal;
   		if (Type.isArray(type)) {
   			//first, the (really trivial) array case...
   			retVal = new HashSet<String>();
   			retVal.add(type); 
   		} else if (Type.isReference(type)) {  
   			//...then, the nonarray case: the substitution 
   			//set is used to determine the compatible classes...
   			retVal = new HashSet<String>();
   			String className = Type.getClassFromReferenceType(type);
   			Set<Environment.Entry> entries =  this.classFileInterface.getEnvironment().classSubstSet(className);
   			for (Entry e : entries) {
   				if (e.matchOrigin(origin) && this.initializedBeforeStart(e.substClass)) {
   					retVal.add(e.substClass);
   				}
   			}

   			//...but if the substitution set does not provide 
   			//any information, the static type of the symbolic reference 
   			//is used, if it does not denote an abstract class or an 
   			//interface
   			if (retVal.size() == 0) { 
   				ClassFile cf = this.getClassFileInterface().getClassFile(className);
   				if (!cf.isAbstract()) {
   					retVal.add(className);
   				}
   			}
    	} else {
    		retVal = null;
    	}
   		return retVal;
    }
        
    public StaticMethodArea getStaticStore() {
    	return this.staticStore;
    }
    
    /**
     * Returns the bytecode of the current method.
     */
    public byte[] getCode() {
        return this.getStack().currentFrame().getCodeArray();
    }
    
    /**
     * Returns the length of the bytecode of the 
     * current method.
     * 
     * @return an <code>int</code> representing the length 
     * in bytes of the bytecode of the current method.
     */
    public int getCodeLength() {
        return this.getCode().length;
    }

    /**
     * Sets the bytecode of the current method.
     * 
     * @param byteCode the new bytecode. 
     */
    public void setCode(byte[] byteCode) {
    	this.getStack().currentFrame().saveFrameCode(byteCode);
    }
    
    /**
     * Returns the instruction in the current method pointed by 
     * the state's current program counter.
     * 
     * @return a <code>byte</code> representing the 
     *         bytecode pointed by the state's current program
     *         counter.
     */
    public byte getInstruction() {
        return(this.getStack().currentFrame().getInstruction());
    }
    
    /**
     * Returns the instruction in the current method pointed by 
     * the state's current program counter plus a displacement.
     * 
     * @param displ a <code>int</code> representing a displacement
     * @return a <code>byte</code> representing the 
     *         bytecode pointed by the state's current program
     *         counter plus <code>displ</code>.
     * @throws CodeOutOfIndexException iff the frame's program
     *         counter plus <code>displ</code> does not point to 
     *         a bytecode.
     */
    public byte getInstruction(int displ) throws CodeOutOfIndexException {
        return(this.getStack().currentFrame().getInstruction(displ));
    }
    
    /**
     * Returns the state's program counter.
     * 
     * @return an <code>int</code> representing the state's 
     *         current program counter.
     */
    public int getPC() {
    	//the execution program counter is the program counter of the 
    	//current frame
        return this.getStack().currentFrame().getProgramCounter();
    }

    /**
     * Increments/decrements the program counter by an arbitrary number.
     * 
     * @param n the <code>int</code> value to be added to the current 
     *          program counter.
     * @throws CodeOutOfIndexException if the incremented program counter
     *         would not point to a valid bytecode in the current method 
     *         (the state's program counter is not changed).
     */
    public void incPC(int n) throws CodeOutOfIndexException {
    	this.setPC(this.getPC() + n);
    }
    
    /**
     * Sets the state's program counter.
     * 
     * @param newPC the new program counter value.
     * @throws CodeOutOfIndexException if <code>newPC</code> does not 
     *         point to a valid bytecode in the current method (the
     *         state's program counter is not changed).
     */
    public void setPC(int newPC) throws CodeOutOfIndexException {
    	this.getStack().currentFrame().saveFrameProgramCounter(newPC);
    }
    
    /**
     * Sets the state's return program counter.
     * 
     * @param offset the offset from the current program counter of 
     *               the return program counter.
     * @throws CodeOutOfIndexException  if the calculated return program 
     *         counter does not point to a valid bytecode in the current 
     *         method (the state's return program counter is not changed).
     */
    public void setReturnPC(int offset) throws CodeOutOfIndexException {
    	this.getStack().currentFrame().saveFrameReturn(offset);
    }
    
    /**
     * Copies the state's return program counter into the program counter.
     * 
     * @throws CodeOutOfIndexException if the return program 
     *         counter does not point to a valid bytecode in the current 
     *         method (the state's program counter is not changed).
     */
    public void useReturnPC() throws CodeOutOfIndexException {
    	this.getStack().currentFrame().useReturnProgramCounter();
    }

    /**
     * Increments the program counter by <code>1</code>.
     * 
     * @throws CodeOutOfIndexException if the incremented program counter
     *         would not point to a valid bytecode in the current method 
     *         (the state's program counter is not changed).
     */
    public void incPC() throws CodeOutOfIndexException {
    	this.incPC(1);
    }

    /**
     * Set the state's path condition.
     * 
     * @param pathCondition the new state path condition.
     */
    public void setPathCondition(PathCondition pathCondition) {
        this.pathCondition = pathCondition;
        this.augmentedPathCondition = false;
    }

    /**
     * Adds a clause to the state's path condition.
     * 
     * @param c the clause which must be added to the state's 
     *          path condition. The method has  
     *          no effect if <code>c == null</code>.
     */
    public void addClauseToPathCondition(Expression c) {
    	if (c == null) { 
    		return;
    	}
        this.pathCondition.concatenate(c);
        this.augmentedPathCondition = true;
    }

    /**
     * Adds a clause to the state's path condition (refinement of 
     * a symbolic reference).
     * 
     * @param r the <code>SESymbolicReference</code> which is refined. It 
     *          must be <code>r != null</code>, or the method has no effect. 
     * @param o the <code>ObjDef</code> with which <code>r</code> is refined, 
     *          iff <code>o</code> is a fresh (i.e., not reachable before) 
     *          symbolic object, or <code>null</code> iff <code>r</code> is 
     *          refined to <code>null</code> or to an already available 
     *          symbolic object.
     */
    public void addClauseToPathCondition(ReferenceSymbolic r, Objekt o) {
        this.pathCondition.concatenate(r, o);
        this.augmentedPathCondition = true;
    }

    /**
     * Returns the state's path condition.
     */
    public PathCondition getPathCondition() {
        return this.pathCondition;
    }
    
    /**
     * Returns the last clause added to the state's path 
     * condition. 
     * 
     * @return a <code>PathCondition.Clause</code> representing 
     * the clause added to the state by the last call to the 
     * <code>addClauseToPathCondition</code> method if
     * <code>augmentedPathCondition</code> is <code>true</code>, 
     * <code>null</code> otherwise. 
     */
    public PathCondition.Clause getLastPathConditionClause() {
    	if (this.augmentedPathCondition()) {
    		return this.pathCondition.getLastAddedClause();
    	}
    	return null;
    }
    
    /**
     * Reset the memorization of the last clause added to the 
     * state's path condition. 
     */
    public void resetLastPathConditionClause() {
    	this.augmentedPathCondition = false;
    }
    
    /**
     * Tests whether a clause has been added to the state's 
     * path condition. 
     * 
     * @return <code>true</code> if a call to the 
     * <code>addClauseToPathCondition</code> method has been 
     * performed since the state's creation or since the last 
     * call to <code>setPathCondition</code>, or 
     * since the last call to <code>resetLastPathConditionClause</code>, 
     * <code>false</code> otherwise.
     */
    public boolean augmentedPathCondition() {
    	return this.augmentedPathCondition;
    }
    
    /**
     * Sets a stuck state caused by a parameterless return. 
     */
    public void setStuck() {
    	this.stuck = true;
    	this.exc = null;
    	this.val = null;
    }
    
    /**
     * Sets a stuck state caused by a return.
     * 
     * @param val the return <code>Value</code>.
     */
    public void setStuckReturn(Value val) {
        this.setStuck();
        this.val = val;
    }
    
    /**
     * Returns the stuck state's return value.
     * 
     * @return the <code>Value</code> set by a 
     *         previous call to <code>setStuckReturn</code>, 
     *         or <code>null</code> if <code>!this.isStuck()</code>
     *         or it has not been set stuck with a call to 
     *         <code>setStuckReturn</code>. 
     */
    public Value getStuckReturn() {
    	return this.val;
    }
    
    /**
     * Sets a stuck state caused by an unhandled throw.
     * 
     * @param exc a <code>SEReference</code> to some instance 
     *            in this <code>State</code>'s heap. 
     */
    public void setStuckException(Reference exc) {
        this.setStuck();
        this.exc = exc;
    }
    
    /**
     * Returns the stuck state's thrown exception.
     * 
     * @return the <code>SEReference</code> set by a 
     *         previous call to <code>setStuckReturn</code>, 
     *         or <code>null</code> if <code>!this.isStuck()</code>
     *         or it has not been set stuck with a call to 
     *         <code>setStuckException</code>. 
     */
    public Reference getStuckException() {
    	return this.exc;
    }
    
    /**
     * Tests for stuck state.
     * 
     * @return <code>true</code> iff the state is stuck, i.e., iff 
	 *         it is a leaf state in the symbolic execution tree. A stuck state is
	 *         the outcome of a halting instruction, or of a return from the topmost
	 *         stack frame, or of an unhandled throw.
     */
    public boolean isStuck() {
        return this.stuck;
    }
    
    /**
     * Increases the state's depth.
     * 
     * @param identifierSuffix 
     *        a <code>String</code> representing the suffix to be 
     *        appended to the state's identifier.
     */
    public void increaseDepth(String identifierSuffix) {
        this.identifier += identifierSuffix;
        this.depth++;
    }
    
    /**
     * Gets the state's identifier.
     * 
     * @return a <code>String</code> representing the 
     *         state's identifier (<code>IDENTIFIER_DEFAULT</code> 
     *         in the case <code>this.setIdentifier</code> has 
     *         not been invoked yet).
     */
    public String getIdentifier() {
        return this.identifier;
    }
    
    /**
     * Gets the state's depth in the symbolic execution tree; the depth 
     * is the number of branches above the state.
     * 
     * @return the depth of the state as an <code>int</code> value
     *         (0 for the topmost state).
     */
    public int getDepth() {
    	return this.depth;
    }

    /**
     * Sets the state's sequence number to <code>0</code>.
     */
	public void resetSequenceNumber() {
		this.sequenceNumber = 0;
	}
	
	/**
	 * Increments the state's sequence number by <code>1</code>.
	 */
	public void incSequenceNumber() {
		this.sequenceNumber++;
	}

	/**
	 * Gets the state's sequence number.
	 * 
	 * @return a nonnegative <code>int</code> representing the 
	 *         state's sequence number.  
	 */
	public int getSequenceNumber() {
		return sequenceNumber;
	}
	
	/**
	 * Refines this state according to some path condition more 
	 * restrictive than its current one.
	 *
	 * @param refiningPathCondition 
	 *        a <code>PathCondition</code>; it must be 
	 *        <code>refiningPathCondition.refines(this.pathCondition)</code>.
	 * @throws CannotRefineException when 
	 *         <code>!refiningPathCondition.refines(this.pathCondition)</code>.
	 */
	public void refine(PathCondition refiningPathCondition) throws CannotRefineException {
		//checks the preconditions
		if (!refiningPathCondition.refines(this.pathCondition)) {
			throw new CannotRefineException();
		}
		
		//defines an iterator for the clauses in the refining path condition
		Iterator<PathCondition.Clause> iRefining = refiningPathCondition.getClauses().iterator();

		//skips the clauses in the state's path condition 
		//(assumed already refined by the execution)
		for (Iterator<?> iPC = this.pathCondition.getClauses().iterator(); iPC.hasNext(); iPC.next()) { 
			iRefining.next();
		}

		//analyzes every additional clause and increases the state's 
		//memory (heap and static store) with all the initial objects;
		//at the same time prepares refs, a map between symbolic references 
		//names and their respective values, and marks all the identifiers 
		//of symbolic values used in the refining instance as already used
		Map<String, ReferenceSymbolic> refs = new HashMap<String, ReferenceSymbolic>();
		while (iRefining.hasNext()) {
			PathCondition.Clause c = iRefining.next();
			if (c.hasRefinement()) {
				ReferenceSymbolic ref = c.getRefinement();
				refs.put(ref.getValue(), ref);
		        Objekt o = (c.hasObject() ? c.getObject().clone() : null);
				if (o != null) {
					this.getHeap().set(ref, o);
					for (int i : o.symIdsPrimitive()) {
						this.consumedSymIdsPrimitive.add(i);
					}
					for (int i : o.symIdsReference()) {
						this.consumedSymIdsReference.add(i);
					}
				}
			} else if (c.hasKlass()) {
				this.staticStore.set(c.getKlass(), c.getClassName());
			}
		}
		this.pathCondition = refiningPathCondition.clone();

		//completes by refining all the symbolic references to 
		//the respective values
		this.staticStore.refine(refs);
		this.heap.refine(refs);
		this.stack.refine(refs);
		if (this.exc != null) {
			this.exc.refine(refs);
		}
		if (this.val != null) {
			this.val.refine(refs);
		}
	}
	
	/**
	 * A factory method for creating symbolic values both for 
	 * references and primitive values.
	 * 
	 * @param descriptor the descriptor of the symbolic value's type.
	 * @return a <code>SESymbolicPrimitive</code> or a <code>SESymbolicReference</code>
	 *         according to the <code>descriptor</code>.
	 */
	public Value createSymbol(String descriptor, String origin) {
		Value retVal;
		
        if (Type.isArray(descriptor) || Type.isReference(descriptor)) {
            retVal = new ReferenceSymbolic(this.nextIdRefSym, descriptor, origin);
            this.consumedSymIdsReference.add(this.nextIdRefSym);
            do {
            	this.nextIdRefSym++;
            } while (consumedSymIdsReference.contains(this.nextIdRefSym));
        } else {
            retVal = new PrimitiveSymbolic(this.nextIdPrimSym, descriptor, origin);
            this.consumedSymIdsPrimitive.add(this.nextIdPrimSym);
            do {
            	this.nextIdPrimSym++;
            } while (consumedSymIdsPrimitive.contains(this.nextIdPrimSym));
        }
        return retVal;
	}


	/**
	 * Unwinds the stack of a <code>State</code> until a handler for a suitable 
	 * throwable <code>Objekt</code> is found.
	 * 
	 * @param exceptionToThrow a <code>Reference</code> to a throwable 
	 *        <code>Objekt</code> in <code>state</code>'s <code>Heap</code>.
	 */
	public void doThrow(Reference exceptionToThrow) {
		/* TODO clarify which should be the state of the operand stack when the 
		 * exception is thrown; perhaps it should be emptied, although it should 
		 * not be strictly necessary as our operand stack has variable size gets 
		 * the class name of the exception. 
		 */
		Objekt myException = this.heap.getObject(exceptionToThrow);
		
		//fills a vector with all the superclass names of the exception
		Vector<String> excTypes = new Vector<String>();
		ClassHierarchy hier = new ClassHierarchy(this.classFileInterface);
		for (ClassFile f : hier.superclasses(myException.getType())) {
			excTypes.add(f.getClassName());
		}
	
		//unwinds the stack
		while (true) {
			Frame frame = this.stack.currentFrame();
			String classToOpen = frame.getName().getClassName();
			ExceptionTable myExTable = this.classFileInterface.getClassFile(classToOpen).getExceptionTable(frame.getName());
			for (int i = 0; i < myExTable.getLength(); i++) {
				ExceptionTableEntry tmpEntry = myExTable.getEntry(i);
				if (excTypes.contains(tmpEntry.getType()) && 
						(this.getPC() >= tmpEntry.getStartPC()) && 
						(this.getPC() <= tmpEntry.getEndPC())) {
					try {
						this.setPC(tmpEntry.getPCHandle());
						frame.push(exceptionToThrow);
					} catch (CodeOutOfIndexException e) {
						//TODO check whether this recursive call is ok!
						Reference myExceRef = Util.loadInstance(this, "java/lang/VerifyError");
						frame.push(myExceRef);
						this.doThrow(myExceRef);
					}
					return;
				}
			}
			this.stack.pop();
			if (this.stack.empty()) {
				this.setStuckException(exceptionToThrow);
				return;
			}
		}
	}

	@Override
    public String toString() {
        String tmp = "[ID:\"" + this.identifier + "[" + this.sequenceNumber + "]\", ";
        if (this.isStuck()) {
        	tmp += "Stuck, ";
        	if (this.exc != null) 
        		tmp += "Raised:" + this.exc.toString() + ", ";
        	else if (this.val != null)
        		tmp += "Return:" + this.val.toString() + ", ";
        } else {
        	tmp += "CurrentMethod:" + this.stack.currentFrame().getName() + ", ";
	        tmp += "ProgramCounter:" + this.getPC() + ", ";
	        tmp += "Stack:" + this.stack.toString() + ", ";
        }
       	tmp += "PathCondition:'" + this.pathCondition.toString() + "', ";
       	tmp += "StaticStore:" + this.staticStore.toString() + ", ";
       	tmp += "Heap:" + this.heap.toString() + "]";
    	return(tmp);        
    }
    
	@Override
    public State clone() {
        State o = null;
        try {
            o = (State) super.clone();
        } catch (CloneNotSupportedException e) {
        	return null;
        }
        
        //stringLiterals
        o.stringLiterals = new HashMap<String, ReferenceConcrete>();
        for (Map.Entry<String, ReferenceConcrete> e : stringLiterals.entrySet()) {
        	o.stringLiterals.put(e.getKey(), (ReferenceConcrete) e.getValue().clone());
        }
        
        //stack
        o.stack = o.stack.clone();
        
        //heap
        o.heap = o.heap.clone();
        
        //staticStore
        o.staticStore = o.staticStore.clone();
        
        //pathCondition
        o.pathCondition = o.pathCondition.clone();

        //exc
        if (o.exc != null) {
        	o.exc = o.exc.clone();
        }
        
        //val
        if (o.val != null) {
        	o.val = o.val.clone();
        }
        
        //consumedSymIdsReference
        o.consumedSymIdsReference = new HashSet<Integer>();
        for (int i : consumedSymIdsReference) {
        	o.consumedSymIdsReference.add(i);
        }
        
        //consumedSymIdsPrimitive
        o.consumedSymIdsPrimitive = new HashSet<Integer>();
        for (int i : consumedSymIdsPrimitive) {
        	o.consumedSymIdsPrimitive.add(i);
        }
        
        //notInitializedClasses
        o.notInitializedClasses = new HashSet<String>();
        for (String s : notInitializedClasses) {
        	o.notInitializedClasses.add(s);
        }
        
        return o;
    }
}