package muse.mem;

import java.util.Vector;

/**
 *Class that contain costants
 */
public final class Type {
    public static final char UNKNOWN        = '?';  //the top, any datum fits (??)
    public static final char VOID           = 'V';  //the bottom, no datum fits (??)
    public static final char BYTE     		= 'B';
    public static final char SHORT    		= 'S';
    public static final char INT      		= 'I';
    public static final char LONG     		= 'J';
    public static final char FLOAT    		= 'F';
    public static final char DOUBLE   		= 'D';
    public static final char CHAR     		= 'C';
    public static final char BOOLEAN  		= 'Z';
    public static final char REFERENCE		= 'L';
    public static final char NULLREF  		= '0';
    public static final char ARRAYOF        = '[';
    public static final char TYPEEND        = ';';

	private static final byte    defByte   = (byte) 0;
	private static final short   defShort  = (short) 0;
	private static final int     defInt    = 0;
	private static final long    defLong   = 0L;
	private static final float   defFloat  = 0.0f;
	private static final double  defDouble = 0.0;
	private static final char    defChar   = '\u0000';
	private static final boolean defBool   = false;
	    
    /**
     * Creates a <code>Value</code> representing the default 
     * as from JVM specification (see JVM specification, 
     * 2nd edition, paragraph 2.5.1).
     *  
     * @param type the type of the <code>Value</code> to be 
     *             created.
     * @return a new instance of <code>Value</code> initialized 
     * with the default for its type, as from JVM specification, 
     * or <code>null</code> if <code>type</code> is 
     * <code>TYPEEND</code>, or does not represent a type.
     */
    public static Value createDefault(char type) {
    	switch (type) {
		case Type.BYTE:
			return new Simplex(defByte);
		case Type.SHORT:
			return new Simplex(defShort);
		case Type.INT:
			return new Simplex(defInt);
		case Type.LONG:
			return new Simplex(defLong);
		case Type.FLOAT:
			return new Simplex(defFloat);
		case Type.DOUBLE:
			return new Simplex(defDouble);
		case Type.CHAR:
			return new Simplex(defChar);
		case Type.BOOLEAN:
			return new Simplex(defBool);
		case Type.REFERENCE:
		case Type.NULLREF:
		case Type.ARRAYOF:
			return Null.getInstance();
		default:
			return null;
    	}
    }
    
    public static boolean isVoid(String s) {
    	if (s == null || s.length() == 0) {
    		return false;
    	}
    	char c = s.charAt(0);
    	return (c == Type.VOID);
    }
    
    public static boolean isPrimitive(String s) {
    	if (s == null || s.length() != 1) {
    		return false;
    	} else {
	    	char c = s.charAt(0);
	    	return isPrimitive(c);
    	}
    }
    
    public static boolean isPrimitive(char c) {
    	return (c == Type.BYTE ||
                c == Type.SHORT ||
                c == Type.INT ||
                c == Type.LONG ||
                c == Type.FLOAT ||
                c == Type.DOUBLE ||
                c == Type.CHAR ||
                c == Type.BOOLEAN);
    }
    
    public static boolean isArray(char c) {
    	return (c == Type.ARRAYOF);
    }
    
    public static boolean isArray(String s) {
    	if (s == null || s.length() == 0) {
    		return false;
    	} else {
	    	char c = s.charAt(0);
	    	return isArray(c);
    	}
    }
    
    public static boolean isReference(char c) {
    	return (c == Type.REFERENCE);
    }
    
    public static boolean isReference(String s) {
    	if (s == null || s.length() == 0) {
    		return false;
    	}
    	char c = s.charAt(0);
    	char cc = s.charAt(s.length() - 1);
    	return (isReference(c) && cc == Type.TYPEEND);
    }
    
    public static boolean isCat_1(char c) {
    	return (c != Type.LONG && c != Type.DOUBLE);
    }

	/**
	 * Given a <code>String</code> descriptor of a method returns an array of 
	 * <code>String</code>s containing the descriptors of its parameters.
	 */
	public static String[] splitParametersDescriptors(String methodDescriptor){
		Vector<String> myVector = new Vector<String>();
		for (int j = 1; j < methodDescriptor.lastIndexOf(')'); j++) {
			if (methodDescriptor.charAt(j) == REFERENCE) {
				int z = j;
				while (methodDescriptor.charAt(j) != TYPEEND)
					j++;
				myVector.add(methodDescriptor.substring(z, j + 1));
			}
			else if (methodDescriptor.charAt(j) == ARRAYOF) {
				int z = j;
				while (methodDescriptor.charAt(j) == ARRAYOF)
					j++;
				if (methodDescriptor.charAt(j) == REFERENCE)
					while (methodDescriptor.charAt(j) != TYPEEND)
						j++;
				myVector.add(methodDescriptor.substring(z, j + 1));
			}
			else
				myVector.add("" + methodDescriptor.charAt(j));
		}
		String[] retString = new String[myVector.size()];
		for (int b = 0; b < myVector.size(); b++)
			retString[b] = myVector.get(b);
		return (retString);
	}

	/**
	 * Given a <code>String</code> descriptor of a method returns a 
	 * <code>String</code> containing the descriptor of its return value, 
	 * or <code>null</code> if the input is not the descriptor of a method.
	 */
	public static String splitReturnValueDescriptor(String methodDescriptor) {
		int index = methodDescriptor.lastIndexOf(')') + 1;
		if (index == 0) {
			return null;
		} else {
			return methodDescriptor.substring(index);
		}
	}
	
	public static String getClassFromReferenceType(String refType) {
		if (isReference(refType)) {
			return refType.substring(1, refType.length() - 1);
		} else {
			return null;
		}
	}
	
	public static boolean isUninitializedReference(Value v) {
		return 
		(Type.isReference(v.getType()) && 
        ((Reference) v).isSymbolic() &&
        !((ReferenceSymbolic) v).isInitialized());
	}
}
