package muse.tree;

import java.util.Comparator;
import java.util.SortedSet;
import java.util.TreeSet;


/**
 * Factory for empty ordered sets of <code>DecisionAlternative</code>s
 * used by <code>Algorithm</code>s to determine the insertion order of 
 * states into the <code>StateTree</code>.
 * 
 * @author pietro
 */
public class DecisionAlternativeSetFactory {
    private Comparator<DecisionAlternativeBoolean> cb;
    private Comparator<DecisionAlternativeComparison> cc;
    private Comparator<DecisionAlternativeSwitch> ci;
	private Comparator<DecisionAlternativeReference> cr;
	private Comparator<DecisionAlternativeAnew> cn; 
	private Comparator<DecisionAlternativeAstore> cs;
	private Comparator<DecisionAlternativeAload> cl;
	
    /**
     * Default comparator for <code>DecisionAlternativeBoolean</code>s.
     * 
     * @return A <code>Comparator&lt;DecisionAlternativeBoolean&gt; yielding
     *         <code>TT</code> &lt; <code>FF</code>.
     */
	private static Comparator<DecisionAlternativeBoolean> defaultComparatorDecisionAlternativeBoolean() {
		return null;
	}

    /**
     * Default comparator for <code>DecisionAlternativeComparison</code>s.
     * 
     * @return A <code>Comparator&lt;DecisionAlternativeComparison&gt; yielding
     *         <code>GT</code> &lt; <code>EQ</code> &lt; <code>LT</code>.
     */
	private static Comparator<DecisionAlternativeComparison> defaultComparatorDecisionAlternativeComparison() {
		return null;
	}

    /**
     * Default comparator for <code>DecisionAlternativeSwitch</code>s.
     * 
     * @return A <code>Comparator&lt;DecisionAlternativeSwitch&gt; where
     *         <code>default</code> &lt; <code>nondefault</code> and 
     *         <code>nondefault1<code> &lt; <code>nondefault2<code>
     *         iff <code>nondefault1.value<code> &gt; <code>nondefault2.value<code>.
     */
	private static Comparator<DecisionAlternativeSwitch> defaultComparatorDecisionAlternativeSwitch() {
	    return new Comparator<DecisionAlternativeSwitch>() {
	        public int compare(DecisionAlternativeSwitch o1, DecisionAlternativeSwitch o2) {
	        	DecisionAlternativeSwitch def = DecisionAlternativeSwitch.alternativeDefault();
	        	boolean o1def = (o1 == def);
	        	boolean o2def = (o2 == def);

	        	//o1 before o2
	        	if ((o1def && !o2def) || 
	        		(!o1def && !o2def && o1.value() > o2.value())) {
	        		return -1;
	        	}
	        	
	        	//o1 after o2
	        	if ((!o1def && o2def) || 
	            		(!o1def && !o2def && o1.value() < o2.value())) {
	            	return 1;
	        	}
	        	
	        	//o1 equals o2
	       		return 0;
	        }
	    };
	}

    /**
     * Default comparator for <code>DecisionAlternativeReference</code>s.
     * 
     * @return A <code>Comparator&lt;DecisionAlternativeReference&gt; where
     *         null &lt; object &lt; class.
     */
	private static Comparator<DecisionAlternativeReference> defaultComparatorDecisionAlternativeReference() { 	
		//TODO this comparator may be wrong! Check if two different references/classes elide when inserted. 
		return new Comparator<DecisionAlternativeReference>() {
			public int compare(DecisionAlternativeReference o1, DecisionAlternativeReference o2) {
				if (o1 instanceof DecisionAlternativeReferenceNull) {
					if (o2 instanceof DecisionAlternativeReferenceNull) {
						return 0;
					} else {
						return -1;
					}
				} else if (o1 instanceof DecisionAlternativeReferenceObject) {
					if (o2 instanceof DecisionAlternativeReferenceNull) {
						return 1;
					} else if (o2 instanceof DecisionAlternativeReferenceObject) {
						return 0;
					} else {
						return -1;
					}
				} else {
					if (o2 instanceof DecisionAlternativeReferenceClass) {
						return 0;
					} else {
						return 1;
					}
				}
			}
		};
	}

    /**
     * Default comparator for <code>DecisionAlternativeAnew</code>s.
     * 
     * @return A <code>Comparator&lt;DecisionAlternativeAnew&gt; yielding
     *         outbound &lt; inbound.
     */
	private static Comparator<DecisionAlternativeAnew> defaultComparatorDecisionAlternativeAnew() {
		//TODO this comparator may be nondetermininstic. Check how to make it deterministic. 
		return new Comparator<DecisionAlternativeAnew>() {
			public int compare(DecisionAlternativeAnew o1, DecisionAlternativeAnew o2) {
				if (o1 instanceof DecisionAlternativeAnewOutbound && o2 instanceof DecisionAlternativeAnewInbound) {
					return -1;
				} else if (o1 instanceof DecisionAlternativeAnewInbound && o2 instanceof DecisionAlternativeAnewOutbound) {
					return 1;
				} else if (o1.equals(o2)) {
					return 0;
				} else {
					return (o1.hashCode() < o2.hashCode() ? -1 : 1);
				}
			}
		};
	}
	
    /**
     * Default comparator for <code>DecisionAlternativeAstore</code>s.
     * 
     * @return A <code>Comparator&lt;DecisionAlternativeAstore&gt; yielding
     *         <code>IN</code> &lt; <code>OUT</code>.
     */
	private static Comparator<DecisionAlternativeAstore> defaultComparatorDecisionAlternativeAstore() {
		return null;
	}
	

    /**
     * Default comparator for <code>DecisionAlternativeAload</code>s.
     * 
     * @return A <code>Comparator&lt;DecisionAlternativeAload&gt; yielding
     *         value &lt; null &lt; instance &lt; class &lt; outbound.
     */
	private static Comparator<DecisionAlternativeAload> defaultComparatorDecisionAlternativeAload() {
		//TODO this comparator may be nondetermininstic. Check how to make it deterministic. 
		return new Comparator<DecisionAlternativeAload>() {
			public int compare(DecisionAlternativeAload o1, DecisionAlternativeAload o2) {
				int o1ord, o2ord;
				if (o1 instanceof DecisionAlternativeAloadValue) {
					o1ord = 0;
				} else if (o1 instanceof DecisionAlternativeAloadReferenceNull) {
					o1ord = 1;
				} else if (o1 instanceof DecisionAlternativeAloadReferenceInstance) {
					o1ord = 2;
				} else if (o1 instanceof DecisionAlternativeAloadReferenceClass) {
					o1ord = 3;
				} else { //o1 instanceof DecisionAlternativeOutOfBound
					o1ord = 4;
				}
				if (o2 instanceof DecisionAlternativeAloadValue) {
					o2ord = 0;
				} else if (o2 instanceof DecisionAlternativeAloadReferenceNull) {
					o2ord = 1;
				} else if (o2 instanceof DecisionAlternativeAloadReferenceInstance) {
					o2ord = 2;
				} else if (o2 instanceof DecisionAlternativeAloadReferenceClass) {
					o2ord = 3;
				} else { //o2 instanceof DecisionAlternativeOutOfBound
					o2ord = 4;
				}
				if (o1ord == o2ord) {
					if (o1.equals(o2)) {
						return 0;
					} else {
						return (o1.hashCode() < o2.hashCode() ? -1 : 1);
					}			
				} else {
					return (o1ord < o2ord ? -1 : 1);
				}
			}
		};
	}
	
	/**
	 * Default constructor.
	 */
    public DecisionAlternativeSetFactory() {
    	this.cb = DecisionAlternativeSetFactory.defaultComparatorDecisionAlternativeBoolean();
    	this.cc = DecisionAlternativeSetFactory.defaultComparatorDecisionAlternativeComparison();
    	this.ci = DecisionAlternativeSetFactory.defaultComparatorDecisionAlternativeSwitch();
    	this.cr = DecisionAlternativeSetFactory.defaultComparatorDecisionAlternativeReference();
    	this.cn = DecisionAlternativeSetFactory.defaultComparatorDecisionAlternativeAnew();
    	this.cs = DecisionAlternativeSetFactory.defaultComparatorDecisionAlternativeAstore();
    	this.cl = DecisionAlternativeSetFactory.defaultComparatorDecisionAlternativeAload();
    }

    /**
     * Constructor allowing to override one or more of the comparators returned 
     * by the <code>DecisionAlternativeComparators</code> getters. 
     * 
     * @param cb 
     *        a <code>Comparator&lt;DecisionAlternativeBoolean&gt;</code> to 
     *        be returned by calls to <code>createEmptyDecisionAlternativeBoolean()</code>
     *        or <code>null</code> iff the default <code>Comparator&lt;DecisionAlternativeBoolean&gt;</code>
     *        must be returned.
     * @param cc
     *        a <code>Comparator&lt;DecisionAlternativeComparison&gt;</code> to 
     *        be returned by calls to <code>createEmptyDecisionAlternativeComparison()</code>
     *        or <code>null</code> iff the default <code>Comparator&lt;DecisionAlternativeComparison&gt;</code>
     *        must be returned.
     * @param ci
     *        a <code>Comparator&lt;DecisionAlternativeSwitch&gt;</code> to 
     *        be returned by calls to <code>createEmptyDecisionAlternativeSwitch()</code>
     *        or <code>null</code> iff the default <code>Comparator&lt;DecisionAlternativeSwitch&gt;</code>
     *        must be returned.
     * @param cr
     *        a <code>Comparator&lt;DecisionAlternativeReference&gt;</code> to 
     *        be returned by calls to <code>createEmptyDecisionAlternativeReference()</code>
     *        or <code>null</code> iff the default <code>Comparator&lt;DecisionAlternativeReference&gt;</code>
     *        must be returned.
     * @param cn
     *        a <code>Comparator&lt;DecisionAlternativeAnew&gt;</code> to 
     *        be returned by calls to <code>createEmptyDecisionAlternativeAnew()</code>
     *        or <code>null</code> iff the default <code>Comparator&lt;DecisionAlternativeAnew&gt;</code>
     *        must be returned.
     * @param cs
     *        a <code>Comparator&lt;DecisionAlternativeAstore&gt;</code> to 
     *        be returned by calls to <code>createEmptyDecisionAlternativeAstore()</code>
     *        or <code>null</code> iff the default <code>Comparator&lt;DecisionAlternativeAstore&gt;</code>
     *        must be returned.
     * @param cl
     *        a <code>Comparator&lt;DecisionAlternativeAload&gt;</code> to 
     *        be returned by calls to <code>createEmptyDecisionAlternativeAload()</code>
     *        or <code>null</code> iff the default <code>Comparator&lt;DecisionAlternativeAload&gt;</code>
     *        must be returned.
     */
    public DecisionAlternativeSetFactory(
    		final Comparator<DecisionAlternativeBoolean> cb, 
			final Comparator<DecisionAlternativeComparison> cc, 
			final Comparator<DecisionAlternativeSwitch> ci,
    		final Comparator<DecisionAlternativeReference> cr,
    		final Comparator<DecisionAlternativeAnew> cn,
    		final Comparator<DecisionAlternativeAstore> cs,
    		final Comparator<DecisionAlternativeAload> cl) {
    	this(); //sets defaults
    	
    	/**
    	 * Comparator inverting the order provided by another comparator
    	 * provided by the constructor.
    	 */
    	class InverseOrderComparator<R> implements Comparator<R> {
    		private Comparator<R> cmp;
    		public InverseOrderComparator(Comparator<R> cmp) { this.cmp = cmp; }
			public int compare(R o1, R o2) {
				return this.cmp.compare(o2, o1);
			}
    	}
    	
		if (cb != null) { this.cb = new InverseOrderComparator<DecisionAlternativeBoolean>(cb); } 
		if (cc != null) { this.cc = new InverseOrderComparator<DecisionAlternativeComparison>(cc); }
		if (ci != null) { this.ci = new InverseOrderComparator<DecisionAlternativeSwitch>(ci); }
		if (cr != null) { this.cr = new InverseOrderComparator<DecisionAlternativeReference>(cr); }
		if (cn != null) { this.cn = new InverseOrderComparator<DecisionAlternativeAnew>(cn); }
		if (cs != null) { this.cs = new InverseOrderComparator<DecisionAlternativeAstore>(cs); }
		if (cl != null) { this.cl = new InverseOrderComparator<DecisionAlternativeAload>(cl); }
	}
    
    /**
     * Factory method for <code>SortedSet&lt;DecisionAlternativeBoolean&gt;</code> 
     * according to the order imposed by the comparator set by the constructor.
     *  
     * @return an empty <code>SortedSet&lt;DecisionAlternativeBoolean&gt;</code>
     */
    public SortedSet<DecisionAlternativeBoolean> createEmptyDecisionAlternativeBoolean() { 
    	return new TreeSet<DecisionAlternativeBoolean>(this.cb); 
    }
    
    /**
     * Factory method for <code>SortedSet&lt;DecisionAlternativeComparison&gt;</code> 
     * according to the order imposed by the comparator set by the constructor.
     *  
     * @return an empty <code>SortedSet&lt;DecisionAlternativeComparison&gt;</code>
     */
    public SortedSet<DecisionAlternativeComparison> createEmptyDecisionAlternativeComparison() { 
    	return new TreeSet<DecisionAlternativeComparison>(this.cc); 
    }

    /**
     * Factory method for <code>SortedSet&lt;DecisionAlternativeSwitch&gt;</code> 
     * according to the order imposed by the comparator set by the constructor.
     *  
     * @return an empty <code>SortedSet&lt;DecisionAlternativeSwitch&gt;</code>
     */
    public SortedSet<DecisionAlternativeSwitch> createEmptyDecisionAlternativeSwitch() { 
    	return new TreeSet<DecisionAlternativeSwitch>(this.ci); 
    }    

    /**
     * Factory method for <code>SortedSet&lt;DecisionAlternativeReference&gt;</code> 
     * according to the order imposed by the comparator set by the constructor.
     *  
     * @return an empty <code>SortedSet&lt;DecisionAlternativeReference&gt;</code>
     */
    public SortedSet<DecisionAlternativeReference> createEmptyDecisionAlternativeReference() { 
    	return new TreeSet<DecisionAlternativeReference>(this.cr); 
    }
    
    /**
     * Factory method for <code>SortedSet&lt;DecisionAlternativeAnew&gt;</code> 
     * according to the order imposed by the comparator set by the constructor.
     *  
     * @return an empty <code>SortedSet&lt;DecisionAlternativeAnew&gt;</code>
     */
    public SortedSet<DecisionAlternativeAnew> createEmptyDecisionAlternativeAnew() { 
    	return new TreeSet<DecisionAlternativeAnew>(this.cn); 
    }    
    
    /**
     * Factory method for <code>SortedSet&lt;DecisionAlternativeAstore&gt;</code> 
     * according to the order imposed by the comparator set by the constructor.
     *  
     * @return an empty <code>SortedSet&lt;DecisionAlternativeAstore&gt;</code>
     */
    public SortedSet<DecisionAlternativeAstore> createEmptyDecisionAlternativeAstore() { 
    	return new TreeSet<DecisionAlternativeAstore>(this.cs); 
    }    
    
    /**
     * Factory method for <code>SortedSet&lt;DecisionAlternativeAload&gt;</code> 
     * according to the order imposed by the comparator set by the constructor.
     *  
     * @return an empty <code>SortedSet&lt;DecisionAlternativeAload&gt;</code>
     */
    public SortedSet<DecisionAlternativeAload> createEmptyDecisionAlternativeAload() { 
    	return new TreeSet<DecisionAlternativeAload>(this.cl); 
    }    
}
