package muse.tree;

import java.util.LinkedList;

import muse.mem.*;


/**
 * Class storing the <code>State</code>s in the symbolic execution
 * tree which have been discovered but not yet analyzed.
 * 
 * @author Pietro Braione
 * @author unknown
 */

public class StateTree {
	/**
	 * A Memento for tree branches.
	 * 
	 * @author pietro
	 */
	public static class BranchPoint { }
	
	/** Buffer of the inserted <code>State</code>s. */
	private LinkedList<State> stateBuffer;

	/** Buffer of the inserted branches. */
	private LinkedList<BranchNode> branchList;

	/** 
	 * Flag indicating whether the tree level has been increased 
	 * since the last inspection by invocation of <code>createdLevel()</code>.
	 */
	private boolean increased;
	
	/** 
	 * Flag indicating whether the next state stored in the initializer 
	 * is the first one (so its identifier is preserved).
	 */
	private boolean firstState;
    
	/** 
	 * Private class of the item of <code>backtrackList</code>.
	 * 
	 * @author Pietro Braione
	 */ 
	private class BranchNode {
        /** A <code>BranchPoint</code>. */
		BranchPoint branch;

        /** The total number of states in the branch identified by <code>branch</code>. */ 
        int total;

        /** The number of states of <code>branch</code> already emitted. */
        int emitted;
        
        /** 
         * Constructor.
         */
        BranchNode() {
            this.total = 0;
            this.emitted = 0;
            this.branch = new BranchPoint();
        }
    }
	
	/**
	 * Constructor.
	 */
    public StateTree() {
        this.stateBuffer = new LinkedList<State>();
        this.branchList = new LinkedList<BranchNode>();
        this.addBranch();
        this.increased = false;
    	this.firstState = true;
    }
	
    /**
     * Adds a new state to the store.
     * 
     * @param s the <code>State</code> to be added.
     */
    public void addState(State s) {
        this.stateBuffer.addFirst(s);
        this.branchList.getFirst().total += 1;
    }

    /**
     * Checks whether some <code>State</code> can be emitted.
     * 
     * @return true iff the store has one or more states 
     *              to emit.
     */
    public boolean hasStates() {
        return !this.stateBuffer.isEmpty();
    }
    
    /**
     * Removes the next state from the store and emits it.
     * 
     * @return the <code>State</code> removed from the store.
     * @throws NoSuchElementException if <code>this.hasStates() == false</code>.
     */
    public State nextState() {
        State s = this.stateBuffer.getFirst();
        BranchNode b = this.branchList.getFirst();
        b.emitted += 1;
        if (!firstState) {
        	s.increaseDepth("." + b.emitted);
        }
        firstState = false;
        s.resetSequenceNumber();
        if (b.emitted == b.total) {
            this.branchList.removeFirst();
        }
        return stateBuffer.removeFirst();
    }
    
    /**
     * Increases by one the level of the tree.
     */
    public void addBranch() {
        this.branchList.addFirst(new BranchNode());
        this.increased = true;
    }
    
    /**
     * Checks whether a new level of the tree has been created 
     * since the last invocation of this method.
     * 
     * @return <code>true</code> if <code>increaseLevel()</code>
     *         has been invoked since the previous invocation 
     *         of <code>createdLevel()</code>, <code>false</code> 
     *         otherwise.
     */
    public boolean createdBranch() {
        boolean retval = this.increased;
        
        this.increased = false;
        return retval;
    }
    
    /**
     * Returns the next backtrack point.
     * 
     * @return the <code>BacktrackPoint</code> associated to the next initial 
     *         state stored in the initializer, as it would be returned 
     *         by a call to <code>nextInitialState</code>, 
     *         or <code>null</code> in the case such state exists.  
     */
    public BranchPoint nextBranch() {
        if (this.branchList.isEmpty()) {
            return null;
        } else {
            return this.branchList.getFirst().branch;
        }
    }    
 }
