package muse.bc;

import static org.junit.Assert.*;
import muse.bc.ClassFile;
import muse.bc.ClassFileFactory;
import muse.bc.ClassFileFactoryJavassist;
import muse.bc.ExceptionTable;
import muse.bc.ExceptionTableEntry;
import muse.bc.InvalidIndexException;
import muse.bc.MethodCodeNotFoundException;
import muse.bc.MethodNotFoundException;
import muse.bc.Signature;
import muse.jvm.Environment;
import muse.mem.Simplex;

import org.junit.*;

import java.util.HashSet;


public class ClassFileFactoryTest {
	ClassFileFactory f;

	@Before
	public void setUp() {
		//environment
		Environment env = new Environment();
		env.addClassPath("tst/muse/bc/testdata/rt.jar");
		env.addClassPath("tst/muse/bc/testdata/");

		//ASM
		//this.f = new ClassFileFactoryASM();

		//Jaba
		//this.f = new ClassFileFactoryJaba("output/");

		//Javassist
		this.f = new ClassFileFactoryJavassist();
		
		f.setEnv(env);
	}

	@Test
	public void testNewClassFile1() {
		ClassFile c = f.newClassFile("java/util/LinkedList$ListItr");
		assertNotNull(c);
	}

	@Test
	public void testGetClassName1() {
		String className = "java/lang/Object";
		ClassFile c = f.newClassFile(className);		
		assertEquals(className, c.getClassName());
	}

	@Test
	public void testGetClassName2() {
		String className = "tsafe/main/SimpleCalculator";
		ClassFile c = f.newClassFile(className);
		assertEquals(className, c.getClassName());
	}

	@Test
	public void testGetClassSignature1() {
		String className = "tsafe/main/SimpleCalculator";
		ClassFile c = f.newClassFile(className);
		int i = 1; //entry 1 in constant pool should be tsafe/main/SimpleCalculator
		assertEquals(className, c.getClassSignature(i));
	}		

	@Test(expected=InvalidIndexException.class)
	public void testGetClassSignature2() {
		String className = "tsafe/engine/TsafeEngine";
		ClassFile c = f.newClassFile(className);
		c.getClassSignature(10);
	}		

	@Test
	public void testGetSuperClassName1() {
		String className = "tsafe/main/SimpleCalculator";
		ClassFile c = f.newClassFile(className);
		assertEquals("tsafe/engine/EngineCalculator", c.getSuperClassName());
	}

	@Test
	public void testGetSuperClassName2() {
		String className = "muse/jvm/Engine";
		ClassFile c = f.newClassFile(className);
		assertEquals("java/lang/Object", c.getSuperClassName());
	}

	@Test
	public void testGetSuperClassName3() {
		String className = "java/lang/Object";
		ClassFile c = f.newClassFile(className);
		assertNull(c.getSuperClassName());
	}

	@Test
	public void testGetSuperInterfaceNames1() {
		String className = "java/util/LinkedList";
		ClassFile c = f.newClassFile(className);

		HashSet<String> ifnsExp = new HashSet<String>();
		ifnsExp.add("java/util/List");
		ifnsExp.add("java/util/Queue");
		ifnsExp.add("java/lang/Cloneable");
		ifnsExp.add("java/io/Serializable");

		HashSet<String> ifnsAct = new HashSet<String>(c.getSuperInterfaceNames());

		assertEquals(ifnsExp, ifnsAct);
	}

	@Test
	public void testGetSuperInterfaceNames2() {
		String className = "java/io/Serializable";
		ClassFile c = f.newClassFile(className);

		HashSet<String> ifnsExp = new HashSet<String>();
		HashSet<String> ifnsAct = new HashSet<String>(c.getSuperInterfaceNames());
		assertEquals(ifnsExp, ifnsAct);
	}

	@Test
	public void testGetSuperInterfaceNames3() {
		String className = "java/lang/Object";
		ClassFile c = f.newClassFile(className);

		HashSet<String> ifnsExp = new HashSet<String>();
		HashSet<String> ifnsAct = new HashSet<String>(c.getSuperInterfaceNames());
		assertEquals(ifnsExp, ifnsAct);
	}

	/**
	 * Method not declared in the class, but declared in some superclass
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testGetExceptionTable1() {
		Signature sig = new Signature("muse/jvm/Engine", "()Ljava/lang/String;", "toString");
		ClassFile c = f.newClassFile("muse/jvm/Engine");
		c.getExceptionTable(sig);
	}

	/**
	 * Method declared, but without code
	 */
	@Test(expected=MethodCodeNotFoundException.class)
	public void testGetExceptionTable2() {
		Signature sig = new Signature("java/lang/Runnable", "()V", "run");
		ClassFile c = f.newClassFile("java/lang/Runnable");
		c.getExceptionTable(sig);
	}

	/**
	 * Method with bytecode.
	 */
	@Test
	public void testGetExceptionTable3() {
		Signature sig = new Signature("muse/jvm/Engine", "()Lmuse/tree/StateTree$BranchPoint;", "step");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		//we use strings because ExceptionTableEntry does not override equals()
		
		HashSet<String> etExp = new HashSet<String>();
		etExp.add("15,240,241,muse/dec/DecisionException");
		
		ExceptionTable et = c.getExceptionTable(sig);
		HashSet<String> etAct = new HashSet<String>();
		for (int k = 0; k < et.getLength(); k++) {
			ExceptionTableEntry e = et.getEntry(k);
			etAct.add(e.getStartPC() + "," + e.getEndPC() + "," + e.getPCHandle() + "," + e.getType());
		}
		assertEquals(etExp, etAct);
	}

	/**
	 * Native method.
	 */
	@Test(expected=MethodCodeNotFoundException.class)
	public void testGetExceptionTable4() {
		Signature sig = new Signature("java/lang/Object", "()V", "notifyAll");
		ClassFile c = f.newClassFile("java/lang/Object");

		c.getExceptionTable(sig);
	}

	@Test
	public void testGetValueFromConstantPool1() {
		ClassFile c = f.newClassFile("java/lang/Object");
		assertEquals(new Simplex(500000), c.getValueFromConstantPool(1));
	}

	@Test
	public void testGetValueFromConstantPool2() {
		ClassFile c = f.newClassFile("java/lang/Math");
		assertEquals(new Simplex(0.5f), c.getValueFromConstantPool(2));
	}

	@Test
	public void testGetValueFromConstantPool3() {
		ClassFile c = f.newClassFile("java/lang/Math");
		assertEquals(new Simplex(Math.E), c.getValueFromConstantPool(73));
	}

	@Test
	public void testGetValueFromConstantPool4() {
		ClassFile c = f.newClassFile("java/lang/Long");
		assertEquals(new Simplex(Long.MAX_VALUE), c.getValueFromConstantPool(150));
	}

	/**
	 * value out of bounds
	 */
	@Test(expected=InvalidIndexException.class)
	public void testGetValueFromConstantPool5() {
		ClassFile c = f.newClassFile("java/lang/Long");
		c.getValueFromConstantPool(300);
	}

	/**
	 * Method not declared in the class, but declared in some superclass.
	 */
	@Test
	public void testHasMethodDeclaration1() {
		Signature sig = new Signature("muse/jvm/Engine", "()Ljava/lang/String;", "toString");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		assertFalse(c.hasMethodDeclaration(sig));
	}

	/**
	 * Method not declared in the class neither in some superclass.
	 */
	@Test
	public void testHasMethodDeclaration2() {
		Signature sig = new Signature("muse/jvm/Engine", "()Z", "foo");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		assertFalse(c.hasMethodDeclaration(sig));
	}

	/**
	 * Method declared in abstract class, but not implemented in it.
	 */
	@Test
	public void testHasMethodDeclaration3() {
		Signature sig = new Signature("muse/bc/ClassFile", "()Z", "isInterface");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertTrue(c.hasMethodDeclaration(sig));
	}

	/**
	 * Method declared and implemented in abstract class.
	 */
	@Test
	public void testHasMethodDeclaration4() {
		Signature sig = new Signature("muse/bc/ClassFile", "(Ljava/lang/Object;)Z", "equals");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertTrue(c.hasMethodDeclaration(sig));
	}

	/**
	 * Method declared and implemented in concrete class.
	 */
	@Test
	public void testHasMethodDeclaration5() {
		Signature sig = new Signature("java/lang/Object", "(Ljava/lang/Object;)Z", "equals");
		ClassFile c = f.newClassFile("java/lang/Object");

		assertTrue(c.hasMethodDeclaration(sig));
	}

	/**
	 * Method declared in interface.
	 */
	@Test
	public void testHasMethodDeclaration6() {
		Signature sig = new Signature("java/lang/Runnable", "()V", "run");
		ClassFile c = f.newClassFile("java/lang/Runnable");

		assertTrue(c.hasMethodDeclaration(sig));
	}

	/**
	 * Native method.
	 */
	@Test
	public void testHasMethodDeclaration7() {
		Signature sig = new Signature("java/lang/Object", "()V", "notifyAll");
		ClassFile c = f.newClassFile("java/lang/Object");

		assertTrue(c.hasMethodDeclaration(sig));
	}

	/**
	 * Method not declared in the class, but declared in some superclass.
	 */
	@Test
	public void testHasMethodImplementation1() {
		Signature sig = new Signature("muse/jvm/Engine", "()Ljava/lang/String;", "toString");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		assertFalse(c.hasMethodImplementation(sig));
	}

	/**
	 * Method not declared in the class neither in some superclass.
	 */
	@Test
	public void testHasMethodImplementation2() {
		Signature sig = new Signature("muse/jvm/Engine", "()Z", "foo");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		assertFalse(c.hasMethodImplementation(sig));
	}

	/**
	 * Method declared in abstract class, but not implemented in it.
	 */
	@Test
	public void testHasMethodImplementation3() {
		Signature sig = new Signature("muse/bc/ClassFile", "()Z", "isInterface");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertFalse(c.hasMethodImplementation(sig));
	}

	/**
	 * Method declared and implemented in abstract class.
	 */
	@Test
	public void testHasMethodImplementation4() {
		Signature sig = new Signature("muse/bc/ClassFile", "(Ljava/lang/Object;)Z", "equals");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertTrue(c.hasMethodImplementation(sig));
	}

	/**
	 * Method declared and implemented in concrete class.
	 */
	@Test
	public void testHasMethodImplementation5() {
		Signature sig = new Signature("java/lang/Object", "(Ljava/lang/Object;)Z", "equals");
		ClassFile c = f.newClassFile("java/lang/Object");

		assertTrue(c.hasMethodImplementation(sig));
	}

	/**
	 * Method declared in interface.
	 */
	@Test
	public void testHasMethodImplementation6() {
		Signature sig = new Signature("java/lang/Runnable", "()V", "run");
		ClassFile c = f.newClassFile("java/lang/Runnable");

		assertFalse(c.hasMethodImplementation(sig));
	}

	/**
	 * Native method.
	 */
	@Test
	public void testHasMethodImplementation7() {
		Signature sig = new Signature("java/lang/Object", "()V", "notifyAll");
		ClassFile c = f.newClassFile("java/lang/Object");

		assertTrue(c.hasMethodImplementation(sig));
	}

	@Test
	public void testIsInterface1() {
		ClassFile c = f.newClassFile("java/lang/Object");
		assertFalse(c.isInterface());
	}

	@Test
	public void testIsInterface2() {
		ClassFile c = f.newClassFile("java/lang/Runnable");		
		assertTrue(c.isInterface());
	}

	@Test
	public void testIsAbstract1() {
		ClassFile c = f.newClassFile("java/lang/Object");
		assertFalse(c.isAbstract());
	}

	@Test
	public void testIsAbstract2() {
		ClassFile c = f.newClassFile("java/lang/Runnable");		
		assertTrue(c.isAbstract());
	}

	@Test
	public void testIsAbstract3() {
		ClassFile c = f.newClassFile("muse/bc/ClassFile");
		assertTrue(c.isAbstract());
	}

	/** 
	 * public class
	 */
	@Test
	public void testIsPublic1() {
		ClassFile c = f.newClassFile("muse/bc/ClassFile");
		assertTrue(c.isPublic());
	}

	/**
	 * private nested class
	 */
	@Test
	public void testIsPublic2() {
		ClassFile c = f.newClassFile("java/util/LinkedList$ListItr");
		assertFalse(c.isPublic());
	}

	/**
	 * public nested class
	 */
	@Test
	public void testIsPublic3() {
		ClassFile c = f.newClassFile("java/util/concurrent/ThreadPoolExecutor$AbortPolicy");
		assertTrue(c.isPublic());
	}

	/**
	 * (package) class
	 */
	@Test
	public void testIsPublic4() {
		ClassFile c = f.newClassFile("java/util/regex/ASCII");
		assertFalse(c.isPublic());
	}
	
	@Test
	public void testIsSuperInvoke1() {
		ClassFile c = f.newClassFile("java/lang/Class");
		assertTrue(c.isSuperInvoke());
	}

	/**
	 * Method not declared in the class, but declared in some superclass.
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodAbstract1() {
		Signature sig = new Signature("muse/jvm/Engine", "()Ljava/lang/String;", "toString");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodAbstract(sig);
	}

	/**
	 * Method not declared in the class neither in some superclass.
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodAbstract2() {
		Signature sig = new Signature("muse/jvm/Engine", "()Z", "foo");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodAbstract(sig);
	}

	/**
	 * Method declared in abstract class, but not implemented in it.
	 */
	@Test
	public void testIsMethodAbstract3() {
		Signature sig = new Signature("muse/bc/ClassFile", "()Z", "isInterface");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertTrue(c.isMethodAbstract(sig));
	}

	/**
	 * Method declared and implemented in abstract class.
	 */
	@Test
	public void testIsMethodAbstract4() {
		Signature sig = new Signature("muse/bc/ClassFile", "(Ljava/lang/Object;)Z", "equals");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertFalse(c.isMethodAbstract(sig));
	}

	/**
	 * Method declared and implemented in concrete class.
	 */
	@Test
	public void testIsMethodAbstract5() {
		Signature sig = new Signature("java/lang/Object", "(Ljava/lang/Object;)Z", "equals");
		ClassFile c = f.newClassFile("java/lang/Object");

		assertFalse(c.isMethodAbstract(sig));
	}

	/**
	 * Method declared in interface.
	 */
	@Test
	public void testIsMethodAbstract6() {
		Signature sig = new Signature("java/lang/Runnable", "()V", "run");
		ClassFile c = f.newClassFile("java/lang/Runnable");

		assertTrue(c.isMethodAbstract(sig));
	}

	/**
	 * Method not declared in the class, but declared in some superclass 
	 * (not native in the superclass).
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodNative1() {
		Signature sig = new Signature("muse/jvm/Engine", "()Ljava/lang/String;", "toString");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodNative(sig);
	}

	/**
	 * Method not declared in the class, but declared in some superclass 
	 * (native in the superclass).
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodNative2() {
		Signature sig = new Signature("muse/jvm/Engine", "()Ljava/lang/Class;", "getClass");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodNative(sig);
	}

	/**
	 * Method not declared in the class neither in some superclass.
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodNative3() {
		Signature sig = new Signature("muse/jvm/Engine", "()Z", "baz");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodNative(sig);
	}

	/**
	 * Method declared in abstract class, but not implemented in it.
	 */
	@Test
	public void testIsMethodNative4() {
		Signature sig = new Signature("muse/bc/ClassFile", "()Z", "isInterface");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertFalse(c.isMethodNative(sig));
	}

	/**
	 * Method declared and implemented in abstract class, non native.
	 */
	@Test
	public void testIsMethodNative5() {
		Signature sig = new Signature("muse/bc/ClassFile", "(Ljava/lang/Object;)Z", "equals");
		ClassFile c = f.newClassFile("muse/bc/ClassFile");

		assertFalse(c.isMethodNative(sig));
	}

	/**
	 * Method declared and implemented in concrete class, native.
	 */
	@Test
	public void testIsMethodNative6() {
		Signature sig = new Signature("java/lang/Shutdown", "()V", "runAllFinalizers");
		ClassFile c = f.newClassFile("java/lang/Shutdown");

		assertTrue(c.isMethodNative(sig));
	}

	/**
	 * Method declared and implemented in concrete class, non native.
	 */
	@Test
	public void testIsMethodNative7() {
		Signature sig = new Signature("java/lang/Class", "([Ljava/lang/Class;)Ljava/lang/String;", "argumentTypesToString");
		ClassFile c = f.newClassFile("java/lang/Class");

		assertFalse(c.isMethodNative(sig));
	}
	
	/**
	 * Method declared in interface.
	 */
	@Test
	public void testIsMethodNative8() {
		Signature sig = new Signature("java/lang/Runnable", "()V", "run");
		ClassFile c = f.newClassFile("java/lang/Runnable");

		assertFalse(c.isMethodNative(sig));
	}

	/**
	 * Static method declared in class.
	 */
	@Test
	public void testIsMethodNative9() {
		Signature sig = new Signature("java/lang/Object", "()V", "registerNatives");
		ClassFile c = f.newClassFile("java/lang/Object");

		assertTrue(c.isMethodNative(sig));
	}
	
	/**
	 * Method not declared in the class, but declared in some superclass
	 * (not static in the superclass).
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodStatic1() {
		Signature sig = new Signature("muse/jvm/Engine", "()Ljava/lang/Class;", "getClass");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodStatic(sig);
	}

	/**
	 * Method not declared in the class, but declared in some superclass
	 * (static in the superclass).
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodStatic2() {
		Signature sig = new Signature("muse/jvm/Engine", "()V", "registerNatives");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodStatic(sig);
	}

	/**
	 * Method not declared in the class neither in some superclass.
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testIsMethodStatic3() {
		Signature sig = new Signature("muse/jvm/Engine", "()Z", "baz");
		ClassFile c = f.newClassFile("muse/jvm/Engine");

		c.isMethodStatic(sig);
	}
	
	/**
	 * Method declared in the class, not static.
	 */
	@Test
	public void testIsMethodStatic4() {
		Signature sig = new Signature("java/lang/Object", "()Ljava/lang/Class;", "getClass");
		ClassFile c = f.newClassFile("java/lang/Object");

		assertFalse(c.isMethodStatic(sig));
	}
	
	/**
	 * Method declared in the class, static.
	 */
	@Test
	public void testIsMethodStatic5() {
		Signature sig = new Signature("java/lang/Class", "()V", "registerNatives");
		ClassFile c = f.newClassFile("java/lang/Class");

		assertTrue(c.isMethodStatic(sig));
	}
	
	@Test
	public void testGetFieldSignature1() {
		Signature sig = new Signature("java/lang/Byte", "B", "value");
		ClassFile c = f.newClassFile("java/lang/Byte");

		assertEquals(sig, c.getFieldSignature(130));
	}
	
	@Test(expected=InvalidIndexException.class)
	public void testGetFieldSignature2() {
		ClassFile c = f.newClassFile("java/lang/Byte");
		c.getFieldSignature(1);
	}
	
	@Test(expected=InvalidIndexException.class)
	public void testGetFieldSignature3() {
		ClassFile c = f.newClassFile("java/lang/Byte");
		c.getFieldSignature(200);
	}

	@Test
	public void testGetMethodSignature1() {
		Signature sig = new Signature("java/io/ObjectOutputStream", "()V", "defaultWriteObject");
		ClassFile c = f.newClassFile("java/util/LinkedList");

		assertEquals(sig, c.getMethodSignature(200));
	}
	
	@Test(expected=InvalidIndexException.class)
	public void testGetMethodSignature2() {
		ClassFile c = f.newClassFile("java/util/LinkedList");
		c.getMethodSignature(0);
	}
	
	@Test
	public void testGetInterfaceMethodSignature1() {
		Signature sig = new Signature("java/util/Collection", "()[Ljava/lang/Object;", "toArray");
		ClassFile c = f.newClassFile("java/util/LinkedList");

		assertEquals(sig, c.getInterfaceMethodSignature(228));
	}
	
	@Test(expected=InvalidIndexException.class)
	public void testGetInterfaceMethodSignature2() {
		ClassFile c = f.newClassFile("java/util/LinkedList");
		c.getInterfaceMethodSignature(227);
	}
	
	@Test
	public void testGetFieldsNonStatic1() {
		ClassFile c = f.newClassFile("java/lang/Boolean");

		//oddly, with HashSet<Signature> it does not work even with 
		//Signature.equals defined, so we use HashSet<String> instead
		
		HashSet<String> fldsExp = new HashSet<String>();
		fldsExp.add("java/lang/Boolean:Z:value");

		Signature[] flds = c.getFieldsNonStatic();
		HashSet<String> fldsAct = new HashSet<String>();
		for (Signature sig: flds) fldsAct.add(sig.toString());

		assertEquals(fldsExp, fldsAct);
	}
	
	@Test
	public void testGetFieldsStatic1() {
		ClassFile c = f.newClassFile("java/lang/Boolean");

		//oddly, with HashSet<Signature> it does not work even with 
		//Signature.equals defined, so we use HashSet<String> instead
		
		HashSet<String> fldsExp = new HashSet<String>();
		fldsExp.add("java/lang/Boolean:Ljava/lang/Boolean;:TRUE");
		fldsExp.add("java/lang/Boolean:Ljava/lang/Boolean;:FALSE");
		fldsExp.add("java/lang/Boolean:Ljava/lang/Class;:TYPE");
		fldsExp.add("java/lang/Boolean:J:serialVersionUID");

		Signature[] flds = c.getFieldsStatic();
		HashSet<String> fldsAct = new HashSet<String>();
		for (Signature sig: flds) fldsAct.add(sig.toString());

		assertEquals(fldsExp, fldsAct);
	}
	
	@Test
	public void testGetLocalVariableLength1() {
		Signature sig = new Signature("java/util/LinkedList", "(Ljava/lang/Object;)V", "addFirst");
		ClassFile c = f.newClassFile("java/util/LinkedList");
		assertEquals(2, c.getLocalVariableLength(sig));
	}
	
	@Test(expected=MethodNotFoundException.class)
	public void testGetLocalVariableLength2() {
		Signature sig = new Signature("java/lang/Object", "()V", "foo");
		ClassFile c = f.newClassFile("java/lang/Object");
		c.getLocalVariableLength(sig);
	}
	
	@Test(expected=MethodCodeNotFoundException.class)
	public void testGetLocalVariableLength3() {
		Signature sig = new Signature("java/util/Collection", "(Ljava/lang/Object;)Z", "remove");
		ClassFile c = f.newClassFile("java/util/Collection");
		c.getLocalVariableLength(sig);
	}
	
	@Test
	public void testGetCodeLength1() {
		Signature sig = new Signature("java/util/LinkedList", "()V", "<init>");
		ClassFile c = f.newClassFile("java/util/LinkedList");
		assertEquals(43, c.getCodeLength(sig));
	}
	
	@Test
	public void testGetCodeLength2() {
		Signature sig = new Signature("java/lang/Object", "()V", "<clinit>");
		ClassFile c = f.newClassFile("java/lang/Object");
		assertEquals(4, c.getCodeLength(sig));
	}

	/**
	 * Native method.
	 */
	@Test(expected=MethodCodeNotFoundException.class)
	public void testGetCodeLength3() {
		Signature sig = new Signature("java/lang/Object", "()V", "notifyAll");
		ClassFile c = f.newClassFile("java/lang/Object");
		c.getCodeLength(sig);
	}

	/**
	 * Method not declared in the class neither in some superclass.
	 */
	@Test(expected=MethodNotFoundException.class)
	public void testGetMethodCodeBySignature1() {
		Signature sig = new Signature("java/lang/Object", "(I)Z", "foo");
		ClassFile c = f.newClassFile("java/lang/Object");
		c.getMethodCodeBySignature(sig);
	}

	/**
	 * Native method.
	 */
	@Test(expected=MethodCodeNotFoundException.class)
	public void testGetMethodCodeBySignature2() {
		Signature sig = new Signature("java/lang/Object", "()V", "notifyAll");
		ClassFile c = f.newClassFile("java/lang/Object");
		c.getMethodCodeBySignature(sig);
	}

	//TODO hasField/getLocalVariableTable/getMethodCodeBySignature (to complete), array classes
}
