package muse.dec;

import static org.junit.Assert.*;

import java.util.SortedSet;


import muse.dec.DecisionException;
import muse.dec.DecisionProcedure;
import muse.dec.DecisionProcedureSicstus;
import muse.mem.Expression;
import muse.mem.Operator;
import muse.mem.Primitive;
import muse.mem.Simplex;
import muse.mem.Term;
import muse.mem.Type;
import muse.tree.DecisionAlternativeBoolean;
import muse.tree.DecisionAlternativeComparison;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

public class DecisionProcedureTest {

	static DecisionProcedure simplifier;

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		simplifier = 
			new DecisionProcedureSicstus("127.0.0.1", 8888, null);
	}

	@AfterClass
	public static void tearDownBeforeClass() throws Exception {
		simplifier.quit();
	}

	@After
	public void tearDown() throws Exception {
		simplifier.resetClause();
	}


	@Test
	public void testSimplifyDecision1() throws DecisionException {
		// 2 > 4
		Expression e = new Expression(new Simplex(2), Operator.GT, new Simplex(4));

		//expected: {FF}
		SortedSet<DecisionAlternativeBoolean> dExp = simplifier.getComparators().createEmptyDecisionAlternativeBoolean();
		dExp.add(DecisionAlternativeBoolean.FF);

		SortedSet<DecisionAlternativeBoolean> d = simplifier.getComparators().createEmptyDecisionAlternativeBoolean();
		simplifier.simplifyDecision(e, d);
		assertEquals(dExp, d);
	}

	@Test
	public void testSimplifyDecision2() throws DecisionException {
		// true |- (A > 0) && (A <= 1)
		Term A = new Term(Type.INT, "A");
		Expression e = (Expression) A.gt(new Simplex(0));
		e = (Expression) e.and(A.le(new Simplex(1)));

		//expected: {TT, FF}
		SortedSet<DecisionAlternativeBoolean> dExp = simplifier.getComparators().createEmptyDecisionAlternativeBoolean();
		dExp.add(DecisionAlternativeBoolean.TT);
		dExp.add(DecisionAlternativeBoolean.FF);

		SortedSet<DecisionAlternativeBoolean> d = simplifier.getComparators().createEmptyDecisionAlternativeBoolean();
		simplifier.simplifyDecision(e, d);
		assertEquals(dExp, d);
	}

	@Test
	public void testSimplifyComparison1() throws DecisionException {
		// true |- 2 ? 5
		Simplex two = new Simplex(2);
		Simplex five = new Simplex(5);

		//expected {LT}
		SortedSet<DecisionAlternativeComparison> dExp = simplifier.getComparators().createEmptyDecisionAlternativeComparison();
		dExp.add(DecisionAlternativeComparison.LT);

		SortedSet<DecisionAlternativeComparison> d = simplifier.getComparators().createEmptyDecisionAlternativeComparison();
		simplifier.simplifyComparison(two, five, d);
		assertEquals(dExp, d);
	}

	@Test
	public void testSimplifyComparison2() throws DecisionException {
		// A >= 0 |- 2 * A ? A
		Term A = new Term(Type.INT, "A");
		Expression Agtzero = (Expression) A.ge(new Simplex(0));
		Expression Atwice = (Expression) A.mul(new Simplex(2));

		//expected {EQ, GT}
		SortedSet<DecisionAlternativeComparison> dExp = simplifier.getComparators().createEmptyDecisionAlternativeComparison();
		dExp.add(DecisionAlternativeComparison.EQ);
		dExp.add(DecisionAlternativeComparison.GT);

		SortedSet<DecisionAlternativeComparison> d = simplifier.getComparators().createEmptyDecisionAlternativeComparison();
		simplifier.pushClause(Agtzero);
		simplifier.simplifyComparison(Atwice, A, d);
		assertEquals(dExp, d);
	}

	@Test
	public void testSimplify1() throws DecisionException {
		// true |- (A >= 0) && (A < 5) && !(A == 1) && !(A == 2) && !(A == 3)
		Term A = new Term(Type.INT, "A");
		Expression e = new Expression(A, Operator.GE, new Simplex(0));
		e = (Expression) e.and(new Expression(A, Operator.LT, new Simplex(5)));
		e = (Expression) e.and(new Expression(A, Operator.EQ, new Simplex(1)).not());
		e = (Expression) e.and(new Expression(A, Operator.EQ, new Simplex(2)).not());
		e = (Expression) e.and(new Expression(A, Operator.EQ, new Simplex(3)).not());

		//expected satisfiable (by A == 4)
		assertTrue(simplifier.isSat(e));
	}

	@Test
	public void testSimplify2() throws DecisionException {
		// true |- ((A >= 0) || (A < -3)) && (A == -1)  
		Term A = new Term(Type.INT, "A");
		Primitive e = new Simplex(true);
		e = e.and(new Expression(A, Operator.GE, new Simplex(0)));
		e = (Expression) e.or(new Expression(A, Operator.LT, new Simplex(-3)));
		e = (Expression) e.and(new Expression(A, Operator.EQ, new Simplex(-1)));

		//expected unsatisfiable
		assertFalse(simplifier.isSat((Expression) e));
	}

	@Test
	public void testSimplify3() throws DecisionException {
		// true |- (X && !Y) || (!X && Y)

		//boolean terms are emulated with satisfiable independent clauses
		Term A = new Term(Type.INT, "A");
		Term B = new Term(Type.INT, "B");
		Primitive X = A.ge(new Simplex(0));
		Primitive Y = B.ge(new Simplex(0));
		Primitive e0 = new Simplex(true);
		e0 = e0.and(X);
		e0 = e0.and(Y.not());
		Primitive e1 = new Simplex(true);
		e1 = e1.and(X.not());
		e1 = e1.and(Y);
		Primitive e = e0.or(e1);

		//expected satisfiable
		assertTrue(simplifier.isSat((Expression) e));
	}

	@Test
	public void testAssumption1() throws DecisionException {
		//assume:
		//(0 > {V12}) && 
		//     ({V14} >= 0) && 
		//     (0 < {V14}) && 
		//     (0 <= {V15}) && 
		//     ({V20} >= 0) && 
		//     (0 < {V20})) && 
		//     ({V12} != -2147483648) && 
		//(- {V12} <= {V15}) && 
		//(- {V12} < 65536) && 
		//(- {V12} - ({X0} * 8.0 + {X0} * 2.0)) == 0)
		Term V12 = new Term(Type.INT, "{V12}");
		//Term V14 = new Term(Type.INT, "{V14}");
		Term V15 = new Term(Type.INT, "{V15}");
		//Term V20 = new Term(Type.INT, "{V20}");
		Term X0 = new Term(Type.INT, "{X0}");
		Simplex zero = new Simplex(0);
		//Simplex bottom = new Simplex(-2147483648);
		Simplex top = new Simplex(65536);
		Simplex eight = new Simplex(8.0);
		Simplex two = new Simplex(2.0);

		Primitive e = new Simplex(true);
		e = e.and(zero.gt(V12));
		//e = e.and(V14.ge(zero));
		//e = e.and(zero.lt(V14));
		//e = e.and(zero.le(V15));
		//e = e.and(V20.ge(zero));
		//e = e.and(zero.lt(V20));
		//e = e.and(V12.ne(bottom));
		e = e.and(V12.neg().le(V15));
		e = e.and(V12.neg().lt(top));
		e = e.and(V12.neg().sub(X0.mul(eight).add(X0.mul(two))).eq(zero));

		//simplifier.pushClause(e);

		//Expression ee = (Expression) X0.ne(zero);

		assertTrue(simplifier.isSat((Expression) e));
	}

	@Test
	public void testAssumption2() throws DecisionException {
		//assume:
		//0 > B &&
		//-B <= C &&
		//B + 10.0 * A = 0 
		Term A = new Term(Type.INT, "{A}");
		Term B = new Term(Type.INT, "{B}");
		Term C = new Term(Type.INT, "{C}");
		Simplex zero = new Simplex(0);
		Simplex ten = new Simplex(10);

		Primitive e = new Simplex(true);
		e = e.and(zero.gt(B));
		e = e.and(B.neg().le(C));
		e = e.and(B.add(A.mul(ten)).eq(zero));

		assertTrue(simplifier.isSat((Expression) e));
		//current error: the Sicstus server first simplifies the expression with the clpqr solver, 
		//which simplifies the third constraint as B - 1/10*C <= 0 , then reuses the simplified constraint
		//to feed the integer solver. The latter apparently solves 1/10 as 0, yielding an unsatisfiable set 
		//of constraints.
	}

	//This boundary value has proven critical for the Sicstus 
	//simplifier, with regression
	@Test
	public void testBoundary() throws DecisionException {
		Expression e = (Expression) new Term(Type.INT, "A").eq(new Simplex(Integer.MIN_VALUE));
		assertTrue(simplifier.isSat(e));
	}
	
	@Test
	public void testOther() throws DecisionException {
		//true |- A/B + (C - C) / D < E  
		Term A = new Term(Type.INT, "{A}");
		Term B = new Term(Type.INT, "{B}");
		Term C = new Term(Type.INT, "{C}");
		Term D = new Term(Type.INT, "{D}");
		Term E = new Term(Type.INT, "{E}");
		
		Primitive e = new Simplex(true);
		e = e.and(A.div(B).add(C.sub(C).div(D)).lt(E));
		
		//expected satisfiable
		assertTrue(simplifier.isSat((Expression) e));
	}
}
